////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: formula_operations.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
// FORMULA OPERATIONS
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "libproof.h"
#include "formula_operations.h"
#include "symbolhlk.h"
//#include "util/combinations.h"
#include "term_operations.h"

#include "printhlk.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD
  ( AtomFormula, Formula, OperationReplaceTerm, f )
{
  TermList *tl = new TermList();
  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
  	Push(i);
	TOperationReplaceTerm op_rep( m_ndb, replacement, position, np );
	Term *t = DataNode::ApplyTo( f->GetTerm( i ), op_rep );
        tl->AddRight( t );
  	Pop();
  }

  AtomFormula *new_f = AtomFormula::New( m_ndb, f->GetSymbol(), tl, f->GetSecondOrderVariable() );

  return FormulaOperation::ApplyToAtomFormula( new_f );
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD ( AtomFormula, Formula, OperationGetTerms, f )
{
#if 0
  std::cerr << "getting terms of formula ";
  HLK::Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl;
#endif

  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    list.AddRight( f->GetTerm( i ) );
    np.Push( i );
    nps.Add( np );
//    OperationGetSubterms op_sub( list, nps, np );
//    DataNode::ApplyTo( f->GetTerm( i ), op_sub );
    np.Pop();
  }

  return f;
}

////////////////////////////////////////////////////////////////////////////////

OperationLambdaSubstitution::OperationLambdaSubstitution(
  ProofDatabase *pdb, const SecondOrderVariable *v, const LambdaTerm *subst )
{
  ASSERT( pdb != NULL );
  m_ndb = pdb->GetNodeDatabase();
  m_pdb = pdb;

  AddSubstitution( v, subst );
}

OperationLambdaSubstitution::OperationLambdaSubstitution(
  NodeDatabase *ndb, const SecondOrderVariable *v, const LambdaTerm *subst )
{
  ASSERT( ndb != NULL );
  m_ndb = ndb;
  m_pdb = NULL;

  AddSubstitution( v, subst );
}

OperationLambdaSubstitution::OperationLambdaSubstitution(
  NodeDatabase *ndb )
{
  ASSERT( ndb != NULL );
  m_ndb = ndb;
  m_pdb = NULL;
}

void OperationLambdaSubstitution::AddSubstitution(
  const SecondOrderVariable *v, const LambdaTerm *subst )
{
  ASSERT( v != NULL );
  ASSERT( subst != NULL );

  substitutions.push_back( Substitution( v, subst ) );
  m_positions.resize( substitutions.size() );
}

const SecondOrderVariable* OperationLambdaSubstitution::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return substitutions[index].first;
}

const LambdaTerm* OperationLambdaSubstitution::GetSubst( size_t index ) const
{
  ASSERT( index < GetCount() );
  return substitutions[index].second;
}

const NodePointers&
OperationLambdaSubstitution::GetPositions( size_t index ) const
{
  ASSERT( index < substitutions.size() );
  ASSERT( index < m_positions.size() );

  return m_positions[index];
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, OperationLambdaSubstitution, v )
{
  for( size_t i = 0; i < substitutions.size(); i++ )
  {
    if( v->GetSecondOrderVariable() != substitutions[i].first )
      continue; // not "this" variable

    if( std::find( bounded.begin(), bounded.end(), substitutions[i].first ) != bounded.end() )
      continue; // variable is bounded

    m_positions[i].Add( m_np );
    
    TermList tl;
    for ( unsigned int j = 0; j < v->GetChildCount(); j++)
    {
      tl.AddRight( v->GetTerm( j ) );
    }
    
    return substitutions[i].second->apply( m_pdb, tl );
  }

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SecondOrderQuantorFormula, Formula, OperationLambdaSubstitution, f )
{
  // don't replace bounded vars!
  bounded.push_back( f->GetSecondOrderVariable() );
#if 0
  if( v == f->GetVariable() )
   return f;
#endif

  ProofLib::Formula *f_1 = FormulaOperation::ApplyToSecondOrderQuantorFormula( f );

  ASSERT( bounded.back() == f->GetSecondOrderVariable() );
  bounded.pop_back();

  return f_1;
}

std::ostream& OperationLambdaSubstitution::ToString( std::ostream &out ) const
{
  typedef std::pair< const SecondOrderVariable*, const SecondOrderVariable* > Substitution;

  for( size_t i = 0; i < substitutions.size(); i++ )
  {
    if( i != 0 ) out << ", ";
    substitutions[i].first->ToString(out);
    out << " <- ";
    substitutions[i].second->ToString(out);
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

Variable* 
OperationGetSecondOrderVariableSubstitutions::getNormalVariable( size_t index, USTRING type_name )
{
  if ( index >= normal_variables.size() )
  {
  	for ( size_t i = normal_variables.size(); i <= index; i++)
    {
  	  normal_variables.push_back( NULL );
    }
  }
  
  if ( normal_variables[ index ] == NULL )
  {  
    // FIXME: get sourcelocation from somewhere else?
    SourceLocation sloc( "___INTERNAL_OperationGetLambdaSubstitutions___",
                          0, 0);
     
    
    const HLK::VariableSymbol* vs = m_symtbl->NewVariable(
        USTRING(),
        type_name,
        true,
        sloc );
    Variable* new_v = Variable::New( m_ndb, vs );
    normal_variables[ index ] = new_v;
  }
  
  return normal_variables[ index ];
}

void OperationGetSecondOrderVariableSubstitutions::Clear()
{
  while( vars.size() > 0 ) vars.clear();
  while( reps.size() > 0 ) reps.clear();
  while( bounded.size() > 0 ) bounded.clear();
//  while( equals.GetChildCount() > 0 ) equals.RemoveAt(0);
  pos.clear();
  np.Clear();
}

bool OperationGetSecondOrderVariableSubstitutions::CheckForSubstitutions(
  DataNode *general, DataNode *instance )
{
  Clear();
  return general->ApplyCompare( *this, instance );
}

const SecondOrderVariable* OperationGetSecondOrderVariableSubstitutions::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return vars[index];
}

const LambdaTerm* OperationGetSecondOrderVariableSubstitutions::GetSubst( size_t index ) const
{
  ASSERT( index < GetCount() );
  return reps[index];
}

const NodePointer& OperationGetSecondOrderVariableSubstitutions::GetVarPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].var;
}

const NodePointers&
OperationGetSecondOrderVariableSubstitutions::GetSubstPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].substs;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( AtomFormula, OperationGetSecondOrderVariableSubstitutions, f )
{
  if ( f == o )
  {
  	// atoms are equal
    return true;
  }

  const AtomFormula *af = o->CastTo<const AtomFormula>();
 
  if( af == NULL )
  {
  	// datanode is not an AtomFormula, so we can't match
  	// by substituting variables
    return false;
  }

  SecondOrderVariable *var = af->GetSecondOrderVariable();
  
  if ( var == NULL )
  {
  	// AtomFormula is constant, so we can't match by
  	// substituting variables
  	return false;
  }

  SecondOrderVariable *var2 = f->GetSecondOrderVariable();

  if ( var2 == NULL )
  {
  	// AtomFormula is constant, so this can't be the
  	// result of a variable substitution
  	return false;
  }

  // variable already found?
  int p = -1;
  for (unsigned int i = 0; i < vars.size(); i++)
  {
  	if ( vars[i] == var2 )
  	  p = i;
  }

  if( p >= 0 )
  {
  	TermList tl;
  	for ( unsigned int i = 0; i < f->GetChildCount(); i++ )
  	{
  	  tl.AddRight( f->GetTerm( i ) );	
  	}
    // variable has to be substituted with the same variable!
#if 0
    std::cerr << "applying substitution: ";
    HLK::Print::PrintFormula( reps[p]->getFormula(), std::cerr );
    std::cerr << std::endl;
#endif
    if( reps[p]->apply( m_pdb, tl ) != af )
      return false;

    pos[p].substs.Add( np );
    return true;
  }
  
  // was there already a position, where this var wasn't substituted?
  if( std::find( equals.begin(), equals.end(), var2 ) != equals.end() ) 
    return false;

  // is this a bounded variable?
  if( std::find( bounded.begin(), bounded.end(), var2 ) != bounded.end() ) 
    return true;

  p = pos.size();
  pos.resize( p + 1 );
  pos[p].var = np;
  pos[p].substs.Add( np );
  vars.push_back( var2 );
  
  // create lambda term representing this variable substitution
  // with an empty list of variables
  TermList b_vars;
  TermList *tl = new TermList();
  for ( unsigned int i = 0; i < af->GetChildCount(); i++)
  {
    // create new variables for lambda term
    Term *t = af->GetTerm(i);
    USTRING type_name;
    Variable *v;
    Function *f;
    Constant *c;
     
    if ( (v = dynamic_cast<Variable*>(t)) != NULL )
    {
     const HLK::VariableSymbol* vs = dynamic_cast<const HLK::VariableSymbol*>(v->GetSymbol());
     ASSERT( vs != NULL );
     type_name = vs->GetType()->GetName();
    }
    else if ( (f = dynamic_cast<Function*>(t)) != NULL)
    {
      const HLK::FunctionSymbol* fs = dynamic_cast<const HLK::FunctionSymbol*>(f->GetSymbol());
      ASSERT( fs != NULL );
      type_name = fs->GetType()->GetName();
    }
    else if ( (c = dynamic_cast<Constant*>(t)) != NULL)
    {
      const HLK::ConstantSymbol* cs = dynamic_cast<const HLK::ConstantSymbol*>(c->GetSymbol());
      ASSERT( cs != NULL );
      type_name = cs->GetType()->GetName();
    }
    else
      ASSERT( false );
    
    
    Variable* new_v = getNormalVariable( i, type_name ); 
    tl->AddRight( new_v );
    new_v->Attach();
    b_vars.AddRight( new_v );
  }

  Formula *form = AtomFormula::New( m_ndb, af->GetSymbol(), tl, af->GetSecondOrderVariable() );
  LambdaTerm *t = LambdaTerm::New( m_ndb, b_vars, form );
#if 0
  std::cerr << "Adding lambdaterm to reps: ";
  HLK::Print::PrintFormula( form, std::cerr );
  std::cerr << std::endl;
#endif
  reps.push_back( t );

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SecondOrderQuantorFormula, OperationGetSecondOrderVariableSubstitutions, f )
{
  SecondOrderVariable *v = f->GetSecondOrderVariable();

  bool already = ( std::find( bounded.begin(), bounded.end(), v ) != bounded.end() );

  if( !already )
    bounded.push_back( v );

  bool ok = FormulaCompareOperation::ApplyToSecondOrderQuantorFormula( f, o );

  if( !already )
  {
    std::vector<SecondOrderVariable*>::iterator pos = 
      std::find( bounded.begin(), bounded.end(), v );
    if ( pos != bounded.end() )
      bounded.erase( pos );
  }

  return ok;
}

////////////////////////////////////////////////////////////////////////////////

void OperationGetLambdaSubstitutions::Clear()
{
  while( vars.size() > 0 ) vars.clear();
  while( reps.size() > 0 ) reps.clear();
  while( bounded.size() > 0 ) bounded.clear();
//  while( equals.GetChildCount() > 0 ) equals.RemoveAt(0);
  pos.clear();
  np.Clear();
}

bool OperationGetLambdaSubstitutions::CheckForSubstitutions(
  DataNode *general, DataNode *instance )
{
  Clear();
  return general->ApplyCompare( *this, instance );
}

const SecondOrderVariable* OperationGetLambdaSubstitutions::GetVar( size_t index ) const
{
  ASSERT( index < GetCount() );
  return vars[index];
}

const std::set<LambdaTerm*> OperationGetLambdaSubstitutions::GetSubstSet( size_t index ) const
{
  ASSERT( index < GetCount() );
  return reps[index];
}

const LambdaTerm* OperationGetLambdaSubstitutions::GetSubst( size_t index ) const
{
  std::set<LambdaTerm*> substs = GetSubstSet( index );
  LambdaTerm *res = NULL;
  int varcount = -1;
  
  for ( std::set<LambdaTerm*>::iterator i = substs.begin();
        i != substs.end(); i++)
  {
  	// most general == most free variable occurences
  	TermList tl;
  	OperationCountFreeVariables op( tl );
  	(*i)->getFormula()->Apply( op );
  	int num = op.GetCount();

  	if ( num > varcount )
  	{
  	  varcount = num;
  	  res = *i;
  	}
  }
  
  ASSERT( res != NULL );
  
  return res;
}

const NodePointer& OperationGetLambdaSubstitutions::GetVarPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].var;
}

const NodePointers&
OperationGetLambdaSubstitutions::GetSubstPos( size_t index ) const
{
  ASSERT( index < GetCount() );
  return pos[index].substs;
}

Variable* 
OperationGetLambdaSubstitutions::getNormalVariable( size_t index, USTRING type_name )
{
  if ( index >= normal_variables.size() )
  {
  	for ( size_t i = normal_variables.size(); i <= index; i++)
    {
  	  normal_variables.push_back( NULL );
    }
  }
  
  if ( normal_variables[ index ] == NULL )
  {  
    // FIXME: get sourcelocation from somewhere else?
    SourceLocation sloc( "___INTERNAL_OperationGetLambdaSubstitutions___",
                          0, 0);
     
    
    const HLK::VariableSymbol* vs = m_symtbl->NewVariable(
        USTRING(),
        type_name,
        true,
        sloc );
    Variable* new_v = Variable::New( m_ndb, vs );
    normal_variables[ index ] = new_v;
  }
  
  return normal_variables[ index ];
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( AtomFormula, OperationGetLambdaSubstitutions, f )
{
  if ( f == o )
  {
  	// atoms are equal
    return true;
  }

  SecondOrderVariable *var2 = f->GetSecondOrderVariable();

  if ( var2 == NULL )
  {
   // AtomFormula f is constant and not equal to af, so we
   // can't substitute 
   return false;
  }

  // collect terms appearing in formula, together with
  // a set of variables with size equal to the arity of
  // the variable atom formula

  TermList formula_terms;
  NodePointers formula_term_positions;
  OperationGetTerms op_gt( formula_terms, formula_term_positions );
  DataNode::ApplyTo( const_cast<DataNode*>(o), op_gt ) ;
#if 0
  std::cerr << "formula_terms.GetChildCount() = " << formula_terms.GetChildCount() << std::endl;
  std::cerr << "formula_term_positions.GetCount() = " << formula_term_positions.GetCount() << std::endl;
#endif
  TermList variables;
  for (unsigned int i = 0; i < f->GetChildCount(); i++)
  {
    // create new variables for lambda term
    Term *t = f->GetTerm(i);
    USTRING type_name;
    Variable *v;
    Function *f;
    Constant *c;
     
    if ( (v = dynamic_cast<Variable*>(t)) != NULL )
    {
     const HLK::VariableSymbol* vs = dynamic_cast<const HLK::VariableSymbol*>(v->GetSymbol());
     ASSERT( vs != NULL );
     type_name = vs->GetType()->GetName();
    }
    else if ( (f = dynamic_cast<Function*>(t)) != NULL)
    {
      const HLK::FunctionSymbol* fs = dynamic_cast<const HLK::FunctionSymbol*>(f->GetSymbol());
      ASSERT( fs != NULL );
      type_name = fs->GetType()->GetName();
    }
    else if ( (c = dynamic_cast<Constant*>(t)) != NULL)
    {
      const HLK::ConstantSymbol* cs = dynamic_cast<const HLK::ConstantSymbol*>(c->GetSymbol());
      ASSERT( cs != NULL );
      type_name = cs->GetType()->GetName();
    }
    else
      ASSERT( false );
    
    
    Variable* new_v = getNormalVariable( i, type_name ); 
    variables.AddRight( new_v );
  }

  TermList arguments;
  
  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
  	arguments.AddRight( f->GetTerm( i ) );
  }

  // create all permutations of terms and variables of
  // length equal to the arity of the variable atom formula

  const Formula *f2 = dynamic_cast<const Formula*>(o);
  
  ASSERT( f2 != NULL );
  
  permutations.clear();
  createPermutedFormulas( formula_term_positions,
                          //NodePointer(),
                          formula_terms,
                          variables,
                          arguments,
                          0,
                          const_cast<Formula*>(f2) );
 
  std::set<LambdaTerm*> lambda_terms;
  for ( size_t i = 0; i < permutations.size(); i++ )
  {
#if 0
        std::cerr << "creating lambdaterm from permutation: ";
	HLK::Print::PrintFormula( permutations[i], std::cerr );
	std::cerr << std::endl;
#endif
  	LambdaTerm *lt = LambdaTerm::New( m_ndb, variables, permutations[i] );
  	lambda_terms.insert( lt );
  }


  // variable already found?
  int p = -1;
  for (unsigned int i = 0; i < vars.size(); i++)
  {
    if ( vars[i] == var2 )
      p = i;
  }

  if( p >= 0 )
  {
    std::vector< std::vector<LambdaTerm*>::iterator > to_erase;
    
    // intersect sets of lambda terms
    std::set<LambdaTerm*> new_lambda_terms;
    std::set_intersection(reps[p].begin(), reps[p].end(),
       lambda_terms.begin(), lambda_terms.end(),
       std::insert_iterator< std::set<LambdaTerm*> >(new_lambda_terms,
                                                     new_lambda_terms.begin() ) );
    reps[p] = new_lambda_terms;
    
    if ( reps[p].empty() )
    {
      return false;
    }

    return true;
  }
  
  // is this a bounded variable?
  if( std::find( bounded.begin(), bounded.end(), var2 ) != bounded.end() ) 
    return true;

  p = pos.size();
  pos.resize( p + 1 );
  pos[p].var = np;
  pos[p].substs.Add( np );
  vars.push_back( var2 );
  reps.resize( p + 1 );
  reps[p] = lambda_terms;

  if ( reps[p].empty() )
    return false;

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SecondOrderQuantorFormula, OperationGetLambdaSubstitutions, f )
{
  SecondOrderVariable *v = f->GetSecondOrderVariable();

  bool already = ( std::find( bounded.begin(), bounded.end(), v ) != bounded.end() );

  if( !already )
    bounded.push_back( v );

  bool ok = FormulaCompareOperation::ApplyToSecondOrderQuantorFormula( f, o );

  if( !already )
  {
    std::vector<SecondOrderVariable*>::iterator pos = 
      std::find( bounded.begin(), bounded.end(), v );
    if ( pos != bounded.end() )
      bounded.erase( pos );
  }

  return ok;
}

void OperationGetLambdaSubstitutions::createPermutedFormulas( NodePointers &term_positions,
                                                              TermList &terms,
							      TermList &variables,
							      TermList &arguments,
							      size_t pos,
							      Formula *f )
{
  std::vector<Formula*>::iterator i;
  i = find( permutations.begin(), permutations.end(), f );
  if ( terms.GetChildCount() == 0  && i == permutations.end() )
  {
#if 0
    std::cerr << "Adding permutation: ";
    HLK::Print::PrintFormula( f, std::cerr );
    std::cerr << std::endl;
#endif
    permutations.push_back( f );
    return;
  }

  for ( size_t i = 0; i < terms.GetChildCount(); i++ )
  {
    Term *t = terms.GetTerm( i );
    NodePointer t_pos = term_positions[ i ];
#if 0
    std::cerr << "considering term ";
    HLK::Print::PrintTerm( t, std::cerr );
    std::cerr << " at pos ";
    t_pos.Dump( std::cerr ) ;
    std::cerr << std::endl;
#endif
    for ( size_t j = 0; j < arguments.GetChildCount(); j++ )
    {
      Term *arg = arguments.GetTerm( j );
#if 0
      std::cerr << "considering argument: ";
      HLK::Print::PrintTerm( arg, std::cerr );
      std::cerr << std::endl;
#endif
      if ( t == arg )
      {
#if 0
	std::cerr << "they are equal" << std::endl;
        std::cerr << "substituting ";
	HLK::Print::PrintTerm( variables.GetTerm( j ), std::cerr );
	std::cerr << " for term" << std::endl;
#endif
        OperationReplaceTerm op_rt( m_ndb, *(variables.GetTerm( j )), t_pos );
  	Formula *new_f = DataNode::ApplyTo( f, op_rt );
#if 0
        std::cerr << "new_f: ";
	HLK::Print::PrintFormula( new_f, std::cerr );
	std::cerr << std::endl;
#endif
	TermList newterms;
	NodePointers newpointers;
#if 0
        std::cerr << "new termlist: ";
#endif
	for ( size_t k = 0; k < terms.GetChildCount(); k++ )
	{
          if ( k != i )
	  {
#if 0
	    HLK::Print::PrintTerm( terms.GetTerm( k ), std::cerr );
	    std::cerr << ", ";
#endif
	    newterms.AddRight( terms.GetTerm( k ) ) ;
	    newpointers.Add( term_positions[ k ] ) ;
	  }
	}
#if 0
	std::cerr << std::endl;
#endif
	createPermutedFormulas( newpointers, newterms, variables, arguments, pos, new_f );
      }

      if ( blacklist.FindTerm( t ) == -1 )
      {
        TermList subterms;
        NodePointers subpointers;
        OperationGetSubterms op_sub( subterms, subpointers, t_pos );
        DataNode::ApplyTo( t, op_sub );
        createPermutedFormulas( subpointers, subterms, variables, arguments, pos, f );
      }
    }
  }
}

#if 0
void OperationGetLambdaSubstitutions::createPermutedFormulas( NodePointers &term_positions,
                                                              //NodePointer &current,
                                                              TermList &terms,
                                                              TermList &variables,
                                                              TermList &arguments,
                                                              size_t pos,
                                                              Formula *f )
{
  if ( term_positions.GetCount() - pos == 1)
  {
  	if ( blacklist.FindTerm( terms.GetTerm( pos ) ) == -1 )
  	{
  	  permutations.push_back( f );
  	}
  	for ( size_t i = 0; i < variables.GetChildCount(); i++ )
  	{
  	  if ( arguments.GetTerm( i ) == terms.GetTerm( pos ) )
  	  {
  	  	OperationReplaceTerm op_rt( m_ndb, *(variables.GetTerm( i )), term_positions[ pos ] );
  	  	Formula *new_f = DataNode::ApplyTo( f, op_rt );
  	  	permutations.push_back( new_f );
  	  }
  	}

  	return;
  }

  for ( size_t i = 0; i < variables.GetChildCount(); i++ )
  {
    if ( arguments.GetTerm( i ) == terms.GetTerm( pos ) )
  	{
  	  OperationReplaceTerm op_rt( m_ndb, *(variables.GetTerm( i )), term_positions[ pos ] );
  	  Formula *new_f = DataNode::ApplyTo( f, op_rt );

	  std::cerr << "new_f: " << std::endl;
	  HLK::Print::PrintFormula( new_f, std::cerr );

          createPermutedFormulas( term_positions, /*term_positions[ pos ],*/ terms, variables, arguments, pos+1, new_f );
  	}
  	if ( blacklist.FindTerm( terms.GetTerm( pos ) ) == -1 )
  	{
  	  createPermutedFormulas( term_positions, /*current,*/ terms, variables, arguments, pos+1, f );
  	}
  }
}
#endif

////////////////////////////////////////////////////////////////////////////////

OperationReplaceFormula::OperationReplaceFormula
  ( NodeDatabase *n, Formula *to_search, Formula *its_replacement )
{
  ASSERT( to_search != NULL );
  ASSERT( its_replacement != NULL );
  ASSERT( n != NULL );

  search = to_search;
  replacement = its_replacement;
  ndb = n;
}

ProofDatabase* OperationReplaceFormula::GetProofDatabase()
{
  ASSERT( false );
  return NULL;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( AtomFormula, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToAtomFormula( f );
}

DECLARE_APPLY_FUNCTION_HEAD
  ( UnaryConFormula, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToUnaryConFormula( f );
}

DECLARE_APPLY_FUNCTION_HEAD
  ( QuantorFormula, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToQuantorFormula( f );
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SecondOrderQuantorFormula, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToSecondOrderQuantorFormula( f );
}

DECLARE_APPLY_FUNCTION_HEAD
  ( BinaryConFormula, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToBinaryConFormula( f );
}

DECLARE_APPLY_FUNCTION_HEAD
  ( FormulaVariable, Formula, OperationReplaceFormula, f )
{
  if( AtSequentScope() && f == search )
    return replacement;

  return FormulaOperation::ApplyToFormulaVariable( f );
}

////////////////////////////////////////////////////////////////////////////////

OperationReplaceFormulaSequence::
OperationReplaceFormulaSequence( NodeDatabase *ndb,
                                 const FormulaList *to_search,
                                 FormulaListElement *its_replacement )
{
  search = to_search;
  replacement = its_replacement;

  this->ndb = ndb;
}

ProofDatabase* OperationReplaceFormulaSequence::GetProofDatabase()
{
  ASSERT( false );
  return NULL;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( FormulaList, FormulaList, OperationReplaceFormulaSequence, l )
{
  return l;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SElementList, SElementList, OperationReplaceFormulaSequence, l )
{
  if( !InSequentScope() )
    return FormulaOperation::ApplyToSElementList( l );

  size_t index = 0;

  while( l->FindSubListAtIndex( search, index ) )
  {
    Push( index );
    
    if( !AtSequentScope() )
    {
      Pop();
      continue;
    }
    
    for( size_t i = index + 1; i < search->GetChildCount() + index; i++ )
    {
      SElement *e = l->GetSElement(i);
      l->RemoveAt(i);
      delete e;
    }

    ProofLib::FormulaListVariable *fv =
      replacement->CastTo<ProofLib::FormulaListVariable>();

    if( fv != NULL )
      l->ReplaceSElement( index, new SFormulaListVariable( fv ) );
    else
      ASSERT( false );

    Pop();
    index++;
  }

  return l;
}
  
////////////////////////////////////////////////////////////////////////////////

OperationReplaceFormulaListElementBySequence::
OperationReplaceFormulaListElementBySequence( NodeDatabase *ndb,
                                   FormulaListElement *to_search,
                                   const FormulaList *its_replacement )
{
  search = to_search;
  replacement = its_replacement;

  this->ndb = ndb;
}

ProofDatabase* OperationReplaceFormulaListElementBySequence::GetProofDatabase()
{
  ASSERT( false );
  return NULL;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SElementList, SElementList,
    OperationReplaceFormulaListElementBySequence, l )
{
  if( !InSequentScope() )
    return l;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Push(i);

    if( !AtSequentScope() )
    {
      Pop();
      continue;
    }

    ProofLib::SFormulaListVariable *sfv =
      l->GetSElement(i)->CastTo<ProofLib::SFormulaListVariable>();

    if( sfv == NULL || sfv->GetFormulaListVariable() != search )
    {
      Pop();
      continue;
    }

    l->RemoveAt(i);
    l->InsertAt( i, replacement );

    Pop();
  }

  return l;
}

////////////////////////////////////////////////////////////////////////////////

}
