////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: formula_operations.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
// FORMULA OPERATIONS
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _FORMULA_OPERATIONS_H_
#define _FORMULA_OPERATIONS_H_

#include <vector>

#include "libproof.h"
#include "operations.h"

namespace HLK
{
  class SymbolTable;
}

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

class OperationReplaceTerm : public FormulaOperation
{
protected:
  NodeDatabase *m_ndb;
  
  Term &replacement;
  const NodePointer &position;
  NodePointer np;


public:
  NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationReplaceTerm( NodeDatabase *ndb, Term &replacement, const NodePointer &position ) :
    m_ndb( ndb ), replacement( replacement ), position( position ) {}

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationGetTerms :
  public FormulaOperation
{
protected:
  TermList &list;
  NodePointers &nps;
  NodePointer np;
public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationGetTerms( TermList &list, NodePointers &nps ) :
    list( list ), nps( nps ) {}

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class LambdaSubstitution
{
public:
  virtual ~LambdaSubstitution() {}

  virtual size_t GetCount() const = 0;

  virtual const SecondOrderVariable* GetVar( size_t index ) const = 0;
  virtual const LambdaTerm* GetSubst( size_t index ) const = 0;

  const LambdaTerm* GetSubstFor( const SecondOrderVariable *v ) const;
};

////////////////////////////////////////////////////////////////////////////////

class OperationLambdaSubstitution :
  public FormulaOperation,
  public LambdaSubstitution
{
  ProofDatabase *m_pdb;
  NodeDatabase  *m_ndb;

  typedef std::pair< const SecondOrderVariable*, const LambdaTerm* > Substitution;

  std::vector< Substitution > substitutions;

  std::vector<SecondOrderVariable*> bounded;

  NodePointer m_np;
  typedef std::vector<NodePointers> PositionList;
  PositionList m_positions;

public:
  OperationLambdaSubstitution( ProofDatabase *pdb,
                                 const SecondOrderVariable *v, const LambdaTerm *subst );
  OperationLambdaSubstitution( NodeDatabase *ndb,
                                 const SecondOrderVariable *v, const LambdaTerm *subst );
  OperationLambdaSubstitution( NodeDatabase *ndb );
  virtual ~OperationLambdaSubstitution() {}

  void AddSubstitution( const SecondOrderVariable *v, const LambdaTerm *subst );

  NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofDatabase* GetProofDatabase() { return m_pdb; }

  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }

  virtual size_t GetCount() const { return substitutions.size(); }

  virtual const SecondOrderVariable* GetVar( size_t index ) const;
  virtual const LambdaTerm* GetSubst( size_t index ) const;

  const NodePointers& GetPositions( size_t index ) const;
  const PositionList& GetPositions() const { return m_positions; }

  virtual std::ostream& ToString( std::ostream &out ) const;

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationGetSecondOrderVariableSubstitutions :
  public FormulaCompareOperation,
  public LambdaSubstitution
{
public:
  typedef struct {
    NodePointer          var;
    NodePointers         substs;
  } Positions;

protected:
  std::vector<SecondOrderVariable*> vars;
  std::vector<LambdaTerm*> reps;
  std::vector<SecondOrderVariable*> bounded;
  std::vector<Positions> pos;
  std::vector<SecondOrderVariable*> equals;

  ProofDatabase *m_pdb;
  NodeDatabase *m_ndb;
  HLK::SymbolTable *m_symtbl;

  NodePointer np;

  std::vector<Variable*> normal_variables;

  Variable* getNormalVariable( size_t index, USTRING type_name );

public:
  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationGetSecondOrderVariableSubstitutions( ProofDatabase *pdb, NodeDatabase *ndb, HLK::SymbolTable *symtbl ) :
    m_pdb( pdb ), m_ndb( ndb ), m_symtbl( symtbl ) {}
  virtual ~OperationGetSecondOrderVariableSubstitutions() {}

  void Clear();

  bool CheckForSubstitutions( DataNode *general, DataNode *instance );

  virtual size_t GetCount() const { return vars.size(); }

  virtual const SecondOrderVariable* GetVar( size_t index ) const;
  virtual const LambdaTerm* GetSubst( size_t index ) const;

  virtual const NodePointer& GetVarPos( size_t index ) const;
  virtual const NodePointers& GetSubstPos( size_t index ) const;

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( AtomFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SecondOrderQuantorFormula );
};

////////////////////////////////////////////////////////////////////////////////

// Generalized variant of OperationGetSecondOrderVariableSubstitutions
// OperationGetSecondOrderVariableSubstitutions finds second order
// variable substitutions, while this operation finds a set of lambda
// terms L for two formulas F, G s.t. F[X<-l]=G iff l \in L (modulo
// alpha conversion) 

class OperationGetLambdaSubstitutions :
  public FormulaCompareOperation,
  public LambdaSubstitution
{
public:
  typedef struct {
    NodePointer          var;
    NodePointers         substs;
  } Positions;

protected:
  std::vector<SecondOrderVariable*> vars;
  std::vector< std::set<LambdaTerm*> > reps;    // sets of possible
                                                // lambdaterms
  std::vector<SecondOrderVariable*> bounded;
  std::vector<Positions> pos;
  std::vector<Formula*> permutations;
  TermList blacklist;   // these variables may not appear
                        // in the computed lambdaterms
                        // to satisfy the condition of the
                        // weak quantor rule 

  std::vector<Variable*> normal_variables;

  ProofDatabase *m_pdb;
  NodeDatabase *m_ndb;
  HLK::SymbolTable *m_symtbl;

  NodePointer np;
  
  Variable* getNormalVariable( size_t index, USTRING type_name );
  
  void createPermutedFormulas( NodePointers &term_positions,
                               //NodePointer &current,
                               TermList &terms,
                               TermList &variables,
                               TermList &arguments,
                               size_t pos,
                               Formula *f );

public:
  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationGetLambdaSubstitutions( ProofDatabase *pdb, NodeDatabase *ndb, HLK::SymbolTable *symtbl, TermList &blacklist ) :
    blacklist( blacklist ), m_pdb( pdb ), m_ndb( ndb ), m_symtbl( symtbl ) {}
  virtual ~OperationGetLambdaSubstitutions() {}

  void Clear();

  bool CheckForSubstitutions( DataNode *general, DataNode *instance );

  virtual size_t GetCount() const { return vars.size(); }

  virtual const SecondOrderVariable* GetVar( size_t index ) const;
  virtual const std::set<LambdaTerm*> GetSubstSet( size_t index ) const;
  virtual const LambdaTerm* GetSubst( size_t index ) const;

  virtual const NodePointer& GetVarPos( size_t index ) const;
  virtual const NodePointers& GetSubstPos( size_t index ) const;

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( AtomFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SecondOrderQuantorFormula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationReplaceFormula :
  public FormulaOperation
{
protected:
  Formula *search;
  Formula *replacement;

  NodeDatabase *ndb;

protected:
  NodeDatabase* GetNodeDatabase() { return ndb; }
  ProofDatabase* GetProofDatabase();

public:
  OperationReplaceFormula( NodeDatabase *ndb,
                           Formula *to_search, Formula *its_replacement );

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
  DECLARE_APPLY_FUNCTION( UnaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( BinaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( FormulaVariable, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationReplaceFormulaSequence :
  public FormulaOperation
{
protected:
  const FormulaList *search;
  FormulaListElement *replacement;

  NodeDatabase *ndb;

protected:
  NodeDatabase* GetNodeDatabase() { return ndb; }
  ProofDatabase* GetProofDatabase();

public:
  OperationReplaceFormulaSequence( NodeDatabase *ndb,
                                   const FormulaList *to_search,
                                   FormulaListElement *its_replacement );

  DECLARE_APPLY_FUNCTION( FormulaList, FormulaList );
  DECLARE_APPLY_FUNCTION( SElementList, SElementList );
};

////////////////////////////////////////////////////////////////////////////////

class OperationReplaceFormulaListElementBySequence :
  public FormulaOperation
{
protected:
  FormulaListElement *search;
  const FormulaList *replacement;

  NodeDatabase *ndb;

protected:
  NodeDatabase* GetNodeDatabase() { return ndb; }
  ProofDatabase* GetProofDatabase();

public:
  OperationReplaceFormulaListElementBySequence( NodeDatabase *ndb,
                                   FormulaListElement *to_search,
                                   const FormulaList *its_replacement );

  DECLARE_APPLY_FUNCTION( SElementList, SElementList );
};

////////////////////////////////////////////////////////////////////////////////
}
#endif

