////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: formulas.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
// FORMULA IMPLEMENTATION
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"
#include "symbolhlk.h"
#include "term_operations.h"

#include <iostream>
#include <typeinfo>
#include "printhlk.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

void Formula::InsertIntoFormulaList( FormulaList *list, size_t index )
{
  list->InsertAt( index, this );
}

void Formula::InsertIntoSElementList( SElementList *list, size_t index )
{
  list->InsertAt( index, new SFormula(this) );
}

////////////////////////////////////////////////////////////////////////////////

LambdaTerm::LambdaTerm( const TermList &v, const Formula *f ) :
	m_form( const_cast<Formula*>(f) ), m_vars( v )
{
	for (unsigned int i = 0; i < v.GetChildCount(); i++)
	{
		// terms must be variables
		Variable *var = dynamic_cast<Variable*>(v.GetTerm(i));
		if ( var == NULL )
		{
		  ASSERT( false );
		}
		for (unsigned int j = i+1; j < v.GetChildCount(); j++)
		{
			// variables must be pairwise distinct
			if(v.GetTerm(i) == v.GetTerm(j))
			{
				// FIXME
				// throw Exception
				ASSERT( false );
			}	
		}
	}
}

Formula *LambdaTerm::apply( ProofDatabase* pdb, const TermList &tv ) const
{
#if 0
	if ( tv.GetChildCount() != m_vars.GetChildCount() )
	{
		// FIXME
		// throw exception
		ASSERT( false );
	}
#endif

    OperationVariableSubstitution op_sub( pdb );
#if 0    
    std::cerr << "Applying lambdaterm ";
    ToString( std::cerr );
    std::cerr << " to: ";
#endif
    // if applied to more arguments than there are variables
    // the additional arguments are ignored
    for ( unsigned int i = 0; i < m_vars.GetChildCount(); i++ )
    {
      op_sub.AddSubstitution( dynamic_cast<Variable*>(m_vars.GetTerm(i)),
                               tv.GetTerm(i) );
#if 0
      HLK::Print::PrintTerm( tv.GetTerm(i), std::cerr );
#endif
    }
    
    Formula *f = ProofLib::DataNode::ApplyTo( m_form, op_sub );
#if 0
    std::cerr << " yielding ";
    HLK::Print::PrintFormula( f, std::cerr );
    std::cerr << std::endl;
#endif
    return f;
}

bool LambdaTerm::DbEqual(const DbNode *n) const
{
  if ( const LambdaTerm *n_t = dynamic_cast<const LambdaTerm *> (n) )
    if ( m_form->DbEqual( n_t->m_form ) )
    {
      for ( unsigned int i = 0; i < m_vars.GetChildCount(); i++ )
      {
      	if ( !m_vars.GetTerm(i)->DbEqual( n_t->m_vars.GetTerm(i) ) )
      	{
      	  return false;	
      	}
      }
      return true;
    }

  return false;
}

size_t LambdaTerm::Hash() const
{
  //TODO compute hash value!
  return 17;
}

LambdaTerm* LambdaTerm::New( NodeDatabase *ndb, const TermList &v, const Formula *f )
{
  LambdaTerm *t = new LambdaTerm( v, f );
  return t->AddToNodeDatabase<LambdaTerm>( ndb );
}

std::ostream& LambdaTerm::ToString( std::ostream &out ) const
{
  m_vars.ToString(out);
  out << " . ";
  m_form->ToString(out);
  return out;
}

DataNode* LambdaTerm::Apply( Operation &op )
{
  ASSERT( false ); // NOT IMPLEMENTED
  return NULL;
} 

bool LambdaTerm::ApplyCompare( CompareOperation &op, const DataNode *other )
  const
{
  ASSERT( false ); //NOT IMPLEMENTED
  return false; 
}

////////////////////////////////////////////////////////////////////////////////

SecondOrderVariable::SecondOrderVariable( const Symbol *e ) : DbSymbolNode( e ) { }

size_t SecondOrderVariable::Hash() const
{
  return GetSymbol()->Hash();
}

SecondOrderVariable* SecondOrderVariable::New( NodeDatabase *ndb, const Symbol *e )
{
  SecondOrderVariable *v = new SecondOrderVariable( e );
  return v->AddToNodeDatabase<SecondOrderVariable>( ndb );
}

std::ostream& SecondOrderVariable::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);
  return out;
}

DataNode* SecondOrderVariable::Apply( Operation &op )
{
  ASSERT( false ); // NOT IMPLEMENTED
  return NULL;
} 

bool SecondOrderVariable::ApplyCompare( CompareOperation &op, const DataNode *other )
  const
{
  ASSERT( false ); //NOT IMPLEMENTED
  return false; 
}

////////////////////////////////////////////////////////////////////////////////

AtomFormula::AtomFormula( const Symbol *e, int arity, SecondOrderVariable* var ) :
  DbSymbolNode( e ),
  StaticTermList( arity ),
  m_var( var )
{
}

bool AtomFormula::DbEqual( const DbNode *n ) const
{
  const AtomFormula *t = dynamic_cast<const AtomFormula*>(n);

  if( t == NULL )
    return false;

  return ( DbSymbolNode::DbEqual( n ) && StaticTermList::DbEqual( n ) );
}

size_t AtomFormula::Hash() const
{
  return (StaticTermList::Hash() << 1) ^ GetSymbol()->Hash();
}

IMPLEMENT_APPLY( AtomFormula, Formula )

AtomFormula* AtomFormula::ReplaceTerms( TermList &terms ) const
{
  return New( GetNodeDatabase(), GetSymbol(), &terms );
}

AtomFormula* AtomFormula::New( NodeDatabase *ndb, const Symbol *e,
                               TermList *terms, SecondOrderVariable* var )
{
  AtomFormula *f;

  if( terms == NULL )
    f = new AtomFormula( e, 0, var );
  else
  {
    f = new AtomFormula( e, terms->GetChildCount(), var );

    f->SetTerms( terms );
  }

  return f->AddToNodeDatabase<AtomFormula>( ndb );
}

std::ostream& AtomFormula::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);

  if( !IsEmpty() )
  {
    out << "(";
    StaticTermList::ToString(out);
    out << ")";
  }

  return out;
}

SecondOrderVariable* AtomFormula::GetSecondOrderVariable() const
{
  return m_var;
}

////////////////////////////////////////////////////////////////////////////////

OperatorFormula::OperatorFormula( const OperatorSymbol *sym )
{
  ASSERT( sym != NULL );

  m_sym = sym;
}

bool OperatorFormula::DbEqual( const DbNode *n ) const
{
  const OperatorFormula *t = dynamic_cast<const OperatorFormula*>(n);

  if( t == NULL )
    return false;

  if( m_sym != t->m_sym )
    return false;

  ASSERT( GetChildCount() == t->GetChildCount() );

  for( size_t i = 0; i < GetChildCount(); i++ )
    if( GetChild(i) != t->GetChild(i) )
      return false;

  return true;
}

////////////////////////////////////////////////////////////////////////////////

UnaryConFormula::UnaryConFormula( const OperatorSymbol *sym, Formula *f )
  : OperatorFormula( sym )
{
  ASSERT( f != NULL );

  m_formula = f;
}

size_t UnaryConFormula::Hash() const
{
  return (m_formula->Hash() << 1) ^
    size_t(GetOperatorSymbol()->GetDisplayChar());
}

DataNode* UnaryConFormula::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return GetFormula();
}

IMPLEMENT_APPLY( UnaryConFormula, Formula )

UnaryConFormula* UnaryConFormula::New( NodeDatabase *ndb,
                                       const OperatorSymbol *sym,
                                       Formula *f0 )
{
  UnaryConFormula *f = new UnaryConFormula( sym, f0 );
  return f->AddToNodeDatabase<UnaryConFormula>( ndb );
}

std::ostream& UnaryConFormula::ToString( std::ostream &out ) const
{
  out << "UnaryConFormula";
  return out;
}

////////////////////////////////////////////////////////////////////////////////

QuantorFormula::QuantorFormula( const OperatorSymbol *sym,
                                Variable *v, Formula *f ) :
  OperatorFormula( sym )
{
  ASSERT( f != NULL );
  ASSERT( v != NULL );

  m_formula = f;
  m_variable = v;
}

size_t QuantorFormula::Hash() const
{
  size_t hash = m_variable->Hash();
  hash = (hash << 1) ^ m_formula->Hash();
  return (hash << 1) ^ size_t(GetOperatorSymbol()->GetDisplayChar());
}

DataNode* QuantorFormula::GetChild( size_t index ) const
{
  ASSERT( index < 2 );

  if( index == 0 )
    return GetVariable();

  return GetFormula();
}

IMPLEMENT_APPLY( QuantorFormula, Formula )

QuantorFormula* QuantorFormula::New( NodeDatabase *ndb,
                                     const OperatorSymbol *sym,
                                     Variable *v,
                                     Formula *f0 )
{
  QuantorFormula *f = new QuantorFormula( sym, v, f0 );
  return f->AddToNodeDatabase<QuantorFormula>( ndb );
}

std::ostream& QuantorFormula::ToString( std::ostream &out ) const
{
  out << "QuantorFormula";
  return out;
}

////////////////////////////////////////////////////////////////////////////////

SecondOrderQuantorFormula::SecondOrderQuantorFormula( const OperatorSymbol *sym,
                                SecondOrderVariable *v, Formula *f ) :
  OperatorFormula( sym )
{
  ASSERT( f != NULL );
  ASSERT( v != NULL );

  m_formula = f;
  m_variable = v;
}

size_t SecondOrderQuantorFormula::Hash() const
{
  // FIXME: change this!?
  size_t hash = m_variable->Hash();
  hash = (hash << 1) ^ m_formula->Hash();
  return (hash << 1) ^ size_t(GetOperatorSymbol()->GetDisplayChar());
}

DataNode* SecondOrderQuantorFormula::GetChild( size_t index ) const
{
  ASSERT( index < 2 );

  if ( index == 0 )
    return GetSecondOrderVariable();

  return GetFormula();
}

IMPLEMENT_APPLY( SecondOrderQuantorFormula, Formula )

SecondOrderQuantorFormula* SecondOrderQuantorFormula::New( NodeDatabase *ndb,
                                     const OperatorSymbol *sym,
                                     SecondOrderVariable *v,
                                     Formula *f0 )
{
  SecondOrderQuantorFormula *f = new SecondOrderQuantorFormula( sym, v, f0 );
  return f->AddToNodeDatabase<SecondOrderQuantorFormula>( ndb );
}

std::ostream& SecondOrderQuantorFormula::ToString( std::ostream &out ) const
{
  out << "SecondOrderQuantorFormula";
  return out;
}

////////////////////////////////////////////////////////////////////////////////


BinaryConFormula::BinaryConFormula( const OperatorSymbol *sym,
                                    Formula *fl, Formula *fr ) :
  OperatorFormula( sym )
{
  ASSERT( fl != NULL );
  ASSERT( fr != NULL );
  m_left = fl;
  m_right = fr;
}

size_t BinaryConFormula::Hash() const
{
  size_t hash = m_left->Hash();
  hash = (hash << 1) ^ m_right->Hash();
  return (hash << 1) ^ size_t(GetOperatorSymbol()->GetDisplayChar());
}

DataNode* BinaryConFormula::GetChild( size_t index ) const
{
  ASSERT( index < 2 );
  return ( index == 0 ? GetLeft() : GetRight() );
}

IMPLEMENT_APPLY( BinaryConFormula, Formula )

BinaryConFormula* BinaryConFormula::New( NodeDatabase *ndb,
                                         const OperatorSymbol *sym,
                                         Formula *fl,
                                         Formula *fr )
{
  BinaryConFormula *f = new BinaryConFormula( sym, fl, fr );
  return f->AddToNodeDatabase<BinaryConFormula>( ndb );
}

std::ostream& BinaryConFormula::ToString( std::ostream &out ) const
{
  out << "BinaryConFormula";
  return out;
}

////////////////////////////////////////////////////////////////////////////////

FormulaVariable::FormulaVariable( const Symbol *e ) :
  DbSymbolNode( e )
{
  m_formula = NULL;
}

size_t FormulaVariable::Hash() const
{
  return GetSymbol()->Hash();
}

void FormulaVariable::SetFormula( Formula *f )
{
  if( f == NULL )
    return;

  ASSERT( m_formula == NULL );

  m_formula = f;
  m_formula->Attach();
}

Formula* FormulaVariable::GetFormula() const
{
  return m_formula;
}

bool FormulaVariable::DbEqual( const DbNode *n ) const
{
  const FormulaVariable *t = dynamic_cast<const FormulaVariable*>(n);

  if( t == NULL )
    return false;

  return ( DbSymbolNode::DbEqual( n ) ); //&& m_formula == t->m_formula );
}

#if 0
DataNode* FormulaVariable::ApplyReplacement( const Replacement *rep )
{
  const FormulaReplacement *fr = dynamic_cast<const FormulaReplacement*>( rep );

  if( fr != NULL && fr->IsOverReplacePosition() && fr->GetSearch() == this )
    return fr->GetReplacement();

  Formula *f = ApplyReplacementTo( m_formula, rep );

  // if formula has changed, formula variable will be substituted.
  if( m_formula != f )
    return f;

  return this;
}
#endif

DataNode* FormulaVariable::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return GetFormula();
}

IMPLEMENT_APPLY( FormulaVariable, Formula )

FormulaVariable*
FormulaVariable::New( NodeDatabase *ndb, const Symbol *sym, Formula *f )
{
  ASSERT( sym != NULL );

  FormulaVariable *fv = new FormulaVariable( sym );

  FormulaVariable *fv_db = fv->AddToNodeDatabase<FormulaVariable>( ndb );

  if( fv_db != fv )
  {
    fv = NULL;

    if( f == NULL )
      return fv_db;
    
    if( fv_db->GetFormula() == NULL )
    {
      fv_db->SetFormula( f );
      return fv_db;
    }

    if( f == fv_db->GetFormula() )
      return fv_db;

    return NULL;
  }

  if( f != NULL )
    fv_db->SetFormula( f );

  return fv_db;
}

std::ostream& FormulaVariable::ToString( std::ostream &out ) const
{
  out << "FormulaVariable";
  return out;
}

////////////////////////////////////////////////////////////////////////////////

void FormulaList::AddLeft( FormulaListElement *f )
{
  ASSERT( f != NULL );

  insert( begin(), f );
}

void FormulaList::AddRight( FormulaListElement *f )
{
  ASSERT( f != NULL );

  push_back( f );
}

void FormulaList::AddLeft( FormulaList *o )
{
  if( o == NULL )
    return;

  insert( begin(), o->begin(), o->end() );
}

void FormulaList::InsertAt( size_t index, FormulaListElement *f )
{
  ASSERT( f != NULL );
  ASSERT( index <= size() );

  insert( begin() + index, f );
}

void FormulaList::InsertAt( size_t index, const FormulaList *o )
{
  ASSERT( index <= size() );

  if( o == NULL )
    return;

  insert( begin() + index, o->begin(), o->end() );
}

FormulaListElement* FormulaList::GetElement( size_t index ) const
{
  ASSERT( index < size() );
  return (*this)[index];
}

bool FormulaList::IsSubListAtIndex( const FormulaList *list,
                                    size_t index ) const
{
  ASSERT( list != NULL );
  ASSERT( list->GetChildCount() > 0 );

  size_t last_index = index + list->GetChildCount() - 1;

  if( last_index >= GetChildCount() )
    return false;

  for( size_t j = 0, i = index; i < last_index + 1; i++, j++ )
  {
    DataNode *oitem = list->GetElement(j);
    DataNode *myitem = GetElement(i);

    if( myitem != oitem )
      return false;
  }

  return true;
}

bool FormulaList::FindSubListAtIndex( const FormulaList *list,
                                      size_t &index ) const
{
  ASSERT( list != NULL );
  ASSERT( list->GetChildCount() > 0 );

  for( size_t i = index; i < GetChildCount(); i++ )
  {
    if( IsSubListAtIndex( list, i ) )
    {
      index = i;
      return true;
    }
  }

  return false;
}

/// Verify index before calling this function!
void FormulaList::ReplaceFormulas( FormulaListVariable *fv,
                                    size_t start_index )
{
  iterator i_start = begin() + start_index;
  iterator i_end = i_start + fv->GetChildCount();

  for( iterator i = i_start; i != i_end; i++ )
  {
    DbNode *dbn = dynamic_cast<DbNode*>( *i );

    ASSERT( dbn != NULL );
    dbn->Detach();
  }

  iterator i_new_pos = erase( i_start, i_end );

  insert( i_new_pos, fv );
  fv->Attach();
}

/// Verify index before calling this function!
void FormulaList::UnsubstFormulaListVar( FormulaListVariable *fv,
                                         size_t index )
{
  DbNode *dbn = dynamic_cast<DbNode*>(*( begin() + index ));
  ASSERT( dbn != NULL );
  dbn->Detach();
  iterator pos = erase( begin() + index );

  for( size_t i = 0; i < fv->GetChildCount(); i++ )
  {
    Formula *f = fv->GetChild(i)->CastTo<Formula>();

    if( f != NULL )
    {
      pos = insert( pos, f) + 1;
      f->Attach();
      continue;
    }

    FormulaListVariable *fv_ =
      fv->GetChild(i)->CastTo<FormulaListVariable>();

    if( fv_ != NULL )
    {
      pos = insert( pos, fv_ ) + 1;
      fv_->Attach();
      continue;
    }

    ASSERT( false );
  }
}

IMPLEMENT_APPLY( FormulaList, Formula )

void FormulaList::InsertIntoFormulaList( FormulaList *list, size_t index )
{
  list->InsertAt( index, this );
}

void FormulaList::InsertIntoSElementList( SElementList *list, size_t index )
{
  list->InsertAt( index, this );
}

std::ostream& FormulaList::ToString( std::ostream &out ) const
{
  out << "FormulaList";
  return out;
}


////////////////////////////////////////////////////////////////////////////////

FormulaListVariable::FormulaListVariable( const Symbol *e )
 : DbSymbolNode( e )
{
  list = NULL;
}

FormulaListVariable::~FormulaListVariable()
{
  if( list != NULL )
    delete list;
}

size_t FormulaListVariable::Hash() const
{
  return GetSymbol()->Hash();
}

bool FormulaListVariable::DbEqual( const DbNode *n ) const
{
  const FormulaListVariable *t = dynamic_cast<const FormulaListVariable*>(n);

  if( t == NULL )
    return false;

  return DbSymbolNode::DbEqual( n );
}

IMPLEMENT_APPLY( FormulaListVariable, Formula )

DataNode* FormulaListVariable::GetChild( size_t index ) const
{
  ASSERT( list != NULL && index == 0 );
  return list;
}

FormulaListVariable*
FormulaListVariable::New( NodeDatabase *ndb, const Symbol *sym,
                          FormulaList *list )
{
  ASSERT( sym != NULL );

  FormulaListVariable *fv = new FormulaListVariable( sym );

  FormulaListVariable *fv_db =
    fv->AddToNodeDatabase<FormulaListVariable>( ndb );

  if( fv_db != fv )
  {
    fv = NULL;

    if( list == NULL || list->GetChildCount() == 0 )
      return fv_db;

    if( fv_db->GetFormulaList() == NULL )
    {
      fv_db->list = new FormulaList( *list );

      for( size_t i = 0; i < list->GetChildCount(); i++ )
      {
        DbNode *c = dynamic_cast<DbNode*>( list->GetChild(i) );
        if( c != NULL )
          c->Attach();
      }

      return fv_db;
    }

    if( fv_db->GetFormulaList()->GetChildCount() != list->GetChildCount() )
      return NULL;

    for( size_t i = 0; i < list->GetChildCount(); i++ )
    {
      if( fv_db->GetFormulaList()->GetElement(i) != list->GetElement(i) )
        return NULL;
    }

    return fv_db;
  }

  if( list != NULL )
  {
    fv_db->list = new FormulaList( *list );

    for( size_t i = 0; i < list->GetChildCount(); i++ )
    {
      DbNode *c = dynamic_cast<DbNode*>( list->GetChild(i) );
      if( c != NULL )
        c->Attach();
    }
  }

  return fv_db;
}

void FormulaListVariable::InsertIntoFormulaList( FormulaList *list,
                                                 size_t index )
{
  list->InsertAt( index, this );
}

void FormulaListVariable::InsertIntoSElementList( SElementList *list,
                                                  size_t index )
{
  list->InsertAt( index, new SFormulaListVariable(this) );
}

std::ostream& FormulaListVariable::ToString( std::ostream &out ) const
{
  out << "FormulaListVariable";
  return out;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

