////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: import/importxml.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "libproof.h"
#include "util/xmlread.h"
#include "util/proofdbdtd.h"
#include "util/proofexceptions.h"
#include "symboltable.h"

#include "importxml.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

ImportXML::ImportXML( ProofLib::ProofDatabase* pdb )
  : ImportModul( pdb )
{
}

////////////////////////////////////////////////////////////////////////////////

void ImportXML::ReadVarDefinitions()
{
  ASSERT( GetXmlReader() != NULL );

  GetXmlReader()->
    AssertCurrentElement( PROOFDBDTD_NODE_VARIABLE_DEFINITIONS );

  size_t depth = GetXmlReader()->CurrentDepth();

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return;

  while(1)
  {
    char* name = GetXmlReader()->GetElementName();

    ASSERT( name != NULL );

    try
    {
      if( strcasecmp( name, PROOFDBDTD_NODE_FORMULALISTVAR_DEF ) == 0 )
        ReadFormulaListVarDef();
      else if( strcasecmp( name, PROOFDBDTD_NODE_FORMULAVAR_DEF ) == 0 )
        ReadFormulaVarDef();
      else
        GetXmlReader()->ThrowUnexpectedElement( name );

      free( name );
    }
    catch( ... )
    {
      free( name );
      throw;
    }

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );
}

////////////////////////////////////////////////////////////////////////////////

void ImportXML::ReadFormulaVarDef()
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FORMULAVAR_DEF );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return;

  Formula *f = NULL;
  (*this) >> &f;

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PREDICATE );

  ASSERT( symbol != NULL );

  FormulaVariable *fv = FormulaVariable::New( GetNodeDatabase(), symbol, f );

  if( fv == NULL )
    throw new FormulaVarExistsException( name );

  // ignore any existing child nodes
  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );
}

////////////////////////////////////////////////////////////////////////////////

void ImportXML::ReadFormulaListVarDef()
{
  GetXmlReader()->
    AssertCurrentElement( PROOFDBDTD_NODE_FORMULALISTVAR_DEF );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return;

  FormulaList list;

  (*this) >> list;

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PREDICATE );

  ASSERT( symbol != NULL );

  FormulaListVariable *fv =
    FormulaListVariable::New( GetNodeDatabase(), symbol, &list );

  if( fv == NULL )
    throw new FormulaListVarExistsException( name );

  // ignore any existing child nodes
  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Variable** v )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_VARIABLE );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_VARIABLE );

  *v = Variable::New( GetNodeDatabase(), symbol );

  // ignore any existing child nodes
  GetXmlReader()->SeekToNextElement( GetXmlReader()->CurrentDepth() );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Constant** c )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_CONSTANT );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_CONSTANT );

  *c = Constant::New( GetNodeDatabase(), symbol );

  // ignore any existing child nodes
  GetXmlReader()->SeekToNextElement( GetXmlReader()->CurrentDepth() );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Function** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FUNCTION );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  TermList l;
  
  (*this) >> l;

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, l.GetChildCount(), SYMBOL_TYPE_FUNCTION );

  *f = Function::New( GetNodeDatabase(), symbol, &l );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Term** t )
{
  char* name = GetXmlReader()->GetElementName();

  ASSERT( name != NULL );

  *t = NULL;

  try
  {
    if( strcasecmp( name, PROOFDBDTD_NODE_VARIABLE ) == 0 )
      (*this) >> (Variable**) t;
    if( strcasecmp( name, PROOFDBDTD_NODE_CONSTANT ) == 0 )
      (*this) >> (Constant**) t;
    if( strcasecmp( name, PROOFDBDTD_NODE_FUNCTION ) == 0 )
      (*this) >> (Function**) t;

    if( *t == NULL )
      GetXmlReader()->ThrowUnexpectedElement( name );
  }
  catch( ... )
  {
    free( name);
    throw;
  }

  free( name);

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( TermList& l )
{
  size_t depth = GetXmlReader()->CurrentDepth();

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return *this;

  while(1)
  {
    Term *t;
    (*this) >> &t;

    l.AddRight( t );

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( AtomFormula** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_CONSTANTATOMFORMULA );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  TermList l;
  
  (*this) >> l;

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, l.GetChildCount(),
                                   SYMBOL_TYPE_PREDICATE );

  *f = AtomFormula::New( GetNodeDatabase(), symbol, &l );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( QuantorFormula** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_QUANTIFIEDFORMULA );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING type = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_TYPE );

  const OperatorSymbol *opsym =
    dynamic_cast<const OperatorSymbol*>(
    GetSymbolTable()->GetOrAdd( type, 1, SYMBOL_TYPE_JUNCTOR ));

  ASSERT( opsym != NULL ); // TODO: auto add or error?

  Variable *v;
  Formula *c;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    GetXmlReader()->ThrowMissingChildException( PROOFDBDTD_NODE_VARIABLE );

  (*this) >> &v;

  if( GetXmlReader()->CurrentDepth() <= depth )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &c;

  *f = QuantorFormula::New( GetNodeDatabase(), opsym, v, c );

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( UnaryConFormula** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_CONJUNCTIVEFORMULA );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING type = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_TYPE );

  const OperatorSymbol *opsym =
    dynamic_cast<const OperatorSymbol*>(
    GetSymbolTable()->GetOrAdd( type, 1, SYMBOL_TYPE_JUNCTOR ));

  ASSERT( opsym != NULL ); // TODO: auto add or error?

  Formula *c;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &c;

  *f = UnaryConFormula::New( GetNodeDatabase(), opsym, c );
  
  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( BinaryConFormula** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_CONJUNCTIVEFORMULA );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING type = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_TYPE );

  const OperatorSymbol *opsym =
    dynamic_cast<const OperatorSymbol*>(
    GetSymbolTable()->GetOrAdd( type, 2, SYMBOL_TYPE_JUNCTOR ));

  ASSERT( opsym != NULL ); // TODO: auto add or error?

  Formula *l, *r;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &l;

  if( GetXmlReader()->CurrentDepth() <= depth )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &r;

  *f = BinaryConFormula::New( GetNodeDatabase(), opsym, l ,r );

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( FormulaVariable** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FORMULAVAR );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PREDICATE );

  ASSERT( symbol != NULL );

  *f = FormulaVariable::New( GetNodeDatabase(), symbol, NULL );

  if( *f == NULL )
    throw new FormulaVarExistsException( name );

  // ignore any existing child nodes
  GetXmlReader()->SeekToNextElement( GetXmlReader()->CurrentDepth() );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Formula** f )
{
  char* name = GetXmlReader()->GetElementName();

  ASSERT( name != NULL );

  *f = NULL;

  try
  {
    if( strcasecmp( name, PROOFDBDTD_NODE_CONSTANTATOMFORMULA ) == 0 )
      (*this) >> (AtomFormula**) f;
    else if( strcasecmp( name, PROOFDBDTD_NODE_QUANTIFIEDFORMULA ) == 0 )
      (*this) >> (QuantorFormula**) f;
    else if( strcasecmp( name, PROOFDBDTD_NODE_CONJUNCTIVEFORMULA ) == 0 )
    {
      USTRING type = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_TYPE );
  
      size_t arity = GetSymbolTable()->GetArity( type, SYMBOL_TYPE_JUNCTOR );
  
      if( arity == 1 )
        (*this) >> (UnaryConFormula**) f;
      else if( arity == 2 )
        (*this) >> (BinaryConFormula**) f;
      else
        ASSERT( false );
    }
    else if( strcasecmp( name, PROOFDBDTD_NODE_FORMULAVAR ) == 0 )
      (*this) >> (FormulaVariable**) f;
  
    if( *f == NULL )
      GetXmlReader()->ThrowUnexpectedElement( name );
  }
  catch( ... )
  {
    free( name );
    throw;
  }

  free( name );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( FormulaListElement** e )
{
  char* name = GetXmlReader()->GetElementName();

  ASSERT( name != NULL );

  *e = NULL;

  try
  {
    if( strcasecmp( name, PROOFDBDTD_NODE_CONSTANTATOMFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_QUANTIFIEDFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_CONJUNCTIVEFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_FORMULAVAR ) == 0 )
    {
      Formula *f = NULL;
      (*this) >> &f;
      *e = f;
    }
    else if( strcasecmp( name, PROOFDBDTD_NODE_FORMULALISTVAR ) == 0 )
    {
      FormulaListVariable *f = NULL;
      (*this) >> &f;
      *e = f;
    }
  
    if( *e == NULL )
      GetXmlReader()->ThrowUnexpectedElement( name );
  }
  catch( ... )
  {
    free( name );
    throw;
  }

  free( name );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( FormulaList& l )
{
#if 0
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FORMULALIST );

  size_t depth = GetXmlReader()->CurrentDepth();

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return *this;
#endif

  size_t depth = GetXmlReader()->CurrentDepth() - 1;

  while(1)
  {
    FormulaListElement *e = NULL;
    (*this) >> &e;

    l.AddRight( e );

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( FormulaListVariable** f )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FORMULALISTVAR );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PREDICATE );

  ASSERT( symbol != NULL );

  *f = FormulaListVariable::New( GetNodeDatabase(), symbol, NULL );

  if( *f == NULL )
    throw new FormulaListVarExistsException( name );

  // ignore any existing child nodes
  GetXmlReader()->SeekToNextElement( GetXmlReader()->CurrentDepth() );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( SElement** e )
{
  char* name = GetXmlReader()->GetElementName();

  ASSERT( name != NULL );

  *e = NULL;

  try
  {
    if( strcasecmp( name, PROOFDBDTD_NODE_CONSTANTATOMFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_QUANTIFIEDFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_CONJUNCTIVEFORMULA ) == 0 ||
        strcasecmp( name, PROOFDBDTD_NODE_FORMULAVAR ) == 0 )
    {
      Formula *f = NULL;
      (*this) >> &f;
      *e = new SFormula( f );
    }
    else if( strcasecmp( name, PROOFDBDTD_NODE_FORMULALISTVAR ) == 0 )
    {
      FormulaListVariable *f = NULL;
      (*this) >> &f;
      *e = new SFormulaListVariable( f );
    }
  
    if( *e == NULL )
      GetXmlReader()->ThrowUnexpectedElement( name );
  }
  catch( ... )
  {
    free( name );
    throw;
  }

  free( name );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( SElementList** l )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_FORMULALIST );

  size_t depth = GetXmlReader()->CurrentDepth();

  *l = new SElementList;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return *this;

  while(1)
  {
    SElement *e = NULL;
    (*this) >> &e;

    (*l)->AddRight( e );

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Sequent** s )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_SEQUENT );

  size_t depth = GetXmlReader()->CurrentDepth();

  SElementList *l = NULL;
  SElementList *r = NULL;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &l;

  if( GetXmlReader()->CurrentDepth() <= depth )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &r;

  *s = new Sequent( l, r );

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Rule** r_ )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_RULE );

  size_t depth = GetXmlReader()->CurrentDepth();

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL, true );

  Rule *r = new Rule;
  *r_ = r;

  if( !name.empty() )
    r->SetName( name );

  Sequent *s;

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    GetXmlReader()->ThrowMissingChildException( NULL );

  (*this) >> &s;
  ASSERT( s != NULL );
  r->SetConclusion( s );

  while(1)
  {
    RuleItem *ri = NULL;

    (*this) >> &ri;
    ASSERT( ri != NULL );


    r->AddRight( ri );

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( ProofLink** pl )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_PROOFLINK );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PROOF );

  Proof *p = GetProofDatabase()->NewProof( symbol );

  if( p == NULL )
    p = GetProofDatabase()->GetProof( symbol );

  ASSERT( p != NULL );

  *pl = p->NewProofLink( GetCurrentProof() );

  // ignore any existing child nodes
  GetXmlReader()->SeekToNextElement( GetXmlReader()->CurrentDepth() );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( RuleItem** ruleitem )
{
  char* name = GetXmlReader()->GetElementName();

  ASSERT( name != NULL );

  *ruleitem = NULL;

  try
  {
    if( strcasecmp( name, PROOFDBDTD_NODE_SEQUENT ) == 0 )
      (*this) >> (Sequent**) ruleitem;
    if( strcasecmp( name, PROOFDBDTD_NODE_RULE ) == 0 )
      (*this) >> (Rule**) ruleitem;
    if( strcasecmp( name, PROOFDBDTD_NODE_PROOFLINK ) == 0 )
      (*this) >> (ProofLink**) ruleitem;

    if( *ruleitem == NULL )
      GetXmlReader()->ThrowUnexpectedElement( name );
  }
  catch( ... )
  {
    free( name );
    throw;
  }

  free( name );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( Proof** p )
{
  *p = NULL;

  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_PROOF );

  USTRING name = GetXmlReader()->GetAttribute( PROOFDBDTD_ATTR_SYMBOL );

  size_t depth = GetXmlReader()->CurrentDepth();

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return *this;

#if 0
cerr << "ProofName: " << name << endl;
#endif
  const Symbol *symbol =
    GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_PROOF );

  *p = GetProofDatabase()->NewProof( symbol );

  if( *p == NULL )
  {
    *p = GetProofDatabase()->GetProof( symbol );

    ASSERT( (*p)->GetEndRule() == NULL );
#if 0
    if( (*p)->GetEndRule() != NULL )
      throw new ProofExistsException( name );
#endif
  }

  SetCurrentProof( *p );

  RuleItem *ri = NULL;

  try
  {
    (*this) >> &ri;

    (*p)->SetEndRule( ri );
  }
  catch( ... )
  {
    delete *p;
    *p = NULL;
    throw;
  }

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

ImportXML& ImportXML::operator >> ( ProofDatabase& pdb )
{
  GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_PROOFDATABASE );

  size_t depth = GetXmlReader()->CurrentDepth();

  if( !GetXmlReader()->SeekToChildElement( depth ) )
    return *this;

  while(1)
  {
    char* name = GetXmlReader()->GetElementName();

    ASSERT( name != NULL );

    try
    {
      if( strcasecmp( name, PROOFDBDTD_NODE_DEFINITIONLIST ) == 0 )
      {
        GetXmlReader()->AssertCurrentElement( PROOFDBDTD_NODE_DEFINITIONLIST );

        size_t depth = GetXmlReader()->CurrentDepth();
        GetXmlReader()->SeekToNextElement( depth );
      }
      else if( strcasecmp( name, PROOFDBDTD_NODE_PROOF ) == 0 )
      {
        Proof *p = NULL;

        (*this) >> &p;
      }
      else if( strcasecmp( name, PROOFDBDTD_NODE_VARIABLE_DEFINITIONS ) == 0 )
        ReadVarDefinitions();
      else
        GetXmlReader()->ThrowUnexpectedElement( name );

      free( name );
    }
    catch( ... )
    {
      free( name );
      throw;
    }

    // next node is another children?
    if( GetXmlReader()->CurrentDepth() <= depth )
      break;
  }

  if( GetXmlReader()->CurrentDepth() > depth )
    GetXmlReader()->SeekToNextElement( depth );

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

}
