////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: term_operations.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _TERM_OPERATIONS_H_
#define _TERM_OPERATIONS_H_

#include <vector>

#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"

namespace ProofLib
{

  ////////////////////////////////////////////////////////////////////////////////

class TOperationReplaceTerm : public TermOperation
{
protected:
  NodeDatabase *m_ndb;
  
  Term &replacement;
  const NodePointer &position;
  NodePointer np;

public:
  NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  TOperationReplaceTerm( NodeDatabase *ndb, Term &replacement, const NodePointer &position, const NodePointer &start ) :
    m_ndb( ndb ), replacement( replacement ), position( position ), np( start ) {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
};

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

class OperationGetVariables :
  public TermOperation
{
protected:
  TermList &list;

public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  OperationGetVariables( TermList &list ) : list( list ) {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

/* Gets the direct subterms of a term and nodpointers to them.
 * If the term is a constant or a variable, the empty list is returned.
 * If the term is a function, the list contains the arguments of the function
 * and pointers to them.
 */

class OperationGetSubterms : public TermOperation
{
protected:
  TermList &list;
  NodePointers &nps;
  NodePointer np;
public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationGetSubterms( TermList &list, NodePointers &nps, NodePointer start ) :
    list( list ), nps( nps ), np( start ) {}

  DECLARE_APPLY_FUNCTION( Function, Term );
};

////////////////////////////////////////////////////////////////////////////////

#if 0
class OperationGetFirstAndSecondOrderVariables :
  public TermOperation
{
protected:
  DefinitionArgumentList &list;

public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  OperationGetFirstAndSecondOrderVariables( DefinitionArgumentList &list ) : list( list ) {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
};
#endif 

////////////////////////////////////////////////////////////////////////////////

class OperationCountFreeVariables :
  public TermOperation
{
protected:
  NodePointer np;
  TermList bounded;
  int varcount;

public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

  int GetCount() const { return varcount; }

public:
  OperationCountFreeVariables( TermList &list ) : varcount( 0 ) {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};


////////////////////////////////////////////////////////////////////////////////

class OperationGetFreeVariablesAndPositions :
  public TermOperation
{
public:
  typedef std::vector<NodePointers> PositionList;

protected:
  TermList &list;
  PositionList pos;
  NodePointer np;

  TermList bounded;

public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

  const NodePointers& GetPositions( size_t index ) const;

public:
  OperationGetFreeVariablesAndPositions( TermList &list ) : list( list ) {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};


////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

class OperationGetFreeVariables :
  public OperationGetVariables
{
public:
  OperationGetFreeVariables( TermList &list )
    : OperationGetVariables( list ) {}

  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////
#if 0
class OperationGetFreeFirstAndSecondOrderVariables :
  public OperationGetFirstAndSecondOrderVariables
{
public:
  OperationGetFreeFirstAndSecondOrderVariables( DefinitionArgumentList &list )
    : OperationGetFirstAndSecondOrderVariables( list ) {}

  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
};
#endif
////////////////////////////////////////////////////////////////////////////////

class OperationGetBoundedVariables :
  public TermOperation
{
protected:
  TermList &list;

public:
  NodeDatabase* GetNodeDatabase() { ASSERT(false); return NULL; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  OperationGetBoundedVariables( TermList &list ) : list( list ) {}

  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class VariableSubstitution
{
public:
  virtual ~VariableSubstitution() {}

  virtual size_t GetCount() const = 0;

  virtual const Variable* GetVar( size_t index ) const = 0;
  virtual const Term* GetSubst( size_t index ) const = 0;

  const Term* GetSubstFor( const Variable *v ) const;
};

////////////////////////////////////////////////////////////////////////////////

class OperationGetVariableSubstitutions :
  public TermCompareOperation,
  public VariableSubstitution
{
public:
  typedef struct {
    NodePointer          var;
    NodePointers         substs;
  } Positions;

protected:
  TermList vars;
  TermList reps;
  TermList bounded;
  TermList equals;
  std::vector<Positions> pos;
  bool equals_too;

  NodePointer np;
public:
  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

public:
  OperationGetVariableSubstitutions( bool equals_too = false )
    { this->equals_too = equals_too; }
  virtual ~OperationGetVariableSubstitutions() {}

  void Clear();

  bool CheckForSubstitutions( DataNode *general, DataNode *instance );

  virtual size_t GetCount() const { return vars.GetChildCount(); }

  virtual const Variable* GetVar( size_t index ) const;
  virtual const Term* GetSubst( size_t index ) const;

  virtual const NodePointer& GetVarPos( size_t index ) const;
  virtual const NodePointers& GetSubstPos( size_t index ) const;

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( QuantorFormula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationVariableSubstitution :
  public TermOperation,
  public VariableSubstitution
{
  ProofDatabase *m_pdb;
  NodeDatabase  *m_ndb;

  typedef std::pair< const Variable*, const Term* > Substitution;

  std::vector< Substitution > substitutions;

  TermList bounded;

  NodePointer m_np;
  typedef std::vector<NodePointers> PositionList;
  PositionList m_positions;

public:
  OperationVariableSubstitution( ProofDatabase *pdb,
                                 const Variable *v, const Term *subst );
  OperationVariableSubstitution( NodeDatabase *ndb,
                                 const Variable *v, const Term *subst );
  OperationVariableSubstitution( NodeDatabase *ndb,
                                 const VariableSubstitution &subst );
  OperationVariableSubstitution( NodeDatabase *ndb );
  OperationVariableSubstitution( ProofDatabase *pdb );
  virtual ~OperationVariableSubstitution() {}

  void AddSubstitution( const Variable *v, const Term *subst );

  NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofDatabase* GetProofDatabase() { return m_pdb; }

  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }

  virtual size_t GetCount() const { return substitutions.size(); }

  virtual const Variable* GetVar( size_t index ) const;
  virtual const Term* GetSubst( size_t index ) const;

  const NodePointers& GetPositions( size_t index ) const;
  const PositionList& GetPositions() const { return m_positions; }

  virtual std::ostream& ToString( std::ostream &out ) const;

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

class OperationTermSubstitution :
  public TermOperation
{
private:
//  ProofDatabase *m_pdb;
  NodeDatabase  *m_ndb;

  const Term *t;
  const Term *subst;

  NodePointers m_positions;

  const NodePointers *m_positions_allowed;

  NodePointer m_np;

public:
//  OperationTermSubstitution( ProofDatabase *pdb,
  OperationTermSubstitution( NodeDatabase *ndb,
                             const Term *t, const Term *subst,
                             const NodePointers* positions_allowed = NULL );
  virtual ~OperationTermSubstitution() {}

  NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index );
  virtual size_t Pop();

  const NodePointers& GetPositions() const { return m_positions; }

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
};

////////////////////////////////////////////////////////////////////////////////

class OperationMGUSimple :
  public TermCompareOperation,
  public VariableSubstitution
{
  typedef std::pair< const Variable*, const Term* > Unification;
  typedef std::vector< Unification > UnificationTable;
  typedef enum { Left, Right } Side;

  UnificationTable m_table;

  NodeDatabase *m_ndb;

public:
  OperationMGUSimple( NodeDatabase *ndb ) { m_ndb = ndb; }

  NodeDatabase* GetNodeDatabase() { return m_ndb; }

  bool Unify( const DataNode *left, const DataNode *right );

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

  virtual std::ostream& ToString( std::ostream &out ) const;

  virtual size_t GetCount() const { return m_table.size(); }

  virtual const Variable* GetVar( size_t index ) const;
  virtual const Term* GetSubst( size_t index ) const;

protected:
  bool AddSubstition( Side s, const Variable *v, const Term *t );

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );
};

////////////////////////////////////////////////////////////////////////////////

class FreshVariableGenerator
{
public:
  virtual ~FreshVariableGenerator() {}

  virtual Variable* GetVariableNum( size_t ) = 0;
  virtual NodeDatabase* GetNodeDatabase() = 0;
};

////////////////////////////////////////////////////////////////////////////////

class FreshVarMap : public VariableSubstitution
{
  typedef std::pair< Variable*, Variable* > VarMap;
  typedef std::vector< VarMap > VarMapping;

  VarMapping m_map;

  FreshVariableGenerator *m_fg;
  size_t m_next_var_num;

public:
  FreshVarMap( FreshVariableGenerator *fg, const DataNode *node,
               size_t start = 0 );
  FreshVarMap( const FreshVarMap &o ) { (*this) = o; }
  virtual ~FreshVarMap();

  void operator=( const FreshVarMap &o );

  void AddVarsFromNode( const DataNode *node );

  size_t GetNextVarNum() const { return m_next_var_num; }

  virtual size_t GetCount() const { return m_map.size(); }
  virtual const Variable* GetVar( size_t index ) const;
  virtual const Variable* GetSubst( size_t index ) const;
};

////////////////////////////////////////////////////////////////////////////////

class OperationMGU :
  public TermCompareOperation
{
public:
  typedef enum { Left, Right } Side;
private:
  typedef std::pair< Variable*, Term* > Unification;

  class UnificationTable : private std::vector< Unification >,
                           public VariableSubstitution
  {
  public:
    UnificationTable() {}
    UnificationTable( const UnificationTable &o ) { (*this) = o; }
    ~UnificationTable();
    void operator=( const UnificationTable &o );

    void AddSubstitution( const Variable* v, const Term* t );
    void ApplyOperation( TermOperation &op );
    virtual size_t GetCount() const { return size(); }
    virtual const Variable* GetVar( size_t index ) const;
    virtual const Term* GetSubst( size_t index ) const;
    virtual std::ostream& ToString( std::ostream &out ) const;
  };

  UnificationTable m_table;

#if 0
  typedef std::pair< const Variable*, const Variable* > VarMap;
  typedef std::vector< VarMap > VarMapping;

  VarMapping m_map_l;
  VarMapping m_map_r;
#endif
  const FreshVarMap *m_map_l;
  const FreshVarMap *m_map_r;

  size_t m_new_var_count;

  FreshVariableGenerator *m_fg;

public:
  OperationMGU( FreshVariableGenerator *fg,
                const FreshVarMap *m1, const FreshVarMap *m2 );
/*
  OperationMGU( const OperationMGU &o ) { (*this) = o; }
  void operator=( const OperationMGU &o );
*/

  NodeDatabase* GetNodeDatabase() { return m_fg->GetNodeDatabase(); }

  bool Unify( const DataNode *left, const DataNode *right );

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

  virtual std::ostream& ToString( std::ostream &out ) const;

  void GetSubstitution( Side s, OperationVariableSubstitution &op ) const;

  virtual void Finalize();

protected:
#if 0
  virtual size_t GetCount() const { return m_table.GetCount(); }

  virtual const Variable* GetVar( size_t index ) const;
  virtual const Term* GetSubst( size_t index ) const;
#endif

  const Variable* MapVar( Side s, const Variable *v );
  bool AddSubstition( Side s, const Variable *v, const Term *t );
  bool AddSubstition( const Variable *v, const Term *t );

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );
};

////////////////////////////////////////////////////////////////////////////////

class Substitution
{
  Variable *var;
  Term     *by;

public:
  Substitution();
  Substitution( const Variable *v, const Term *r );
  Substitution( const Substitution &o );
  ~Substitution();

  void operator=( const Substitution &o );

  const Variable* GetVar() const { return var; }
  const Term* GetSubst() const { return by; }

  std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class Substitutions : public std::vector<Substitution>
{
public:
  const Term* GetReplacementFor( const Variable *var ) const;

  bool CheckSubstitution( const Variable *var, const Term *rep );

  /// subset of s?
  bool operator < ( const Substitutions &s ) const;

  std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class SubstPossibilities : public std::vector<Substitutions>
{
public:
  void Add( const Substitutions &s );
  bool Merge( const SubstPossibilities &o );

  std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

}
#endif
