////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: terms.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// TERMS IMPLEMENTATION
////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
#include <iostream>
#include <set>
#include <stack>
#include "util/hash.h"
#endif

#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"

#include <typeinfo>

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

std::ostream& TermListBase::ToString( std::ostream &out ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    if( i != 0 ) out << ",";

    GetTerm(i)->ToString(out);
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

TermList::TermList( size_t arity )
{
}

TermList::TermList( const StaticTermList& stl )
{
  for( size_t i = 0; i < stl.GetChildCount(); i++ )
    AddRight( stl.GetTerm(i) );
}

void TermList::AddLeft( Term *t )
{
  t->Attach();
  insert( begin(), t );
}

void TermList::AddRight( Term *t )
{
  t->Attach();
  push_back( t );
}

size_t TermList::GetChildCount() const
{
  return size();
}

Term* TermList::GetTerm( size_t index ) const
{
  ASSERT( index < size() );

  return (*this)[index];
}

void TermList::SetTerms( TermList *other )
{
  clear();

  insert( begin(), other->begin(), other->end() );

  other->clear();
}

int TermList::FindTerm( const Term *t ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
    if( GetTerm(i) == t )
      return i;

  return -1;
}

void TermList::RemoveAt( size_t index )
{
  ASSERT( index < size() );

  Term *t = *(begin() + index);

  erase( begin() + index );

  t->Detach();
}

////////////////////////////////////////////////////////////////////////////////

StaticTermList::StaticTermList( size_t arity )
{
  typedef Term* TermPtr;

  m_arity = arity;

  if( m_arity == 0 )
    m_terms = NULL;
  else
    m_terms = new TermPtr[arity];
}

StaticTermList::StaticTermList( const TermListBase *terms )
{
  typedef Term* TermPtr;

  m_arity = 0;

  if( terms != NULL )
    m_arity = terms->GetChildCount();

  if( m_arity == 0 )
    m_terms = NULL;
  else
  {
    m_terms = new TermPtr[m_arity];

    SetTerms( terms );
  }
}

StaticTermList::~StaticTermList()
{
  if( m_terms != NULL )
    delete m_terms;
}

size_t StaticTermList::Hash() const
{
  size_t hash = 0;

  for( size_t i = 0; i < m_arity; i++ )
  {
    //size_t h = m_terms[i]->Hash();
    size_t h = size_t(m_terms[i]); // >> 8;
    //h = h & 0xffff;

    //hash = (hash << 3) ^ h;
    if( i % 2 )
      hash = (hash << 1) ^ h;
    else
      hash = (hash << 1) ^ (h << 16);
  }

  return hash;
}

void StaticTermList::SetTerms( const TermListBase *tl )
{
  ASSERT( tl->GetChildCount() == GetChildCount() );

  for( size_t i = 0; i < m_arity; i++ )
  {
    m_terms[i] = tl->GetTerm(i);
    m_terms[i]->Attach();
  }
}

Term* StaticTermList::GetTerm( size_t index ) const
{
  ASSERT( index < m_arity );

  ASSERT( m_terms != NULL );

  return m_terms[index];
}

int StaticTermList::FindTerm( const Term *t ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
    if( GetTerm(i) == t )
      return i;

  return -1;
}

bool StaticTermList::DbEqual( const DbNode *n ) const
{
  const StaticTermList *t = dynamic_cast<const StaticTermList*>(n);

  if( t == NULL )
    return false;

  if( m_arity != t->m_arity )
    return false;

  for( size_t i = 0; i < m_arity; i++ )
  {
    if( m_terms[i] != t->m_terms[i] )
      return false;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////

Variable::Variable( const Symbol *e ) : DbSymbolNode( e ) { }

size_t Variable::Hash() const
{
  return GetSymbol()->Hash();
}

IMPLEMENT_APPLY( Variable, Term )

Variable* Variable::New( NodeDatabase *ndb, const Symbol *e )
{
  Variable *v = new Variable( e );
  return v->AddToNodeDatabase<Variable>( ndb );
}

std::ostream& Variable::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Constant::Constant( const Symbol *e ) : DbSymbolNode( e ) { }

size_t Constant::Hash() const
{
  return GetSymbol()->Hash();
}

IMPLEMENT_APPLY( Constant, Term )

Constant* Constant::New( NodeDatabase *ndb, const Symbol *e )
{
  Constant *c = new Constant( e );
  return c->AddToNodeDatabase<Constant>( ndb );
}

std::ostream& Constant::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Function::Function( const Symbol *e, int arity ) :
  DbSymbolNode( e ),
  StaticTermList( arity )
{
}

bool Function::DbEqual( const DbNode *n ) const
{
  const Function *t = dynamic_cast<const Function*>(n);

  if( t == NULL )
    return false;

  return ( DbSymbolNode::DbEqual( n ) && StaticTermList::DbEqual( n ) );
}

size_t Function::Hash() const
{
  return (StaticTermList::Hash() << 1) ^ GetSymbol()->Hash();
}

IMPLEMENT_APPLY( Function, Term )

Function* Function::ReplaceTerms( TermList &terms ) const
{
  return New( GetNodeDatabase(), GetSymbol(), &terms );
}

Function* Function::New( NodeDatabase *ndb, const Symbol *e,
                         TermList *terms )
{
  ASSERT( terms != NULL );

  Function *f = new Function( e, terms->GetChildCount() );

  f->SetTerms( terms );

  return f->AddToNodeDatabase<Function>( ndb );
}

bool Function::HaveEqualSymbol( const Function *o ) const
{
  return GetSymbol() == o->GetSymbol();
}

std::ostream& Function::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);

  if( !IsEmpty() )
  {
    out << "(";
    StaticTermList::ToString(out);
    out << ")";
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

