////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: main.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <getopt.h>

#include "libproof/libproof.h"
#include "libproof/util/exceptions.h"
#include "libproof/util/myassert.h"
#include "libproof/chars/latexchars.h"
#include "libproof/export/savedbxml.h"
#include "libproof/export/exportflatproof.h"
#include "libproof/export/exportlatex.h"

#include "config.h"
#include "readhlk.h"
#include "typehlk.h"
#include "compilehlk.h"
#include "transhlk.h"
#include "printhlk.h"
#include "ceresexport.h"

////////////////////////////////////////////////////////////////////////////////

bool do_transform_to_regular_lk = true;
typedef enum { proofdb, flat, ceres, latex, flat_latex } OutputType;
OutputType output_type = proofdb;
const char *filename;
const char *output_filename = "/dev/stdout";
USTRING proofname;
const char *programname;
std::ofstream *axioms_outfile = NULL;

////////////////////////////////////////////////////////////////////////////////

void Usage()
{
  std::cerr << "Usage: " << programname << " [options] <filename> <proof-name>"
            << std::endl << std::endl;

  std::cerr << "Options:" << std::endl;
  std::cerr << "  --version          print the version of hlk." << std::endl;
  std::cerr << "  -o <filename>      output file (default /dev/stdout)."
            << std::endl;
  std::cerr << "  --flat             output flat proof." << std::endl;
  std::cerr << "  --ceres            output ceres input." << std::endl;
  std::cerr << "  --slk              skeleton lk output." << std::endl;
  std::cerr << "  --latex            output proof(s) as LaTeX source."
            << std::endl;
//  std::cerr << "  --utf8             utf8 output on console." << std::endl;
  std::cerr << "  --sim-undef        simulate undef predicate rules."
            << std::endl;
  std::cerr << "  --axioms[=<file>]  will output the initial sequents to file,"
            << std::endl
            << "                     default ist /dev/stderr."
            << std::endl;
  exit(1);
}

void Version()
{
  std::cerr << "HLK version " << VERSION
            << std::endl << std::endl;

  exit(0);
}

////////////////////////////////////////////////////////////////////////////////

void check_options( int argc, char *argv[] )
{
  //int digit_optind = 0;

  while (1)
  {
//    int this_option_optind = optind ? optind : 1;
    int option_index = 0;

    static struct option long_options[] =
    {
#if 0
      {"add", 1, 0, 0},
      {"append", 0, 0, 0},
      {"delete", 1, 0, 0},
      {"verbose", 0, 0, 0},
      {"create", 1, 0, 'c'},
      {"file", 1, 0, 0},
#endif
      { "slk", no_argument, 0, 0 },
      { "utf8", no_argument, 0, 0 },
      { "flat", no_argument, 0, 0 },
      { "sim-undef", no_argument, 0, 0 },
      { "axioms", optional_argument, 0, 0 },
      { "ceres", no_argument, 0, 0 },
      { "version", no_argument, 0, 0 },
      { "latex", no_argument, 0, 0 },
      { 0, 0, 0, 0 }
    };

    // "abc:d:012"
    int c = getopt_long( argc, argv, "o:", long_options, &option_index );

    if( c == -1 )
      break;

    switch(c)
    {
    case 0:
      switch( option_index )
      {
      case 0: // slk
        do_transform_to_regular_lk = false;
        break;
      case 1: // utf-8
        HLK::Print::use_utf8 = true;
        break;
      case 2: // flat
        if( output_type == latex )
          output_type = flat_latex;
        else
          output_type = flat;
        break;
      case 3: // simulate pred undefs
        HLK::Compiler::simulate_pred_undefs = true;
        break;
      case 4: // output of axioms
        if( optarg != NULL )
        {
          axioms_outfile = new std::ofstream( optarg );

          if( !(*axioms_outfile) )
          {
            std::cerr << "Can not open file '" << optarg << "' for writting."
                      << std::endl;
            exit( 1 );
          }

          HLK::Transformer::axiom_writer = axioms_outfile;
        }
        else
          HLK::Transformer::axiom_writer = &std::cerr;
        break;
      case 5: // ceres
        if( output_type == flat )
        {
          std::cerr << "Choose either 'flat' or 'ceres' output, not both."
                    << std::endl;
          exit( 1 );
        }
        output_type = ceres;
        break;
      case 6: // version
        Version();
        break;
      case 7: // latex
        if( output_type == flat )
          output_type = flat_latex;
        else
          output_type = latex;
        break;

      default:
        printf ("Option %s", long_options[option_index].name);
        if (optarg)
          printf (" with arg %s", optarg);
        printf ("\n");
      }
      break;

    case 'o':
      output_filename = optarg;
      break;

#if 0
    case '0':
    case '1':
    case '2':
      if (digit_optind != 0 && digit_optind != this_option_optind)
        printf ("digits occur in two different argv-elements.\n");
      digit_optind = this_option_optind;
      printf ("Option %c\n", c);
      break;

    case 'a':
      printf ("Option a\n");
      break;

    case 'b':
      printf ("Option b\n");
      break;

    case 'd':
      printf ("Option d mit Wert `%s'\n", optarg);
      break;
#endif

    case '?':
      Usage();

    default:
      ASSERT( false );
    }
  }

  if( argc - optind != 2 )
    Usage();

  filename = argv[optind++];
  proofname.FromUtf8( argv[optind++] );
  LatexConvertor.KnownLatexSymbolsToUnicode( proofname );
}

////////////////////////////////////////////////////////////////////////////////

int main( int argc, char *argv[] )
{
  programname = argv[0];

  check_options( argc, argv );

  HLK::Reader *reader = HLK::Reader::New();

  HLK::Document *doc = new HLK::Document();

#ifdef DEBUG
  std::cerr << "START PARSING INPUT ..." << std::endl;
#endif

  try
  {
    reader->Open( filename );

    reader->Read( doc );
  }
  catch( Exception* e )
  {
    e->PrintAll( std::cerr );
    std::cerr << std::endl;
    delete e;
  
    reader->Close();

    // delete doc;
  
    return 1;
  }
  catch( ... )
  {
    std::cerr << "CAUGHT UNKNOW EXCEPTION IN MAIN!" << std::endl;
  
    reader->Close();

    // delete doc;
  
    return 1;
  }

  reader->Close();

#ifdef DEBUG
  std::cerr << "PARSING INPUT SUCCESSFULLY FINISHED." << std::endl;

  std::cerr << "BEGIN TRANSLATION ..." << std::endl;
#endif

  HLK::Compiler *comp = HLK::Compiler::NewCompiler();

  ProofLib::ProofDatabase *pdb = NULL;

  try
  {
    pdb = comp->GenerateLK( doc, proofname );

    if( do_transform_to_regular_lk )
    {
      HLK::Transformer *trans =
        HLK::Transformer::NewTransformer( doc->GetAxiomTable() );

      trans->CompleteProof( pdb->GetDefaultProof(), pdb );
    }
  }
  catch( Exception* e )
  {
    e->PrintAll( std::cerr );
    std::cerr << std::endl;
    delete e;
  
    // delete doc;
    delete comp;
  
    return 1;
  }
  catch( ... )
  {
    std::cerr << "CAUGHT UNKNOW EXCEPTION IN MAIN!" << std::endl;
  
    // delete doc;
    delete comp;
  
    return 1;
  }

#ifdef DEBUG
  std::cerr << "TRANSLATION SUCCESSFULLY COMPLETED." << std::endl;
#endif

  ExportModul *out;
  ExportCeres *out_ceres = NULL;

  if( output_type == flat )
    out = ExportFlatProofXML::New( true );
  else if( output_type == ceres )
    out = out_ceres = ExportCeres::New();
  else if( output_type == latex )
    out = ExportLaTeX::New();
  else if( output_type == flat_latex )
    out = ExportLaTeX::New( true );
  else
    out = SaveDbXML::New();

  try
  {
    out->Open( output_filename );

    if( out_ceres != NULL )
      out_ceres->Write( pdb, comp->GetUsedDefinitionsList(), doc );
    else
      out->Write( pdb );

    out->Close();
    delete out;
  }
  catch( Exception* e )
  {
    e->PrintAll( std::cerr );
    std::cerr << std::endl;
    delete e;
  
    delete out;
    delete pdb;
    // delete doc;
    delete comp;
  
    return 1;
  }
  catch( ... )
  {
    std::cerr << "CAUGHT UNKNOW EXCEPTION IN MAIN!" << std::endl;
  
    delete out;
    delete pdb;
    // delete doc;
    delete comp;
  
    return 1;
  }

  //delete pdb;

  // delete doc;

  delete comp;

  if( axioms_outfile != NULL )
  {
    axioms_outfile->close();
    delete axioms_outfile;
  }

  return 0;
}

