////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: printhlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>

#include "libproof/util/myassert.h"
#include "libproof/libproof.h"
#include "libproof/term_operations.h"

#include "symbolhlk.h"
#include "printhlk.h"
#include "recursive.h"

namespace HLK
{

bool Print::use_utf8 = false;

void Print::PrintTerm( const ProofLib::Term *t, std::ostream &out,
                       bool need_prnth )
{
  using namespace ProofLib;

  const Variable *v = t->CastTo<Variable>();

  if( v != NULL )
  {
    out << v->GetSymbol()->GetName();

    const RecursiveVariable *v_r = t->CastTo<RecursiveVariable>();

    if( v_r != NULL )
    {
      out << "[ ";
      PrintRecursiveExpr( v_r->GetRecursiveExpr(), out );
      out << " ]";
    }

    return;
  }

  const Constant *c = t->CastTo<Constant>();

  if( c != NULL )
  {
    out << c->GetSymbol()->GetName();

    const RecursiveConstant *c_r = t->CastTo<RecursiveConstant>();

    if( c_r != NULL )
    {
      out << "[ ";
      PrintRecursiveExpr( c_r->GetRecursiveExpr(), out );
      out << " ]";
    }

    return;
  }

  const Function *f = t->CastTo<Function>();

  if( f != NULL )
  {
    const FunctionSymbol *fs =
      dynamic_cast<const FunctionSymbol*>( f->GetSymbol() );

    const RecursiveFunction *f_r = t->CastTo<RecursiveFunction>();

    if( fs->IsInfix() )
    {
      ASSERT( f_r == NULL );

      if( need_prnth )
        out << "( ";
      PrintTerm( f->GetTerm(0), out, true );
      out << " " << f->GetSymbol()->GetName() << " ";
      PrintTerm( f->GetTerm(1), out, true );
      if( need_prnth )
        out << " )";
    }
    else
    {
      out << f->GetSymbol()->GetName();

      if( f_r != NULL )
      {
        out << "[ ";
        PrintRecursiveExpr( f_r->GetRecursiveExpr(), out );
        out << " ]";
      }

      out << "( ";
  
      for( size_t i = 0; i < f->GetChildCount(); i++ )
      {
        if( i > 0 ) out << ", ";
        PrintTerm( f->GetTerm(i), out );
      }
  
      out << " )";
    }

    return;
  }
}

void Print::PrintTerm( const ProofLib::Term *t, std::wostream &out,
                       bool need_prnth )
{
  using namespace ProofLib;

  const Variable *v = t->CastTo<Variable>();

  if( v != NULL )
  {
    out << v->GetSymbol()->GetName();

    const RecursiveVariable *v_r = t->CastTo<RecursiveVariable>();

    if( v_r != NULL )
    {
      out << "[ ";
      PrintRecursiveExpr( v_r->GetRecursiveExpr(), out );
      out << " ]";
    }

    return;
  }

  const Constant *c = t->CastTo<Constant>();

  if( c != NULL )
  {
    out << c->GetSymbol()->GetName();

    const RecursiveConstant *c_r = t->CastTo<RecursiveConstant>();

    if( c_r != NULL )
    {
      out << "[ ";
      PrintRecursiveExpr( c_r->GetRecursiveExpr(), out );
      out << " ]";
    }

    return;
  }

  const Function *f = t->CastTo<Function>();

  if( f != NULL )
  {
    const FunctionSymbol *fs =
      dynamic_cast<const FunctionSymbol*>( f->GetSymbol() );

    const RecursiveFunction *f_r = t->CastTo<RecursiveFunction>();

    if( fs->IsInfix() )
    {
      ASSERT( f_r == NULL );

      if( need_prnth )
        out << "( ";
      PrintTerm( f->GetTerm(0), out, true );
      out << " " << f->GetSymbol()->GetName() << " ";
      PrintTerm( f->GetTerm(1), out, true );
      if( need_prnth )
        out << " )";
    }
    else
    {
      out << f->GetSymbol()->GetName();

      if( f_r != NULL )
      {
        out << "[ ";
        PrintRecursiveExpr( f_r->GetRecursiveExpr(), out );
        out << " ]";
      }

      out << "( ";
  
      for( size_t i = 0; i < f->GetChildCount(); i++ )
      {
        if( i > 0 ) out << ", ";
        PrintTerm( f->GetTerm(i), out );
      }
  
      out << " )";
    }

    return;
  }
}

void Print::PrintRecursiveExpr( const ProofLib::RecursiveExpr *expr,
                                std::ostream &out )
{
  if( expr->GetVariableName() == NULL )
  {
    out << expr->GetNum();
    return;
  }

  out << expr->GetVariableName();

  if( expr->GetNum() != 0 )
    out << " " << expr->GetOp() << " " << expr->GetNum();
}

void Print::PrintRecursiveExpr( const ProofLib::RecursiveExpr *expr,
                                std::wostream &out )
{
  if( expr->GetVariableName() == NULL )
  {
    out << expr->GetNum();
    return;
  }

  out << expr->GetVariableName();

  if( expr->GetNum() != 0 )
    out << " " << expr->GetOp() << " " << expr->GetNum();
}

void Print::PrintFormula( const ProofLib::Formula *f, std::ostream &out,
                          bool need_prnth )
{
  using namespace ProofLib;

  const AtomFormula *a = f->CastTo<AtomFormula>();

  if( a != NULL )
  {
    const PredicateSymbol *ps =
      dynamic_cast<const PredicateSymbol*>( a->GetSymbol() );

    if( ps->IsInfix() )
    {
      ASSERT( !ps->IsRecursive() );

      if( need_prnth )
        out << "( ";
      PrintTerm( a->GetTerm(0), out, true );
      out << " " << a->GetSymbol()->GetName() << " ";
      PrintTerm( a->GetTerm(1), out, true );
      if( need_prnth )
        out << " )";
    }
    else
    {
      out << a->GetSymbol()->GetName();

      if( ps->IsRecursive() )
      {
        const RecursivePredicate *a_r = f->CastTo<RecursivePredicate>();

        out << "[ ";
        PrintRecursiveExpr( a_r->GetRecursiveExpr(), out );
        out << " ]";
      }
  
      if( a->GetChildCount() == 0 )
        return;
  
      out << "( ";
  
      for( size_t i = 0; i < a->GetChildCount(); i++ )
      {
        if( i > 0 ) out << ", ";
        PrintTerm( a->GetTerm(i), out );
      }
  
      out << " )";
    }

    return;
  }

  const UnaryConFormula *u = f->CastTo<UnaryConFormula>();

  if( u != NULL )
  {
    if( use_utf8 )
    {
      std::string sym;
      UTF8FromUChar( u->GetOperatorSymbol()->GetDisplayChar(), sym );
      
      out << sym;
    }
    else
      out << u->GetOperatorSymbol()->GetName() << " ";

/*
    bool need_pth_here =
      ( dynamic_cast<ProofLib::BinaryConFormula*>( u->GetFormula() ) != NULL );

    if( need_pth_here )
      out << "( ";
*/

    PrintFormula( u->GetFormula(), out, true );

/*
    if( need_pth_here )
      out << " )";
*/

    return;
  }

  const QuantorFormula *q = f->CastTo<QuantorFormula>();

  if( q != NULL )
  {
    out << "(";

    if( use_utf8 )
    {
      std::string sym;
      UTF8FromUChar( q->GetOperatorSymbol()->GetDisplayChar(), sym );
      
      out << sym;
    }
    else
      out << q->GetOperatorSymbol()->GetName() << " ";

    PrintTerm( q->GetVariable(), out );
    out << ")";

    PrintFormula( q->GetFormula(), out );

    return;
  }

  const SecondOrderQuantorFormula *qs = f->CastTo<SecondOrderQuantorFormula>();

  if( qs != NULL )
  {
    out << "(";

    if( use_utf8 )
    {
      std::string sym;
      UTF8FromUChar( qs->GetOperatorSymbol()->GetDisplayChar(), sym );
      
      out << sym;
    }
    else
      out << qs->GetOperatorSymbol()->GetName() << " ";

    qs->GetSecondOrderVariable()->ToString( out );
    out << ")";

    PrintFormula( qs->GetFormula(), out );

    return;
  }

  const BinaryConFormula *b = f->CastTo<BinaryConFormula>();

  if( b != NULL )
  {
    if( need_prnth )
      out << "( ";

    PrintFormula( b->GetLeft(), out, true );

    out << " ";

    if( use_utf8 )
    {
      std::string sym;
      UTF8FromUChar( b->GetOperatorSymbol()->GetDisplayChar(), sym );
      
      out << sym;
    }
    else
      out << b->GetOperatorSymbol()->GetName();

    out << " ";

    PrintFormula( b->GetRight(), out, true );

    if( need_prnth )
      out << " )";

    return;
  }
}

void Print::PrintFormulaList( const ProofLib::FormulaList *l,
                              std::ostream &out )
{
  using namespace ProofLib;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Formula *f = l->GetElement(i)->CastTo<Formula>();
    ASSERT( f != NULL );

    if( i > 0 ) out << ", ";

    PrintFormula( f, out );
  }
}

void Print::PrintSequent( const ProofLib::Sequent *s, std::ostream &out )
{
  ASSERT( s != NULL );

  ProofLib::SElementList *l = s->GetLeft();

  if( l != NULL )
  {
    size_t i = 0;
    for( ; i < l->GetChildCount(); i++ )
    {
      if( i > 0 )
        out << ", ";

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
        ( l->GetSElement(i) );

      ASSERT( sf != NULL );

      PrintFormula( sf->GetFormula(), out );
    }

    if( i > 0 )
      out << " ";
  }

  out << ":-";

  l = s->GetRight();

  if( l != NULL )
  {
    size_t i = 0;
    for( ; i < l->GetChildCount(); i++ )
    {
      if( i > 0 )
        out << ", ";
      else
        out << " ";

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
        ( l->GetSElement(i) );

      ASSERT( sf != NULL );

      PrintFormula( sf->GetFormula(), out );
    }
  }

}

void Print::PrintSequentAsList( const ProofLib::Sequent *s, std::ostream &out )
{
  ASSERT( s != NULL );

  ProofLib::SElementList *l = s->GetLeft();

  if( l != NULL )
  {
    size_t i = 0;
    for( ; i < l->GetChildCount(); i++ )
    {
      out << i << ": ";

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
        ( l->GetSElement(i) );

      ASSERT( sf != NULL );

      PrintFormula( sf->GetFormula(), out );

      out << std::endl;
    }
  }

  out << ":-";
  out << std::endl;

  l = s->GetRight();

  if( l != NULL )
  {
    size_t i = 0;
    for( ; i < l->GetChildCount(); i++ )
    {
      out << i << ": ";

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
        ( l->GetSElement(i) );

      ASSERT( sf != NULL );

      PrintFormula( sf->GetFormula(), out );

      out << std::endl;
    }
  }

}
void Print::PrintVariableSubstitution(
  const ProofLib::VariableSubstitution *u, std::ostream &out )
{
  if( u == NULL ) return;

  for( size_t i = 0; i < u->GetCount(); i++ )
  {
    if( i > 0 ) out << ", ";

    PrintTerm( u->GetVar(i), out );
    out << " <- ";
    PrintTerm( u->GetSubst(i), out );
  }
}

void Print::PrintNodePointer( const ProofLib::NodePointer *np,
                              std::ostream &out )
{
  ASSERT( np != NULL );

  out << '(';

  for( size_t i = 0; i < np->GetLength(); i++ )
  {
    out << ' ';
    out << (*np)[i];
  }

  if( !np->IsEmpty() )
    out << ' ';
  
  out << ')';
}

////////////////////////////////////////////////////////////////////////////////

void Print::PrintSLoc( const ProofLib::SourceLocation &sloc, std::ostream &out )
{
  if( sloc.GetFile() != NULL )
    out << sloc.GetFile() << ":";

  out << sloc.GetLine();

  if( sloc.GetChar() > 0 )
    out << "," << sloc.GetChar();
}

}

