////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: proofhlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <vector>

#include "libproof/util/myassert.h"
#include "libproof/util/hash.h"
#include "libproof/libproof.h"
#include "libproof/operations.h"
#include "libproof/term_operations.h"

#include "reswordshlk.h"
#include "symbolhlk.h"
#include "proofhlk.h"
#include "printhlk.h"
#include "recursive.h"

////////////////////////////////////////////////////////////////////////////////

namespace stlhash
{
  template<> struct hash<const HLK::ProofSymbol*> {
    size_t operator()(const HLK::ProofSymbol* x) const
      { return (size_t) x >> 2; }
  };
  template<> struct hash<HLK::ProofSymbol*> {
    size_t operator()(HLK::ProofSymbol* x) const { return (size_t) x >> 2; }
  };

#if 0
  template<> struct hash<const ProofLib::Symbol*> {
    size_t operator()(const ProofLib::Symbol* x) const { return size_t(x) >> 2; }
  };
  template<> struct hash<ProofLib::Symbol*> {
    size_t operator()(ProofLib::Symbol* x) const { return size_t(x) >> 2; }
  };
#endif
}

////////////////////////////////////////////////////////////////////////////////

namespace HLK
{

////////////////////////////////////////////////////////////////////////////////

class ProofTableStore : public stlhash::hash_map
  < const ProofSymbol*, ProofBase* > //, stlhash::hash<const char*>, eqstr >
{
public:
  ~ProofTableStore()
  {
    for( iterator i = begin(); i != end(); i++ )
      delete i->second;
  }
};

////////////////////////////////////////////////////////////////////////////////

ProofTable::ProofTable()
{
  m_store = new ProofTableStore;
}

ProofTable::~ProofTable()
{
  delete m_store;
}

void ProofTable::AddProof( ProofBase *p )
{
  (*m_store)[p->GetSymbol()] = p; 
}

const ProofBase* ProofTable::GetProof( const ProofSymbol *s )
{
  ProofTableStore::const_iterator i = m_store->find( s );

  if( i == m_store->end() )
    return NULL;

  return i->second;
}

////////////////////////////////////////////////////////////////////////////////

class ProofRuleList : public std::vector<const ProofRule*>
{
public:
  ~ProofRuleList()
  {
    for( iterator i = begin(); i != end(); i++ )
      delete *i;
  }
};

////////////////////////////////////////////////////////////////////////////////

void ProofMetaParams::AddSymbol( const Symbol *s )
{
  m_params.push_back( s );
}

const Symbol* ProofMetaParams::GetSymbol( size_t index ) const
{
  ASSERT( index < m_params.size() );
  return m_params[index];
}

bool ProofMetaParams::IsFormulaAt( size_t index ) const { return false; }
bool ProofMetaParams::IsTermAt( size_t index ) const { return false; }

////////////////////////////////////////////////////////////////////////////////

ProofBody::ProofBody()
{
  m_end_sequent = NULL;
  m_rule_store = new ProofRuleList;
}

ProofBody::~ProofBody()
{
  if( m_end_sequent != NULL );

  delete m_end_sequent;
  delete m_rule_store;
}

void ProofBody::SetEndSequent( ProofLib::Sequent *seq )
{
  ASSERT( seq != NULL );
  m_end_sequent = seq;
}

void ProofBody::AddRule( ProofRule *rule )
{
  m_rule_store->push_back( rule );
//  rule->m_proof = this;
}

size_t ProofBody::GetRuleCount() const
{
  return m_rule_store->size();
}

const ProofRule* ProofBody::GetRule( size_t index ) const
{
  ASSERT( GetRuleCount() > index );
  return (*m_rule_store)[index];
}

////////////////////////////////////////////////////////////////////////////////

ProofBase::ProofBase( const ProofSymbol *sym,
              SymbolTable *symtab,
              const ProofMetaParams *meta_params )
  : m_meta_params( *meta_params )
{
  ASSERT( sym != NULL );

  m_symbol = sym;
  m_symbol_table = symtab;
}

ProofBase::ProofBase( const ProofBase &o )
  : m_meta_params( o.m_meta_params )
{
  m_symbol = o.m_symbol;
  m_symbol_table = o.m_symbol_table;
}

ProofBase::~ProofBase()
{
}

////////////////////////////////////////////////////////////////////////////////

Proof::Proof( const ProofSymbol *sym,
              SymbolTable *symtab,
              const ProofMetaParams *meta_params )
  : ProofBase( sym, symtab, meta_params )
{
}

// used by ProofInstance
Proof::Proof( const Proof *p )
  : ProofBase( *p )
{
}

// used by ProofInstance
Proof::Proof( const RecursiveProof *p )
  : ProofBase( *p )
{
}

Proof::~Proof()
{
}

const Proof* Proof::GetProofInstance( const ProofMetaArguments &args,
                                      const ProofLib::SourceLocation &sloc,
                                      ProofLib::NodeDatabase *ndb,
                                      unsigned int level ) const
{
  if( args.size() != GetProofMetaParams().GetCount() )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "Proof '" << GetSymbol()->GetName() << "' has "
                << GetProofMetaParams().GetCount()
                << " parameters but is called with "
                << args.size()
                << " arguments.";
    throw e;
  }

  return new ProofInstance( this, args, sloc, ndb, level );
}

void Proof::ReleaseProofInstance()
{
}

////////////////////////////////////////////////////////////////////////////////

RecursiveProof::RecursiveProof( const ProofSymbol *sym,
              SymbolTable *symtab,
              const ProofMetaParams *meta_params )
  : ProofBase( sym, symtab, meta_params )
{
}

// used by ProofInstance
RecursiveProof::RecursiveProof( const Proof *p )
  : ProofBase( *p )
{
  //SetEndSequent( p->GetEndSequent()->Clone() );
}

RecursiveProof::~RecursiveProof()
{
  for( size_t i = 0; i < m_bodies.size(); i++ )
    delete m_bodies[i];
}

const Proof* RecursiveProof::GetProofInstance(
  const ProofMetaArguments &args,
  const ProofLib::SourceLocation &sloc,
  ProofLib::NodeDatabase *ndb,
  unsigned int level ) const
{
  if( args.size() != GetProofMetaParams().GetCount() )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "Proof '" << GetSymbol()->GetName() << "' has "
                << GetProofMetaParams().GetCount()
                << " parameters but is called with "
                << args.size()
                << " arguments.";
    throw e;
  }

  return new ProofInstance( this, args, sloc, ndb, level );
}

void RecursiveProof::ReleaseProofInstance()
{
}

ProofBody* RecursiveProof::AddLevel( ProofLib::RecursiveExpr *expr )
{
  expr->Attach();
  m_exprs.push_back( expr );
  m_bodies.push_back( new ProofBody() );

  ASSERT( m_exprs.size() == m_bodies.size() );

  return m_bodies[m_bodies.size()-1];
}

const ProofBody* RecursiveProof::GetProofAtLevel( size_t &level ) const
{
  bool found = false;
  size_t i = FindLevel( m_exprs, level, found, *this );

  if( !found )
    return NULL;

  return m_bodies[i];
}

////////////////////////////////////////////////////////////////////////////////

class ProofParamArgMapImpl :
  public ProofParamArgMap, public ProofLib::TermOperation
{
  typedef std::pair< ProofLib::DbNode*, ProofLib::StaticTermList* > Arg;
  typedef stlhash::hash_map< const ProofLib::Symbol*, Arg > Map;
  typedef std::vector< Arg > List;

  Map m_map;
  List m_list;

  ProofLib::NodeDatabase *m_ndb;

public:
  ProofParamArgMapImpl( ProofLib::NodeDatabase *ndb ) { m_ndb = ndb; }
  virtual ~ProofParamArgMapImpl() {}

  ProofLib::NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofLib::ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

  virtual size_t GetCount() const
  {
    return m_list.size();
  }

  virtual ProofLib::DbNode* GetArg( size_t index ) const
  {
    ASSERT( index < m_list.size() );
    return m_list[index].first;
  }

public:
  virtual void AddMapping( const Symbol *s, ProofLib::DbNode *node,
                           ProofLib::TermList &vars )
  {
    m_map[s] =
      Arg( node, new ProofLib::StaticTermList( &vars ) );
    m_list.push_back( m_map[s] );
  }

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
};

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, ProofParamArgMapImpl, v )
{
  if( m_map.find( v->GetSymbol() ) != m_map.end() )
  {
    ProofLib::Term *t =
     dynamic_cast<ProofLib::Term*>( m_map[v->GetSymbol()].first );

    return t;
  }

  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, ProofParamArgMapImpl, f )
{
  // function f is a meta term of the proof?
  if( m_map.find( f->GetSymbol() ) != m_map.end() )
  {
    // ok, in that case we have the meta formula instance f( x_0, ..., x_n ),
    // for which the argument to the proof was a term
    // t( y_0, ..., y_n ), so the replacement will be t( x_0, ..., x_n ).
    ProofLib::Term *t =
      dynamic_cast<ProofLib::Term*>( m_map[f->GetSymbol()].first );

    ProofLib::OperationVariableSubstitution op( GetNodeDatabase() );

    ProofLib::StaticTermList *term_list =
      m_map[f->GetSymbol()].second;

    for( size_t i = 0; i < f->GetChildCount(); i++ )
    {
      ASSERT( i < term_list->GetChildCount() );

      ProofLib::Variable *v =
        term_list->GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );
      
      op.AddSubstitution( v, f->GetTerm(i) );
    }

    t = ProofLib::DataNode::ApplyTo( t, op );

    // replace now all x_0, ..., x_n that are meta terms or contain meta terms.
    return ProofLib::DataNode::ApplyTo( t, *this );
  }

  return ProofLib::TermOperation::ApplyToFunction( f );
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, ProofParamArgMapImpl, f )
{
  // atomformula f is a meta formula of the proof?
  if( m_map.find( f->GetSymbol() ) != m_map.end() )
  {
    // ok, in that case we have the meta formula instance P( x_0, ..., x_n ),
    // for which the argument to the proof was a (complex) formula
    // F( y_0, ..., y_n ), so the replacement will be F( x_0, ..., x_n ).
    ProofLib::Formula *f_0 =
      dynamic_cast<ProofLib::Formula*>( m_map[f->GetSymbol()].first );

    ProofLib::OperationVariableSubstitution op( GetNodeDatabase() );

    ProofLib::StaticTermList *term_list =
      m_map[f->GetSymbol()].second;

    for( size_t i = 0; i < f->GetChildCount(); i++ )
    {
      ASSERT( i < term_list->GetChildCount() );

      ProofLib::Variable *v =
        term_list->GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );
      
      op.AddSubstitution( v, f->GetTerm(i) );
    }

    f_0 = ProofLib::DataNode::ApplyTo( f_0, op );

    // replace now all x_0, ..., x_n that are meta terms or contain meta terms.
    return ProofLib::DataNode::ApplyTo( f_0, *this );
  }

#if 0
  Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl;
#endif
  return ProofLib::FormulaOperation::ApplyToAtomFormula( f );
}

////////////////////////////////////////////////////////////////////////////////

ProofRecursionsExpander::ProofRecursionsExpander(
    ProofLib::NodeDatabase *ndb,
    unsigned int level,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc )
    : m_sloc( sloc )
{
  m_ndb = ndb;
  m_level = level;
  m_symbol_table = symbol_table;

  ASSERT( m_ndb != NULL );
  ASSERT( m_symbol_table != NULL );
}

ProofLib::ProofDatabase* ProofRecursionsExpander::GetProofDatabase()
{
  ASSERT(false);
  return NULL;
}

ProofLib::RecursiveExpr*
ProofRecursionsExpander::CalcRecursiveExpr( ProofLib::RecursiveExpr* expr )
{
  // constant expresion?
  if( expr->GetVariableName() == NULL )
    return expr;

  // calculate the expersion at this level
  unsigned int level = expr->GetValueAtLevel( m_level );

  return ProofLib::RecursiveExpr::New( GetNodeDatabase(), NULL, '+', level );
}

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, ProofRecursionsExpander, v_0 )
{
  ProofLib::RecursiveVariable *v =
    dynamic_cast<ProofLib::RecursiveVariable*>( v_0 );

  if( v == NULL )
    return v_0;

  const VariableSymbol *vs =
    dynamic_cast<const VariableSymbol*>( v->GetSymbol() );

  ASSERT( vs != NULL );

  unsigned int level = v->GetRecursiveExpr()->GetValueAtLevel( m_level );

  std::wostringstream str;

  str << v->GetSymbol()->GetName() << "_{" << level << "}";

#if 0
  std::cerr << "recursive variable " << str.str() << std::endl;
#endif

  ASSERT( GetSymbolTable() != NULL );
  ASSERT( GetSymbolTable()->GetTopParent() != NULL );

  SymbolTable *symt = GetSymbolTable()->GetTopParent();

  const Symbol *s = symt->GetSymbol( str.str().c_str() );

  if( s == NULL )
  {
    symt->NewVariable( str.str().c_str(), vs->GetType()->GetName(),
                       true, m_sloc );

    return ProofLib::Variable::New( GetNodeDatabase(),
                                    symt->GetSymbol( str.str().c_str() ) );
  }

  if( dynamic_cast<const VariableSymbol*>( s ) == NULL )
  {
    throw new SymbolExistsException( s );
  }

  return ProofLib::Variable::New( GetNodeDatabase(), s );
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, ProofRecursionsExpander, c_0 )
{
  ProofLib::RecursiveConstant *c =
    dynamic_cast<ProofLib::RecursiveConstant*>( c_0 );

  if( c == NULL )
    return c_0;

  const ConstantSymbol *cs =
    dynamic_cast<const ConstantSymbol*>( c->GetSymbol() );

  ASSERT( cs != NULL );

  unsigned int level = c->GetRecursiveExpr()->GetValueAtLevel( m_level );

  std::wostringstream str;

  str << cs->GetName() << "_{" << level << "}";

#if 0
  std::cerr << "recursive constant " << str.str() << std::endl;
#endif

  ASSERT( GetSymbolTable() != NULL );
  ASSERT( GetSymbolTable()->GetTopParent() != NULL );

  SymbolTable *symt = GetSymbolTable()->GetTopParent();

  const Symbol *s = symt->GetSymbol( str.str().c_str() );

  if( s == NULL )
  {
    symt->NewConstant( str.str().c_str(), cs->GetType()->GetName(), m_sloc );

    return ProofLib::Constant::New( GetNodeDatabase(),
                                    symt->GetSymbol( str.str().c_str() ) );
  }

  if( dynamic_cast<const ConstantSymbol*>( s ) == NULL )
  {
    throw new SymbolExistsException( s );
  }

  return ProofLib::Constant::New( GetNodeDatabase(), s );
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, ProofRecursionsExpander, f_0 )
{
  const ProofLib::RecursiveFunction *f =
    dynamic_cast<const ProofLib::RecursiveFunction*>( f_0 );

  if( f == NULL )
  { 
#if 0
Print::PrintTerm( f_0, std::cerr );
std::cerr << std::endl;
  const FunctionSymbol *fs =
    dynamic_cast<const FunctionSymbol*>( f_0->GetSymbol() );
  ASSERT( !fs->IsRecursive() );
std::cerr << "recursive expanding func " << fs->GetName() << " at level " << m_level << std::endl;
#endif
    return ProofLib::TermOperation::ApplyToFunction( f_0 );
  }

  // constant expresion?
  if( f->GetRecursiveExpr()->GetVariableName() == NULL )
    return f_0;

  // calculate the expersion at this level
  unsigned int level = f->GetRecursiveExpr()->GetValueAtLevel( m_level );

  ProofLib::RecursiveExpr* expr =
    ProofLib::RecursiveExpr::New( GetNodeDatabase(), NULL, '+', level );

  const FunctionSymbol *fs =
    dynamic_cast<const FunctionSymbol*>( f->GetSymbol() );

  ProofLib::TermList terms( *f );

  return ProofLib::TermOperation::ApplyToFunction(
    ProofLib::RecursiveFunction::New( GetNodeDatabase(), fs, &terms, expr ) );
#if 0
  return ProofLib::RecursiveFunction::New( GetNodeDatabase(), fs,
                                           &terms, expr );
#endif
#if 0
  unsigned int level = m_level;
  
  m_level = f->GetRecursiveExpr()->GetValueAtLevel( level );

  const FunctionSymbol *fs =
    dynamic_cast<const FunctionSymbol*>( f->GetSymbol() );

#if 0
std::cerr << "Expand rec func " << fs->GetName() << " at level " << m_level << ": ";
Print::PrintTerm( f, std::cerr );
std::cerr << std::endl;
#endif

  // get the term of case for the level
  ProofLib::Term* t;

  if( m_level == 0 )
    t = fs->GetTermAt0();
  else
    t = fs->GetTermAtN();

  ASSERT( t != NULL );

  // replace arguments in the term
  ProofLib::OperationVariableSubstitution op( GetNodeDatabase() );

  const ProofLib::StaticTermList* var_list = fs->GetVariableMapping();

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    ASSERT( i < var_list->GetChildCount() );

    ProofLib::Variable *v =
      var_list->GetTerm(i)->CastTo<ProofLib::Variable>();
    ASSERT( v != NULL );
    
    op.AddSubstitution( v, f->GetTerm(i) );
  }

  t = ProofLib::DataNode::ApplyTo( t, op );

  // next recusrion step.
  t = ProofLib::DataNode::ApplyTo( t, *this );

  m_level = level;

#if 0
std::cerr << "exit level " << m_level << std::endl;
#endif

  return t;
#endif
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula,
                             ProofRecursionsExpander, f_0 )
{
  const ProofLib::RecursivePredicate *f =
    dynamic_cast<const ProofLib::RecursivePredicate*>( f_0 );

  if( f == NULL )
  { 
#if 0
Print::PrintTerm( f_0, std::cerr );
std::cerr << std::endl;
  const PredicateSymbol *fs =
    dynamic_cast<const PredicateSymbol*>( f_0->GetSymbol() );
  ASSERT( !fs->IsRecursive() );
std::cerr << "recursive expanding pred " << fs->GetName() << " at level " << m_level << std::endl;
#endif
    return ProofLib::TermOperation::ApplyToAtomFormula( f_0 );
  }

  // constant expresion?
  if( f->GetRecursiveExpr()->GetVariableName() == NULL )
    return f_0;

  // calculate the expersion at this level
  unsigned int level = f->GetRecursiveExpr()->GetValueAtLevel( m_level );

  ProofLib::RecursiveExpr* expr =
    ProofLib::RecursiveExpr::New( GetNodeDatabase(), NULL, '+', level );

  const PredicateSymbol *ps =
    dynamic_cast<const PredicateSymbol*>( f->GetSymbol() );

  ProofLib::TermList terms( *f );

  return ProofLib::RecursivePredicate::New( GetNodeDatabase(), ps,
                                            &terms, expr );

#if 0
  unsigned int level = m_level;
  
  m_level = f->GetRecursiveExpr()->GetValueAtLevel( level );

  const PredicateSymbol *ps =
    dynamic_cast<const PredicateSymbol*>( f->GetSymbol() );

#if 0
std::cerr << "Expand rec pred " << ps->GetName() << " at level " << m_level
          << std::endl;
#endif

  // get the term of case for the level
  ProofLib::Formula* t;

  if( m_level == 0 )
    t = ps->GetFormulaAt0();
  else
    t = ps->GetFormulaAtN();

  ASSERT( t != NULL );

  // replace arguments in the term
  ProofLib::OperationVariableSubstitution op( GetNodeDatabase() );

  const ProofLib::StaticTermList* var_list = ps->GetVariableMapping();

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    ASSERT( i < var_list->GetChildCount() );

    ProofLib::Variable *v =
      var_list->GetTerm(i)->CastTo<ProofLib::Variable>();
    ASSERT( v != NULL );
    
    op.AddSubstitution( v, f->GetTerm(i) );
  }

  t = ProofLib::DataNode::ApplyTo( t, op );


  // next recusrion step.
  t = ProofLib::DataNode::ApplyTo( t, *this );

  m_level = level;

#if 0
std::cerr << "exit level " << m_level << std::endl;
#endif

  return t;
#endif
}

////////////////////////////////////////////////////////////////////////////////

ProofInstance::ProofInstance( const Proof *p, const ProofMetaArguments &args,
                              const ProofLib::SourceLocation &sloc,
                              ProofLib::NodeDatabase *ndb,
                              unsigned int level )
  : Proof( p )
{
  SetSourceLocation( *p );
  m_arg_map = new ProofParamArgMapImpl( ndb );
  m_rec_expander =
    new ProofRecursionsExpander( ndb, level, GetSymbolTable(), *GetSymbol() );

  CheckArguments( args, sloc );

  InstantiateProofBody( p, sloc );
}

ProofInstance::ProofInstance( const RecursiveProof *p,
                              const ProofMetaArguments &args,
                              const ProofLib::SourceLocation &sloc,
                              ProofLib::NodeDatabase *ndb,
                              unsigned int level )
  : Proof( p )
{
  SetSourceLocation( *p );
  m_arg_map = new ProofParamArgMapImpl( ndb );

  CheckArguments( args, sloc );

  size_t body_level = level;

  const ProofBody *body = p->GetProofAtLevel( body_level );

  m_rec_expander =
    new ProofRecursionsExpander( ndb, body_level,
                                 GetSymbolTable(), *GetSymbol() );

  InstantiateProofBody( body, sloc );
/*
#warning "TODO: InstantiateProofBody"
ASSERT( false );
  if( level == 0 )
    InstantiateProofBody( p->GetProofLevel0(), sloc );
  else
    InstantiateProofBody( p->GetProofLevelN(), sloc );
*/
}

ProofInstance::~ProofInstance()
{
  delete m_arg_map;
  delete m_rec_expander;
}

void ProofInstance::ReleaseProofInstance()
{
  delete this; // kill me
}

void ProofInstance::InstantiateProofBody( const ProofBody *p,
                                          const ProofLib::SourceLocation &sloc )
{
  try
  {
    // instantiate end sequent
    SetEndSequent( p->GetEndSequent()->Clone() );
    GetEndSequent()->Apply( dynamic_cast<ProofParamArgMapImpl&>(*m_arg_map) );
    GetEndSequent()->Apply( *m_rec_expander );

    // instantiate rules
    for( size_t i = 0; i < p->GetRuleCount(); i++ )
      AddRule( p->GetRule( i )->
                 GetInstance( *m_arg_map, *m_rec_expander, sloc ) );
  }
  catch( Exception *e )
  {
    throw new ProofLib::SLocCoverException( &sloc, e );
  }
}

void ProofInstance::CheckArguments( const ProofMetaArguments &args,
                                    const ProofLib::SourceLocation &sloc )
{
  // check arguments
  for( size_t i = 0; i < GetProofMetaParams().GetCount(); i++ )
  {
    const Symbol *s = GetProofMetaParams().GetSymbol(i);

    ProofLib::Term *term = dynamic_cast<ProofLib::Term*>(args[i]);
    ProofLib::Formula *formula = dynamic_cast<ProofLib::Formula*>(args[i]);

    ASSERT( term != NULL || formula != NULL );

    const VariableSymbol *vs = dynamic_cast<const VariableSymbol*>( s );
    const FunctionSymbol *fs = dynamic_cast<const FunctionSymbol*>( s );
    const PredicateSymbol *ps = dynamic_cast<const PredicateSymbol*>( s );

    ASSERT( vs != NULL || fs != NULL || ps != NULL );

    if( term == NULL && ( vs != NULL || fs != NULL ) )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
      e->Stream() << "Proof term parameter '" << s->GetName()
                  << "' is instantiated by formula '";
      Print::PrintFormula( formula, e->Stream() );
      e->Stream() << "'.";
      throw e;
    }

    if( formula == NULL && ps != NULL )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
      e->Stream() << "Proof formula parameter '" << s->GetName()
                  << "' is instantiated by term '";
      Print::PrintTerm( term, e->Stream() );
      e->Stream() << "'.";
      throw e;
    }

    ProofLib::TermList vars;
    ProofLib::OperationGetFreeVariables op( vars );

    if( term != NULL )
    {
      term->Apply( op );

      const SymbolWithType *swt;

      if( vs != NULL )
        swt = vs;
      else
        swt = fs;

      const TypeSymbol* ts = FunctionSymbol::GetTermType( term );

      if( !swt->GetType()->IsAnyType() && !ts->IsAnyType() &&
          swt->GetType() != ts )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Proof term parameter '" << s->GetName()
                    << "' is of type " << swt->GetType()->GetName()
                    << ", but is instantiated by the term '";
        Print::PrintTerm( term, e->Stream() );
        e->Stream() << "' which is of type " << ts->GetName() << ".";
        throw e;
      }
    }
    else
      formula->Apply( op );

    if( vs != NULL )
    {
#if OLD_VAR_MAPPING
      if( vars.GetChildCount() > 1 )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Proof term parameter '" << s->GetName()
                    << "' has one variable, but is instantiated by the term '";
        Print::PrintTerm( term, e->Stream() );
        e->Stream() << "' which has " << vars.GetChildCount() << " variables.";
        throw e;
      }
#endif

      m_arg_map->AddMapping( vs, term, vars );
    }

    if( fs != NULL || ps != NULL )
    {
      const SymbolWithArguments *swa;

      if( fs != NULL )
        swa = fs;
      else
        swa = ps;

      if( vars.GetChildCount() != swa->GetArity() )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Proof " << ( fs == NULL ? "term" : "formula" )
                    << " parameter '" << s->GetName()
                    << "' has " << swa->GetArity()
                    << " variables, but is instantiated by the "
                    << ( fs == NULL ? "term" : "formula" )
                    << " '";
        if( fs == NULL )
          Print::PrintFormula( formula, e->Stream() );
        else
          Print::PrintTerm( term, e->Stream() );
        e->Stream() << "' which has " << vars.GetChildCount() << " variables.";
        throw e;
      }

      for( size_t j = 0; j < vars.GetChildCount(); j++ )
      {
        const TypeSymbol* ts = FunctionSymbol::GetTermType( vars.GetTerm(j) );

        if( !swa->CheckArgType( j, ts ) )
        {
          ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
          e->Stream() << "Proof " << ( fs == NULL ? "term" : "formula" )
                      << " parameter '" << s->GetName()
                      << "' variable #" << j
                      << " is of type " << swa->GetArgType( j )
                      << ", but is instantiated by the variable '";
          Print::PrintTerm( vars.GetTerm(j), e->Stream() );
          e->Stream() << "' which is of type " << ts->GetName() << ".";
          throw e;
        }
      }

      if( fs == NULL )
        m_arg_map->AddMapping( ps, formula, vars );
      else
        m_arg_map->AddMapping( fs, term, vars );
    }
  }
}
////////////////////////////////////////////////////////////////////////////////

ProofRule::ProofRule( int type, ProofLib::Sequent *seq )
{
  ASSERT( type == RESWORD_AUTO );

  m_type = type;
  m_bLeft = false;
  m_seq = seq;
  m_seq_gen = NULL;
  m_cut_formula = NULL;
  m_ref_l = NULL;
  m_ref_r = NULL;
  m_ts_l = NULL;
  m_ts_r = NULL;
  m_rec_expr = NULL;
  m_lambda_term = NULL;
  m_taut = NULL;
}

ProofRule::ProofRule( const ProofRef *ref )
{
  ASSERT( ref != NULL );

  m_type = RESWORD_CONTINUED;
  m_bLeft = false;
  m_seq = NULL;
  m_seq_gen = NULL;
  m_cut_formula = NULL;
  m_ref_l = ref;
  m_ref_r = NULL;
  m_ts_l = NULL;
  m_ts_r = NULL;
  m_rec_expr = NULL;
  m_lambda_term = NULL;
  m_taut = NULL;
}

ProofRule::ProofRule( ProofLib::Sequent *seq, ProofLib::Sequent *seq_gen )
{
  ASSERT( seq != NULL );

  m_type = RESWORD_AXIOM;
  m_bLeft = false;
  m_seq = seq;
  m_seq_gen = seq_gen;
  m_cut_formula = NULL;
  m_ref_l = NULL;
  m_ref_r = NULL;
  m_ts_l = NULL;
  m_ts_r = NULL;
  m_rec_expr = NULL;
  m_lambda_term = NULL;
  m_taut = NULL;
}

ProofRule::ProofRule( int type, bool left, ProofLib::Sequent *seq,
                      const USTRING &def_ident,
                      const TypeSymbol *ts_l, const TypeSymbol *ts_r,
                      ProofLib::RecursiveExpr* rec_expr,
                      ProofLib::LambdaTerm* lambda_term )
{
  ASSERT( seq != NULL );

  m_type = type;
  m_bLeft = left;
  m_seq = seq;
  m_seq_gen = NULL;
  m_def_ident = def_ident;
  m_cut_formula = NULL;
  m_ref_l = NULL;
  m_ref_r = NULL;
  m_ts_l = ts_l;
  m_ts_r = ts_r;
  m_rec_expr = rec_expr;
  m_lambda_term = lambda_term;
  m_taut = NULL;

  if( m_rec_expr != NULL )
    m_rec_expr->Attach();

  if( m_lambda_term != NULL )
  {
    std::cerr << "Got Lambda Term in ProofRule: ";
    m_lambda_term->ToString( std::cerr );
    std::cerr << std::endl;

    m_lambda_term->Attach();
  }
}

ProofRule::ProofRule( int type, bool left, ProofLib::Sequent *seq,
                      ProofLib::Formula *cut_formula,
                      const ProofRef *ref_l, const ProofRef *ref_r )
{
  ASSERT( cut_formula != NULL || type != RESWORD_CUT );
  ASSERT( seq != NULL || ref_l != NULL );
  ASSERT( seq != NULL || ref_r != NULL );
  ASSERT( ( ref_l == NULL || ref_r == NULL ) || seq == NULL );

  m_type = type;
  m_bLeft = left;
  m_seq = seq;
  m_seq_gen = NULL;
  m_cut_formula = cut_formula;
  m_ref_l = ref_l;
  m_ref_r = ref_r;
  m_ts_l = NULL;
  m_ts_r = NULL;
  m_rec_expr = NULL;
  m_lambda_term = NULL;
  m_taut = NULL;
  
  if( m_cut_formula != NULL )
    m_cut_formula->Attach();
    
}

// for ProofRuleInstance
ProofRule::ProofRule( const ProofRule *org )
  : ProofLib::SourceLocation( *org )
{
  m_type = org->m_type;
  m_bLeft = org->m_bLeft;
  m_seq = ( org->m_seq != NULL ? org->m_seq->Clone() : NULL );
  m_seq_gen = ( org->m_seq_gen != NULL ? org->m_seq_gen->Clone() : NULL );
  m_def_ident = org->m_def_ident;
  m_cut_formula = org->m_cut_formula;
  m_ref_l = NULL;
  m_ref_r = NULL;
  m_ts_l = org->m_ts_l;
  m_ts_r = org->m_ts_r;
  m_rec_expr = org->m_rec_expr;
  m_lambda_term = org->m_lambda_term;
  m_taut = ( org->m_taut != NULL ? org->m_taut->Clone() : NULL );

  if( m_rec_expr != NULL )
    m_rec_expr->Attach();
  if( m_cut_formula != NULL )
    m_cut_formula->Attach();
  if( m_lambda_term != NULL )
    m_lambda_term->Attach();
}

ProofRule::~ProofRule()
{
  if( m_seq != NULL )
    delete m_seq;

  if( m_seq_gen != NULL )
    delete m_seq_gen;

  if( m_cut_formula != NULL )
    m_cut_formula->Detach();

  if( m_ref_l != NULL )
    delete m_ref_l;

  if( m_ref_r != NULL )
    delete m_ref_r;

  if( m_rec_expr != NULL )
    m_rec_expr->Detach();

  if( m_lambda_term != NULL )
    m_lambda_term->Detach();

  if ( m_taut != NULL )
    delete m_taut;
}

const USTRING& ProofRule::GetDefIndent() const
{
  return m_def_ident;
}

bool ProofRule::IsBinary() const { return m_ref_l != NULL && m_ref_r != NULL; }
bool ProofRule::IsContinue() const { return m_type == RESWORD_CONTINUED; }
bool ProofRule::IsAxiom() const { return m_type == RESWORD_AXIOM; }

ProofRuleInstance* ProofRule::GetInstance(
  ProofParamArgMap &arg_map,
  ProofRecursionsExpander &rec_expander,
  const ProofLib::SourceLocation &sloc ) const
{
  return new ProofRuleInstance( this, arg_map, rec_expander, sloc );
}

////////////////////////////////////////////////////////////////////////////////

ProofRuleInstance::ProofRuleInstance(
  const ProofRule *org,
  ProofParamArgMap &arg_map_0,
  ProofRecursionsExpander &rec_expander,
  const ProofLib::SourceLocation &sloc )
  : ProofRule( org )
{
  ProofParamArgMapImpl &arg_map =
    dynamic_cast<ProofParamArgMapImpl&>(arg_map_0);

  if( m_seq != NULL )
  {
    m_seq->Apply( arg_map );
    m_seq->Apply( rec_expander );
  }

  if( m_seq_gen != NULL )
  {
    m_seq_gen->Apply( arg_map );
    m_seq_gen->Apply( rec_expander );
  }

  if( m_cut_formula != NULL )
  {
    m_cut_formula = ProofLib::DataNode::ApplyTo( m_cut_formula, arg_map );
    m_cut_formula = ProofLib::DataNode::ApplyTo( m_cut_formula, rec_expander );
  }

  if( m_rec_expr != NULL )
    m_rec_expr = rec_expander.CalcRecursiveExpr( m_rec_expr );

  if( org->GetRefLeft() != NULL )
    m_ref_l = org->GetRefLeft()->GetInstance( arg_map, rec_expander, sloc );
  if( org->GetRefRight() != NULL )
    m_ref_r = org->GetRefRight()->GetInstance( arg_map, rec_expander, sloc );

  if( org->GetTautology() != NULL )
  {
    m_taut->Apply( arg_map );
    m_taut->Apply( rec_expander );
  }
  
}

////////////////////////////////////////////////////////////////////////////////

ProofRef::ProofRef()
{
  m_expr = NULL;
  m_level = 0;
  m_seq = NULL;
  m_taut = NULL;
}

ProofRef::ProofRef( const USTRING &to )
{
  m_to = to;
  m_expr = NULL;
  m_level = 0;
  m_seq = NULL;
  m_taut = NULL;
}

// for ProofRefInstance
ProofRef::ProofRef( const ProofRef *o )
  : ProofLib::SourceLocation( *o ), m_meta_args( o->m_meta_args )
{
  m_to = o->m_to;
  m_expr = o->m_expr;
  if( m_expr != NULL )
    m_expr->Attach();
  m_level = o->m_level;
  if( o->m_seq != NULL )
    m_seq = o->m_seq->Clone();
  else
    m_seq = NULL;
  if( o->m_taut != NULL )
    m_taut = o->m_taut->Clone();
  else
    m_taut = NULL;
}

ProofRef::~ProofRef()
{
  if( m_expr != NULL ) m_expr->Detach();
  if( m_seq != NULL ) delete m_seq;
  if( m_taut != NULL ) delete m_taut;
}

const Proof* ProofRef::GetProof( const SymbolTable *st, ProofTable* pt,
                                 ProofLib::NodeDatabase *ndb ) const
{
  ASSERT( st != NULL );
  ASSERT( pt != NULL );

  if( IsAutoProp() || m_seq != NULL )
    return NULL;

  ASSERT( !GetName().empty() );
  // check if proof exists and retrieve proof symbol table entry
  const ProofSymbol *ps =
    dynamic_cast<const ProofSymbol*>( st->GetSymbol( GetName() ) );

  if( ps == NULL )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException(this);
    e->Stream() << "No proof '" << GetName() << "' found!";
    throw e;
  }

  if( dynamic_cast<const RecursiveProof*>( pt->GetProof( ps ) ) != NULL &&
      m_expr == NULL )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException(this);
    e->Stream() << "Proof '" << GetName() << "' is recursive but "
                   "no recursive expresion found!";
    throw e;
  }

  unsigned int level = 0;

  if( m_expr != NULL )
    level = m_expr->GetValueAtLevel( m_level );

  return pt->GetProof( ps )->GetProofInstance(
    m_meta_args, *this,
    ndb, level );
}

void ProofRef::SetRecursiveExpr( ProofLib::RecursiveExpr* expr )
{
  ASSERT( expr != NULL );
  m_expr = expr;
  m_expr->Attach();
}

void ProofRef::SetExplicitAxiom( ProofLib::Sequent *seq )
{
  ASSERT( m_to.empty() );
  ASSERT( seq != NULL );
  m_seq = seq;
}

void ProofRef::SetTautology( ProofLib::Sequent *seq )
{
  ASSERT( m_to.empty() );
  ASSERT( seq != NULL );
  m_taut = seq;
}

void ProofRef::AddMetaArgument( ProofLib::DbNode *arg )
{
  m_meta_args.push_back( arg );
}

ProofRefInstance* ProofRef::GetInstance(
  ProofParamArgMap &arg_map,
  ProofRecursionsExpander &rec_expander,
  const ProofLib::SourceLocation &sloc ) const
{
  return new ProofRefInstance( this, arg_map, rec_expander, sloc );
}

////////////////////////////////////////////////////////////////////////////////

ProofRefInstance::ProofRefInstance(
  const ProofRef* org, ProofParamArgMap &arg_map_0,
  ProofRecursionsExpander &rec_expander,
  const ProofLib::SourceLocation &sloc )
  : ProofRef( org )
{
  ProofParamArgMapImpl &arg_map =
    dynamic_cast<ProofParamArgMapImpl&>(arg_map_0);

  for( size_t i = 0; i < m_meta_args.size(); i++ )
  {
    ProofLib::Term *t = dynamic_cast<ProofLib::Term*>( m_meta_args[i] );
    ProofLib::Formula *f = dynamic_cast<ProofLib::Formula*>( m_meta_args[i] );

    if( t != NULL )
    {
      t = ProofLib::DataNode::ApplyTo( t, arg_map );
      m_meta_args[i] = ProofLib::DataNode::ApplyTo( t, rec_expander );
    }
    else if( f != NULL )
    {
      f = ProofLib::DataNode::ApplyTo( f, arg_map );
      m_meta_args[i] = ProofLib::DataNode::ApplyTo( f, rec_expander );
    }
    else
      ASSERT( false );
  }
  
  if( m_seq != NULL )
  {
    m_seq->Apply( arg_map );
    m_seq->Apply( rec_expander );
  }

  if ( m_taut != NULL )
  {
    m_taut->Apply( arg_map );
    m_taut->Apply( rec_expander );
  }

  m_level = rec_expander.GetLevel();
}

////////////////////////////////////////////////////////////////////////////////

}

