////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: proofhlk.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _PROOFHLK_H_
#define _PROOFHLK_H_

#include <string>
#include <vector>

#include <util/sloc.h>
#include <term_operations.h>

namespace ProofLib
{
  class DbNode;
  class Formula;
  class Sequent;
  class RecursiveExpr;
}

////////////////////////////////////////////////////////////////////////////////

namespace HLK
{

class Proof;
class RecursiveProof;
class ProofBase;
class ProofSymbol;
class ProofTableStore;
class ProofRule;
class ProofRuleList;
class ProofRef;
class ProofParamArgMap;
class ProofRecursionsExpander;
class ProofRuleInstance;
class ProofRefInstance;

////////////////////////////////////////////////////////////////////////////////

class ProofTable
{
  ProofTableStore       *m_store;

public:
  ProofTable();
  ~ProofTable();

  void AddProof( ProofBase *p );
  const ProofBase* GetProof( const ProofSymbol *sym );
};

////////////////////////////////////////////////////////////////////////////////

class ProofMetaParams
{
  typedef const Symbol* MetaParam;

  std::vector< MetaParam > m_params;

public:
  ProofMetaParams() {}

  void AddSymbol( const Symbol *s );

  size_t GetCount() const { return m_params.size(); }
  
  const Symbol* GetSymbol( size_t index ) const;

  bool IsFormulaAt( size_t index ) const;
  bool IsTermAt( size_t index ) const;
};

class ProofParamArgMap
{
public:
  virtual ~ProofParamArgMap() {}

  virtual size_t GetCount() const = 0;
  virtual ProofLib::DbNode* GetArg( size_t index ) const = 0;
  virtual void AddMapping( const Symbol *s, ProofLib::DbNode *node,
                           ProofLib::TermList &vars ) = 0;
};

////////////////////////////////////////////////////////////////////////////////

typedef std::vector< ProofLib::DbNode* > ProofMetaArguments;

////////////////////////////////////////////////////////////////////////////////

class ProofBody
{
  ProofLib::Sequent *m_end_sequent;
  ProofRuleList     *m_rule_store;

public:
  ProofBody();
  virtual ~ProofBody();

  ProofLib::Sequent* GetEndSequent() const { return m_end_sequent; }
  
  void SetEndSequent( ProofLib::Sequent *seq );
  void AddRule( ProofRule *rule );
  size_t GetRuleCount() const;
  const ProofRule* GetRule( size_t index ) const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofBase : public ProofLib::SourceLocation
{
  const ProofSymbol *m_symbol;

  SymbolTable       *m_symbol_table;
  ProofMetaParams    m_meta_params;

public:
  ProofBase( const ProofSymbol *sym, SymbolTable *symtab,
             const ProofMetaParams *meta_params );
  ProofBase( const ProofBase &o );
  virtual ~ProofBase();

  const ProofSymbol* GetSymbol() const { return m_symbol; }
  
  bool IsMetaProof() const { return GetProofMetaParams().GetCount() > 0; }

protected:
  const ProofMetaParams& GetProofMetaParams() const { return m_meta_params; }
  SymbolTable* GetSymbolTable() const { return m_symbol_table; }

public:
  virtual const Proof* GetProofInstance(
    const ProofMetaArguments &args,
    const ProofLib::SourceLocation &sloc,
    ProofLib::NodeDatabase *ndb,
    unsigned int level ) const = 0;
  virtual void ReleaseProofInstance() = 0;
};

////////////////////////////////////////////////////////////////////////////////

class Proof : public ProofBase, public ProofBody
{
protected:
  Proof( const Proof *p ); // for ProofInstance
  Proof( const RecursiveProof *p ); // for ProofInstance

public:
  Proof( const ProofSymbol *sym, SymbolTable *symtab,
         const ProofMetaParams *meta_params );
  virtual ~Proof();

  virtual const Proof* GetProofInstance( const ProofMetaArguments &args,
                                         const ProofLib::SourceLocation &sloc,
                                         ProofLib::NodeDatabase *ndb,
                                         unsigned int level ) const;
  virtual void ReleaseProofInstance();
};

////////////////////////////////////////////////////////////////////////////////

class RecursiveProof : public ProofBase
{
  std::vector<ProofBody*> m_bodies;
  std::vector<ProofLib::RecursiveExpr*> m_exprs;

protected:
  RecursiveProof( const Proof *p ); // for ProofInstance

public:
  RecursiveProof( const ProofSymbol *sym, SymbolTable *symtab,
                  const ProofMetaParams *meta_params );
  virtual ~RecursiveProof();

  ProofBody* AddLevel( ProofLib::RecursiveExpr *expr );

  ProofBody* GetProofAtLevel( size_t &level );
  const ProofBody* GetProofAtLevel( size_t &level ) const;

  virtual const Proof* GetProofInstance( const ProofMetaArguments &args,
                                         const ProofLib::SourceLocation &sloc,
                                         ProofLib::NodeDatabase *ndb,
                                         unsigned int level ) const;
  virtual void ReleaseProofInstance();
};

////////////////////////////////////////////////////////////////////////////////

class ProofInstance : public Proof
{  
  ProofParamArgMap           *m_arg_map;
  ProofRecursionsExpander    *m_rec_expander;

public:
  ProofInstance( const Proof *p, const ProofMetaArguments &args,
                 const ProofLib::SourceLocation &sloc, ProofLib::NodeDatabase *ndb,
                 unsigned int level );
  ProofInstance( const RecursiveProof *p, const ProofMetaArguments &args,
                 const ProofLib::SourceLocation &sloc, ProofLib::NodeDatabase *ndb,
                 unsigned int level );
  virtual ~ProofInstance();

  virtual void ReleaseProofInstance();

  const ProofParamArgMap* GetProofParamArgMap() const
    { return m_arg_map; }
  const ProofRecursionsExpander* GetProofRecursionsExpander() const
    { return m_rec_expander; }

private:
  void CheckArguments( const ProofMetaArguments &args,
                       const ProofLib::SourceLocation &sloc );

  void InstantiateProofBody( const ProofBody *p, const ProofLib::SourceLocation &sloc );
};

////////////////////////////////////////////////////////////////////////////////

class ProofRule : public ProofLib::SourceLocation
{
  int                 m_type;
  bool                m_bLeft;
  USTRING         m_def_ident;
  const TypeSymbol   *m_ts_l;
  const TypeSymbol   *m_ts_r;

//  const Proof        *m_proof;

//  friend class ProofBody;

protected:
  ProofLib::Sequent  *m_seq;
  ProofLib::Sequent  *m_seq_gen;
  ProofLib::Sequent  *m_taut;
  ProofLib::Formula  *m_cut_formula;
  ProofLib::RecursiveExpr* m_rec_expr;
  ProofLib::LambdaTerm* m_lambda_term;
  const ProofRef     *m_ref_l;
  const ProofRef     *m_ref_r;

protected:
  ProofRule( const ProofRule *pr );

public:
  ProofRule( int type, ProofLib::Sequent *seq = NULL );
  ProofRule( const ProofRef *ref );
  ProofRule( ProofLib::Sequent *seq, ProofLib::Sequent *seq_gen );
  ProofRule( int type, bool left, ProofLib::Sequent *seq,
             const USTRING &def_ident,
             const TypeSymbol *ts_l, const TypeSymbol *ts_r,
             ProofLib::RecursiveExpr* rec_expr,
             ProofLib::LambdaTerm *lambda_term = NULL );
  ProofRule( int type, bool left, ProofLib::Sequent *seq,
             ProofLib::Formula *cut_formula,
             const ProofRef *ref_l, const ProofRef *ref_r );
  ~ProofRule();

  int GetType() const { return m_type; }
  const ProofLib::Sequent* GetSequent() const { return m_seq; }
  const ProofLib::Formula* GetCutFormula() const { return m_cut_formula; }
  const ProofLib::Formula* GetByFormula() const { return m_cut_formula; }
  const ProofLib::LambdaTerm* GetLambdaTerm() const { return m_lambda_term; }
  ProofLib::RecursiveExpr* GetRecursiveExpr() const { return m_rec_expr; }
  const ProofRef* GetRef() const { return m_ref_l; }
  const ProofRef* GetRefLeft() const { return m_ref_l; }
  const ProofRef* GetRefRight() const { return m_ref_r; }
  const USTRING& GetDefIndent() const;
  const TypeSymbol* GetTypeSymbolLeft() const { return m_ts_l; }
  const TypeSymbol* GetTypeSymbolRight() const { return m_ts_r; }
  
  void SetTautology( ProofLib::Sequent *taut ) { m_taut = taut; }
  ProofLib::Sequent* GetTautology() const { return m_taut; }

  bool IsLeftRule() const { return m_bLeft; }

  bool IsBinary() const;
  bool IsContinue() const;
  bool IsAxiom() const;

  ProofRuleInstance* GetInstance( ProofParamArgMap &arg_map,
                                  ProofRecursionsExpander &rec_expander,
                                  const ProofLib::SourceLocation &sloc ) const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofRuleInstance : public ProofRule
{
public:
  ProofRuleInstance( const ProofRule *org,
                     ProofParamArgMap &arg_map,
                     ProofRecursionsExpander &rec_expander,
                     const ProofLib::SourceLocation &sloc );
};

////////////////////////////////////////////////////////////////////////////////

class ProofRef : public ProofLib::SourceLocation
{
  USTRING m_to;

protected:
  ProofMetaArguments m_meta_args;
  ProofLib::RecursiveExpr *m_expr;
  ProofLib::Sequent *m_seq;
  ProofLib::Sequent *m_taut;
  unsigned int m_level;

private:
  const USTRING& GetName() const { return m_to; }

protected:
  ProofRef( const ProofRef *o );

public:
  ProofRef();
  ProofRef( const USTRING& to );
  ~ProofRef();

  const Proof* GetProof( const SymbolTable *st, ProofTable* pt,
                         ProofLib::NodeDatabase *ndb ) const;

  bool IsAutoProp() const { return m_to.empty() && m_seq == NULL; }
  bool IsExplicitAxiom() const { return m_seq != NULL; }

  ProofLib::Sequent* GetExplicitAxiom() const { return m_seq; }
  ProofLib::Sequent* GetTautology() const { return m_taut; }

  void SetRecursiveExpr( ProofLib::RecursiveExpr *expr );
  void SetExplicitAxiom( ProofLib::Sequent *seq );
  void SetTautology( ProofLib::Sequent *seq );
  void AddMetaArgument( ProofLib::DbNode *arg );

  ProofRefInstance* GetInstance( ProofParamArgMap &arg_map,
                                 ProofRecursionsExpander &rec_expander,
                                 const ProofLib::SourceLocation &sloc ) const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofRefInstance : public ProofRef
{
public:
  ProofRefInstance( const ProofRef* org, ProofParamArgMap &arg_map,
                    ProofRecursionsExpander &rec_expander,
                    const ProofLib::SourceLocation &sloc );
};

////////////////////////////////////////////////////////////////////////////////

class ProofRecursionsExpander : public ProofLib::TermOperation
{
  ProofLib::NodeDatabase *m_ndb;
  SymbolTable            *m_symbol_table;
  
  unsigned int m_level;

  /// for better error reporting:
  ProofLib::SourceLocation m_sloc;

public:
  ProofRecursionsExpander(
    ProofLib::NodeDatabase *ndb,
    unsigned int level,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc );
  virtual ~ProofRecursionsExpander() {}

  ProofLib::NodeDatabase* GetNodeDatabase() { return m_ndb; }
  ProofLib::ProofDatabase* GetProofDatabase();
  SymbolTable* GetSymbolTable() const { return m_symbol_table; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  unsigned int GetLevel() const { return m_level; }

  ProofLib::RecursiveExpr* CalcRecursiveExpr( ProofLib::RecursiveExpr* expr );

public:
  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
};

////////////////////////////////////////////////////////////////////////////////

}

#endif
