////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: readhlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <fstream>
#include <stack>
#include <list>
#include <vector>

#include "libproof/util/myassert.h"
#include "libproof/util/exceptions.h"
#include "libproof/chars/latexchars.h"
#include "libproof/libproof.h"
#include "libproof/term_operations.h"

#include "reswordshlk.h"
#include "readhlk.h"
#include "typehlk.h"
#include "symbolhlk.h"
#include "proofhlk.h"
#include "printhlk.h"
#include "recursive.h"

#ifdef DEBUG
//#define DEBUG_PRINT_READ_INFO
#endif

#define COMMENT_START           '#'
#define COMMA                   ','
#define EOC                     ';'
#define POINT                   '.'

#define BO                      '('
#define BC                      ')'
#define BO_2                    '['
#define BC_2                    ']'
#define BO_TERM                 BO
#define BC_TERM                 BC
#define BO_FUNCTION             BO
#define BC_FUNCTION             BC
#define BO_FORMULA              BO
#define BC_FORMULA              BC
#define BO_ATOMF                BO
#define BC_ATOMF                BC
#define BO_PROOF                BO
#define BC_PROOF                BC
#define BO_RECURSIVE_EXPR       BO_2
#define BC_RECURSIVE_EXPR       BC_2

namespace HLK
{

////////////////////////////////////////////////////////////////////////////////
//
// 1. Exceptions
//
////////////////////////////////////////////////////////////////////////////////

class UnexpectedCharException : public ProofLib::SLocException
{
  char                   m_got;
  char                   m_expected;

public:
  UnexpectedCharException( ProofLib::SourceLocation *sloc, char got, char expected )
    : SLocException( sloc )
  {
    m_got = got;
    m_expected = expected;
  }

  UnexpectedCharException( ProofLib::SourceLocation *sloc, char got )
    : SLocException( sloc )
  {
    m_got = got;
    m_expected = '\0';
  }

  virtual void Print( std::ostream &out ) const
  {
    SLocException::Print( out );
    out << "UnexpectedCharException: " << m_got;
    if( m_expected != '\0' )
      out << " expected: " << m_expected;
  }
};

////////////////////////////////////////////////////////////////////////////////

class UnexpectedTokenException : public ProofLib::SLocException
{
  USTRING            m_got;
  USTRING            m_expected;

public:
  UnexpectedTokenException( ProofLib::SourceLocation *sloc,
                            const USTRING &got,
                            const USTRING &expected = USTRING() )
    : SLocException( sloc )
  {
    m_got = got;
    m_expected = expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    SLocException::Print( out );
    out << "UnexpectedTokenException: " << m_got;

    if( !m_expected.empty() )
      out << " expected: " << m_expected;
  }
};

////////////////////////////////////////////////////////////////////////////////

class TypeMismatchException : public ProofLib::SLocException
{
  const TypeSymbol      *m_got;
  const TypeSymbol      *m_expected;

public:
  TypeMismatchException( ProofLib::SourceLocation *sloc,
                         const TypeSymbol* got,
                         const TypeSymbol* expected = NULL )
    : SLocException( sloc )
  {
    m_got = got;
    m_expected = expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    SLocException::Print( out );
    out << "TypeMismatchException: " << m_got->GetName();

    if( m_expected != NULL )
      out << " expected: " << m_expected->GetName();
  }
};

////////////////////////////////////////////////////////////////////////////////

class ArgCountMismatchException : public ProofLib::SLocException
{
  size_t                 m_got;
  size_t                 m_expected;

public:
  ArgCountMismatchException( ProofLib::SourceLocation *sloc, size_t got, size_t expected )
    : SLocException( sloc )
  {
    m_got = got;
    m_expected = expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    SLocException::Print( out );
    out << "ArgCountMismatchException: " << m_got;

    if( m_expected != 0 )
      out << " expected: " << m_expected;
  }
};

////////////////////////////////////////////////////////////////////////////////
//
// Basic Reader
//
////////////////////////////////////////////////////////////////////////////////

/*
  void AssertString( const char *str )
  {
    size_t len = strlen( str );

    char *cmp = new char[len+1];

    ReadString( cmp, len+1 );

    if( strcmp( str, cmp ) != 0 )
      throw new UnexpectedTokenException( m_lineno, m_charno, cmp, str );

    delete cmp;
  }

  void ReadInteger( size_t &i )
  {
    is() >> i;
    m_charno += is().gcount();
  }
*/

class BasicReader : public ProofLib::SourceLocation
{
  std::istream  *m_in;
  bool           m_isopen;
  bool           m_bDeleteIstream;

private:
  std::istream& is() const { return *m_in; };

public:
  BasicReader()
  {
    m_isopen = false;
    m_in = NULL;
    m_bDeleteIstream = false;
  }
  
  //////////////////////////////////////////////////////////////////////////////

  class ReaderState : public ProofLib::SourceLocation
  {
  public:
    std::istream::pos_type pos;
  };

  class FileState : public ProofLib::SourceLocation
  {
  public:
    std::istream  *in;
    bool           isopen;
    bool           bDeleteIstream;
  };
  
private:
  std::stack<FileState> files;

  //////////////////////////////////////////////////////////////////////////////

public:
  void PushFileState()
  {
    FileState state;

    state.in = m_in;
    state.isopen = m_isopen;
    state.bDeleteIstream = m_bDeleteIstream;
    state.SetSourceLocation( *this );

    files.push( state );
  }

  void PopFileState()
  {
    FileState state = files.top();

    m_in = state.in;
    m_isopen = state.isopen;
    m_bDeleteIstream = state.bDeleteIstream;
    SetSourceLocation( state );

    files.pop();
  }

  //////////////////////////////////////////////////////////////////////////////

  ReaderState SetRollbackPoint()
  {
    ReaderState state;

    state.pos = m_in->tellg();
    state.SetSourceLocation( *this );

    return state;
  }

  void Rollback( const ReaderState &state )
  {
    ASSERT( strcmp( state.GetFile(), GetFile() ) == 0 );
    m_in->seekg( state.pos );
    SetSourceLocation( state.GetLine(), state.GetChar() );
  }

  void AssertChar( char c )
  {
    char r;

    if( is().eof() )
      r = '\0';
    else
      is().get( r );

    if( r == '\n' )
      SetSourceLocation( GetLine()+1, 1 );
    else
      SetSourceLocation( GetChar()+1 );

    if( r != c )
      throw new UnexpectedCharException( this, r, c );
  }

  char CurrentChar()
  {
    char c;

    if ( is().eof() ) return '\0';
    if ( is().fail() ) 
      throw new GenericException();

    c = is().peek();

    return c;
  }

  bool IsWhiteSpace( char c )
  {
    return ( c == ' ' || c == '\t' || c == '\n' || c == '\r' );
  }

  bool IsWordEnd( char c )
  {
    if( c == COMMA || c == EOC || c == '\0' || c == BO || c == BC ||
        c == BO_2 || c == BC_2 )
      return true;

    return IsWhiteSpace( c );
  }

  unsigned int ReadPositiveInteger()
  {
    unsigned int i;

    is() >> i;

    SetSourceLocation( GetChar()+is().gcount() );

    return i;
  }

  char ReadChar()
  {
    char c;
    if( is().eof() ) return '\0';

    is().get( c );

    if( c == '\n' )
      SetSourceLocation( GetLine()+1, 0 );
    else
      SetSourceLocation( GetChar()+1 );

    return c;
  }

  void ReadWord( USTRING &str )
  {
    str.clear();

    while( !IsWordEnd( CurrentChar() ) )
    {
      char c = ReadChar();

      str += c;
    }

    str = LatexConvertor.KnownLatexSymbolsToUnicode( str );
  }

  void EatWhiteSpaces()
  {
    while( true )
    {
      if( !IsWhiteSpace( CurrentChar() ) )
        break;

      ReadChar();
    }
  }

  void EatComment()
  {
    if( CurrentChar() != COMMENT_START )
      return;

    char c = ReadChar();

    while( c != '\n' && c != '\0'  )
      c = ReadChar();
  }

  void EatWhiteSpacesAndComments()
  {
    while( true )
    {
      unsigned int l = GetLine();
      unsigned int c = GetChar();

      EatWhiteSpaces();
      EatComment();

      if( l == GetLine() && c == GetChar() )
        return;
    }
  }

  void Open( std::istream &in )
  {
    ASSERT( m_isopen == false );
  
    m_in = &in;
    m_isopen = true;
  }
  
  void Open( const char *sourcename )
  {
    ASSERT( !m_isopen );
  
    m_in = new std::ifstream( sourcename );
    
    if( m_in == NULL || m_in->fail() )
    {
      GenericException *e = new GenericException;
      e->Stream() << "Can not open input file '" << sourcename << "'.";
      throw e;
    }

    SetSourceLocation( sourcename, 1, 1 );
    m_bDeleteIstream = true;
    m_isopen = true;
  }
  

  void SubOpen( const char *sourcename )
  {
    ASSERT( m_isopen );
    PushFileState();

    SetSourceLocation( sourcename, 1, 1 );
    m_in = new std::ifstream( sourcename );
    m_bDeleteIstream = true;

    if( !(*m_in) )
    {
      PopFileState();

      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
      e->Stream() << "Can not open file '" << sourcename << "'.";
      throw e;
    }
  }

  void Close()
  {
    if( m_bDeleteIstream && m_in != NULL )
    {
      delete m_in;
      m_in = NULL;
    }
      
    if( !files.empty() )
      PopFileState();
    else
      m_isopen = false;
  }

  bool IsEOF() const
  {
    if( !m_isopen) return true;

    return m_in->eof();
  }
};

////////////////////////////////////////////////////////////////////////////////
//
//
//
////////////////////////////////////////////////////////////////////////////////

class ReaderImpl : public Reader, public BasicReader
{
  Document   *m_doc;
  SymbolTable *m_symtab;


public:
  ReaderImpl()
  {
    m_doc = NULL;
    m_symtab = NULL;
  }
  
  virtual ~ReaderImpl() {}

  //////////////////////////////////////////////////////////////////////////////

  virtual void Open( std::istream &in )
  {
    BasicReader::Open( in );
  }

  virtual void Open( const char *sourcename )
  {
    BasicReader::Open( sourcename );
  }

  virtual void Close()
  {
    BasicReader::Close();
  }

  //////////////////////////////////////////////////////////////////////////////

  SymbolTable* GetSymbolTable() { return m_symtab; }
  AxiomTable* GetAxiomTable() { return m_doc->GetAxiomTable(); }

  void  SetActiveSymbolTable( SymbolTable *symtab)
  {
    ASSERT( symtab != NULL );
    m_symtab = symtab;
  }

  //////////////////////////////////////////////////////////////////////////////

  void AssertReservedWord( int no )
  {
    USTRING str;

    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    if( ReservedWord( str ) != no )
      throw new UnexpectedTokenException( this, str, reserved_words[no] );
  }

  int ReservedWord( const USTRING &str )
  {
    if( str.empty() )
      return -1;

    std::list<int> possiblities;

    int pos = 0;

    for( int i = 0; reserved_words[i] != NULL; i++ )
    {
      if( reserved_words[i][pos] == str[pos] )
        possiblities.push_back(i);
    }

    while( (unsigned int) ++pos < str.size() )
    {
      std::list<int>::iterator i = possiblities.begin();

      while( i != possiblities.end() )
      {
        if( reserved_words[*i][pos] != str[pos] )
        {
          std::list<int>::iterator j = i;

          i++;
          possiblities.erase( j );
        }
        else
          i++;
      } 
    }

    std::list<int>::const_iterator i;

    for( i = possiblities.begin(); i != possiblities.end(); i++ )
    {
      if( reserved_words[*i][pos] == '\0' )
        return *i;
    }

    return -1;
  }

  int ReadReservedWord()
  {
    std::list<int> possiblities;

    char c = ReadChar();

    for( int i = 0; reserved_words[i] != NULL; i++ )
    {
      if( reserved_words[i][0] == c )
        possiblities.push_back(i);
    }

    int pos = 0;

    while( ( c = ReadChar() ) != '\0' )
    {
      pos++;

      if( IsWhiteSpace(c) )
      {
        std::list<int>::const_iterator i;

        for( i = possiblities.begin(); i != possiblities.end(); i++ )
        {
          if( reserved_words[*i][pos] == '\0' )
            return *i;
        }

        break;
      }
      
      std::list<int>::iterator i = possiblities.begin();

      while( i != possiblities.end() )
      {
        if( reserved_words[*i][pos] != c )
        {
          std::list<int>::iterator j = i;

          i++;
          possiblities.erase( j );
        }
        else
          i++;
      } 
    }

    return -1;
  }

  void ReadIdentifier( USTRING &str )
  {
    ReadWord( str );

    if( str.empty() )
    {
      //throw new UnexpectedCharException( this, CurrentChar() );
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
      e->Stream() << "Unexcepted character "<< CurrentChar()
                  << ", excepted <identifier>";
      throw e;
    }

    if( ReservedWord( str ) != NO_RESWORD )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
      e->Stream() << str << " is a reserved word.";
      throw e;
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  virtual void Read( Document* doc )
  {
    try
    {
      m_doc = doc;
      SetActiveSymbolTable( m_doc->GetSymbolTable() );
    
      while( !IsEOF() )
      {
        EatWhiteSpacesAndComments();
    
        if( IsEOF() )
          break;
    
        USTRING str;
    
        ReadWord( str );
    
        if( str.empty() )
          throw new UnexpectedCharException( this, CurrentChar() );
    
        switch( ReservedWord( str ) )
        {
        case RESWORD_REFERENCE:
          ReadReference();
          break;
        case RESWORD_DEFINE:
          ReadDefine();
          break;
        default:
          throw new UnexpectedTokenException( this, str );
        }
      }
    }
    catch( ProofLib::SLocException *e )
    {
      throw e;
    }
    catch( Exception *e )
    {
      ProofLib::SLocCoverException *ex =
        new ProofLib::SLocCoverException( this, e );
      throw ex;
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////

private:
  std::vector<USTRING> read_files;

protected:
  void ReadReference()
  {
    EatWhiteSpacesAndComments();

    USTRING str;
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    EatWhiteSpacesAndComments();
    AssertChar( EOC );

    bool found = false;
    for( size_t i = 0; i < read_files.size(); i++ )
    {
      if( read_files[i] == str )
      {
        found = true;
        break;
      }
    }

    if( found )
      return;

    read_files.push_back( str );

    SubOpen( str.utf8().c_str() );
    Read( m_doc );
    Close();
  }

  //////////////////////////////////////////////////////////////////////////////

  Document* ReadDefine()
  {
    bool bReadFixWord = false;
    bool bPrefix = true;
    bool bRecursive = false;

    USTRING str;

    EatWhiteSpacesAndComments();

    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    int w = ReservedWord( str );

    if( w == RESWORD_INFIX || w == RESWORD_PREFIX || w == RESWORD_RECURSIVE )
    {
      if( w == RESWORD_INFIX || w == RESWORD_PREFIX )
      {
        if( w == RESWORD_INFIX )
          bPrefix = false;
  
        bReadFixWord = true;
      }
      else if( w == RESWORD_RECURSIVE )
        bRecursive = true;

      EatWhiteSpacesAndComments();

      ReadWord( str );

      if( str.empty() )
        throw new UnexpectedCharException( this, CurrentChar() );

      w = ReservedWord( str );
    }

    switch( w )
    {
    case RESWORD_TYPE:
      if( bReadFixWord || bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );
      ReadTypeDef();
      break;
    case RESWORD_VARIABLE:
      if( bReadFixWord || bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );
      ReadVariableDef();
      break;
    case RESWORD_CONSTANT:
      if( bReadFixWord || bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );
      ReadConstantDef();
      break;
    case RESWORD_AXIOM:
      if( bReadFixWord || bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );
      ReadAxiomDef();
      break;
    case RESWORD_PROOF:
      if( bReadFixWord )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );

      if( bRecursive )
        ReadRecursiveProofDef();
      else
        ReadProofDef();
      break;
    case RESWORD_FUNCTION:
      if( bRecursive )
      {
        if( bReadFixWord )
          throw new
          UnexpectedTokenException( this, reserved_words[w] );

        ReadRecursiveFunctionDef();
      }
      else
        ReadFunctionDef( bPrefix, bReadFixWord );
      break;
    case RESWORD_ATOM:
      if( bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );

      ReadAtomPredicateDef( bPrefix );
      break;
    case RESWORD_PREDICATE:
      if( bRecursive )
      {
        if( bReadFixWord )
          throw new
          UnexpectedTokenException( this, reserved_words[w] );

        ReadRecursivePredicateDef();
      }
      else
        ReadPredicateDef( bPrefix, bReadFixWord );
      break;
    case RESWORD_EQUALITY:
      if( bReadFixWord || bRecursive )
        throw new
        UnexpectedTokenException( this, reserved_words[w] );
      ReadEqualityDef();
      break;
    default:
      throw new UnexpectedTokenException( this, str );
    }

    return NULL;
  }

  void ReadIdentifierList( std::vector<USTRING> &names,
                           std::vector<unsigned int> &cnos,
                           bool allow_resword_any = false,
                           size_t max_count = 0 )
  {
    // just for "backwar" compability!

    std::vector<ProofLib::SourceLocation> slocs;

    ReadIdentifierList( names, slocs, allow_resword_any, max_count );

    for( size_t i = 0; i < slocs.size(); i++ )
      cnos.push_back( slocs[i].GetChar() );
  }

  void ReadIdentifierList( std::vector<USTRING> &names,
                           std::vector<ProofLib::SourceLocation> &slocs,
                           bool allow_resword_any = false,
                           size_t max_count = 0 )
  {
    while( true )
    {
      EatWhiteSpacesAndComments();

      ProofLib::SourceLocation here( *this );

      USTRING str;
      ReadWord( str );

      if( str.empty() )
        throw new UnexpectedCharException( &here, CurrentChar() );

      int resword = ReservedWord( str );
      if( resword != NO_RESWORD &&
          ( !allow_resword_any || resword != RESWORD_ANY ) )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &here );
        e->Stream() << str
                    << " is a reserved word and may not be used as identifier.";
        throw e;
      }

      names.push_back( str );
      slocs.push_back( here );

      if( max_count > 0 && names.size() >= max_count )
        break;

      EatWhiteSpacesAndComments();

      if( CurrentChar() != COMMA )
        break;

      AssertChar( COMMA );
    }
  }

  void ReadTypeList( std::vector<USTRING> &names,
                     std::vector<unsigned int> &cnos,
                     size_t max_count = 0 )
  {
    ReadIdentifierList( names, cnos, true, max_count );
  }

  void ReadAxiomDef()
  {
    USTRING name;

    ProofLib::Sequent *seq = ReadSequent();

    EatWhiteSpacesAndComments();
  
    if( CurrentChar() != EOC )
    {
      AssertReservedWord( RESWORD_NAMED );
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_AS );
      EatWhiteSpacesAndComments();

      ProofLib::SourceLocation before_name( *this );

      ReadWord( name );

      if( name.empty() )
        throw new UnexpectedCharException( this, CurrentChar() );

      if( ReservedWord( name ) != NO_RESWORD )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( this );
        e->Stream() << name << " is a reserved word and may not be used as "
                    << "axiom name.";
        throw e;
      }

      GetSymbolTable()->NewAxiom( name, before_name );

      EatWhiteSpacesAndComments();
    }

    GetAxiomTable()->AddAxiom( seq, ( name.empty() ? NULL : name.c_str() ) );

    AssertChar( EOC );
  }

  void ReadTypeDef()
  {
    USTRING type;
    USTRING fresh;
    //unsigned int cno = GetChar();

    EatWhiteSpacesAndComments();

    ProofLib::SourceLocation here( *this );

    ReadIdentifier( type );
/*
    ReadWord( type );

    if( type.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    if( ReservedWord( type ) != NO_RESWORD )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
      e->Stream() << type << " is a reserved word and may not be used as type.";
      throw e;
    }
*/

    EatWhiteSpacesAndComments();

    if( CurrentChar() != EOC )
    {
      AssertReservedWord( RESWORD_FRESH );
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_VARIABLE );
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_STARTS );
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_WITH );
      EatWhiteSpacesAndComments();

      ReadIdentifier( fresh );
    }

    EatWhiteSpacesAndComments();

    AssertChar( EOC );

    GetSymbolTable()->NewType( type, fresh, here );
  }

  void ReadVariableDef()
  {
    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    AssertReservedWord( RESWORD_OF );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TYPE );

    USTRING type;

    EatWhiteSpacesAndComments();
    ReadWord( type );

    if( type.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    while( !names.empty() )
    {
      GetSymbolTable()->NewVariable(
        names.front().c_str(), type.c_str(),
        true,
        ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) );

      names.erase( names.begin() );
      cnos.erase( cnos.begin() );
    }

    EatWhiteSpacesAndComments();

    AssertChar( EOC );
  }

  void ReadConstantDef()
  {
    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    AssertReservedWord( RESWORD_OF );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TYPE );

    USTRING type;

    EatWhiteSpacesAndComments();
    ReadWord( type );

    if( type.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    while( !names.empty() )
    {
      GetSymbolTable()->NewConstant(
        names.front().c_str(), type.c_str(),
        ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) );

      names.erase( names.begin() );
      cnos.erase( cnos.begin() );
    }

    EatWhiteSpacesAndComments();

    AssertChar( EOC );
  }

  void ReadFunctionDef( bool bPrefix, bool bReadFix )
  {
    std::vector<USTRING> names;
    std::vector<ProofLib::SourceLocation> slocs;
    ProofLib::TermList var_map;

    bool haveFunctionHead = false;

    EatWhiteSpacesAndComments();

    if( (!bReadFix || !bPrefix) && CurrentChar() == BO_FUNCTION )
    {
      AssertChar( BO_FUNCTION );
      EatWhiteSpacesAndComments();

      var_map.AddRight( ReadVariable() );
      EatWhiteSpacesAndComments();

      ProofLib::SourceLocation here( *this );
      USTRING str;
      ReadIdentifier( str );

      names.push_back( str );
      slocs.push_back( here );

      var_map.AddRight( ReadVariable() );
      EatWhiteSpacesAndComments();

      AssertChar( BC_FUNCTION );
      EatWhiteSpacesAndComments();

      haveFunctionHead = true;
      bPrefix = false;
    }
    else
    {
      ReadIdentifierList( names, slocs );
  
      if( (!bReadFix || bPrefix) && CurrentChar() == BO_FUNCTION )
      {
        AssertChar( BO_FUNCTION );
        EatWhiteSpacesAndComments();
  
        ReadSubTermList( var_map );
  
        AssertChar( BC_FUNCTION );
        EatWhiteSpacesAndComments();
  
        haveFunctionHead = true;
        bPrefix = true;
      }
    }

    ProofLib::SourceLocation here(*this);

    USTRING str;
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( &here, CurrentChar() );

    if( ReservedWord( str ) == RESWORD_BY ||
        ReservedWord( str ) == RESWORD_THROUGH )
    {
      if( names.size() > 1 )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &slocs[1] );
        e->Stream() << "Only one identifier allowed.";
        throw e;
      }
    }

    if( ReservedWord( str ) == RESWORD_BY )
    {
      ReadFunctionDefByTerm( bPrefix, names.front().c_str(), slocs.front(),
                             var_map );
      return;
    }

    if( ReservedWord( str ) == RESWORD_THROUGH )
    {
      if( haveFunctionHead )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &slocs[0] );
        e->Stream() << "Function prototype not allowed.";
        throw e;
      }
        
      ReadFunctionDefByFormula( bPrefix, names.front().c_str(), slocs.front() );
      return;
    }

    std::vector<USTRING> types;

    if( haveFunctionHead )
    {
      if( ReservedWord( str ) != RESWORD_TO )
        throw new UnexpectedTokenException( &here, str,
                                            reserved_words[RESWORD_TO] );

      for( size_t i = 0; i < var_map.GetChildCount(); i++ )
      {
        types.push_back(
          FunctionSymbol::GetTermType( var_map.GetTerm(i) )->GetName() );
      }
    }
    else
    {
      if( ReservedWord( str ) != RESWORD_OF )
        throw new UnexpectedTokenException( &here, str,
                                            reserved_words[RESWORD_OF] );

      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_TYPE );
  
      std::vector<unsigned int> t_cnos;
  
      ReadTypeList( types, t_cnos );

      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_TO );
    }

    USTRING type;

    EatWhiteSpacesAndComments();
    here = *this;
    ReadWord( type );

    if( type.empty() )
      throw new UnexpectedCharException( &here, CurrentChar() );

    unsigned int weight = 0;

    if( !bPrefix )
      weight = ReadWeightClausel();

    while( !names.empty() )
    {
      GetSymbolTable()->NewFunction(
        names.front().c_str(), bPrefix, types, type.c_str(),
        weight, slocs.front() );

      names.erase( names.begin() );
      slocs.erase( slocs.begin() );
    }

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  unsigned int ReadWeightClausel()
  {
    EatWhiteSpacesAndComments();

    try
    {
      AssertReservedWord( RESWORD_WITH );
    }
    catch( ProofLib::SLocException *e0 )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( *e0 );

      e->Stream() << "weight clause needed for infix definition.";

      delete e0;

      throw e;
    }
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_WEIGHT );
    EatWhiteSpacesAndComments();
    return ReadPositiveInteger();
  }

  void ReadFunctionDefByTerm( bool bPrefix, const wchar_t *name,
                              const ProofLib::SourceLocation &here,
                              ProofLib::TermList &var_map )
  {
    ProofLib::Term *t = ReadTerm();

#ifdef DEBUG_PRINT_READ_INFO
    std::cerr << "TERM: ";
    Print::PrintTerm( t, std::cerr );
    std::cerr << std::endl;
#endif 

    unsigned int weight = 0;

//    ReadVarMapClause( var_map );

    if( !bPrefix )
      weight = ReadWeightClausel();

    GetSymbolTable()->NewFunction(
      name, bPrefix, t, &var_map, weight, here );

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  ProofLib::LambdaTerm* ReadLambdaTerm()
  {
  	EatWhiteSpacesAndComments();
  	AssertReservedWord( RESWORD_LAMBDA );
  	ProofLib::TermList vars;
  	
  	// read variables, separated by whitespace
  	// until '.' is read
  	// then read formula
  	ReaderState state;
  	try
  	{
  	  while ( true )
  	  {
  	  	state = SetRollbackPoint();
  	    EatWhiteSpacesAndComments();
  	    vars.AddRight( ReadVariable() );
  	  }
  	}
  	catch ( ... )
  	{
  	  Rollback( state );
  	}

    EatWhiteSpacesAndComments();
    AssertChar( POINT );
    
    ProofLib::Formula *f = ReadFormula();
    return ProofLib::LambdaTerm::New( m_doc->GetNodeDatabase(),
                                      vars, f );
  }

#if 0
  void ReadVarMapClause( ProofLib::TermList &var_map )
  {
    ReaderState state = SetRollbackPoint();

    USTRING str;
    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() || ReservedWord( str ) != RESWORD_WITH )
    {
      Rollback( state );
      return;
    }

    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() || ReservedWord( str ) != RESWORD_EXPLICIT )
    {
      Rollback( state );
      return;
    }

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_VARIABLE );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_ORDER );

    ReadSubTermList( var_map );
  }
#endif

  void ReadSubTermList( ProofLib::TermList &var_map )
  {
    while( true )
    {
      ProofLib::Term* t = ReadSubTerm();

      var_map.AddRight( t );

      EatWhiteSpacesAndComments();

      if( CurrentChar() != COMMA )
        break;

      AssertChar( COMMA );
    }
  }

  void ReadFunctionDefByFormula( bool bPrefix, const wchar_t *name,
                                 const ProofLib::SourceLocation &here )
  {
    ProofLib::Formula *through = ReadFormula( false, name );

#ifdef DEBUG_PRINT_READ_INFO
    std::cerr << "FORMULA: ";
    Print::PrintFormula( through, std::cerr );
    std::cerr << std::endl;
#endif 

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_IS );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_DEFINED );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    ProofLib::Formula *by = ReadFormula();

#ifdef DEBUG_PRINT_READ_INFO
    std::cerr << "FORMULA: ";
    Print::PrintFormula( by, std::cerr );
    std::cerr << std::endl;
#endif 

    unsigned int weight = 0;

    if( !bPrefix )
      weight = ReadWeightClausel();

    GetSymbolTable()->NewFunction(
      name, bPrefix, through, by, weight, here );

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  void ReadRecursiveFunctionDef()
  {
    // read the function symbol

    unsigned int cno = GetChar();
    USTRING name, type;

    EatWhiteSpacesAndComments();
    ReadWord( name );

    if( name.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    // read the variable names, if any

    ProofLib::TermList var_map;

    EatWhiteSpacesAndComments();
    if( CurrentChar() == BO_FUNCTION )
    {
      AssertChar( BO_FUNCTION );

      ReadSubTermList( var_map );

      EatWhiteSpacesAndComments();
      AssertChar( BC_FUNCTION );
    }

    // read the return type

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TO );

    EatWhiteSpacesAndComments();
    ReadWord( type );

    if( type.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    // read the recursive levels

    SymbolTable *symtab_above = GetSymbolTable();
    SymbolTable *symtab = symtab_above;

    try
    {
      FunctionSymbol *fs = GetSymbolTable()->NewRecursiveFunction(
        name.c_str(), symtab, type.c_str(), &var_map,
        ProofLib::SourceLocation( GetFile(), GetLine(), cno ) );

      while( true )
      {
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_AT );
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_LEVEL );
  
        EatWhiteSpacesAndComments();
        ProofLib::RecursiveExpr *expr = ReadRecLevelTerm();
  
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_BY );

        symtab = new SymbolTable( symtab, expr->GetVariableName() );
  
        SetActiveSymbolTable( symtab );
  
        ProofLib::Term *t = ReadTerm();

        fs->AddRecursiveLevel( expr, t );
  
        EatWhiteSpacesAndComments();
        if( CurrentChar() == EOC )
          break;
      }

#if 0 //def DEBUG_PRINT_READ_INFO
      std::cerr << "TERM AT LVL " << rec_var << ": ";
      Print::PrintTerm( t_n, std::cerr );
      std::cerr << std::endl;
#endif 

      AssertChar( EOC );

      SetActiveSymbolTable( symtab_above );
    }
    catch( ... )
    {
      if( symtab != NULL ) delete symtab;
      SetActiveSymbolTable( symtab_above );
      throw;
    }
#if 0
    // read the level 0 term

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_AT );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_LEVEL );

    EatWhiteSpacesAndComments();
    AssertChar( '0' );

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    ProofLib::Term *t_0 = ReadTerm();

#if 0 //def DEBUG_PRINT_READ_INFO
    std::cerr << "TERM AT LVL 0: ";
    Print::PrintTerm( t_0, std::cerr );
    std::cerr << std::endl;
#endif 

    // read the level n term

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_AT );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_LEVEL );

    USTRING rec_var;
    EatWhiteSpacesAndComments();
    ReadWord( rec_var );

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    SymbolTable *symtab_above = GetSymbolTable();
    SymbolTable *symtab = new SymbolTable( symtab_above, rec_var.c_str() );

    FunctionSymbol *fs = GetSymbolTable()->NewRecursiveFunction(
      name.c_str(), symtab, type.c_str(), &var_map,
      ProofLib::SourceLocation( GetFile(), GetLine(), cno ) );

    SetActiveSymbolTable( symtab );

    ProofLib::Term *t_n = NULL;

    try
    {
      t_n = ReadTerm();

#if 0 //def DEBUG_PRINT_READ_INFO
      std::cerr << "TERM AT LVL " << rec_var << ": ";
      Print::PrintTerm( t_n, std::cerr );
      std::cerr << std::endl;
#endif 

      EatWhiteSpacesAndComments();
      AssertChar( EOC );

      SetActiveSymbolTable( symtab_above );

      fs->SetRecursiveTerms( t_0, t_n );
    }
    catch( ... )
    {
      if( symtab != NULL ) delete symtab;
      SetActiveSymbolTable( symtab_above );
      throw;
    }
#endif
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadAtomPredicateDef( bool bPrefix )
  {
    EatWhiteSpacesAndComments();
    USTRING str;
    ReadWord( str );
    int w = ReservedWord( str );
    bool bVariable = false;

    if ( w == RESWORD_PREDICATE )
    {
      // constant predicate
    } else if ( w == RESWORD_VARIABLE )
    {
      // variable predicate
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_PREDICATE );
      bVariable = true;
    }

    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    // propositional atoms
    if( CurrentChar() == EOC )
    {
      while( !names.empty() )
      {
        GetSymbolTable()->NewAtomPredicate(
          names.front().c_str(), bPrefix,
          ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ),
	  bVariable );
  
        names.erase( names.begin() );
        cnos.erase( cnos.begin() );
      }

      AssertChar( EOC );

      return;
    }

    AssertReservedWord( RESWORD_OF );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TYPE );

    std::vector<USTRING> types;
    std::vector<unsigned int> t_cnos;

    ReadTypeList( types, t_cnos );

#if 0
    const char **atypes = new const char*[types.size()+1];

    int i = 0;

    while( !types.empty() )
    {
      atypes[i++] = types.front().c_str();
      types.erase( types.begin() );
    }

    atypes[i] = NULL;
#endif

    while( !names.empty() )
    {
      GetSymbolTable()->NewAtomPredicate(
        names.front().c_str(), bPrefix, types, 0,
        ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ), bVariable );

      names.erase( names.begin() );
      cnos.erase( cnos.begin() );
    }

//    delete [] atypes;

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadPredicateDef( bool bPrefix, bool bReadFix )
  {
    std::vector<USTRING> names;
    std::vector<ProofLib::SourceLocation> slocs;
    //std::vector<unsigned int> cnos;
    ProofLib::TermList var_map;

    bool haveFunctionHead = false;

    EatWhiteSpacesAndComments();

    if( (!bReadFix || !bPrefix) && CurrentChar() == BO_FUNCTION )
    {
      AssertChar( BO_FUNCTION );
      EatWhiteSpacesAndComments();

      var_map.AddRight( ReadVariable() );
      EatWhiteSpacesAndComments();

      ProofLib::SourceLocation here( *this );
      USTRING str;
      ReadIdentifier( str );

      names.push_back( str );
      slocs.push_back( here );

      var_map.AddRight( ReadVariable() );
      EatWhiteSpacesAndComments();

      AssertChar( BC_FUNCTION );
      EatWhiteSpacesAndComments();

      haveFunctionHead = true;
      bPrefix = false;
    }
    else
    {
      ReadIdentifierList( names, slocs );
  
      if( (!bReadFix || bPrefix) && CurrentChar() == BO_FUNCTION )
      {
        AssertChar( BO_FUNCTION );
        EatWhiteSpacesAndComments();
  
        ReadSubTermList( var_map );
  
        AssertChar( BC_FUNCTION );
        EatWhiteSpacesAndComments();
  
        haveFunctionHead = true;
        bPrefix = true;
      }
    }

/*
    ReadIdentifierList( names, cnos );
*/

    if( names.size() != 1 )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
      e->Stream() << "Only one identifier allowed for by-clausel.";
      throw e;
    }

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    ProofLib::Formula *f = ReadFormula();

#ifdef DEBUG_PRINT_READ_INFO
    std::cerr << "FORMULA: ";
    Print::PrintFormula( f, std::cerr );
    std::cerr << std::endl;
#endif 

/*
    ProofLib::TermList var_map;

    ReadVarMapClause( var_map );
*/

    GetSymbolTable()->NewPredicate(
      names.front().c_str(), bPrefix, f, &var_map, 0, slocs.front() );
//      ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) );

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadRecursivePredicateDef()
  {
    // read the function symbol

    unsigned int cno = GetChar();
    USTRING name;

    EatWhiteSpacesAndComments();
    ReadWord( name );

    if( name.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    // read the variable names, if any

    ProofLib::TermList var_map;

    EatWhiteSpacesAndComments();
    if( CurrentChar() == BO_FUNCTION )
    {
      AssertChar( BO_FUNCTION );

      ReadSubTermList( var_map );

      EatWhiteSpacesAndComments();
      AssertChar( BC_FUNCTION );
    }

    // read the recursive levels

    SymbolTable *symtab_above = GetSymbolTable();
    SymbolTable *symtab = symtab_above;

    try
    {
      PredicateSymbol *ps = GetSymbolTable()->NewRecursivePredicate(
        name.c_str(), symtab, &var_map,
        ProofLib::SourceLocation( GetFile(), GetLine(), cno ) );

      while( true )
      {
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_AT );
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_LEVEL );
  
        EatWhiteSpacesAndComments();
        ProofLib::RecursiveExpr *expr = ReadRecLevelTerm();
  
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_BY );

        symtab = new SymbolTable( symtab, expr->GetVariableName() );
  
        SetActiveSymbolTable( symtab );
  
        ProofLib::Formula *f = ReadFormula();

        ps->AddRecursiveLevel( expr, f );
  
        EatWhiteSpacesAndComments();
        if( CurrentChar() == EOC )
          break;
      }

#if 0 //def DEBUG_PRINT_READ_INFO
      std::cerr << "TERM AT LVL " << rec_var << ": ";
      Print::PrintTerm( t_n, std::cerr );
      std::cerr << std::endl;
#endif 

      AssertChar( EOC );

      SetActiveSymbolTable( symtab_above );
    }
    catch( ... )
    {
      if( symtab != NULL ) delete symtab;
      SetActiveSymbolTable( symtab_above );
      throw;
    }

#if 0
    // read the level 0 term

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_AT );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_LEVEL );

    EatWhiteSpacesAndComments();
    AssertChar( '0' );

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    ProofLib::Formula *f_0 = ReadFormula();

#if 0 //def DEBUG_PRINT_READ_INFO
    std::cerr << "FORM AT LVL 0: ";
    Print::PrintFormula( f_0, std::cerr );
    std::cerr << std::endl;
#endif 

    // read the level n term

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_AT );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_LEVEL );

    USTRING rec_var;
    EatWhiteSpacesAndComments();
    ReadWord( rec_var );

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_BY );

    SymbolTable *symtab_above = GetSymbolTable();
    SymbolTable *symtab = new SymbolTable( symtab_above, rec_var.c_str() );

    PredicateSymbol *ps = GetSymbolTable()->NewRecursivePredicate(
      name.c_str(), symtab, &var_map,
      ProofLib::SourceLocation( GetFile(), GetLine(), cno ) );

    SetActiveSymbolTable( symtab );

    ProofLib::Formula *f_n = NULL;

    try
    {
      f_n = ReadFormula();

#if 0 //def DEBUG_PRINT_READ_INFO
      std::cerr << "FORM AT LVL " << rec_var << ": ";
      Print::PrintFormula( f_n, std::cerr );
      std::cerr << std::endl;
#endif 

      EatWhiteSpacesAndComments();
      AssertChar( EOC );

      SetActiveSymbolTable( symtab_above );

      ps->SetRecursiveFormulas( f_0, f_n );
    }
    catch( ... )
    {
      if( symtab != NULL ) delete symtab;
      SetActiveSymbolTable( symtab_above );
      throw;
    }
#endif
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadEqualityDef()
  {
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_FOR );

    ProofLib::Variable *v[2];

    for( size_t i = 0; i < 2; i++ )
    {
      ProofLib::SourceLocation before_term( *this );
      ProofLib::Term *t = ReadSubTerm();

      v[i] = dynamic_cast<ProofLib::Variable*>( t );

      if( v[i] == NULL )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_term );
        e->Stream() << "Variable expected, but read term '";
        Print::PrintTerm( t, e->Stream() );
        e->Stream() << "'.";
        throw e;
      }


      const VariableSymbol *vs =
        dynamic_cast<const VariableSymbol*>( v[i]->GetSymbol() );

      if( vs->GetType()->IsAnyType() )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_term );
        e->Stream() << "Variable '";
        Print::PrintTerm( v[i], e->Stream() );
        e->Stream() << "' of type any is not allowed here.";
        throw e;
      }

      if( i == 0 )
      {
        EatWhiteSpacesAndComments();
        AssertChar('=');
      }
    }

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_AS );

    ProofLib::SourceLocation sloc( GetFile(), GetLine(), 0 );

    ProofLib::Formula *f = ReadFormula();

    EatWhiteSpacesAndComments();
    AssertChar( EOC );

    GetSymbolTable()->AddEqualityDefinition( v[0], v[1], f, sloc );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadProofDef()
  {
    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    if( names.size() != 1 )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
      e->Stream() << "Only one identifier allowed for proofs.";
      throw e;
    }

    SymbolTable *symtab_above = GetSymbolTable();
    ASSERT( symtab_above != NULL );

    ProofMetaParams meta_params;
    SymbolTable *symtab = ReadMetaVarDecls( meta_params );

    Proof *p = NULL;

    try
    {
      GetSymbolTable()->NewProof(
        names.front().c_str(),
        ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) );
  
      const ProofSymbol *ps = dynamic_cast<const ProofSymbol*>
        ( GetSymbolTable()->GetSymbol( names.front().c_str() ) );
      ASSERT( ps != NULL );
  
      if( symtab != NULL )
        SetActiveSymbolTable( symtab );

      p = new Proof( ps, GetSymbolTable(), &meta_params );

      p->SetSourceLocation( *this );

      ReadProofBody( p );

#if 0
      // begin "proves" clause
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_PROVES );
  
      ProofLib::Sequent *seq = ReadSequent();
  
      EatWhiteSpacesAndComments();
      AssertChar( EOC );
      // end "proves" clause
  
      p->SetEndSequent( seq );

      p->SetSourceLocation( *this );

      ProofRule *last_rule = NULL;

      while( true )
      {
        EatWhiteSpacesAndComments(); // to get right line # in next line
        unsigned int hl = GetLine();

        ProofRule *rule = ReadRule();

        if( rule == NULL )
          break;

        if( last_rule != NULL &&
            ( last_rule->IsBinary() || last_rule->IsContinue() ||
              last_rule->IsAxiom() ) )
        {
          ProofLib::GenericSLocException *e =
            new ProofLib::GenericSLocException( this );
  
          e->Stream() << "After a binary, continue and axiom rule "
                         "no futher rule may be added.";
  
          throw e;
        }

        rule->SetSourceLocation( GetFile(), hl, 0 );
        p->AddRule( rule );

        last_rule = rule;
      }
#endif
    }
    catch( ... )
    {
      SetActiveSymbolTable( symtab_above );
      //if( p != NULL ) delete p;
      throw;
    }

    EatWhiteSpacesAndComments();
    AssertChar( EOC );

    SetActiveSymbolTable( symtab_above );

    m_doc->GetProofTable()->AddProof( p );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadRecursiveProofDef()
  {
    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    if( names.size() != 1 )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
      e->Stream() << "Only one identifier allowed for proofs.";
      throw e;
    }

    SymbolTable *symtab_above = GetSymbolTable();

    ProofMetaParams meta_params;
    SymbolTable *symtab = ReadMetaVarDecls( meta_params );

    RecursiveProof *p = NULL;

    try
    {
      GetSymbolTable()->NewProof(
        names.front().c_str(),
        ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) );
  
      const ProofSymbol *ps = dynamic_cast<const ProofSymbol*>
        ( GetSymbolTable()->GetSymbol( names.front().c_str() ) );
      ASSERT( ps != NULL );
  
      if( symtab != NULL )
        SetActiveSymbolTable( symtab );
      else
        symtab = symtab_above;

      p = new RecursiveProof( ps, GetSymbolTable(), &meta_params );

      p->SetSourceLocation( *this );

      // read levels
      while( true )
      {
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_AT );
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_LEVEL );

        EatWhiteSpacesAndComments();
        ProofLib::RecursiveExpr *expr = ReadRecLevelTerm();

        ProofBody* body = p->AddLevel( expr );

        symtab = new SymbolTable( symtab, expr->GetVariableName() );
  
        SetActiveSymbolTable( symtab );
  
        ReadProofBody( body );

        EatWhiteSpacesAndComments();
        if( CurrentChar() == EOC )
          break;
      }
    }
    catch( ... )
    {
      SetActiveSymbolTable( symtab_above );
      if( p != NULL ) delete p;
      throw;
    }

    EatWhiteSpacesAndComments();
    AssertChar( EOC );

    SetActiveSymbolTable( symtab_above );

    m_doc->GetProofTable()->AddProof( p );
  }

  //////////////////////////////////////////////////////////////////////////////

  SymbolTable* ReadMetaVarDecls( ProofMetaParams &meta_params )
  {
    SymbolTable* symtab = NULL;

    try
    {
      while( true )
      {
        ReaderState state = SetRollbackPoint();
  
        USTRING str;
  
        EatWhiteSpacesAndComments();
        ReadWord( str );
  
        if( str.empty() || ReservedWord( str ) != RESWORD_WITH )
        {
          Rollback( state );
          break;
        }
  
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_META );
  
        EatWhiteSpacesAndComments();
        ReadWord( str );
  
        if( str.empty() )
          throw new UnexpectedCharException( this, CurrentChar() );
  
        bool bReadFixWord = false;
        bool bPrefix = true;
  
        int w = ReservedWord( str );
    
        if( w == RESWORD_INFIX || w == RESWORD_PREFIX )
        {
          if( w == RESWORD_INFIX )
            bPrefix = false;
    
          bReadFixWord = true;
    
          EatWhiteSpacesAndComments();
    
          ReadWord( str );
    
          if( str.empty() )
            throw new UnexpectedCharException( this, CurrentChar() );
    
          w = ReservedWord( str );
        }
  
        if( symtab == NULL )
          symtab = new SymbolTable( GetSymbolTable() );
    
        switch( w )
        {
        case RESWORD_TERM:
          ReadMetaTermDef( bPrefix, bReadFixWord, symtab, meta_params );
          break;
        case RESWORD_FORMULA:
          ReadMetaFormulaDef( bPrefix, symtab, meta_params );
          break;
        default:
          throw new UnexpectedTokenException( this, str );
        }
      }
    }
    catch( ... )
    {
      if( symtab != NULL ) delete symtab;
      throw;
    }

    return symtab;
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadMetaTermDef( bool bPrefix, bool bReadFixWord, SymbolTable* symtab,
                        ProofMetaParams &meta_params )
  {
    std::vector<USTRING> names;
    std::vector<ProofLib::SourceLocation> slocs;

    ReadIdentifierList( names, slocs );

    USTRING str;

    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    if( ReservedWord( str ) == RESWORD_BY )
    {
      if( names.size() != 1 )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( this );
        e->Stream() << "Only one identifier allowed for by-clausel.";
        throw e;
      }

      ProofLib::TermList var_map;
      ReadFunctionDefByTerm( bPrefix, names.front().c_str(), slocs.front(),
                             var_map );

      return;
    }

    if( ReservedWord( str ) == RESWORD_THROUGH )
    {
      if( names.size() != 1 )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( this );
        e->Stream() << "Only one identifier allowed for through-clausel.";
        throw e;
      }

      ReadFunctionDefByFormula( bPrefix, names.front().c_str(), slocs.front() );

      return;
    }

    if( ReservedWord( str ) != RESWORD_OF )
      throw new UnexpectedTokenException( this, str,
                                          reserved_words[RESWORD_OF] );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TYPE );

    std::vector<USTRING> types;
    std::vector<unsigned int> t_cnos;

    ReadTypeList( types, t_cnos );

    if( CurrentChar() == EOC )
    {
      if( bReadFixWord )
        throw new UnexpectedTokenException( this,
          reserved_words[( bPrefix ? RESWORD_PREFIX : RESWORD_INFIX )] );

      if( types.size() != 1 )
      {
        t_cnos.erase( t_cnos.begin() );
        ProofLib::SourceLocation sloc( GetFile(), GetLine(), t_cnos.front() );
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Expected '" << reserved_words << "' keyword.";
        throw e;
      }

      while( !names.empty() )
      {
        meta_params.AddSymbol(
          symtab->NewVariable(
            names.front().c_str(), types.front().c_str(),
            false, slocs.front() ) );
  
        names.erase( names.begin() );
        slocs.erase( slocs.begin() );
      }
    }
    else
    {
      AssertReservedWord( RESWORD_TO );
  
      USTRING type;
  
      EatWhiteSpacesAndComments();
      ReadWord( type );
  
      if( type.empty() )
        throw new UnexpectedCharException( this, CurrentChar() );
      unsigned int weight = 0;
  
      if( !bPrefix )
        weight = ReadWeightClausel();
  
#if 0
      const char **atypes = new const char*[types.size()+1];
  
      int i = 0;
  
      while( !types.empty() )
      {
        atypes[i++] = types.front().c_str();
        types.erase( types.begin() );
      }
  
      atypes[i] = NULL;
#endif
  
      while( !names.empty() )
      {
        meta_params.AddSymbol(
          symtab->NewFunction(
            names.front().c_str(), bPrefix, types, type.c_str(),
            weight, slocs.front() ) );
  
        names.erase( names.begin() );
        slocs.erase( slocs.begin() );
      }
  
//      delete [] atypes;
    }

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadMetaFormulaDef( bool bPrefix, SymbolTable *symtab,
                           ProofMetaParams &meta_params )
  {
    std::vector<USTRING> names;
    std::vector<unsigned int> cnos;

    ReadIdentifierList( names, cnos );

    // propositional atoms
    if( CurrentChar() == EOC )
    {
      while( !names.empty() )
      {
        meta_params.AddSymbol(
          symtab->NewAtomPredicate(
            names.front().c_str(), bPrefix,
            ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) ) );
  
        names.erase( names.begin() );
        cnos.erase( cnos.begin() );
      }

      AssertChar( EOC );

      return;
    }

    AssertReservedWord( RESWORD_OF );
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_TYPE );

    std::vector<USTRING> types;
    std::vector<unsigned int> t_cnos;

    ReadTypeList( types, t_cnos );

#if 0
    const char **atypes = new const char*[types.size()+1];

    int i = 0;

    while( !types.empty() )
    {
      atypes[i++] = types.front().c_str();
      types.erase( types.begin() );
    }

    atypes[i] = NULL;
#endif

    while( !names.empty() )
    {
      meta_params.AddSymbol(
        symtab->NewAtomPredicate(
          names.front().c_str(), bPrefix, types, 0,
          ProofLib::SourceLocation( GetFile(), GetLine(), cnos.front() ) ) );

      names.erase( names.begin() );
      cnos.erase( cnos.begin() );
    }

//    delete [] atypes;

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
  }

  //////////////////////////////////////////////////////////////////////////////

  void ReadProofBody( ProofBody *p )
  {
    // begin "proves" clause
    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_PROVES );

    ProofLib::Sequent *seq = ReadSequent();

    EatWhiteSpacesAndComments();
    AssertChar( EOC );
    // end "proves" clause

    p->SetEndSequent( seq );

    ProofRule *last_rule = NULL;

    while( true )
    {
      EatWhiteSpacesAndComments(); // to get right line # in next line
      unsigned int hl = GetLine();

      ProofRule *rule = ReadRule();

      if( rule == NULL )
        break;

      if( last_rule != NULL &&
          ( last_rule->IsBinary() || last_rule->IsContinue() ||
            last_rule->IsAxiom() ) )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( this );

        e->Stream() << "After a binary, continue and axiom rule "
                       "no futher rule may be added.";

        throw e;
      }

      rule->SetSourceLocation( GetFile(), hl, 0 );
      p->AddRule( rule );

      last_rule = rule;
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofRule* ReadRule()
  {
    USTRING str;

    EatWhiteSpacesAndComments();

    if( CurrentChar() == EOC )
      return NULL;

    ReaderState state = SetRollbackPoint();
    ReadWord( str );

    if( ReservedWord( str ) == RESWORD_AT )
    {
      Rollback( state );
      return NULL;
    }

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    switch( ReservedWord( str ) )
    {
    case RESWORD_WITH:
    {
      EatWhiteSpacesAndComments();
      ReadWord( str );

      if( str.empty() )
        throw new UnexpectedCharException( this, CurrentChar() );

      int rule_word = ReservedWord( str );

      bool bLeft = true;
      bool bBinary = false;
      USTRING def_ident;
      const TypeSymbol *ts_l = NULL;
      const TypeSymbol *ts_r = NULL;
      ProofLib::Formula *cut_formula = NULL;
      ProofLib::RecursiveExpr* rec_expr = NULL;
      ProofLib::LambdaTerm *lambda_term = NULL;

      switch( rule_word )
      {
      case RESWORD_NOT:
      case RESWORD_ALL:
      case RESWORD_EX:
      case RESWORD_AND:
      case RESWORD_OR:
      case RESWORD_IMPL:
      case RESWORD_ALL2:
      case RESWORD_EX2:
        EatWhiteSpacesAndComments();
        ReadWord( str );

        if( str.empty() )
          throw new UnexpectedCharException( this, CurrentChar() );

        switch( ReservedWord( str ) )
        {
        case RESWORD_RIGHT:
          bLeft = false;
          if ( rule_word == RESWORD_EX2 )
          {
          	// read optional lambda term ?
          	ReaderState s = SetRollbackPoint();
          	EatWhiteSpacesAndComments();
          	ReadWord( str );
          	if ( ReservedWord( str ) == RESWORD_WITH )
          	{ 
          	  lambda_term = ReadLambdaTerm();
          	} else
          	  Rollback( s );
          }
          break;
        case RESWORD_LEFT:
          if ( rule_word == RESWORD_ALL2 )
          {
          	// read optional lambda term ?
          	ReaderState s = SetRollbackPoint();
          	EatWhiteSpacesAndComments();
          	ReadWord( str );
          	if ( ReservedWord( str ) == RESWORD_WITH )
          	{ 
          	  lambda_term = ReadLambdaTerm();
          	} else
          	  Rollback( s );
          }
          break;
        default:
          throw new UnexpectedTokenException( this, str );
        }

        switch( rule_word )
        {
        case RESWORD_AND:
          bBinary = !bLeft;
          break;
        case RESWORD_OR:
        case RESWORD_IMPL:
          bBinary = bLeft;
          break;
        }
        break;
//    case RESWORD_DEF:
      case RESWORD_UNDEF:
        EatWhiteSpacesAndComments();
        ReadWord( def_ident );

        if( ReservedWord( def_ident ) == RESWORD_EQUALITY )
        {
          def_ident.clear();

          EatWhiteSpacesAndComments();
          AssertReservedWord( RESWORD_OF );
          EatWhiteSpacesAndComments();
          AssertReservedWord( RESWORD_TYPE );

          std::vector<USTRING> types;
          std::vector<unsigned int> t_cnos;

          ReadTypeList( types, t_cnos, 2 );

          ts_l = dynamic_cast<const TypeSymbol*>(
            GetSymbolTable()->GetSymbol( types.front().c_str() ) );

          if( ts_l == NULL )
            throw new NoSuchTypeException( types.front().c_str(),
              ProofLib::SourceLocation( GetFile(), GetLine(), t_cnos.front() ) );

          types.erase( types.begin() );
          t_cnos.erase( t_cnos.begin() );

          ts_r = dynamic_cast<const TypeSymbol*>(
            GetSymbolTable()->GetSymbol( types.front().c_str() ) );

          if( ts_r == NULL )
            throw new NoSuchTypeException( types.front().c_str(),
              ProofLib::SourceLocation( GetFile(), GetLine(), t_cnos.front() ) );
        }
        else if( def_ident.empty() )
            throw new UnexpectedCharException( this, CurrentChar() );
        else if( def_ident ==
                 GetSymbolTable()->GetEqualPredicateSymbol()->GetName() )
        {
          ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
          e->Stream() <<
            "Equality must be undefined by 'undef equality' expression.";
          throw e;
        }
        else
        {
          EatWhiteSpacesAndComments();

          if( CurrentChar() == BO_RECURSIVE_EXPR )
          {
            AssertChar( BO_RECURSIVE_EXPR );

            rec_expr = ReadRecursiveExpression();

            EatWhiteSpacesAndComments();
            AssertChar( BC_RECURSIVE_EXPR );
          }
          
          EatWhiteSpacesAndComments();
        }
        break;
      case RESWORD_CUT:
        bBinary = true;
        cut_formula = ReadFormula();
        break;
      case RESWORD_PARAMOD:
        bBinary = true;
        break;
      case RESWORD_AUTO:
        EatWhiteSpacesAndComments();
        AssertReservedWord( RESWORD_PROPOSITIONAL );
        break;
      default:
        throw new UnexpectedTokenException( this, str );
      }
  
      if( bBinary )
      {
        // process binary rule
        EatWhiteSpacesAndComments();
        ReaderState state = SetRollbackPoint();
        ReadWord( str );

        ProofRef *ref_l = NULL;
        ProofRef *ref_r = NULL;
        ProofLib::Sequent *seq = NULL;
        ProofLib::Formula *by = NULL;

// new way
        try
        {
          if( rule_word == RESWORD_PARAMOD &&
              ReservedWord( str ) == RESWORD_BY )
          {
            by = ReadFormula();
  
            EatWhiteSpacesAndComments();
            state = SetRollbackPoint();
            ReadWord( str );
          }
  
          bool bNeedToReadLeftOrRight = false;
          bool bNeedToReadLeft = false;
          bool bNeedToReadRight = false;

          if( by != NULL )
          {
            if( ReservedWord( str ) != RESWORD_RIGHT )
              throw new UnexpectedTokenException( this, str,
                reserved_words[RESWORD_RIGHT]);

            seq = ReadSequent();
          }
          else if( ReservedWord( str ) == RESWORD_LEFT )
          {
            ref_l = ReadProofRef();

            bNeedToReadRight = true;
          }
          else if( ReservedWord( str ) == RESWORD_RIGHT )
          {
            ref_r = ReadProofRef();

            bNeedToReadLeft = true;
          }
          else
          {
            if( cut_formula != NULL )
            {
              if( str.empty() )
                throw new UnexpectedCharException( this, CurrentChar() );

              throw new UnexpectedTokenException( this, str );
            }
              
            Rollback( state );
            seq = ReadSequent();

            bNeedToReadLeftOrRight = true;
          }
  
          if( bNeedToReadLeftOrRight || bNeedToReadLeft || bNeedToReadRight )
          {
            EatWhiteSpacesAndComments();
            ReadWord( str );
    
            if( cut_formula == NULL || !str.empty() )
            {
              if( str.empty() )
                throw new UnexpectedCharException( this, CurrentChar() );
  
              if( (bNeedToReadLeftOrRight || bNeedToReadLeft) &&
                  ReservedWord( str ) == RESWORD_LEFT )
                ref_l = ReadProofRef();
              else if( (bNeedToReadLeftOrRight || bNeedToReadRight) &&
                       ReservedWord( str ) == RESWORD_RIGHT )
                ref_r = ReadProofRef();
  
              if( bNeedToReadLeftOrRight && ref_r == NULL && ref_l == NULL )
                throw new UnexpectedTokenException( this, str );
              if( bNeedToReadLeft && ref_l == NULL )
                throw new UnexpectedTokenException( this, str,
                  reserved_words[RESWORD_LEFT] );
              if( bNeedToReadRight && ref_r == NULL )
                throw new UnexpectedTokenException( this, str,
                  reserved_words[RESWORD_RIGHT] );
            }
          }
        }
        catch( ... )
        {
          if( ref_l != NULL ) delete ref_l;
          if( seq != NULL ) delete seq;
          if( ref_r != NULL ) delete ref_r;
          if( by != NULL ) delete by;
          throw;
        }
        
        EatWhiteSpacesAndComments();
        AssertChar( EOC );
#if 0
        if( str.empty() )
          throw new UnexpectedCharException( this, CurrentChar() );

        if( ReservedWord( str ) == RESWORD_LEFT )
          ref_l = ReadProofRef();
        else if( ReservedWord( str ) == RESWORD_RIGHT )
          ref_r = ReadProofRef();
        else if( rule_word == RESWORD_PARAMOD &&
                 ReservedWord( str ) == RESWORD_BY )
          by = ReadFormula();
        else
          throw new UnexpectedTokenException( this, str );

        try
        {
          EatWhiteSpacesAndComments();

          if( by == NULL && ref_r == NULL )
          {
            ReaderState state = SetRollbackPoint();
            ReadWord( str );

            if( str.empty() )
              throw new UnexpectedCharException( this, CurrentChar() );

            if( ReservedWord( str ) != RESWORD_RIGHT )
            {
              Rollback( state );
            }
            else
            {
              ref_r = ReadProofRef();
      
              EatWhiteSpacesAndComments();
            }
          }

          if( ref_l == NULL || ref_r == NULL )
          {
            EatWhiteSpacesAndComments();
            seq = ReadSequent();
          }

          AssertChar( EOC );
        }
        catch( ... )
        {
          if( ref_l != NULL ) delete ref_l;
          if( seq != NULL ) delete seq;
          if( ref_r != NULL ) delete ref_r;
          throw;
        }
#endif
        if( cut_formula != NULL && ( ref_l == NULL || ref_r == NULL ) )
        {
          seq = new ProofLib::Sequent( NULL, NULL );

          if( ref_l == NULL )
            seq->AddRight( cut_formula );
          else
            seq->AddLeft( cut_formula );
        }

        ASSERT( by == NULL || cut_formula == NULL );

        if( by != NULL )
          cut_formula = by;

        return new ProofRule( rule_word, bLeft, seq,
                              cut_formula, ref_l, ref_r );
      }

      // read sequent of unary rule
      ProofLib::Sequent *seq = ReadSequent();

      try
      {
        EatWhiteSpacesAndComments();
        AssertChar( EOC );
      }
      catch( ... )
      {
        delete seq;
        throw;
      }

      return new ProofRule( rule_word, bLeft, seq, def_ident, ts_l, ts_r,
                            rec_expr, lambda_term );
    }
    case RESWORD_AUTO:
    {
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_PROPOSITIONAL );

      EatWhiteSpacesAndComments();
      ProofLib::Sequent *seq = ReadSequent();
      ProofRule *rule = new ProofRule( RESWORD_AUTO );
      rule->SetTautology( seq );

      EatWhiteSpacesAndComments();
      AssertChar( EOC );

      return rule;
    }
    case RESWORD_CONTINUED:
    {
      ProofRef *ref = ReadProofRef();

      try
      {
        EatWhiteSpacesAndComments();
        AssertChar( EOC );
      }
      catch( ... )
      {
        delete ref;
        throw;
      }

      return new ProofRule( ref );
    }
    case RESWORD_EXPLICIT:
    {
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_AXIOM );

      // read axiom sequent
      ProofLib::Sequent *seq = ReadSequent();
      ProofLib::Sequent *seq_gen = NULL;

      try
      {
#if 0
        if( CurrentChar() == EOC )
#endif
        {
          EatWhiteSpacesAndComments();
          AssertChar( EOC );
        }
#if 0
        else
        {
          EatWhiteSpacesAndComments();
          AssertReservedWord( RESWORD_INSTANCE );
          EatWhiteSpacesAndComments();
          AssertReservedWord( RESWORD_OF );

          seq_gen = ReadSequent();

          EatWhiteSpacesAndComments();
          AssertChar( EOC );
        }
#endif
      }
      catch( ... )
      {
        delete seq;
        if( seq_gen != NULL ) delete seq_gen;
        throw;
      }

      return new ProofRule( seq, seq_gen );
    }
    default:
      throw new UnexpectedTokenException( this, str );
    }

    return NULL;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofRef* ReadProofRef()
  {
    USTRING str;

    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    switch( ReservedWord( str ) )
    {
    case RESWORD_BY:
      break;
    case RESWORD_AUTO:
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_PROPOSITIONAL );
      
      {
        EatWhiteSpacesAndComments();
        ProofLib::Sequent *seq = ReadSequent();
        ProofRef *ref = new ProofRef;
        ref->SetSourceLocation( *this );
	ref->SetTautology( seq );
        return ref;
      }
    case RESWORD_EXPLICIT:
      EatWhiteSpacesAndComments();
      AssertReservedWord( RESWORD_AXIOM );
      
      {
        EatWhiteSpacesAndComments();
        ProofLib::Sequent *seq = ReadSequent();
        ProofRef *ref = new ProofRef;
        ref->SetSourceLocation( *this );
        ref->SetExplicitAxiom( seq );
        return ref;
      }
    default:
      throw new UnexpectedTokenException( this, str );
    }

    EatWhiteSpacesAndComments();
    AssertReservedWord( RESWORD_PROOF );

    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    ProofRef *ref = new ProofRef( str.c_str() );
    ref->SetSourceLocation( *this );

    EatWhiteSpacesAndComments();

    if( CurrentChar() == BO_RECURSIVE_EXPR )
    {
      AssertChar( BO_RECURSIVE_EXPR );

      ref->SetRecursiveExpr( ReadRecursiveExpression() );

      EatWhiteSpacesAndComments();
      AssertChar( BC_RECURSIVE_EXPR );
      EatWhiteSpacesAndComments();
    }

    if( CurrentChar() == BO_PROOF )
    {
      AssertChar( BO_PROOF );

      while( true )
      {
        ReaderState state = SetRollbackPoint();
        
        ProofLib::DbNode *node = NULL;

        try
        {
          node = ReadTerm();
        }
        catch( Exception *e )
        {
          delete e; // ignore it
          Rollback( state );
        }

        if( node == NULL )
          node = ReadFormula();

        ref->AddMetaArgument( node );

        EatWhiteSpacesAndComments();

        if( CurrentChar() != COMMA )
          break;

        AssertChar( COMMA );
      }

      AssertChar( BC_PROOF );
    }

    return ref;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Term* ReadTerm( unsigned int weight = 0,
                            const wchar_t *deffunc = NULL,
                            const wchar_t *deftype = NULL )
  {
    ASSERT( deffunc == NULL || deftype != NULL );

    EatWhiteSpacesAndComments();

    ProofLib::Term* t = NULL;

    // read left side
    if( CurrentChar() == BO_TERM )
    {
      AssertChar( BO_TERM );
      t = ReadTerm( 0, deffunc, deftype );
      EatWhiteSpacesAndComments();
      AssertChar( BC_TERM );
    }
    else
      t = ReadSubTerm( deffunc, deftype );

    while( true )
    {
      USTRING str;
  
      EatWhiteSpacesAndComments();

      ReaderState state = SetRollbackPoint();
  
      ReadWord( str );
  
      const Symbol *s = GetSymbolTable()->GetSymbol( str.c_str() );
      const FunctionSymbol* f = dynamic_cast<const FunctionSymbol*>( s );
  
      bool bIsDefFunc = false;

      if( deffunc != NULL )
        bIsDefFunc = ( str == deffunc && s == NULL );

      if( !bIsDefFunc && f == NULL )
      {
        Rollback( state );
        return t;
      }
  
      if( !bIsDefFunc && !f->IsInfix() )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( this );
  
        e->Stream() << "Function '" << str << "' is not notated infix.";
  
        throw e;
      }
      
      if( !bIsDefFunc && f->GetWeight() < weight )
      {
        Rollback( state );
        return t;
      }
  
      ProofLib::TermList terms;
  
      const TypeSymbol *t1 = FunctionSymbol::GetTermType( t );
      terms.AddRight( t );
  
      if( !bIsDefFunc && !f->CheckArgType( 0, t1 ) )
        throw new TypeMismatchException( &state, t1, f->GetArgType( 0 ) );

      if( bIsDefFunc )
        t = ReadSubTerm();
      else
        t = ReadTerm( f->GetWeight() );
  
      const TypeSymbol *t2 = FunctionSymbol::GetTermType( t );
      terms.AddRight( t );
  
      if( !bIsDefFunc && !f->CheckArgType( 1, t2 ) )
        throw new TypeMismatchException( this, t2, f->GetArgType( 1 ) );

      if( bIsDefFunc )
      {
#if 0
        const char *types[3];

        types[0] = t1->GetName();
        types[1] = t2->GetName();
        types[2] = NULL;
#endif
        std::vector<USTRING> types;
        types.push_back( t1->GetName() );
        types.push_back( t2->GetName() );
        
         
        GetSymbolTable()->NewFunction(
          deffunc, false, types, deftype, 0, *this );
        s = GetSymbolTable()->GetSymbol( str.c_str() );
        f = dynamic_cast<const FunctionSymbol*>( s );
        ASSERT( f != NULL );
      }

      t = ProofLib::Function::New( m_doc->GetNodeDatabase(), f, &terms );
    }

    return t;
  }

  ProofLib::Term* ReadSubTerm( const wchar_t *deffunc = NULL,
                               const wchar_t *deftype = NULL )
  {
    ASSERT( deffunc == NULL || deftype != NULL );

    EatWhiteSpacesAndComments();

    ProofLib::Term* t = NULL;

    USTRING str;

    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    const Symbol *s = GetSymbolTable()->GetSymbol( str.c_str() );

    int type = 0;
    bool bIsDefFunc = false;

    if( s == NULL )
    {
      if( deffunc == NULL || str != deffunc )
        throw new UnexpectedTokenException( this, str );
      bIsDefFunc = true;
      type = SYMBOL_TYPE_FUNCTION;
    }
    else
      type = s->GetSymbolType();

    
    switch( type )
    {
    case SYMBOL_TYPE_VARIABLE:
    {
      EatWhiteSpacesAndComments();

      ProofLib::RecursiveExpr* expr = NULL;

      if( CurrentChar() == BO_RECURSIVE_EXPR )
      {
        AssertChar( BO_RECURSIVE_EXPR );

        expr = ReadRecursiveExpression();

        EatWhiteSpacesAndComments();
        AssertChar( BC_RECURSIVE_EXPR );
        EatWhiteSpacesAndComments();

        t = ProofLib::RecursiveVariable::
          New( m_doc->GetNodeDatabase(), s, expr );

        break;
      }

      t = ProofLib::Variable::New( m_doc->GetNodeDatabase(), s );

      break;
    }
    case SYMBOL_TYPE_CONSTANT:
    {
      EatWhiteSpacesAndComments();

      ProofLib::RecursiveExpr* expr = NULL;

      if( CurrentChar() == BO_RECURSIVE_EXPR )
      {
        AssertChar( BO_RECURSIVE_EXPR );

        expr = ReadRecursiveExpression();

        EatWhiteSpacesAndComments();
        AssertChar( BC_RECURSIVE_EXPR );
        EatWhiteSpacesAndComments();

        t = ProofLib::RecursiveConstant::
          New( m_doc->GetNodeDatabase(), s, expr );

        break;
      }

      t = ProofLib::Constant::New( m_doc->GetNodeDatabase(), s );

      break;
    }
    case SYMBOL_TYPE_FUNCTION:
    {
      const FunctionSymbol *f = dynamic_cast<const FunctionSymbol*>( s );
      ASSERT( f != NULL || bIsDefFunc );

      EatWhiteSpacesAndComments();

      ProofLib::RecursiveExpr* expr = NULL;

      if( f != NULL && f->IsRecursive() )
      {
        AssertChar( BO_RECURSIVE_EXPR );

        expr = ReadRecursiveExpression();

        EatWhiteSpacesAndComments();
        AssertChar( BC_RECURSIVE_EXPR );
        EatWhiteSpacesAndComments();
      }

      ProofLib::TermList terms;
  
      if( expr == NULL || f->GetArity() > 0 )
      {
        AssertChar( BO_FUNCTION );
      
        size_t argc = 0;
  
        while( true )
        {
          if( !bIsDefFunc && argc >= f->GetArity() )
            throw new ArgCountMismatchException( this, argc+1, f->GetArity() );
  
          EatWhiteSpacesAndComments();
          
          ReaderState before_term = SetRollbackPoint();
  
          ProofLib::Term *t1 = ReadTerm();
          const TypeSymbol *type1 = FunctionSymbol::GetTermType( t1 );
          terms.AddRight( t1 );
    
          if( !bIsDefFunc && !f->CheckArgType( argc, type1 ) )
            throw new TypeMismatchException( &before_term,
                                             type1, f->GetArgType( argc ) );
  
          EatWhiteSpacesAndComments();
  
          if( CurrentChar() == BC_FUNCTION )
            break;
  
          argc++;
  
          AssertChar( COMMA );
        }
  
        AssertChar( BC_FUNCTION );
  
        if( !bIsDefFunc && argc+1 != f->GetArity() )
          throw new ArgCountMismatchException( this, argc+1, f->GetArity() );
  
        if( bIsDefFunc )
        {
  #if 0
          const char **types = new const char*[terms.GetChildCount()];
  
          size_t i = 0;
          for( ; i < terms.GetChildCount(); i++ )
            types[i] = FunctionSymbol::GetTermType( terms.GetTerm(i) )->GetName();
   
          types[i] = NULL;
  #endif
          std::vector<USTRING> types;
           
          for( size_t i = 0; i < terms.GetChildCount(); i++ )
            types.push_back(
              FunctionSymbol::GetTermType( terms.GetTerm(i) )->GetName() );
  
          GetSymbolTable()->NewFunction(
            deffunc, true, types, deftype, 0, *this );
          s = GetSymbolTable()->GetSymbol( str.c_str() );
          f = dynamic_cast<const FunctionSymbol*>( s );
          ASSERT( f != NULL );
  
  //        delete [] types;
        }
      }

      if( expr != NULL )
      {
        t = ProofLib::RecursiveFunction::New( m_doc->GetNodeDatabase(), s,
                                              &terms, expr );
        break;
      }

      t = ProofLib::Function::New( m_doc->GetNodeDatabase(), s, &terms );

      break;
    }
    default:
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );

      e->Stream() << "Unexpected symbol '" << str << "' of type "
                  << s->GetSymbolTypeName();
      throw e;
    }

    return t;
  }

  ProofLib::Variable* ReadVariable( ProofLib::SourceLocation *sloc = NULL )
  {
    ProofLib::SourceLocation before_term( *this );
    ProofLib::Term *t = ReadSubTerm();

    ProofLib::Variable* v = dynamic_cast<ProofLib::Variable*>( t );

    if( v == NULL )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_term );
      e->Stream() << "Variable expected, but read term '";
      Print::PrintTerm( t, e->Stream() );
      e->Stream() << "'.";
      throw e;
    }

    if( sloc != NULL )
      *sloc = before_term;

    return v;
  }

  ProofLib::RecursiveExpr* ReadRecursiveExpression()
  {
    USTRING str;
    unsigned int num = 0;

    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    // check rec var

    if( GetSymbolTable()->GetRecursionVar().empty() ||
        str != GetSymbolTable()->GetRecursionVar() )
    {
      for( size_t i = 0; i < str.length(); i++ )
      {
        if( !isdigit( str[i] ) )
        {
          ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
          e->Stream() << "Unknown recursion variable '" << str << "'.";
          throw e;
        }

        num = num * 10 + ( str[i] - '0' );
      }

      
      ProofLib::RecursiveExpr* expr =
        ProofLib::RecursiveExpr::New( m_doc->GetNodeDatabase(), NULL,
                                      '+', num );

      return expr;
    }
        
    if( str != GetSymbolTable()->GetRecursionVar() )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );
      e->Stream() << "Unexpected recursion variable '" << str << "'.";
      throw e;
    }

    // read expr

    char op;

    EatWhiteSpacesAndComments();
    switch( CurrentChar() )
    {
    case '+':
    case '-':
    case '*':
      op = CurrentChar();
      AssertChar( op );
      num = ReadPositiveInteger();
      break;
    default:
      op = '+';
      num = 0;
    }

    ProofLib::RecursiveExpr* expr =
      ProofLib::RecursiveExpr::New( m_doc->GetNodeDatabase(), str.c_str(),
                                    op, num );

    return expr;
  }

  ProofLib::RecursiveExpr* ReadRecLevelTerm()
  {
    USTRING str;
    unsigned int num = 0;

    EatWhiteSpacesAndComments();
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    // check if 'str' is a positive integer

    bool str_is_num = true;

    for( size_t i = 0; i < str.length(); i++ )
    {
      if( !isdigit( str[i] ) )
      {
        str_is_num = false;
        break;
      }

      num = num * 10 + ( str[i] - '0' );
    }

    if( str_is_num )
    {
      ProofLib::RecursiveExpr* expr =
        ProofLib::RecursiveExpr::New( m_doc->GetNodeDatabase(), NULL,
                                      '+', num );

      return expr;
    }

    num = 0;
        
    // read expr

    char op;

    EatWhiteSpacesAndComments();
    switch( CurrentChar() )
    {
    case '+':
    case '*':
      op = CurrentChar();
      AssertChar( op );
      EatWhiteSpacesAndComments();
      num = ReadPositiveInteger();
      break;
    default:
      op = '+';
      num = 0;
    }

    ProofLib::RecursiveExpr* expr =
      ProofLib::RecursiveExpr::New( m_doc->GetNodeDatabase(), str.c_str(),
                                    op, num );

    return expr;
  }

  ProofLib::Formula* ReadRestOfBinaryFormula( ProofLib::Formula *f1 )
  {
    USTRING str;

    EatWhiteSpacesAndComments();
    ReaderState state = SetRollbackPoint();
    ReadWord( str );

    if( str.empty() )
    {
      Rollback( state );
      return f1;
    }

    const ProofLib::OperatorSymbol *os = NULL;

    switch( ReservedWord( str ) )
    {
    case RESWORD_AND:
      os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_AND );
      break;
    case RESWORD_OR:
      os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_OR );
      break;
    case RESWORD_IMPL:
      os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_IMPL );
      break;
    default:
      Rollback( state );
      return f1;
    }

    ProofLib::Formula *f2 = ReadFormula( true );

    return
      ProofLib::BinaryConFormula::New( m_doc->GetNodeDatabase(), os, f1, f2 );
  }

  ProofLib::Formula* ReadFormula( bool subformula = false,
                                  const wchar_t* deffunc = NULL )
  {
    ProofLib::Formula* f = NULL;

    USTRING str;

    EatWhiteSpacesAndComments();
    ReaderState state = SetRollbackPoint();

    try
    {
      f = ReadInfixAtomFormula( deffunc );

      if( subformula )
        return f;
      else
        return ReadRestOfBinaryFormula( f );
    }
    catch( ... )
    {
      // rewind
      Rollback( state );
    }

    ReadWord( str );

    // Formula in parentisis
    if( str.empty() )
    {
      AssertChar( BO_FORMULA );

      f = ReadFormula( false, deffunc );
#if 0
      Print::PrintFormula( f, std::cerr );
      std::cerr << std::endl;
#endif

      EatWhiteSpacesAndComments();

#if 0
      bool read_p = true;
      if( CurrentChar() == BC_FORMULA )
#endif
        AssertChar( BC_FORMULA );
#if 0
      else
        read_p = false;
#endif

      if( !subformula )
        f = ReadRestOfBinaryFormula( f );

#if 0
      EatWhiteSpacesAndComments();

      if( !read_p )
        AssertChar( BC_FORMULA );
#endif

      return f;
    }

    // Negated Formula
    if( ReservedWord( str ) == RESWORD_NOT )
    {
#if 0
      if( subformula )
        throw new
          UnexpectedTokenException( this, str );
#endif

      f = ReadFormula( true );
      
      const ProofLib::OperatorSymbol *os =
        GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_NEG );
      ASSERT( os != NULL );

      f = ProofLib::UnaryConFormula::New( m_doc->GetNodeDatabase(), os, f );
#if 0
    std::cerr << "READ NOT: ";
    Print::PrintFormula( f, std::cerr );
    std::cerr << std::endl;
#endif

      if( subformula )
        return f;
      else
        return ReadRestOfBinaryFormula( f );
    }

    // Quantor Formula
    if( ReservedWord( str ) == RESWORD_ALL ||
        ReservedWord( str ) == RESWORD_EX )
    {
#if 0
      if( subformula )
        throw new
          UnexpectedTokenException( this, str );
#endif

      // Determine whether this is a first- or second-order
      // quantor formula
      ReaderState before = SetRollbackPoint();
      bool first_order = false;
      
      EatWhiteSpacesAndComments();
      
      try
      {
      	ProofLib::Term *t = ReadSubTerm();
      	ProofLib::Variable *v = dynamic_cast<ProofLib::Variable*>( t );
      	if ( v != NULL )
      	  first_order = true;
      }
      catch ( ... ) {}

      Rollback( before );

      if ( first_order )
      {
      	// First-order quantor formula
	    const ProofLib::OperatorSymbol *os = NULL;
	
	    if( ReservedWord( str ) == RESWORD_ALL )
	      os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_ALL );
	    else
	      os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_EX );
	
	    ASSERT( os != NULL );
	
	    //ReaderState before_term = SetRollbackPoint();
	    ProofLib::SourceLocation before_term( *this );
	    ProofLib::Term *t = ReadSubTerm();
	
	    ProofLib::Variable *v = dynamic_cast<ProofLib::Variable*>( t );

	    if( v == NULL )
	    {
	      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_term );
	      e->Stream() << "Variable expected, but read term '";
	      Print::PrintTerm( t, e->Stream() );
	      e->Stream() << "'.";
	      throw e;
	    }
	
	    const VariableSymbol *vs =
	      dynamic_cast<const VariableSymbol*>( v->GetSymbol() );
	    ASSERT( vs != NULL );
	
	    if( !vs->MayBeUsedAsBounded() )
	    {
	      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_term );
	      e->Stream() << "Variable '";
	      Print::PrintTerm( v, e->Stream() );
	      e->Stream() << "' may not be used as a bounded variable.";
	      throw e;
	    }
	
	    f = ReadFormula( true );
	    
	    f = ProofLib::QuantorFormula::
	      New( m_doc->GetNodeDatabase(), os, v, f );
	
	    if( subformula )
	      return f;
	    else
	      return ReadRestOfBinaryFormula( f );
      } else
      {
      	// Second-order quantor formula
      	const ProofLib::OperatorSymbol *os = NULL;
      
        if( ReservedWord( str ) == RESWORD_ALL )
          os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_ALL2 );
        else
          os = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_EX2 );

        ASSERT( os != NULL );

        ProofLib::SourceLocation before_variable( *this );
        // read predicate variable
        EatWhiteSpacesAndComments();
        ReadWord( str );
        const Symbol *s = GetSymbolTable()->GetSymbol( str.c_str() );

        if( s != NULL && s->GetSymbolType() == SYMBOL_TYPE_PREDICATE )
        {
          const PredicateSymbol *ps = dynamic_cast<const PredicateSymbol*>( s );
          ASSERT( ps != NULL );
	      if ( !ps->IsVariable() )
	      {
            ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &before_variable );
	        e->Stream() << "Predicate '"
	          << ps->GetName() << "' is not variable.";
	        throw e;
	      }

	      ProofLib::SecondOrderVariable* v = ProofLib::SecondOrderVariable::
	                                         New( m_doc->GetNodeDatabase(),
			                                      ps );

          f = ReadFormula( true );
      
          f = ProofLib::SecondOrderQuantorFormula::
              New( m_doc->GetNodeDatabase(), os, v, f );

          if( subformula )
            return f;
          else
            return ReadRestOfBinaryFormula( f );
        }
        else
        {
	      throw new UnexpectedTokenException( this, str );
        }
      }
    }

    // Atomformula
    const Symbol *s = GetSymbolTable()->GetSymbol( str.c_str() );

    if( s != NULL && s->GetSymbolType() == SYMBOL_TYPE_PREDICATE )
    {
      const PredicateSymbol *ps = dynamic_cast<const PredicateSymbol*>( s );
      ASSERT( ps != NULL );

      ProofLib::RecursiveExpr* expr = NULL;

      if( ps->IsRecursive() )
      {
        AssertChar( BO_RECURSIVE_EXPR );

        expr = ReadRecursiveExpression();

        EatWhiteSpacesAndComments();
        AssertChar( BC_RECURSIVE_EXPR );
        EatWhiteSpacesAndComments();
      }

    ProofLib::SecondOrderVariable *var = NULL;
#if 0
    std::cerr << "is predicate symbol " << ps->GetName() << " variable?";
#endif
    if ( ps->IsVariable() )
    {
#if 0
      std::cerr << " yes";
#endif
      var = ProofLib::SecondOrderVariable::New( m_doc->GetNodeDatabase(),
                                                ps );
    }
    else
    {
#if 0
      std::cerr << " no";
#endif
    }
#if 0
    std::cerr << std::endl;
#endif

      ProofLib::TermList terms;

      if( ps->GetArity() == 0 )
      {
        if( expr == NULL )
          f = ProofLib::AtomFormula::
            New( m_doc->GetNodeDatabase(), ps, &terms, var );
        else
          f = ProofLib::RecursivePredicate::
            New( m_doc->GetNodeDatabase(), ps, &terms, expr );

        if( subformula )
          return f;
        else
          return ReadRestOfBinaryFormula( f );
      }

      EatWhiteSpacesAndComments();
      AssertChar( BO_ATOMF );
    
      size_t argc = 0;

      while( true )
      {
        if( argc >= ps->GetArity() )
          throw new ArgCountMismatchException( this, argc+1, ps->GetArity() );

        EatWhiteSpacesAndComments();
        
        ReaderState before_term = SetRollbackPoint();

        ProofLib::Term *t = ReadTerm( 0, deffunc,
                              ps->GetArgType( argc )->GetName().c_str() );
        const TypeSymbol *t_type = FunctionSymbol::GetTermType( t );
        terms.AddRight( t );
  
        if( !ps->CheckArgType( argc, t_type ) )
          throw new TypeMismatchException( &before_term,
                                           t_type, ps->GetArgType( argc ) );

        EatWhiteSpacesAndComments();

        if( CurrentChar() == BC_ATOMF )
          break;

        argc++;

        AssertChar( COMMA );
      }

      AssertChar( BC_ATOMF );

      if( argc+1 != ps->GetArity() )
        throw new ArgCountMismatchException( this, argc+1, ps->GetArity() );

      if( expr == NULL )
        f = ProofLib::AtomFormula::
          New( m_doc->GetNodeDatabase(), ps, &terms, var );
      else
        f = ProofLib::RecursivePredicate::
          New( m_doc->GetNodeDatabase(), ps, &terms, expr );

      if( subformula )
        return f;
      else
        return ReadRestOfBinaryFormula( f );
    }

    if( subformula )
      throw new
        UnexpectedTokenException( this, str );

    // rewind
    Rollback( state );

    return ReadInfixAtomFormula( deffunc );
  }

  ProofLib::Formula* ReadInfixAtomFormula( const wchar_t *deffunc )
  {
    // Infix Atomformula
    ReaderState before_term = SetRollbackPoint();
    USTRING type;
    type.FromUtf8( "any" );
    ProofLib::Term *t1 = ReadTerm( 0, deffunc, type.c_str() );
#if 0
    Print::PrintTerm( t1, std::cerr );
    std::cerr << std::endl;
#endif

    EatWhiteSpacesAndComments();
    USTRING str;
    ReadWord( str );

    if( str.empty() )
      throw new UnexpectedCharException( this, CurrentChar() );

    const Symbol *s = GetSymbolTable()->GetSymbol( str.c_str() );
    const PredicateSymbol *ps = dynamic_cast<const PredicateSymbol*>( s );

    if( ps == NULL )
      throw new UnexpectedTokenException( this, str );

    if( !ps->IsInfix() )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
  
      e->Stream() << "Predicate '" << str << "' is not notated infix.";
  
      throw e;
    }

    const TypeSymbol *t_type_l = FunctionSymbol::GetTermType( t1 );
    if( !ps->CheckArgType( 0, t_type_l ) )
      throw new TypeMismatchException( &before_term,
                                       t_type_l, ps->GetArgType( 0 ) );
      
    before_term = SetRollbackPoint();
    ProofLib::Term *t2 = ReadTerm( 0, deffunc,
                                   ps->GetArgType( 1 )->GetName().c_str() );

    const TypeSymbol *t_type_r = FunctionSymbol::GetTermType( t2 );
    if( !ps->CheckArgType( 1, t_type_r ) )
      throw new TypeMismatchException( &before_term,
                                       t_type_r, ps->GetArgType( 1 ) );
      
    if( ps->IsEqualPred() &&
        t_type_l != t_type_r &&
        !t_type_l->IsAnyType() &&
        !t_type_r->IsAnyType() )
    {
      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );
  
      e->Stream() << "Both arguments of '=' must be of the same type.";
  
      throw e;
    }

    ProofLib::TermList terms;

    terms.AddRight( t1 );
    terms.AddRight( t2 );

    ProofLib::SecondOrderVariable *var = NULL;
#if 0
    std::cerr << "is predicate symbol " << ps->GetName() << " variable?";
#endif
    if ( ps->IsVariable() )
    {
#if 0
      std::cerr << "yes";
#endif
      var = ProofLib::SecondOrderVariable::New( m_doc->GetNodeDatabase(),
                                                ps );
    }
    else
    {
#if 0
      std::cerr << "no";
#endif
    }
#if 0
    std::cerr << std::endl;
#endif

#if 0
    ProofLib::Formula *f = m_doc->GetNodeDatabase()->NewAtomFormula( ps, &terms );
    Print::PrintFormula( f, std::cerr );
    std::cerr << std::endl;
    return f;
#else
    return ProofLib::AtomFormula::New( m_doc->GetNodeDatabase(), ps, &terms, var );
#endif
  }

  ProofLib::Sequent* ReadSequent()
  {
    ProofLib::SElementList *l = NULL;
    ProofLib::SElementList *r = NULL;

    EatWhiteSpacesAndComments();
    ReaderState state = SetRollbackPoint();

    USTRING str;
    ReadWord( str );

    if( str != reserved_words[RESWORD_SEQ] )
    {
      Rollback( state );
  
      while( true )
      {
        if( l == NULL )
          l = new ProofLib::SElementList;

        l->AddRight( new ProofLib::SFormula( ReadFormula() ) );
  
        EatWhiteSpacesAndComments();
        if( CurrentChar() != COMMA )
          break;
  
        AssertChar( COMMA );
      }

      AssertReservedWord( RESWORD_SEQ );
    }

    EatWhiteSpacesAndComments();

    bool done = false;

    ReaderState state_1 = SetRollbackPoint();
    ReadWord( str );
    if( ReservedWord( str ) == RESWORD_LEFT ||
        ReservedWord( str ) == RESWORD_RIGHT )
      done = true;
    Rollback( state_1 );

    EatWhiteSpacesAndComments();

    if( !done && CurrentChar() != EOC )
    {
      while( true )
      {
        if( r == NULL )
          r = new ProofLib::SElementList;

        r->AddRight( new ProofLib::SFormula( ReadFormula() ) );
  
        EatWhiteSpacesAndComments();
        if( CurrentChar() != COMMA )
          break;
  
        AssertChar( COMMA );
      }
    }

#ifdef DEBUG_PRINT_READ_INFO
    ProofLib::Sequent *s = new ProofLib::Sequent( l, r );
    Print::PrintSequent( s, std::cerr );
    std::cerr << std::endl;
    return s;
#else
    return new ProofLib::Sequent( l, r );
#endif
  }
};

////////////////////////////////////////////////////////////////////////////////

/// factory function for ReaderImpl
/// \return an ReaderImpl obj usable by the general interface
/// ReaderImpl (user have to delete non null pointer)
Reader* Reader::New()
{
  return new ReaderImpl();
}

}

