////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: symbolhlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <vector>

#include "libproof/util/exceptions.h"
#include "libproof/util/myassert.h"
#include "libproof/util/hash.h"

#include "libproof/libproof.h"
#include "libproof/recursive.h"
#include "libproof/term_operations.h"

#include "symbolhlk.h"
#include "readhlk.h"
#include "reswordshlk.h"
#include "proofhlk.h"
#include "printhlk.h"

//#define DEBUG_PRINT_DEFINE

namespace HLK
{

const wchar_t *symbol_type_names[] =
{
  L"type",
  L"variable",
  L"constant",
  L"function",
  L"predicate",
  L"rule",
  L"proof",
  NULL
};

////////////////////////////////////////////////////////////////////////////////

static void CheckVarMap( const ProofLib::TermList *var_map,
                         ProofLib::TermList &vars,
                         const ProofLib::SourceLocation &sloc )
{
  if( var_map->GetChildCount() != vars.GetChildCount() )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );

    e->Stream() << "The number of variables in the variable order clause ("
                << var_map->GetChildCount() << ") differs to the one of the "
                << "defining term (" << vars.GetChildCount() << ").";

    throw e;
  }

  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    bool found = false;

    for( size_t j = 0; j < var_map->GetChildCount(); j++ )
    {
      if( vars.GetTerm(i) == var_map->GetTerm(j) )
      {
        found = true;
        break;
      }
    }

    if( !found )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );

      e->Stream() << "Variable '";
      Print::PrintTerm( vars.GetTerm(i), e->Stream() );
      e->Stream() << "' is missing in the prototype.";

      throw e;
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

SymbolExistsException::SymbolExistsException( const Symbol *s )
{
  ASSERT( s != NULL );

  m_symbol = s;
}

void SymbolExistsException::Print( std::ostream &out ) const
{
  out << "Symbol " << m_symbol->GetName()
      << " was already defined as " << USTRING(m_symbol->GetSymbolTypeName());

  if( m_symbol->GetFile() != NULL )
  {
    out << " in file '" << m_symbol->GetFile()
        << "' on line " << m_symbol->GetLine()
        << " char " << m_symbol->GetChar();
  }
}

////////////////////////////////////////////////////////////////////////////////

NoSuchTypeException::NoSuchTypeException(
  const USTRING &type, const ProofLib::SourceLocation &sloc )
  : SLocException( &sloc )
{
  m_type = type;
}

void NoSuchTypeException::Print( std::ostream &out ) const
{
  SLocException::Print( out );
  out << "Unknown type " << m_type;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

size_t Symbol::Hash() const
{
  stlhash::hash<USTRING> h;
  return h( GetName() );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class EqualityDefinition : public ProofLib::SourceLocation
{
public:
  ProofLib::Variable *l, *r;
  const TypeSymbol *ts_l, *ts_r;
  ProofLib::Formula *f;
  USTRING id;

  EqualityDefinition(
    ProofLib::Variable *v1, ProofLib::Variable *v2,
    const TypeSymbol *ts_l, const TypeSymbol *ts_r,
    ProofLib::Formula *f, const USTRING& id,
    const ProofLib::SourceLocation &sloc )
    : ProofLib::SourceLocation( sloc )
  {
    l = v1;
    r = v2;
    this->ts_l = ts_l;
    this->ts_r = ts_r;
    this->f = f;
    this->id = id;
  }
};

class EqualityDefinitionManager
{
  std::vector< EqualityDefinition > m_defs;

public:
  void AddEqualityDefinition( ProofLib::Variable *l, ProofLib::Variable *r,
    ProofLib::Formula *f, const ProofLib::SourceLocation &sloc )
  {
    const VariableSymbol *vs_l =
      dynamic_cast<const VariableSymbol*>( l->GetSymbol() );

    const TypeSymbol *ts_l = vs_l->GetType(); 

    const VariableSymbol *vs_r =
      dynamic_cast<const VariableSymbol*>( r->GetSymbol() );

    const TypeSymbol *ts_r = vs_r->GetType(); 

    // check if this equality is already defined.
    for( size_t i = 0; i < m_defs.size(); i++ )
    {
      if( m_defs[i].ts_l == ts_l && m_defs[i].ts_r == ts_r )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Equality of type '" << ts_l->GetName() << " = "
                    << ts_r->GetName() << "' is already defined in "
                    << m_defs[i].GetFile() << ":" << m_defs[i].GetLine()
                    << " as ";
        Print::PrintFormula( m_defs[i].f, e->Stream() );
        e->Stream() << ".";
              
        throw e;
      }
    }

    std::wstringstream str;
    str << L"=";
    str << L"_(" << ts_l->GetName() << L"x" << ts_r->GetName() << L")";

    m_defs.push_back( EqualityDefinition( l, r, ts_l, ts_r, f,
                                          str.str().c_str(), sloc ) );
  }

  size_t GetEqualityDefinitionCount() const { return m_defs.size(); }

  ProofLib::Formula* GetEqualityDefinitionAt( size_t index,
    const TypeSymbol **l, const TypeSymbol **r, USTRING *id ) const
  {
    ASSERT( index < m_defs.size() );
    ASSERT( l != NULL && r != NULL );

    *l = m_defs[index].ts_l;
    *r = m_defs[index].ts_r;
    if( id != NULL )
      (*id) = m_defs[index].id;

    return m_defs[index].f;
  }

  const UCHAR* GetEqualityDefinitionId(
    const TypeSymbol *l, const TypeSymbol *r ) const
  {
    for( size_t i = 0; i < m_defs.size(); i++ )
    {
      if( m_defs[i].ts_l == l && m_defs[i].ts_r == r )
        return m_defs[i].id.c_str();
    }

    return NULL;
  }

  ProofLib::Formula* GetEqualityInstance(
    ProofLib::Term *l, ProofLib::Term *r,
    ProofLib::NodeDatabase *ndb ) const
  {
    ASSERT( l != NULL );
    ASSERT( r != NULL );

    const TypeSymbol *ts_l = FunctionSymbol::GetTermType( l );
    const TypeSymbol *ts_r = FunctionSymbol::GetTermType( r );

    for( size_t i = 0; i < m_defs.size(); i++ )
    {
      if( m_defs[i].ts_l == ts_l && m_defs[i].ts_r == ts_r )
      {
        ProofLib::OperationVariableSubstitution tr( ndb );

        tr.AddSubstitution( m_defs[i].l, l );
        tr.AddSubstitution( m_defs[i].r, r );

        return ProofLib::DataNode::ApplyTo( m_defs[i].f, tr );
      }
    }

    GenericException *e = new GenericException;
    e->Stream() << "Equality of type '" << ts_l->GetName() << " = "
                << ts_r->GetName() << "' is not defined.";
              
    throw e;
  }

  ProofLib::Formula* GetEqualityDefinitionOf(
    const TypeSymbol *l, const TypeSymbol *r ) const
  {
    ASSERT( l != NULL );
    ASSERT( r != NULL );

    for( size_t i = 0; i < m_defs.size(); i++ )
    {
      if( m_defs[i].ts_l == l && m_defs[i].ts_r == r )
        return m_defs[i].f;
    }

    return NULL;
  }

  ProofLib::Formula* GetEqualityInstance(
    const TypeSymbol *l, const TypeSymbol *r,
    const ProofLib::VariableSubstitution& sigma,
    ProofLib::NodeDatabase *ndb,
    const PredicateSymbol *ps_equal ) const
  {
    ASSERT( l != NULL );
    ASSERT( r != NULL );

    for( size_t i = 0; i < m_defs.size(); i++ )
    {
      if( m_defs[i].ts_l == l && m_defs[i].ts_r == r )
      {
        ProofLib::TermList terms;
        terms.AddRight( m_defs[i].l );
        terms.AddRight( m_defs[i].r );
        ProofLib::AtomFormula *af =
          ProofLib::AtomFormula::New( ndb, ps_equal, &terms );

#if 0
  Print::PrintTerm( m_defs[i].l, std::cerr );
  std::cerr << "." << std::endl;
  Print::PrintTerm( m_defs[i].r, std::cerr );
  std::cerr << "." << std::endl;
  Print::PrintFormula( af, std::cerr );
  std::cerr << "." << std::endl;
#endif

        ProofLib::OperationVariableSubstitution tr( ndb, sigma );

        return ProofLib::DataNode::ApplyTo( af, tr );
      }
    }

    return NULL;
  }
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class SymbolTableStore :
  public stlhash::hash_map< USTRING, Symbol* >,
  public EqualityDefinitionManager
{
  USTRING rec_var;

public:
  std::vector<OperatorSymbol*> m_op_map;

public:
  SymbolTableStore( const USTRING &rec_var )
    : m_op_map( OP_SYMBOL_USER )
  {
    this->rec_var = rec_var;
  }

  const USTRING& GetRecursionVar() const { return rec_var; }
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

SymbolTable::SymbolTable( SymbolTable *parent, const USTRING &rec_var )
{
  m_store = new SymbolTableStore( rec_var );

  m_parent = parent;

  if( parent == NULL )
  {
    TypeSymbol *any = new TypeSymbol( reserved_words[RESWORD_ANY], L"" );
    m_AnyTypeSymbol = any;
    AddSymbol( any );

    std::vector<const TypeSymbol*> arg_types;
    arg_types.push_back( m_AnyTypeSymbol );
    arg_types.push_back( m_AnyTypeSymbol );
    USTRING name;
    name.FromUtf8( "=" );
    PredicateSymbol *equal = new PredicateSymbol( name, false, arg_types,
                                                  NULL, NULL, this );
    m_EqualPredicateSymbol = equal;
    AddSymbol( equal );
  
#define ADD_RULE( r ) AddSymbol( new RuleSymbol( reserved_words[r], r ) )
    ADD_RULE( RESWORD_NOT );
    ADD_RULE( RESWORD_AND );
    ADD_RULE( RESWORD_OR );
    ADD_RULE( RESWORD_IMPL );
    ADD_RULE( RESWORD_ALL );
    ADD_RULE( RESWORD_EX );
    ADD_RULE( RESWORD_AUTO );
    ADD_RULE( RESWORD_UNDEF );
    ADD_RULE( RESWORD_DEF );
    ADD_RULE( RESWORD_CUT );
    ADD_RULE( RESWORD_PARAMOD );
#undef ADD_RULE

#define ADD_OP( NAME, TYPE, CHAR, ARITY ) \
    { \
      OperatorSymbol *o = \
        new OperatorSymbol( NAME, TYPE, CHAR, ARITY ); \
      m_store->m_op_map[TYPE] = o; \
      AddSymbol( o ); \
    }
    ADD_OP( L"and", OP_SYMBOL_AND, CHAR_AND, 2 );
    ADD_OP( L"or", OP_SYMBOL_OR, CHAR_OR, 2 );
    ADD_OP( L"impl", OP_SYMBOL_IMPL, CHAR_IMPL, 2 );
    ADD_OP( L"neg", OP_SYMBOL_NEG, CHAR_NEG, 1 );
    ADD_OP( L"all", OP_SYMBOL_ALL, CHAR_ALL, 1 );
    ADD_OP( L"exists", OP_SYMBOL_EX, CHAR_EXISTS, 1 );
    ADD_OP( L"all2", OP_SYMBOL_ALL2, CHAR_ALL, 1 );
    ADD_OP( L"exists2", OP_SYMBOL_EX2, CHAR_EXISTS, 1 );
#undef ADD_OP
  }
  else
  {
    m_AnyTypeSymbol = m_parent->m_AnyTypeSymbol;
    m_EqualPredicateSymbol = m_parent->m_EqualPredicateSymbol;
  }
}

SymbolTable* SymbolTable::GetTopParent()
{
  SymbolTable* s = this;

  while( s->GetParent() != NULL )
    s = s->GetParent();

  return s;
}

SymbolTable::~SymbolTable()
{
  while( m_store->size() > 0 )
  {
    Symbol* s = m_store->begin()->second;

    m_store->erase( m_store->begin() );

    delete s;
  }
   
  delete m_store;
}

const Symbol* SymbolTable::GetSymbol( const USTRING &s,
                                      bool parent_lookup ) const
{
  SymbolTableStore::const_iterator i = m_store->find( s );

  if( i == m_store->end() )
  {
    if( parent_lookup && m_parent != NULL )
      return m_parent->GetSymbol( s );

    return NULL;
  }

  return i->second;
}

const USTRING& SymbolTable::GetRecursionVar() const
{
  return m_store->GetRecursionVar();
}

void SymbolTable::AddSymbol( Symbol *s )
{
  (*m_store)[s->GetName()] = s; 
}

void SymbolTable::NewProof( const USTRING &proof, const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " proof: " << proof
            << std::endl;
#endif

  const Symbol *s = GetSymbol( proof );

  if( s != NULL )
    throw new SymbolExistsException( s );

  ProofSymbol *p = new ProofSymbol( proof );
  p->SetSourceLocation( sloc );

  AddSymbol( p );
}

void SymbolTable::NewType( const USTRING &type, const USTRING &fresh,
                           const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " type: " << type
            << std::endl;
#endif

  const Symbol *s = GetSymbol( type );

  if( s != NULL )
    throw new SymbolExistsException( s );

  TypeSymbol *t = new TypeSymbol( type, fresh );
  t->SetSourceLocation( sloc );

  AddSymbol( t );
}

const VariableSymbol* SymbolTable::NewVariable(
  const USTRING &var_0, const USTRING &type,
  bool usable_as_bounded,
  const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " variable: " << var << " of type " << type
            << std::endl;
#endif
  USTRING var = var_0;
  //char num_str[32];

  const TypeSymbol *t = dynamic_cast<const TypeSymbol*>( GetSymbol( type ) );

  if( t == NULL )
    throw new NoSuchTypeException( type,
      ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );


  if( var.empty() ) // generate a name!
  {
    const Symbol *s = NULL;
    size_t num = 1;
    do
    {
      std::wostringstream str;
      str << t->GetFreshStart() << "_{" << num++ << "}";
    
      var = str.str().c_str();
      
      s = GetSymbol( var );
    }
    while( s != NULL );
  }
  else
  {
    const Symbol *s = GetSymbol( var, false );

    if( s != NULL )
      throw new SymbolExistsException( s );
  }

  VariableSymbol *v = new VariableSymbol( var, t, usable_as_bounded );
  v->SetSourceLocation( sloc );

  AddSymbol( v );

  return v;
}

void SymbolTable::NewConstant( const USTRING &con, const USTRING &type,
                               const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " constant: " << con << " of type " << type
            << std::endl;
#endif

  const Symbol *s = GetSymbol( con, false );

  if( s != NULL )
    throw new SymbolExistsException( s );

  const TypeSymbol *t = dynamic_cast<const TypeSymbol*>( GetSymbol( type ) );

  if( t == NULL )
    throw new NoSuchTypeException( type,
      ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );

  ConstantSymbol *v = new ConstantSymbol( con, t );
  v->SetSourceLocation( sloc );

  AddSymbol( v );
}

const FunctionSymbol* SymbolTable::NewFunction(
  const USTRING &func, bool prefix,
  std::vector<USTRING> &types,
  const USTRING &rtype, unsigned int weight,
  const ProofLib::SourceLocation &sloc )
{
  const TypeSymbol *t = dynamic_cast<const TypeSymbol*>( GetSymbol( rtype ) );

  if( t == NULL )
    throw new NoSuchTypeException( rtype,
      ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );

  std::vector<const TypeSymbol*> arg_types;

  for( size_t i = 0; i < types.size(); i++ )
  {
    const TypeSymbol *t =
      dynamic_cast<const TypeSymbol*>( GetSymbol( types[i].c_str() ) );

    if( t == NULL )
      throw new NoSuchTypeException( types[i].c_str(),
        ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );

    arg_types.push_back( t );
  }

  return NewFunction( func, prefix, arg_types, t, weight,
                      NULL, NULL,
                      NULL, NULL,
                      sloc );
}

void SymbolTable::NewFunction( const USTRING &func, bool prefix,
  ProofLib::Term *def, const ProofLib::TermList *var_map,
  unsigned int weight,
  const ProofLib::SourceLocation &sloc )
{
  const TypeSymbol *type = FunctionSymbol::GetTermType( def );

  ProofLib::TermList vars;
  ProofLib::OperationGetVariables op( vars );
  def->Apply( op );
//  def->GetVariables( vars );

  std::vector<const TypeSymbol*> arg_types;

  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    const TypeSymbol *type = FunctionSymbol::GetTermType( vars.GetTerm(i) );

    arg_types.push_back( type );
  }

  if( var_map == NULL || var_map->GetChildCount() == 0 )
    var_map = &vars;
  else
    CheckVarMap( var_map, vars, sloc );

  NewFunction( func, prefix, arg_types, type, weight, def,
               NULL, NULL, var_map, sloc );
}

void SymbolTable::NewFunction( const USTRING &func, bool prefix,
  ProofLib::Formula *through, ProofLib::Formula *by,
  unsigned int weight,
  const ProofLib::SourceLocation &sloc )
{
  ASSERT( through != NULL );
  ASSERT( by != NULL );

  FunctionSymbol *f =
    dynamic_cast<FunctionSymbol*>( (Symbol*) GetSymbol( func ) );

  if( f == NULL )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "function '" << func << "' appears not in through-clause.";
    throw e;
  }

  if( f->IsInfix() != !prefix )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "function '" << func << "' was declared "
                << ( !prefix ? "infix" : "prefix" )
                << " but notated in through-clause otherwise.";
    throw e;
  }

  f->SetThroughByData( through, by, NULL, weight );
}

const FunctionSymbol* SymbolTable::NewFunction(
    const USTRING &func, bool prefix,
    const std::vector<const TypeSymbol*> &types, const TypeSymbol *rtype,
    unsigned int weight,
    ProofLib::Term *def,
    ProofLib::Formula *through, ProofLib::Formula *by,
    const ProofLib::TermList *var_map,
    const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " function: " << func << " of type ";
  if( types.empty() )
    std::cerr << "'empty'";
  else for( size_t i = 0; i < types.size(); i++ )
  {
    if( i != 0 )
      std::cerr << " X ";
    std::cerr << types[i]->GetName();
  }
  std::cerr << " -> " << rtype->GetName()
            << " " << ( prefix ? "prefix" : "infix" ) << " notated"
            << " with weight " << weight
            << std::endl;
#endif

  ASSERT( rtype != NULL );
  ASSERT( def != NULL || types.size() > 0 );

  const Symbol *s = GetSymbol( func, false );

  if( s != NULL )
    throw new SymbolExistsException( s );

  if( types.size() != 2 && !prefix )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "infix functions must have 2 arguments.";
    throw e;
  }

  FunctionSymbol *f = new FunctionSymbol( func, prefix, rtype, types, weight,
                                          def, through, by, var_map );
  f->SetSourceLocation( sloc );

  AddSymbol( f );

  return f;
}


FunctionSymbol* SymbolTable::NewRecursiveFunction(
  const USTRING &func, SymbolTable *symtab,
  const USTRING &rtype,
  const ProofLib::TermList *var_map,
  const ProofLib::SourceLocation &sloc )
{
#if 0 //def DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " recursive function: " << func << std::endl;
#endif

  const TypeSymbol *type = dynamic_cast<const TypeSymbol*>( GetSymbol( rtype ) );

  if( type == NULL )
    throw new NoSuchTypeException( rtype,
      ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );

  std::vector<const TypeSymbol*> arg_types;

  for( size_t i = 0; i < var_map->GetChildCount(); i++ )
  {
    if( dynamic_cast<ProofLib::Variable*>( var_map->GetTerm(i) ) == NULL )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
      e->Stream() << "expression ";
      Print::PrintTerm( var_map->GetTerm(i), e->Stream() );
      e->Stream() << " is not a variable.";
      throw e;
    }

    const TypeSymbol *t = FunctionSymbol::GetTermType( var_map->GetTerm(i) );

    ASSERT( t != NULL );

    arg_types.push_back( t );
  }

  ASSERT( symtab != NULL );

  const Symbol *s = GetSymbol( func, false );

  if( s != NULL )
    throw new SymbolExistsException( s );

  FunctionSymbol *f =
    new FunctionSymbol( func, symtab, type, arg_types, var_map );

  f->SetSourceLocation( sloc );

  AddSymbol( f );

  return f;
}

const PredicateSymbol* SymbolTable::NewAtomPredicate(
  const USTRING &pred, bool prefix,
  std::vector<USTRING> &types,
  unsigned int weight,
  const ProofLib::SourceLocation &sloc,
  bool variable )
{
  std::vector<const TypeSymbol*> typesyms;

  for( size_t i = 0; i < types.size(); i++ )
  {
    const TypeSymbol *t =
      dynamic_cast<const TypeSymbol*>( GetSymbol( types[i].c_str() ) );

    if( t == NULL )
      throw new NoSuchTypeException( types[i].c_str(),
        ProofLib::SourceLocation( sloc.GetFile(), sloc.GetLine(), 0 ) );

    typesyms.push_back( t );
  }

  return NewPredicate( pred, prefix, typesyms, NULL, NULL, weight, sloc, variable );
}

const PredicateSymbol* SymbolTable::NewAtomPredicate(
  const USTRING &pred, bool prefix,
  const ProofLib::SourceLocation &sloc, bool variable )
{
  std::vector<const TypeSymbol*> typesyms;

  return NewPredicate( pred, prefix, typesyms, NULL, NULL, 0, sloc,
                       variable );
}

void SymbolTable::NewPredicate(
  const USTRING &pred, bool prefix,
  ProofLib::Formula *def, const ProofLib::TermList *var_map,
  unsigned int weight,
  const ProofLib::SourceLocation &sloc )
{
  ProofLib::TermList vars;
  ProofLib::OperationGetFreeVariables op( vars );
  def->Apply( op );

  std::vector<const TypeSymbol*> arg_types;

  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    const TypeSymbol *type = FunctionSymbol::GetTermType( vars.GetTerm(i) );

    arg_types.push_back( type );
  }

  if( var_map == NULL || var_map->GetChildCount() == 0 )
    var_map = &vars;
  else
    CheckVarMap( var_map, vars, sloc );

  NewPredicate( pred, prefix, arg_types, def, var_map, weight,
                sloc );
}

const PredicateSymbol* SymbolTable::NewPredicate(
  const USTRING &pred, bool prefix,
  const std::vector<const TypeSymbol*> &arg_types,
  ProofLib::Formula *def, const ProofLib::TermList *var_map,
  unsigned int weight,
  const ProofLib::SourceLocation &sloc, bool variable )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " predicate: " << pred << " of type ";
  if( arg_types.empty() )
    std::cerr << "'empty'";
  for( size_t i = 0; i < arg_types.size(); i++ )
  {
    if( i != 0 )
      std::cerr << " X ";
    std::cerr << arg_types[i]->GetName();
  }
  std::cerr << " " << ( prefix ? "prefix" : "infix" ) << " notated"
            << std::endl;
#endif

  const Symbol *s = GetSymbol( pred, false );

  if( s != NULL )
    throw new SymbolExistsException( s );

  if( arg_types.size() != 2 && !prefix )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
    e->Stream() << "infix predicates must have 2 arguments.";
    throw e;
  }

  PredicateSymbol *p;
  if ( variable )
  {
    // call constructor for variable predicate symbol
    p = new PredicateSymbol( pred, prefix, arg_types, var_map );
  }
  else
  {
    p = new PredicateSymbol( pred, prefix, arg_types,
                                            def, var_map );
  }
  p->SetSourceLocation( sloc );

  AddSymbol( p );

  return p;
}

PredicateSymbol* SymbolTable::NewRecursivePredicate(
  const USTRING &pred, SymbolTable *symtab,
  const ProofLib::TermList *var_map,
  const ProofLib::SourceLocation &sloc )
{
#if 0 //def DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " recursive predicate: " << pred << std::endl;
#endif

  std::vector<const TypeSymbol*> arg_types;

  for( size_t i = 0; i < var_map->GetChildCount(); i++ )
  {
    if( dynamic_cast<ProofLib::Variable*>( var_map->GetTerm(i) ) == NULL )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
      e->Stream() << "expression ";
      Print::PrintTerm( var_map->GetTerm(i), e->Stream() );
      e->Stream() << " is not a variable.";
      throw e;
    }

    const TypeSymbol *t = FunctionSymbol::GetTermType( var_map->GetTerm(i) );

    ASSERT( t != NULL );

    arg_types.push_back( t );
  }

  ASSERT( symtab != NULL );

  const Symbol *s = GetSymbol( pred, false );

  if( s != NULL )
    throw new SymbolExistsException( s );

  PredicateSymbol *p =
    new PredicateSymbol( pred, symtab, arg_types, var_map );

  p->SetSourceLocation( sloc );

  AddSymbol( p );

  return p;
}

void SymbolTable::NewAxiom( const USTRING &name, const ProofLib::SourceLocation &sloc )
{
#ifdef DEBUG_PRINT_DEFINE
  Print::PrintSLoc( sloc, std::cerr );
  std::cerr << " axiom: " << type
            << std::endl;
#endif

  const Symbol *s = GetSymbol( name );

  if( s != NULL )
    throw new SymbolExistsException( s );

  AxiomSymbol *a = new AxiomSymbol( name );
  a->SetSourceLocation( sloc );

  AddSymbol( a );
}

void SymbolTable::AddEqualityDefinition(
  ProofLib::Variable *v1, ProofLib::Variable *v2,
  ProofLib::Formula *f, const ProofLib::SourceLocation &sloc )
{
  m_store->AddEqualityDefinition( v1, v2, f, sloc );
}

size_t SymbolTable::GetEqualityDefinitionCount() const
{
  return m_store->GetEqualityDefinitionCount();
}

ProofLib::Formula* SymbolTable::GetEqualityDefinitionAt( size_t index,
  const TypeSymbol **l, const TypeSymbol **r, USTRING *id ) const
{
  return m_store->GetEqualityDefinitionAt( index, l, r, id );
}

const UCHAR* SymbolTable::GetEqualityDefinitionId(
    const TypeSymbol *l, const TypeSymbol *r ) const
{
  return m_store->GetEqualityDefinitionId( l, r );
}

ProofLib::Formula* SymbolTable::GetEqualityInstance(
    ProofLib::Term *l, ProofLib::Term *r,
    ProofLib::NodeDatabase *ndb ) const
{
  return m_store->GetEqualityInstance( l, r, ndb );
}

ProofLib::Formula* SymbolTable::GetEqualityDefinitionOf(
  const TypeSymbol *l, const TypeSymbol *r ) const
{
  return m_store->GetEqualityDefinitionOf( l, r );
}

ProofLib::Formula* SymbolTable::GetEqualityInstance(
  const TypeSymbol *l, const TypeSymbol *r,
  const ProofLib::VariableSubstitution& sigma,
  ProofLib::NodeDatabase *ndb ) const
{
  return m_store->
    GetEqualityInstance( l, r, sigma, ndb, GetEqualPredicateSymbol() );
}

const ProofLib::Symbol* SymbolTable::Get( const USTRING &name,
                                          size_t arity,
                                          ProofLib::SymbolType type )
{
  const ProofLib::Symbol *s = GetSymbol( name );

  ASSERT( s != NULL );
  ASSERT( s->GetArity() == arity );
  ASSERT( s->GetSymbolType() == type );

  return s;
}

const ProofLib::Symbol* SymbolTable::Add( const USTRING &name,
                                          size_t arity,
                                          ProofLib::SymbolType type )
{
  ASSERT( false );

  return NULL;
}

const ProofLib::Symbol* SymbolTable::GetOrAdd( const USTRING &name,
                                               size_t arity,
                                               ProofLib::SymbolType type )
{
  const ProofLib::Symbol *s = GetSymbol( name );

  ASSERT( s != NULL );
  ASSERT( s->GetArity() == arity );
  ASSERT( s->GetSymbolType() == type );

  return s;
}

const ProofLib::OperatorSymbol*
  SymbolTable::GetOperatorSymbol( ProofLib::OpSymbolType type )
{
  if( GetTopParent() != this )
    return GetTopParent()->GetOperatorSymbol( type );

  ASSERT( type < m_store->m_op_map.size() );
  return m_store->m_op_map[type];
}

size_t SymbolTable::GetArity( const USTRING &name,
                              ProofLib::SymbolType type ) const
{
  const ProofLib::Symbol *s = GetSymbol( name );

  ASSERT( s != NULL );

  return s->GetArity();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Symbol::Symbol( const USTRING &s )
  : m_symbol( s )
{
  ASSERT( !s.empty() );
}

////////////////////////////////////////////////////////////////////////////////

TypeSymbol::TypeSymbol( const USTRING &s, const USTRING &fresh )
  : Symbol( s )
{
  if( !fresh.empty() )
    m_fresh = fresh;
  else
    m_fresh = s;
}

bool TypeSymbol::IsAnyType() const
{
  return GetName() == reserved_words[RESWORD_ANY];
}

////////////////////////////////////////////////////////////////////////////////

SymbolWithType::SymbolWithType( const USTRING &s, const TypeSymbol *type )
  : Symbol( s )
{
  ASSERT( type != NULL );

  m_type = type;
}

////////////////////////////////////////////////////////////////////////////////

VariableSymbol::VariableSymbol( const USTRING &s, const TypeSymbol *type,
                                bool usable_as_bounded )
  : SymbolWithType( s, type )
{
  m_usable_as_bounded = usable_as_bounded;
}

////////////////////////////////////////////////////////////////////////////////

ConstantSymbol::ConstantSymbol( const USTRING &s, const TypeSymbol *type )
  : SymbolWithType( s, type )
{
}

////////////////////////////////////////////////////////////////////////////////

SymbolWithArguments::SymbolWithArguments(
  const std::vector<const TypeSymbol*> &arg_types )
{
  m_arity = arg_types.size();

  m_arg_types = new const TypeSymbol*[m_arity];

  std::vector<const TypeSymbol*>::const_iterator i = arg_types.begin();

  unsigned int j = 0;

  while( i != arg_types.end() )
  {
    m_arg_types[j++] = *i;
    i++;
  }
}

SymbolWithArguments::~SymbolWithArguments()
{
  delete [] m_arg_types;
}

const TypeSymbol* SymbolWithArguments::GetArgType( unsigned int index ) const
{
  ASSERT( index < m_arity );

  return m_arg_types[index];
}

bool SymbolWithArguments::CheckArgType( unsigned int index,
                                        const TypeSymbol *t ) const
{
  const TypeSymbol *my = GetArgType( index );

  if( my->IsAnyType() || t->IsAnyType() )
    return true;

  return ( my == t );
}

////////////////////////////////////////////////////////////////////////////////

FunctionSymbol::FunctionSymbol(
  const USTRING &s, bool bPrefix,
  const TypeSymbol *type,
  const std::vector<const TypeSymbol*> &arg_types,
  unsigned int weight,
  ProofLib::Term *def,
  ProofLib::Formula *through, ProofLib::Formula *by,
  const ProofLib::TermList *var_map )
  : SymbolWithType( s, type ), SymbolWithArguments( arg_types )
{
  ASSERT( def != NULL || arg_types.size() > 0 );

  m_bPrefix = bPrefix;
  m_weight = weight;
  m_bRecursive = false;
  
  m_def = def;

  if( m_def != NULL )
    m_def->Attach();

  m_through = through;
  m_by = by;

  ASSERT( m_def == NULL || ( m_through == NULL && m_by == NULL ) );
  ASSERT( m_through != NULL || m_by == NULL );
  ASSERT( m_by != NULL || m_through == NULL );

  if( m_through != NULL )
  {
    m_through->Attach();
    m_by->Attach();
  }

  if( m_def != NULL )
  {
    if( var_map != NULL )
      m_var_map = new ProofLib::StaticTermList( var_map );
    else if( m_def != NULL )
    {
      ProofLib::TermList list;
      ProofLib::OperationGetVariables op( list );
      m_def->Apply( op );
     // m_def->GetVariables( list );
      m_var_map = new ProofLib::StaticTermList( &list );
    }
  }
  else
    m_var_map = NULL;

  ASSERT( var_map == NULL || var_map->GetChildCount() == arg_types.size() );
}

FunctionSymbol::FunctionSymbol( const USTRING &s, SymbolTable *symtab,
  const TypeSymbol *type,
  const std::vector<const TypeSymbol*> &arg_types,
  const ProofLib::TermList *var_map )
  : SymbolWithType( s, type ), SymbolWithArguments( arg_types )
{
  m_bPrefix = true;
  m_bRecursive = true;
  m_weight = 0;

  m_def = NULL;
  m_through = NULL;
  m_by = NULL;

  if( var_map->GetChildCount() > 0 )
    m_var_map = new ProofLib::StaticTermList( var_map );
  else
    m_var_map = NULL;

  ASSERT( var_map == NULL || var_map->GetChildCount() == arg_types.size() );

  ASSERT( symtab != NULL );
  m_symtab = symtab;
}

FunctionSymbol::~FunctionSymbol()
{
//if( m_def != NULL ) m_def->Detach();
  if( m_var_map != NULL ) delete m_var_map;
}

const TypeSymbol* FunctionSymbol::GetTermType( const ProofLib::Term *t )
{
  const ProofLib::SymbolNode *s =
    dynamic_cast<const ProofLib::SymbolNode*>( t );

  ASSERT( s != NULL );

  const SymbolWithType *w =
    dynamic_cast<const SymbolWithType*>( s->GetSymbol() );

  return w->GetType();
}

void FunctionSymbol::SetThroughByData(
  ProofLib::Formula *through, ProofLib::Formula *by,
  const ProofLib::TermList *var_map, unsigned int weight )
{
  ASSERT( through != NULL );
  ASSERT( by != NULL );
  ASSERT( m_through == NULL );
  ASSERT( m_by == NULL );
  ASSERT( m_def == NULL );

  m_bRecursive = false;
  m_symtab = NULL;

  m_weight = weight;

  m_through = through;
  m_by = by;
  
  if( var_map != NULL )
    m_var_map = new ProofLib::StaticTermList( var_map );
  else
  {
    ProofLib::TermList vars;
    ProofLib::OperationGetFreeVariables op( vars );
/*
  Print::PrintFormula( m_through, std::cerr );
  std::cerr << std::endl;
  Print::PrintFormula( m_by, std::cerr );
  std::cerr << std::endl;
*/
    m_through->Apply( op );
    m_var_map = new ProofLib::StaticTermList( &vars );
  }
}

ProofLib::Term* FunctionSymbol::CreateDefineInstance(
  const ProofLib::VariableSubstitution *sigma,
  ProofLib::NodeDatabase *ndb,
  ProofLib::RecursiveExpr* expr /* = NULL */ ) const
{
  ProofLib::TermList args;

  ASSERT( m_var_map != NULL || IsRecursive() );

  if( !IsRecursive() || m_var_map != NULL )
    for( size_t i = 0; i < m_var_map->GetChildCount(); i++ )
  {
    bool already = false;
    for( size_t j = 0; j < sigma->GetCount(); j++ )
    {
      if( sigma->GetVar(j) == m_var_map->GetTerm(i) )
      {
        if( already ) return NULL;
  
        already = true;
        args.AddRight( (ProofLib::Term*) sigma->GetSubst(j) );
      }
    }
    if( !already ) return NULL; // have to find it
  }

  if( IsRecursive() )
  {
    ASSERT( expr != NULL );
    return ProofLib::RecursiveFunction::New( ndb, this, &args, expr );
  }
  else
    return ProofLib::Function::New( ndb, this, &args );
}

ProofLib::Term* FunctionSymbol::CreateBodyInstance(
  ProofLib::Term *d_0,
  ProofLib::NodeDatabase *ndb,
  std::vector<ProofLib::NodePointers> &positions,
  unsigned int level /*= 0*/,
  HLK::SymbolTable* symtab /* = NULL */,
  const ProofLib::SourceLocation* sloc /* = NULL */ ) const
{
  ProofLib::Function *f_f = d_0->CastTo<ProofLib::Function>();
  ASSERT( f_f != NULL );
  ASSERT( f_f->GetSymbol() == this );

  ProofLib::OperationVariableSubstitution tr( ndb );

  if( GetVariableMapping() == NULL )
  {
    ASSERT( f_f->GetChildCount() == 0 );
    ASSERT( IsRecursive() );
  }
  else
  {
    ASSERT( GetVariableMapping()->GetChildCount() == f_f->GetChildCount() );
  
    for( size_t i = 0; i < GetVariableMapping()->GetChildCount(); i++ )
    {
      tr.AddSubstitution(
        GetVariableMapping()->GetTerm(i)->CastTo<ProofLib::Variable>(),
        f_f->GetTerm(i) );
    }
  }

  ProofLib::Term *T = NULL;

  if( IsRecursive() )
  {
    ASSERT( symtab != NULL );

    T = GetTermAtLevel( level, ndb, symtab, *sloc );
  }
  else
    T = GetFunctionDefinition();

  ProofLib::Term* ret = ProofLib::DataNode::ApplyTo( T, tr );

  positions = tr.GetPositions();

  return ret;
}

ProofLib::Formula* FunctionSymbol::CreateThroughInstance(
  const ProofLib::VariableSubstitution *sigma,
  ProofLib::NodeDatabase *ndb ) const
{
  ASSERT( sigma != NULL );
  ASSERT( m_through != NULL );

  ProofLib::Formula *f = m_through;

#if 0
  for( size_t j = 0; j < sigma->GetCount(); j++ )
  {
    ProofLib::OperationVariableSubstitution
      tr( ndb, (ProofLib::Variable*) sigma->GetVar(j),
               (ProofLib::Term*) sigma->GetSubst(j) );

    f = ProofLib::DataNode::ApplyTo( f, tr );
  }
#endif
  ProofLib::OperationVariableSubstitution tr( ndb );

  for( size_t j = 0; j < sigma->GetCount(); j++ )
    tr.AddSubstitution( sigma->GetVar(j), sigma->GetSubst(j) );

  f = ProofLib::DataNode::ApplyTo( f, tr );

  return f;
}

ProofLib::Formula* FunctionSymbol::CreateByInstance(
  const ProofLib::VariableSubstitution *sigma,
  ProofLib::NodeDatabase *ndb,
  std::vector<ProofLib::NodePointers> &positions ) const
{
  ASSERT( sigma != NULL );
  ASSERT( m_by != NULL );

  ProofLib::Formula *f = m_by;

#if 0
  for( size_t j = 0; j < sigma->GetCount(); j++ )
  {
    ProofLib::OperationVariableSubstitution
      tr( ndb, (ProofLib::Variable*) sigma->GetVar(j),
               (ProofLib::Term*) sigma->GetSubst(j) );

    f = ProofLib::DataNode::ApplyTo( f, tr );
  }
#endif
  ProofLib::OperationVariableSubstitution tr( ndb );

  for( size_t j = 0; j < sigma->GetCount(); j++ )
    tr.AddSubstitution( sigma->GetVar(j), sigma->GetSubst(j) );

  f = ProofLib::DataNode::ApplyTo( f, tr );

  positions = tr.GetPositions();

  return f;
}

void FunctionSymbol::AddRecursiveLevel(
  ProofLib::RecursiveExpr *expr,
  ProofLib::Term *t )
{
  if( FunctionSymbol::GetTermType( t ) != GetType() )
  {
    ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );

    e->Stream() << "recursive function " << GetName()
                << ": recursive term is of type "
                << FunctionSymbol::GetTermType( t )->GetName()
                << " while function as defined of type " << GetType()->GetName()
                << ".";

    throw e;
  }

  ProofLib::TermList vars;
  ProofLib::OperationGetVariables op( vars );
  t->Apply( op );

  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    if( m_var_map == NULL || m_var_map->FindTerm( vars.GetTerm(i) ) < 0 )
    {
      ProofLib::Variable *v = vars.GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );

      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );

      e->Stream() << "recursive function " << GetName()
                  << ": term contains a variable which is not "
                  << "defined as argument: "
                  << v->GetSymbol()->GetName() << ".";

      throw e;
    }
  }

  expr->Attach();
  t->Attach();
  m_rec_exprs.push_back( expr );
  m_rec_terms.push_back( t );

  ASSERT( m_rec_exprs.size() == m_rec_terms.size() );
}

size_t FindLevel(
  const std::vector<ProofLib::RecursiveExpr*> &exprs,
  size_t &level,
  bool &found,
  const ProofLib::SourceLocation &sloc )
{
  found = false;

  for( size_t i = 0; i < exprs.size(); i++ )
  {
    if( exprs[i]->GetVariableName() == NULL )
      if( level == exprs[i]->GetNum() )
      {
        found = true;
        return i;
      }
      else
        continue;

    switch( exprs[i]->GetOp() )
    {
    case '+':
      if( level >= exprs[i]->GetNum() )
      {
        level -= exprs[i]->GetNum();
        found = true;
        return i;
      }
      break;
    case '*':
      if( exprs[i]->GetNum() == 0 )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
        e->Stream() << "Multiplation by 0 is not allowed.";
        throw e;
      }

      if( level % exprs[i]->GetNum() == 0 )
      {
        level /= exprs[i]->GetNum();
        found = true;
        return i;
      }
      break;

    default:
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
      e->Stream() << "Invalid recursion operator '" << exprs[i]->GetOp();
      e->Stream() << "'.";
      throw e;
    }
  }

  ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( &sloc );
  e->Stream() << "No suitable body found for level " << level;
  e->Stream() << ".";
  throw e;

  return 0;
}

ProofLib::Term* FunctionSymbol::GetTermAtLevel(
    size_t level,
    ProofLib::NodeDatabase *ndb,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc ) const
{
#if 0
  std::cerr << "level: " << level << std::endl;
  Print::PrintTerm( m_t0, std::cerr );
  std::cerr << "." << std::endl;
  Print::PrintTerm( m_tn, std::cerr );
  std::cerr << "." << std::endl;
#endif
  bool found = false;
  size_t index = FindLevel( m_rec_exprs, level, found, sloc );

  if( !found )
    return NULL;

  ProofRecursionsExpander expander( ndb, level, symbol_table, sloc );

  return ProofLib::DataNode::ApplyTo( m_rec_terms[index], expander );
}

bool FunctionSymbol::IsDefined( const ProofLib::TermListBase *terms ) const
{
  return ( m_def != NULL || m_through != NULL || m_rec_terms.size() > 0 );
}

const UCHAR*
FunctionSymbol::GetDefinitionId( const ProofLib::TermListBase *terms ) const
{
  return GetName().c_str();
}


////////////////////////////////////////////////////////////////////////////////

PredicateSymbol::PredicateSymbol(
  const USTRING &s, bool bPrefix,
  const std::vector<const TypeSymbol*> &arg_types,
  ProofLib::Formula *def, const ProofLib::TermList *var_map,
  SymbolTable *symtab )
  : Symbol( s ), SymbolWithArguments( arg_types )
{
  m_bRecursive = false;
  m_symtab = symtab;
  m_bVariable = false;

  ASSERT( def != NULL || var_map == NULL );

  m_bPrefix = bPrefix;

  m_def = def;

  if( m_def != NULL )
  {
    m_def->Attach();

    if( var_map != NULL )
      m_var_map = new ProofLib::StaticTermList( var_map );
    else
    {
      ProofLib::TermList list;
      //m_def->GetFreeVariables( list );
      ProofLib::OperationGetFreeVariables op( list );
      m_def->Apply( op );

      m_var_map = new ProofLib::StaticTermList( &list );
    }
  }
  else
    m_var_map = NULL;

  ASSERT( var_map == NULL || var_map->GetChildCount() == arg_types.size() );
}

PredicateSymbol::PredicateSymbol( const USTRING &s, bool bPrefix,
                   const std::vector<const TypeSymbol*> &arg_types,
		   const ProofLib::TermList *var_map,
		   SymbolTable *symtab )
                   : Symbol( s ), SymbolWithArguments( arg_types )
{
  m_bRecursive = false;
  m_symtab = symtab;
  m_def = NULL;
  m_bPrefix = bPrefix;
  m_var_map = new ProofLib::StaticTermList( var_map );
  m_bVariable = true;

  ASSERT( var_map == NULL || var_map->GetChildCount() == arg_types.size() );
}

PredicateSymbol::PredicateSymbol(
  const USTRING &s, SymbolTable *symtab,
  const std::vector<const TypeSymbol*> &arg_types,
  const ProofLib::TermList *var_map )
  : Symbol( s ), SymbolWithArguments( arg_types )
{
  m_bRecursive = true;
  m_bPrefix = true;
  m_def = NULL;
  m_bVariable = false;

  ASSERT( var_map != NULL );
  ASSERT( symtab != NULL );

  m_var_map = new ProofLib::StaticTermList( var_map );

  m_symtab = symtab;

  ASSERT( var_map == NULL || var_map->GetChildCount() == arg_types.size() );
}

void PredicateSymbol::AddRecursiveLevel(
  ProofLib::RecursiveExpr *expr,
  ProofLib::Formula *f )
{
  ProofLib::TermList vars;
  ProofLib::OperationGetFreeVariables op( vars );
  f->Apply( op );

  for( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    if( m_var_map == NULL || m_var_map->FindTerm( vars.GetTerm(i) ) < 0 )
    {
      ProofLib::Variable *v = vars.GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );

      ProofLib::GenericSLocException *e =
        new ProofLib::GenericSLocException( this );

      e->Stream() << "recursive predicate " << GetName()
                  << ": formula contains a variable which is not "
                  << "defined as argument: "
                  << v->GetSymbol()->GetName() << ".";

      throw e;
    }
  }

  expr->Attach();
  f->Attach();
  m_rec_exprs.push_back( expr );
  m_rec_formulas.push_back( f );

  ASSERT( m_rec_exprs.size() == m_rec_formulas.size() );
}

#if 0
void PredicateSymbol::SetRecursiveFormulas( ProofLib::Formula *f_0,
                                            ProofLib::Formula *f_n )
{
  ASSERT( f_0 != NULL );
  ASSERT( f_n != NULL );

  ProofLib::TermList vars_0;
  ProofLib::OperationGetFreeVariables op_0( vars_0 );
  f_0->Apply( op_0 );

  for( size_t i = 0; i < vars_0.GetChildCount(); i++ )
  {
    if( m_var_map->FindTerm( vars_0.GetTerm(i) ) < 0 )
    {
      ProofLib::Variable *v = vars_0.GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );

      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );

      e->Stream() << "recursive function " << GetName()
                  << ": term at level 0 contains a variable which is not "
                  << "defined as argument: "
                  << v->GetSymbol()->GetName() << ".";

      throw e;
    }
  }

  ProofLib::TermList vars_n;
  ProofLib::OperationGetFreeVariables op_n( vars_n );
  f_n->Apply( op_n );

  for( size_t i = 0; i < vars_n.GetChildCount(); i++ )
  {
    if( m_var_map->FindTerm( vars_n.GetTerm(i) ) < 0 )
    {
      ProofLib::Variable *v = vars_n.GetTerm(i)->CastTo<ProofLib::Variable>();
      ASSERT( v != NULL );

      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( this );

      e->Stream() << "recursive function " << GetName()
                  << ": term at level n contains a variable which is not "
                  << "defined as argument: "
                  << v->GetSymbol()->GetName() << ".";

      throw e;
    }
  }

  m_f0 = f_0;
  m_fn = f_n;
}
#endif


PredicateSymbol::~PredicateSymbol()
{
//if( m_def != NULL ) m_def->Detach();
  if( m_var_map != NULL ) delete m_var_map;
}

bool PredicateSymbol::IsEqualPred() const
{
  USTRING equal;
  equal.FromUtf8( "=" );
  return GetName() == equal;
}

ProofLib::Formula* PredicateSymbol::CreateDefineInstance(
  const ProofLib::VariableSubstitution *sigma,
  ProofLib::NodeDatabase *ndb,
  ProofLib::RecursiveExpr* expr /* = NULL */ ) const
{
  ProofLib::TermList args;

  for( size_t i = 0; i < m_var_map->GetChildCount(); i++ )
  {
    bool already = false;
    for( size_t j = 0; j < sigma->GetCount(); j++ )
    {
      if( sigma->GetVar(j) == m_var_map->GetTerm(i) )
      {
        ASSERT( !already ); // oops, it appears twice in sigma!
  
        already = true;
        args.AddRight( (ProofLib::Term*) sigma->GetSubst(j) );
      }
    }
    if( !already ) // default value
     args.AddRight( m_var_map->GetTerm(i) );
  }

  if( IsRecursive() )
  {
    ASSERT( expr != NULL );
    return ProofLib::RecursivePredicate::New( ndb, this, &args, expr );
  }
  else
    return ProofLib::AtomFormula::New( ndb, this, &args );
}

ProofLib::Formula* PredicateSymbol::CreateBodyInstance(
  ProofLib::Formula *D_0,
  ProofLib::NodeDatabase *ndb,
  std::vector<ProofLib::NodePointers> &positions ) const
{
  ProofLib::AtomFormula *D_f = D_0->CastTo<ProofLib::AtomFormula>();
  ASSERT( D_f != NULL );
  ASSERT( D_f->GetSymbol() == this );

  ProofLib::OperationVariableSubstitution tr( ndb );

  ASSERT( GetVariableMapping()->GetChildCount() == D_f->GetChildCount() );

  for( size_t i = 0; i < GetVariableMapping()->GetChildCount(); i++ )
  {
    tr.AddSubstitution(
      GetVariableMapping()->GetTerm(i)->CastTo<ProofLib::Variable>(),
      D_f->GetTerm(i) );
  }

  ProofLib::Formula* ret = 
    ProofLib::DataNode::ApplyTo( GetDefinedFormula(), tr );

  positions = tr.GetPositions();

  return ret;
}

ProofLib::Formula* PredicateSymbol::GetFormulaAtLevel(
    size_t level,
    ProofLib::NodeDatabase *ndb,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc ) const
{
#if 0
  ProofRecursionsExpander expander( ndb, level, symbol_table, sloc );

  if( level == 0 )
    return ProofLib::DataNode::ApplyTo( m_f0, expander );

  return ProofLib::DataNode::ApplyTo( m_fn, expander );
#endif
  bool found = false;
  size_t index = FindLevel( m_rec_exprs, level, found, sloc );

  if( !found )
    return NULL;

  ProofRecursionsExpander expander( ndb, level, symbol_table, sloc );

  return ProofLib::DataNode::ApplyTo( m_rec_formulas[index], expander );
}

bool PredicateSymbol::IsDefined( const ProofLib::TermListBase *terms ) const
{
#if 0
  std::cerr << "FUNCTION: " << GetName();
  std::cerr << std::endl;
  std::cerr << "m_def is " << ( m_def == NULL ? "" : "not " ) << "null.";
  std::cerr << std::endl;
  std::cerr << "m_through is " << ( m_through == NULL ? "" : "not " ) << "null.";
  std::cerr << std::endl;
  std::cerr << "m_t0 is " << ( m_t0 == NULL ? "" : "not " ) << "null.";
  std::cerr << std::endl;
#endif
  if( IsEqualPred() )
  {
    ASSERT( terms != NULL );
    ASSERT( terms->GetChildCount() == 2 );
    ASSERT( m_symtab != NULL );
    
    const TypeSymbol *l = FunctionSymbol::GetTermType( terms->GetTerm(0) );
    const TypeSymbol *r = FunctionSymbol::GetTermType( terms->GetTerm(1) );

    return m_symtab->GetEqualityDefinitionOf( l, r ) != NULL;
  }

  return ( m_def != NULL || m_rec_formulas.size() > 0 );
}

const UCHAR*
PredicateSymbol::GetDefinitionId( const ProofLib::TermListBase *terms ) const
{
  if( IsEqualPred() )
  {
    ASSERT( terms != NULL );
    ASSERT( terms->GetChildCount() == 2 );
    ASSERT( m_symtab != NULL );
    
    const TypeSymbol *l = FunctionSymbol::GetTermType( terms->GetTerm(0) );
    const TypeSymbol *r = FunctionSymbol::GetTermType( terms->GetTerm(1) );

    return m_symtab->GetEqualityDefinitionId( l, r );
  }

  return GetName().c_str();
}


////////////////////////////////////////////////////////////////////////////////

AxiomSymbol::AxiomSymbol( const USTRING &s ) : Symbol( s ) { }

////////////////////////////////////////////////////////////////////////////////

RuleSymbol::RuleSymbol( const USTRING &s, int ruleid )
  : Symbol( s )
{
  m_ruleid = ruleid;
}

////////////////////////////////////////////////////////////////////////////////

OperatorSymbol::OperatorSymbol( const USTRING &name,
                                ProofLib::OpSymbolType type,
                                UCHAR c, size_t arity )
  : HLK::Symbol( name )
{
  m_type = type;
  m_arity = arity;
  m_dc = c;
}

bool OperatorSymbol::IsAndJunctor() const
{
  return m_type == OP_SYMBOL_AND;
}

bool OperatorSymbol::IsOrJunctor() const
{
  return m_type == OP_SYMBOL_OR;
}

bool OperatorSymbol::IsImplJunctor() const
{
  return m_type == OP_SYMBOL_IMPL;
}

bool OperatorSymbol::IsNegJunctor() const
{
  return m_type == OP_SYMBOL_NEG;
}

bool OperatorSymbol::IsAllQuantor() const
{
  return m_type == OP_SYMBOL_ALL;
}

bool OperatorSymbol::IsExistsQuantor() const
{
  return m_type == OP_SYMBOL_EX;
}

bool OperatorSymbol::IsSecondOrderAllQuantor() const
{
  return m_type == OP_SYMBOL_ALL2;
}

bool OperatorSymbol::IsSecondOrderExistsQuantor() const
{
  return m_type == OP_SYMBOL_EX2;
}

////////////////////////////////////////////////////////////////////////////////

ProofSymbol::ProofSymbol( const USTRING &s )
  : Symbol( s )
{
}

////////////////////////////////////////////////////////////////////////////////

bool TypeSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool ProofSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool ConstantSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool VariableSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool FunctionSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool AxiomSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool RuleSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool PredicateSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}

bool OperatorSymbol::operator == ( const ProofLib::Symbol& o ) const
{
  ASSERT(false);
}
}
