////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: symbolhlk.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _SYMBOLHLK_H_
#define _SYMBOLHLK_H_

#include <vector>
#include <util/sloc.h>
#include "symboltable.h"

////////////////////////////////////////////////////////////////////////////////

#define SYMBOL_TYPE_TYPE                    SYMBOL_TYPE_USER
#define SYMBOL_TYPE_AXIOM                   SYMBOL_TYPE_USER + 1

////////////////////////////////////////////////////////////////////////////////

namespace ProofLib
{
  class Variable;
  class Term;
  class TermList;
  class StaticTermList;
  class Formula;
  class NodeDatabase;
  class VariableSubstitution;
  class RecursiveExpr;
  class NodePointers;
}

////////////////////////////////////////////////////////////////////////////////

namespace HLK
{

extern const wchar_t *symbol_type_names[];

class Symbol;
class SymbolTableStore;
class TypeSymbol;
class SymbolWithType;
class VariableSymbol;
class ConstantSymbol;
class FunctionSymbol;
class PredicateSymbol;

////////////////////////////////////////////////////////////////////////////////

size_t FindLevel(
  const std::vector<ProofLib::RecursiveExpr*> &exprs,
  size_t &level,
  bool &found,
  const ProofLib::SourceLocation &sloc );

////////////////////////////////////////////////////////////////////////////////

class SymbolExistsException : public Exception
{
  const Symbol *m_symbol;

public:
  SymbolExistsException( const Symbol *s );

  virtual void Print( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class NoSuchTypeException : public ProofLib::SLocException
{
  USTRING m_type;

public:
  NoSuchTypeException( const USTRING &type, const ProofLib::SourceLocation &sloc );

  virtual void Print( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class SymbolTable : public ProofLib::SymbolTable
{
  SymbolTableStore           *m_store;

  const TypeSymbol           *m_AnyTypeSymbol;
  const PredicateSymbol      *m_EqualPredicateSymbol;

  SymbolTable                *m_parent;

private:
  SymbolTable( const SymbolTable& ) {}
  SymbolTable& operator=( const SymbolTable& ) { return *this; }

public:
  SymbolTable( SymbolTable *parent = NULL, const USTRING &rev_var = NULL );
  virtual ~SymbolTable();

  SymbolTable* GetParent() const { return m_parent; }
  SymbolTable* GetTopParent(); // may be itself, so NOT const

  virtual const ProofLib::Symbol *Get( const USTRING &name,
                                       size_t arity,
                                       ProofLib::SymbolType type );
  virtual const ProofLib::Symbol *Add( const USTRING &name,
                                       size_t arity,
                                       ProofLib::SymbolType type );
  virtual const ProofLib::Symbol *GetOrAdd( const USTRING &name,
                                            size_t arity,
                                            ProofLib::SymbolType type );
  virtual const ProofLib::OperatorSymbol*
    GetOperatorSymbol( ProofLib::OpSymbolType type );
  virtual size_t GetArity( const USTRING &name,
                           ProofLib::SymbolType type ) const;

  const Symbol* GetSymbol( const USTRING &s, bool parent_lookup = true ) const;
  const USTRING& GetRecursionVar() const;

  void NewType( const USTRING &type, const USTRING &fresh,
                const ProofLib::SourceLocation &sloc );
  const VariableSymbol* NewVariable(
    const USTRING &var, const USTRING &type,
    bool usable_as_bounded,
    const ProofLib::SourceLocation &sloc );
  void NewConstant( const USTRING &con, const USTRING &type,
                    const ProofLib::SourceLocation &sloc );
  const FunctionSymbol* NewFunction(
    const USTRING &func, bool prefix,
    std::vector<USTRING> &types,
    const USTRING &rtype,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc );
  void NewFunction( const USTRING &func, bool prefix,
    ProofLib::Term *def, const ProofLib::TermList *var_map,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc );
  void NewFunction( const USTRING &func, bool prefix,
    ProofLib::Formula *through, ProofLib::Formula *by,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc );
  FunctionSymbol* NewRecursiveFunction( const USTRING &func,
    SymbolTable *symtab,
    const USTRING &rtype,
    const ProofLib::TermList *var_map,
    const ProofLib::SourceLocation &sloc );
  const PredicateSymbol* NewAtomPredicate(
    const USTRING &pred, bool prefix,
    const ProofLib::SourceLocation &sloc,
    bool variable = false);
  const PredicateSymbol* NewAtomPredicate(
    const USTRING &pred, bool prefix,
    std::vector<USTRING> &types,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc,
    bool variable = false);
  void NewPredicate(
    const USTRING &pred, bool prefix,
    ProofLib::Formula *def, const ProofLib::TermList *var_map,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc );
  PredicateSymbol* NewRecursivePredicate( const USTRING &pred,
    SymbolTable *symtab,
    const ProofLib::TermList *var_map,
    const ProofLib::SourceLocation &sloc );
  void NewProof( const USTRING &proof,
                 const ProofLib::SourceLocation &sloc );
  void NewAxiom( const USTRING &name, const ProofLib::SourceLocation &sloc );

  void AddEqualityDefinition(
    ProofLib::Variable *v1, ProofLib::Variable *v2,
    ProofLib::Formula *f, const ProofLib::SourceLocation &sloc );
  ProofLib::Formula* GetEqualityDefinitionOf(
    const TypeSymbol *l, const TypeSymbol *r ) const;
  ProofLib::Formula* GetEqualityInstance(
    const TypeSymbol *l, const TypeSymbol *r,
    const ProofLib::VariableSubstitution& sigma,
    ProofLib::NodeDatabase *ndb ) const;
  ProofLib::Formula* GetEqualityInstance(
    ProofLib::Term *l, ProofLib::Term *r,
    ProofLib::NodeDatabase *ndb ) const;

  size_t GetEqualityDefinitionCount() const;
  ProofLib::Formula* GetEqualityDefinitionAt( size_t index,
    const TypeSymbol **l, const TypeSymbol **r, USTRING *id ) const;
  const UCHAR* GetEqualityDefinitionId(
    const TypeSymbol *l, const TypeSymbol *r ) const;

  const TypeSymbol* GetAnyTypeSymbol()
    { return m_AnyTypeSymbol; }
  const PredicateSymbol* GetEqualPredicateSymbol() const
    { return m_EqualPredicateSymbol; }

private:
  const FunctionSymbol* NewFunction(
    const USTRING &func, bool prefix,
    const std::vector<const TypeSymbol*> &types, const TypeSymbol *rtype,
    unsigned int weight,
    ProofLib::Term *def,
    ProofLib::Formula *through, ProofLib::Formula *by,
    const ProofLib::TermList *var_map,
    const ProofLib::SourceLocation &sloc );
  const PredicateSymbol* NewPredicate(
    const USTRING &pred, bool prefix,
    const std::vector<const TypeSymbol*> &arg_types,
    ProofLib::Formula *def, const ProofLib::TermList *var_map,
    unsigned int weight,
    const ProofLib::SourceLocation &sloc,
    bool variable = false );
  void AddSymbol( Symbol *s );
};

////////////////////////////////////////////////////////////////////////////////

class Symbol : public ProofLib::Symbol, public ProofLib::SourceLocation
{
  USTRING m_symbol;

public:
  Symbol( const USTRING &s );
  virtual ~Symbol() {}

  const USTRING& GetName() const { return m_symbol; }

  virtual ProofLib::SymbolType GetSymbolType() const = 0;
  const wchar_t* GetSymbolTypeName() const
    { return symbol_type_names[GetSymbolType()]; }
  
  virtual size_t GetArity() const { return 0; }

  size_t Hash() const;
};

////////////////////////////////////////////////////////////////////////////////

class TypeSymbol : public Symbol
{
  USTRING m_fresh;

protected:
  TypeSymbol( const USTRING &s, const USTRING &fresh );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_TYPE; }

  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  bool IsAnyType() const;
  const USTRING& GetFreshStart() const { return m_fresh; }
};

////////////////////////////////////////////////////////////////////////////////

class SymbolWithType : public Symbol
{
  const TypeSymbol *m_type;

protected:
  SymbolWithType( const USTRING &s, const TypeSymbol *type );

  friend class SymbolTable;

public:
  const TypeSymbol* GetType() const { return m_type; }
};

////////////////////////////////////////////////////////////////////////////////

class VariableSymbol : public SymbolWithType
{
  bool m_usable_as_bounded;

protected:
  VariableSymbol( const USTRING &s, const TypeSymbol *type,
                  bool usable_as_bounded = true );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_VARIABLE; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  bool MayBeUsedAsBounded() const { return m_usable_as_bounded; }
};

////////////////////////////////////////////////////////////////////////////////

class ConstantSymbol : public SymbolWithType
{
protected:
  ConstantSymbol( const USTRING &s, const TypeSymbol *type );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_CONSTANT; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;
};

////////////////////////////////////////////////////////////////////////////////

class SymbolWithArguments
{
  const TypeSymbol **m_arg_types;
  unsigned int m_arity;

protected:
  SymbolWithArguments( const std::vector<const TypeSymbol*> &arg_types );
  virtual ~SymbolWithArguments();

  friend class SymbolTable;

public:
  size_t GetArity() const { return m_arity; }

  const TypeSymbol* GetArgType( unsigned int index ) const;
  bool CheckArgType( unsigned int index, const TypeSymbol *t ) const;
};

////////////////////////////////////////////////////////////////////////////////

class FunctionSymbol : public SymbolWithType, public SymbolWithArguments
{
  bool m_bPrefix;
  bool m_bRecursive;
  unsigned int m_weight;

  ProofLib::Term *m_def;
  ProofLib::Formula *m_through;
  ProofLib::Formula *m_by;
  ProofLib::StaticTermList *m_var_map;

  std::vector<ProofLib::Term*> m_rec_terms;
  std::vector<ProofLib::RecursiveExpr*> m_rec_exprs;

  SymbolTable *m_symtab;

protected:
  FunctionSymbol( const USTRING &s, bool bPrefix, const TypeSymbol *type,
                  const std::vector<const TypeSymbol*> &arg_types,
                  unsigned int weight,
                  ProofLib::Term *def,
                  ProofLib::Formula *through, ProofLib::Formula *by,
                  const ProofLib::TermList *var_map );
  FunctionSymbol( const USTRING &s, SymbolTable *symtab,
                  const TypeSymbol *type,
                  const std::vector<const TypeSymbol*> &arg_types,
                  const ProofLib::TermList *var_map );
  virtual ~FunctionSymbol();

  void SetThroughByData(
    ProofLib::Formula *through, ProofLib::Formula *by,
    const ProofLib::TermList *var_map, unsigned int weight );

  friend class SymbolTable;

public:
  virtual bool IsDefined( const ProofLib::TermListBase *terms = NULL ) const;
  virtual const UCHAR* GetDefinitionId(
                          const ProofLib::TermListBase *terms = NULL ) const;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_FUNCTION; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  virtual size_t GetArity() const { return SymbolWithArguments::GetArity(); }

  virtual bool IsInfix() const { return !m_bPrefix; }
  unsigned int GetWeight() const { return m_weight; }

  bool IsRecursive() const { return m_bRecursive; }

  ProofLib::Term* GetFunctionDefinition() const { return m_def; }
  const ProofLib::StaticTermList* GetVariableMapping() const
    { return m_var_map; }
  ProofLib::Formula* GetThroughFormula() const { return m_through; }
  ProofLib::Formula* GetByFormula() const { return m_by; }

  void AddRecursiveLevel( ProofLib::RecursiveExpr *expr, ProofLib::Term *t );
  ProofLib::Term* GetTermAtLevel(
    size_t level,
    ProofLib::NodeDatabase *ndb,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc ) const;

  ProofLib::Term* CreateDefineInstance(
    const ProofLib::VariableSubstitution *sigma,
    ProofLib::NodeDatabase *ndb,
    ProofLib::RecursiveExpr* expr = NULL ) const;
  ProofLib::Term* CreateBodyInstance(
    ProofLib::Term *d_0,
    ProofLib::NodeDatabase *ndb,
    std::vector<ProofLib::NodePointers> &positions,
    unsigned int level = 0,
    HLK::SymbolTable* symtab = NULL,
    const ProofLib::SourceLocation* sloc = NULL ) const;
  ProofLib::Formula* CreateByInstance(
    const ProofLib::VariableSubstitution *sigma,
    ProofLib::NodeDatabase *ndb,
    std::vector<ProofLib::NodePointers> &positions ) const;
  ProofLib::Formula* CreateThroughInstance(
    const ProofLib::VariableSubstitution *sigma,
    ProofLib::NodeDatabase *ndb ) const;


public:
  static const TypeSymbol* GetTermType( const ProofLib::Term *t );
};

////////////////////////////////////////////////////////////////////////////////

class PredicateSymbol : public Symbol, public SymbolWithArguments
{
  bool m_bPrefix;
  bool m_bRecursive;
  bool m_bVariable;

  ProofLib::Formula *m_def;
  ProofLib::StaticTermList *m_var_map;

#if 0
  ProofLib::Formula *m_f0;
  ProofLib::Formula *m_fn;
#endif
  std::vector<ProofLib::Formula*> m_rec_formulas;
  std::vector<ProofLib::RecursiveExpr*> m_rec_exprs;

  SymbolTable *m_symtab;

protected:
  PredicateSymbol( const USTRING &s, bool bPrefix,
                   const std::vector<const TypeSymbol*> &arg_types,
                   ProofLib::Formula *def, const ProofLib::TermList *var_map,
                   SymbolTable *symtab = NULL );
  PredicateSymbol( const USTRING &s, SymbolTable *symtab,
                   const std::vector<const TypeSymbol*> &arg_types,
                   const ProofLib::TermList *var_map );
  // constructor for predicate variables
  PredicateSymbol( const USTRING &s, bool bPrefix,
                   const std::vector<const TypeSymbol*> &arg_types,
		   const ProofLib::TermList *var_map,
		   SymbolTable *symtab = NULL );
  virtual ~PredicateSymbol();

  friend class SymbolTable;

public:
  virtual bool IsDefined( const ProofLib::TermListBase *terms = NULL ) const;
  virtual const UCHAR* GetDefinitionId(
                          const ProofLib::TermListBase *terms = NULL ) const;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_PREDICATE; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  virtual size_t GetArity() const { return SymbolWithArguments::GetArity(); }

  const ProofLib::StaticTermList* GetVariableMapping() const
    { return m_var_map; }
  ProofLib::Formula* GetDefinedFormula() const { return m_def; }
  ProofLib::Formula* CreateDefineInstance(
    const ProofLib::VariableSubstitution *sigma,
    ProofLib::NodeDatabase *ndb,
    ProofLib::RecursiveExpr* expr = NULL ) const;
  ProofLib::Formula* CreateBodyInstance(
    ProofLib::Formula *D,
    ProofLib::NodeDatabase *ndb,
    std::vector<ProofLib::NodePointers> &positions ) const;

  virtual bool IsInfix() const { return !m_bPrefix; }
  bool IsEqualPred() const;
  bool IsVariable() const { return m_bVariable; }

  void AddRecursiveLevel( ProofLib::RecursiveExpr *expr, ProofLib::Formula *f );
#if 0
public:
  void SetRecursiveFormulas( ProofLib::Formula *f_0, ProofLib::Formula *f_n );

  ProofLib::Formula* GetFormulaAt0() const { return m_f0; }
  ProofLib::Formula* GetFormulaAtN() const { return m_fn; }
#endif
  ProofLib::Formula* GetFormulaAtLevel(
    size_t level,
    ProofLib::NodeDatabase *ndb,
    SymbolTable *symbol_table,
    const ProofLib::SourceLocation &sloc ) const;

  bool IsRecursive() const { return m_bRecursive; }

};

////////////////////////////////////////////////////////////////////////////////

class AxiomSymbol : public Symbol
{
protected:
  AxiomSymbol( const USTRING &s );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_AXIOM; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;
};

////////////////////////////////////////////////////////////////////////////////

class RuleSymbol : public Symbol
{
  int m_ruleid;

protected:
  RuleSymbol( const USTRING &s, int ruleid );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_RULE; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  int GetRuleId() const { return m_ruleid; }
};

////////////////////////////////////////////////////////////////////////////////

class OperatorSymbol : public HLK::Symbol, public ProofLib::OperatorSymbol
{
  ProofLib::OpSymbolType    m_type;
  size_t                    m_arity;
  UCHAR                     m_dc;

protected:
  OperatorSymbol( const USTRING &name, ProofLib::OpSymbolType type,
                  UCHAR c, size_t arity );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_JUNCTOR; }

  virtual bool operator == ( const ProofLib::Symbol& o ) const;

  virtual const USTRING& GetName() const { return HLK::Symbol::GetName(); }
  virtual size_t GetArity() const { return m_arity; }

  virtual size_t Hash() const { return HLK::Symbol::Hash(); }

  virtual UCHAR GetDisplayChar() const { return m_dc; }

  virtual bool IsAndJunctor() const;
  virtual bool IsOrJunctor() const;
  virtual bool IsImplJunctor() const;
  virtual bool IsNegJunctor() const;

  virtual bool IsAllQuantor() const;
  virtual bool IsExistsQuantor() const;
  
  virtual bool IsSecondOrderAllQuantor() const;
  virtual bool IsSecondOrderExistsQuantor() const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofSymbol : public Symbol
{
protected:
  ProofSymbol( const USTRING &s );

  friend class SymbolTable;

public:
  virtual ProofLib::SymbolType GetSymbolType() const
    { return SYMBOL_TYPE_PROOF; }
  virtual bool operator == ( const ProofLib::Symbol& o ) const;
};

////////////////////////////////////////////////////////////////////////////////

}

#endif
