////////////////////////////////////////////////////////////////////////////////
//
// Handy LK --------------------------------------------------------------------
//
// File: compilehlk.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <queue>
#include <stack>

#include "libproof/util/myassert.h"
#include "libproof/util/exceptions.h"
#include "libproof/util/hash.h"
#include "libproof/chars/latexchars.h"
#include "libproof/libproof.h"
#include "libproof/recursive.h"
#include "libproof/term_operations.h"
#include "libproof/formula_operations.h"

#include "cereslk.h"

#include "reswordshlk.h"
#include "compilehlk.h"
#include "typehlk.h"
#include "symbolhlk.h"
#include "proofhlk.h"
#include "printhlk.h"

#ifdef DEBUG
#define DEBUG_PRINT_INFO
#endif

#define NEW_AUTO_PROP

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

class TermDiffernces : public TermCompareOperation
{
protected:
  TermList terms;
  TermList reps;

  std::vector< NodePointers > m_nps;

  NodePointer m_np;

public:
  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }

public:
  TermDiffernces() {}
  virtual ~TermDiffernces() {}

  void Clear()
  {
    while( terms.GetChildCount() > 0 ) terms.RemoveAt(0);
    while( reps.GetChildCount() > 0 ) reps.RemoveAt(0);
    m_nps.clear();
    m_np.Clear();
  }
  bool BuildDifferences( Formula *original, Formula *modified )
  {
    Clear();
    return original->ApplyCompare( *this, modified );
  }

  size_t GetCount() const { return terms.GetChildCount(); }

  const Term* GetTerm( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return terms.GetTerm(index);
  }

  const Term* GetSubst( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return reps.GetTerm(index);
  }

  const NodePointers& GetOccurrences( size_t index )
  {
    ASSERT( index < GetCount() );
    return m_nps[index];
  }

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );

protected:
  bool CheckTerms( Term *l, DataNode *r_ )
  {
    if( l == r_ )
      return true;

    Term *r = r_->CastTo<Term>();
  
    if( r == NULL )
      return false;
  
    int pos = terms.FindTerm( l );

    if( pos >= 0 && GetSubst( pos ) == r )
    {
      m_nps[pos].Add( m_np );
      return true;
    }

#if 0
HLK::Print::PrintTerm( l, std::cerr );
std::cerr << " <- ";
HLK::Print::PrintTerm( r, std::cerr );
std::cerr << std::endl;
#endif

    terms.AddRight( (Term*) l );
    reps.AddRight( (Term*) r );
    m_nps.resize( m_nps.size() + 1 );
    m_nps.back().Add( m_np );

    return true;
  }
};

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, TermDiffernces, t )
{
  return CheckTerms( (Term*) t, (DataNode*) o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Constant, TermDiffernces, t )
{
  return CheckTerms( (Term*) t, (DataNode*) o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Function, TermDiffernces, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Function *f_o = o->CastTo<const Function>();

/*
  const HLK::FunctionSymbol *fs =
    dynamic_cast< const HLK::FunctionSymbol*>( t->GetSymbol() );
  ASSERT( fs != NULL );

  bool rec_levels_diff = false;

  if( fs->IsRecursive() )
  {
    const ProofLib::RecursiveFunction *f_r = NULL;
    const ProofLib::RecursiveFunction *f_o_r = NULL;

    f_r = t->CastTo<const ProofLib::RecursiveFunction>();
    f_o_r = o->CastTo<const ProofLib::RecursiveFunction>();

    ASSERT( f_r != NULL && f_o_r != NULL );

    ASSERT( f_r->GetRecursiveExpr()->GetVariableName() == NULL );
    ASSERT( f_o_r->GetRecursiveExpr()->GetVariableName() == NULL );

    rec_levels_diff = f_r->GetRecursiveExpr()->GetValueAtLevel( 0 ) !=
                      f_o_r->GetRecursiveExpr()->GetValueAtLevel( 0 );
  }
#if 0
HLK::Print::PrintTerm( t, std::cerr );
std::cerr << " =?= ";
HLK::Print::PrintTerm( o->CastTo<Term>(), std::cerr );
std::cerr << std::endl;
#endif
  
  if( f_o == NULL || fs != f_o->GetSymbol() ||
      ( fs->IsRecursive() && rec_levels_diff ) )
    return CheckTerms( (Term*) t, (DataNode*) o );
*/
  if( f_o == NULL || !t->HaveEqualSymbol( f_o ) )
    return CheckTerms( (Term*) t, (DataNode*) o );

#if 0
std::cerr << "schame";
std::cerr << std::endl;
#endif
  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////

class ParamodulationSubstitutions : public TermCompareOperation
{
protected:
  class Substitution
  {
  public:
    Substitution( Term *l, Term *r, const NodePointer &np )
      : l(l), r(r), np(np) {}

    Term *l;
    Term *r;
    NodePointer np;
  };
  typedef std::vector<Substitution> Substitutions;
  Substitutions m_substs;
#if 0
  TermList terms;
  TermList reps;
  std::vector<NodePointer> m_nps;
#endif

  TermList bounded;

  NodePointer m_np;

public:
  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }

public:
  ParamodulationSubstitutions() {}
  virtual ~ParamodulationSubstitutions() {}

  void Clear()
  {
    m_substs.clear();
#if 0
    while( terms.GetChildCount() > 0 ) terms.RemoveAt(0);
    while( reps.GetChildCount() > 0 ) reps.RemoveAt(0);
    m_nps.clear();
#endif
    while( bounded.GetChildCount() > 0 ) bounded.RemoveAt(0);
    m_np.Clear();
  }

  bool CheckForSubstitutions( Formula *original, Formula *modified )
  {
    Clear();
    bool ret = original->ApplyCompare( *this, modified );
#if 0
    for( size_t i = 0; i < GetCount(); i++ )
    {
      std::cerr << i << ": ";
      HLK::Print::PrintTerm( m_substs[i].l, std::cerr );
      std::cerr << " <--- ";
      HLK::Print::PrintTerm( m_substs[i].r, std::cerr );
      std::cerr << " @ ";
      HLK::Print::PrintNodePointer( &m_substs[i].np, std::cerr );
      std::cerr << std::endl;
    }
#endif
    FoldSubstitionsTogether( original, modified );
#if 0
    for( size_t i = 0; i < GetCount(); i++ )
    {
      std::cerr << i << ": ";
      HLK::Print::PrintTerm( m_substs[i].l, std::cerr );
      std::cerr << " <--- ";
      HLK::Print::PrintTerm( m_substs[i].r, std::cerr );
      std::cerr << " @ ";
      HLK::Print::PrintNodePointer( &m_substs[i].np, std::cerr );
      std::cerr << std::endl;
    }
#endif
    return ret;
  }

  void FoldSubstitionsTogether( Formula *original, Formula *modified )
  {
    if( GetCount() < 2 )
      return;

    for( size_t i = 0; i < GetCount(); i++ )
    {
      NodePointer common_np( GetNodePointer(i) );

      for( size_t j = i + 1; j < GetCount(); )
      {
        if( GetTerm(i) == GetTerm(j) && GetSubst(i) == GetSubst(j) )
        {
          j++;
          continue;
        }
      
        NodePointer np_h = common_np.GetCommonBeginning( GetNodePointer(j) );

        Term *t_l = np_h.GetTypedDataNode<Term>( original );
        
        if( t_l == NULL )
        {
          j++;
          continue;
        }

        Term *t_r = np_h.GetTypedDataNode<Term>( modified );
        
        ASSERT( t_r != NULL );

        // remove j
        m_substs.erase( m_substs.begin() + j );

        // replace i
        m_substs[i] = Substitution( t_l, t_r, np_h );

        // update common_np
        common_np = np_h;
#if 0
        terms.RemoveAt(j);
        reps.RemoveAt(j);
        m_nps.erase( m_nps.begin() + j );
      
        // update i
        terms.RemoveAt(j);
        reps.RemoveAt(j);
        m_nps.erase( m_nps.begin() + j );
#endif
#if 0
      if( sigma->GetTerm(0) != sigma->GetTerm(i) ||
          sigma->GetSubst(0) != sigma->GetSubst(i) )
      {
        delete sigma;
        sigma = NULL;
        break;
      }
#endif
      }
    }
  }

  size_t GetCount() const { return m_substs.size(); }
#if 0
  size_t GetCount() const { return terms.GetChildCount(); }
#endif

  const Term* GetTerm( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return m_substs[index].l;
#if 0
    return terms.GetTerm(index);
#endif
  }

  const Term* GetSubst( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return m_substs[index].r;
#if 0
    return reps.GetTerm(index);
#endif
  }

  const NodePointer& GetNodePointer( size_t index )
  {
#if 0
HLK::Print::PrintNodePointer( &m_nps[index], std::cerr );
std::cerr << std::endl;
#endif
    ASSERT( index < GetCount() );
    return m_substs[index].np;
#if 0
    return m_nps[index];
#endif
  }

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( QuantorFormula );

protected:
  bool CheckTerms( Term *l, DataNode *r_ )
  {
    if( l == r_ )
      return true;

    Term *r = r_->CastTo<Term>();
  
    if( r == NULL )
      return false;
  
    ProofLib::TermList t_vars;

    ProofLib::OperationGetVariables op( t_vars );
    l->Apply( op );
    r->Apply( op );
    
    for( size_t i = 0; i < bounded.GetChildCount(); i++ )
    {
      // are in l and r any bounded variables?
      if( t_vars.FindTerm( bounded.GetTerm(i) ) >= 0 )
        return false;
    }

#if 0
HLK::Print::PrintTerm( l, std::cerr );
std::cerr << " <- ";
HLK::Print::PrintTerm( r, std::cerr );
std::cerr << ": ";
HLK::Print::PrintNodePointer( &m_np, std::cerr );
std::cerr << std::endl;
#endif

    m_substs.push_back( Substitution( l, r, m_np ) );
#if 0
    terms.AddRight( (Term*) l );
    reps.AddRight( (Term*) r );
    m_nps.push_back( m_np );
#endif

    return true;
  }
};

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, ParamodulationSubstitutions, t )
{
  return CheckTerms( (Term*) t, (DataNode*) o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Constant, ParamodulationSubstitutions, t )
{
  return CheckTerms( (Term*) t, (DataNode*) o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Function, ParamodulationSubstitutions, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Function *f_o = o->CastTo<const Function>();

  if( f_o == NULL || t->GetSymbol() != f_o->GetSymbol() )
    return CheckTerms( (Term*) t, (DataNode*) o );

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( QuantorFormula, ParamodulationSubstitutions, f )
{
  if( o == f )
    return true;

  const QuantorFormula *f_o = o->CastTo<const QuantorFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol()
      || f->GetVariable() != f_o->GetVariable() )
    return false;

  Variable *v = f->GetVariable();

  bool already = ( bounded.FindTerm( v ) >= 0 );

  if( !already )
    bounded.AddRight( v );

  bool ok = FormulaCompareOperation::ApplyToQuantorFormula( f, o );

  if( !already )
  {
    int pos = bounded.FindTerm( v );
    bounded.RemoveAt( pos );
  }

  return ok;
}

////////////////////////////////////////////////////////////////////////////////

class OperationFormulasPosAndPol : public FormulaOperation
{
public:
  class PositionAndPolarity
  {
  public:
    PositionAndPolarity( const NodePointer& np, bool left, size_t ff_offset )
      : np( np ), left( left ), first_formula_offset( ff_offset ) {}

    NodePointer np;
    bool        left;
    size_t      first_formula_offset;
  };

  //typedef std::pair< NodePointer, bool > PositionAndPolarity;
  typedef std::vector< PositionAndPolarity > Occurrences;
  typedef stlhash::hash_map< Formula*, Occurrences > FormulaPositions;

public:
  FormulaPositions m_positions;

protected:
  NodePointer m_np;
  size_t m_ffo; // offset in np for first formula.
  bool m_polarity;

  bool m_inside_quantors;
  bool m_inside_binary_rules;
  bool m_only_atom_formulas;

public:
  OperationFormulasPosAndPol()
  {
    m_polarity = false;
    m_inside_quantors = false;
    m_inside_binary_rules = false;
    m_only_atom_formulas = false;
  }

  virtual ~OperationFormulasPosAndPol() {}

  virtual NodeDatabase* GetNodeDatabase() { return NULL; }
  virtual ProofDatabase* GetProofDatabase() { return NULL; }

  /// operations for nodepointers
  virtual void Push( size_t index ) { m_np.Push( index ); }
  virtual size_t Pop() { return m_np.Pop(); }

  void SetSearchBinaryRule( bool on ) { m_inside_binary_rules = on; }
  /// if we do not search inside quantor formulas, they will treathed as
  /// atoms as well.
  void SetCollectOnlyAtoms( bool on ) { m_only_atom_formulas = on; }

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
  DECLARE_APPLY_FUNCTION( UnaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( BinaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( Sequent, Sequent );
  DECLARE_APPLY_FUNCTION( SFormula, FormulaListInsertable );
};

DECLARE_APPLY_FUNCTION_HEAD( Sequent, Sequent, OperationFormulasPosAndPol, s )
{
  if( s->GetLeft() != NULL )
  {
    Push(0);
    m_polarity = true;
    DataNode::ApplyTo( s->GetLeft(), *this );
    Pop();
  }
  
  if( s->GetRight() != NULL )
  {
    Push(1);
    m_polarity = false;
    DataNode::ApplyTo( s->GetRight(), *this );
    Pop();
  }

  return s;
}

DECLARE_APPLY_FUNCTION_HEAD( SFormula, FormulaListInsertable, OperationFormulasPosAndPol, s )
{
  m_ffo = m_np.GetLength() + 1;
  FormulaOperation::ApplyToSFormula( s );
  return s;
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, OperationFormulasPosAndPol, f )
{
  m_positions[f].push_back( PositionAndPolarity( m_np, m_polarity, m_ffo ) );

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( UnaryConFormula, Formula, OperationFormulasPosAndPol, f )
{
  if( !m_only_atom_formulas )
    m_positions[f].push_back( PositionAndPolarity( m_np, m_polarity, m_ffo ) );

  bool pol = m_polarity;
  m_polarity = !m_polarity;

  Push(0);
  DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  m_polarity = pol;

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD( QuantorFormula, Formula, OperationFormulasPosAndPol, f )
{
  if( !m_only_atom_formulas || !m_inside_quantors )
    m_positions[f].push_back( PositionAndPolarity( m_np, m_polarity, m_ffo ) );

  if( !m_inside_quantors )
    return f;

  FormulaOperation::ApplyToQuantorFormula( f );

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( BinaryConFormula, Formula, OperationFormulasPosAndPol, f )
{
  if( !m_only_atom_formulas )
    m_positions[f].push_back( PositionAndPolarity( m_np, m_polarity, m_ffo ) );

  if( !m_inside_binary_rules )
  {
    if( ( f->GetOperatorSymbol()->IsAndJunctor() && !m_polarity ) ||
        ( f->GetOperatorSymbol()->IsOrJunctor() && m_polarity ) ||
        ( f->GetOperatorSymbol()->IsImplJunctor() && m_polarity ) )
      return f;
  }

  bool pol = m_polarity;

  if( f->GetOperatorSymbol()->IsImplJunctor() )
    m_polarity = !m_polarity;

  Push(0);
  DataNode::ApplyTo( f->GetLeft(), *this );
  Pop();

  m_polarity = pol;

  Push(1);
  DataNode::ApplyTo( f->GetRight(), *this );
  Pop();

  return f;
}

////////////////////////////////////////////////////////////////////////////////

// search the first function that is defined.
class OperationFindFirstDefinedFunction : public TermOperation
{
  Function *f;

public:
  OperationFindFirstDefinedFunction()
  {
    f = NULL;
  }

  virtual ~OperationFindFirstDefinedFunction() {}

  virtual NodeDatabase* GetNodeDatabase() { return NULL; }
  virtual ProofDatabase* GetProofDatabase() { return NULL; }

  /// operations for nodepointers
  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

  Function* GetFunction() { return f; }

  DECLARE_APPLY_FUNCTION( Function, Term );
};

DECLARE_APPLY_FUNCTION_HEAD
  ( Function, Term, OperationFindFirstDefinedFunction, f0 )
{
  if( f != NULL )
    return f0;

  const HLK::FunctionSymbol *fs =
    dynamic_cast< const HLK::FunctionSymbol*>( f0->GetSymbol() );
  ASSERT( fs != NULL );

  if( fs->GetFunctionDefinition() != NULL )
    f = f0;
  if( fs->GetThroughFormula() != NULL )
    f = f0;
  if( fs->IsRecursive() )
    f = f0;

  return TermOperation::ApplyToFunction( f0 );
}

////////////////////////////////////////////////////////////////////////////////


class OperationFindFirstFuncSymbolOccurence : public TermOperation
{
  const Symbol *tofind;

  NodePointer np, occ;

public:
  OperationFindFirstFuncSymbolOccurence( const Symbol *s )
  {
    ASSERT( s != NULL );
    tofind = s;
  }

  virtual ~OperationFindFirstFuncSymbolOccurence() {}

  virtual NodeDatabase* GetNodeDatabase() { return NULL; }
  virtual ProofDatabase* GetProofDatabase() { return NULL; }

  /// operations for nodepointers
  virtual void Push( size_t index ) { np.Push(index); }
  virtual size_t Pop() { return np.Pop(); }

  const NodePointer& GetOccurence() { return occ; }

  DECLARE_APPLY_FUNCTION( Function, Term );
};

DECLARE_APPLY_FUNCTION_HEAD
  ( Function, Term, OperationFindFirstFuncSymbolOccurence, f )
{
  if( f->GetSymbol() == tofind )
  {
    occ = np;
    return f;
  }

  return TermOperation::ApplyToFunction( f );
}

}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

namespace HLK
{

class UsedDefinitionsListImpl : public UsedDefinitionsList
{
  typedef stlhash::hash_map<const ProofLib::Symbol*, std::set<size_t> >
    Definitions;
  Definitions definitions;

public:
  UsedDefinitionsListImpl()
  {
  }
  
public:
  virtual size_t GetCount() const { return definitions.size(); }
  virtual const ProofLib::Symbol* GetSymbol( size_t index ) const
  {
    ASSERT( index < GetCount() );
    Definitions::const_iterator i = definitions.begin();
    for( ; index > 0; index-- ) i++;
    return i->first;
  }
  virtual std::set<size_t> GetRecursionLevels( size_t index ) const
  {
    ASSERT( index < GetCount() );
    Definitions::const_iterator i = definitions.begin();
    for( ; index > 0; index-- ) i++;
    return i->second;
  }

public:
  void AddSymbol( const ProofLib::Symbol* sym, size_t level )
  {
  //  std::cerr << sym->GetName() << ", " << level;
  //  std::cerr << std::endl;
    definitions[sym].insert( level );
  }
};

////////////////////////////////////////////////////////////////////////////////

class CompilerImpl : public Compiler
{
  const Document          *m_doc;
  ProofLib::Proof         *m_lk_proof;
  ProofLib::Proof         *m_lk_root_proof;
  ProofLib::ProofDatabase *m_lk_pdb;
  
  UsedDefinitionsListImpl  *m_useddefs;

  std::vector<const Proof*> m_proof_stack;

private:
  CompilerImpl( const CompilerImpl& ) {};

public:
  CompilerImpl()
  {
    m_doc = NULL;
    m_lk_proof = NULL;
    m_lk_root_proof = NULL;
    m_lk_pdb = NULL;
    m_useddefs = new UsedDefinitionsListImpl;
  }
  virtual ~CompilerImpl()
  {
    delete m_useddefs;
  }

  const UsedDefinitionsList* GetUsedDefinitionsList() const { return m_useddefs; }

  //////////////////////////////////////////////////////////////////////////////

  const Document* GetDoc() const { return m_doc; }
  SymbolTable* GetSymbolTable() const { return GetDoc()->GetSymbolTable(); }
  UsedDefinitionsListImpl* GetUsedDefinitionsListImpl() { return m_useddefs; }

  //////////////////////////////////////////////////////////////////////////////

  virtual ProofLib::ProofDatabase* GenerateLK( const Document *doc,
                                               const ProofSymbol *ps )
  {
    ASSERT( doc != NULL );
    ASSERT( ps != NULL );

    m_doc = doc;
    m_lk_pdb = new ProofLib::ProofDatabase( doc->GetNodeDatabase(),
                                            doc->GetSymbolTable() );

    ProofMetaArguments meta_args;

    const Proof *p = GetDoc()->GetProofTable()->GetProof( ps )->
      GetProofInstance( meta_args, *ps, doc->GetNodeDatabase(), 0 );

    ASSERT( p != NULL );

    if( p->IsMetaProof() )
    {
      GenericException *e = new GenericException;

      e->Stream() << "Can not start with proof '" << ps->GetName() << "' "
                  << "because it is a meta proof.";

      throw e;
    }

    m_proof_stack.push_back( p );
    
    m_lk_root_proof = m_lk_pdb->NewProof( p->GetSymbol() );

    ProcessProof( m_lk_root_proof, p );

    m_proof_stack.pop_back();

    //p->ReleaseProofInstance();

    return m_lk_pdb;
  }

  //////////////////////////////////////////////////////////////////////////////

  virtual ProofLib::ProofDatabase* GenerateLK( const Document *doc,
                                               const USTRING &proof )
  {
    ASSERT( doc != NULL );
    ASSERT( !proof.empty() );

    const ProofSymbol *ps = dynamic_cast<const ProofSymbol*>
      ( doc->GetSymbolTable()->GetSymbol( proof ) );

    if( ps == NULL )
    {
      GenericException *e = new GenericException;

      e->Stream() << "No proof '" << proof << "' found!";

      throw e;
    }

    return GenerateLK( doc, ps );
  }

  //////////////////////////////////////////////////////////////////////////////

  void SetRuleName( ProofLib::Rule *rule, int rule_name,
                    std::string s_param = std::string() )
  {
    USTRING name;
    name.FromUtf8( rule_names[rule_name].latex_code );
    LatexConvertor.KnownLatexSymbolsToUnicode( name );

    rule->SetName( name );

    USTRING type;
    type.FromUtf8( rule_names[rule_name].xml_name );

    rule->SetType( type );

    if( !s_param.empty() )
    {
      USTRING param;
      param.FromUtf8( s_param.c_str() );

      rule->SetParam( param );
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  static std::string& PrintXMLNodePointer( const ProofLib::NodePointer &np,
                                           std::string &str )
  {
    std::ostringstream stream;

    PrintXMLNodePointer( np, stream );

    str += stream.str();

    return str;
  }

  static std::ostream& PrintXMLNodePointer( const ProofLib::NodePointer &np,
                                            std::ostream &stream )
  {
//std::cerr << "++PrintXMLNodePointer" << std::endl;

    stream << '(';
  
    for( size_t i = 0; i < np.GetLength(); i++ )
    {
      if( i > 0 )
        stream << ' ';
      stream << long(np[i] + 1);
    }
  
    stream << ')';

//std::cerr << "--PrintXMLNodePointer" << std::endl;

    return stream;
  }

  //////////////////////////////////////////////////////////////////////////////

  void AddToGlobalEndSequent( ProofLib::Formula* f, bool left )
  {
    ASSERT( m_lk_root_proof != NULL );
    ASSERT( m_lk_root_proof->GetEndRule() != NULL );

    ProofLib::Rule *rule =
      m_lk_root_proof->GetEndRule()->CastTo<ProofLib::Rule>();
    ASSERT( rule != NULL );

    ProofLib::Sequent *seq = rule->GetConclusion();
    ASSERT( seq != NULL );

    ProofLib::SElementList *list = ( left ? seq->GetLeft() : seq->GetRight() );

    if( list != NULL )
    {
      bool already = false;

      for( size_t i = 0; i < list->GetChildCount(); i++ )
      {
        ProofLib::SFormula *sf =
          dynamic_cast<ProofLib::SFormula*>( list->GetSElement(i) );
  
        ASSERT( sf != NULL );

        if( sf->GetFormula() == f )
        {
          already = true;
          break;
        }
      }

      if( already )
        return;
    }

    if( left )
      seq->AddLeft( new ProofLib::SFormula( f ) );
    else
      seq->AddRight( new ProofLib::SFormula( f ) );
  }

  //////////////////////////////////////////////////////////////////////////////

  void UpdateConculsionForUsedParent(
    ProofLib::Rule *rule,
    ProofLib::Formula *f,
    bool on_left_side )
  {
    ASSERT( rule != NULL );
    ASSERT( f != NULL );

    if( rule == m_lk_root_proof->GetEndRule() )
    {
      return;
    }

    ProofLib::Sequent *seq = rule->GetConclusion();
    ASSERT( seq != NULL );

    if( on_left_side )
    {
      if( seq->GetLeft() != NULL && seq->GetLeft()->GetChildCount() > 0 )
      {
        ASSERT( seq->GetLeft()->GetChildCount() == 1 );

        ProofLib::SFormula *sf =
          dynamic_cast<ProofLib::SFormula*>( seq->GetLeft()->GetSElement(0) );

        ASSERT( sf != NULL );

        if( f == sf->GetFormula() )
          return;
      }

      seq->AddLeft( new ProofLib::SFormula( f ) );
    }
    else
    {
      if( seq->GetRight() != NULL && seq->GetRight()->GetChildCount() > 0 )
      {
        ASSERT( seq->GetRight()->GetChildCount() == 1 );

        ProofLib::SFormula *sf =
          dynamic_cast<ProofLib::SFormula*>( seq->GetRight()->GetSElement(0) );

        ASSERT( sf != NULL );

        if( f == sf->GetFormula() )
          return;
      }

      seq->AddRight( new ProofLib::SFormula( f ) );
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  void ProcessProof( ProofLib::Proof *p_lk, const Proof *p,
                     ProofLib::Sequent *world = NULL,
                     ProofLib::Sequent *end_seq = NULL,
                     const ProofLib::SourceLocation *sloc = NULL )
  {
    ASSERT( p_lk != NULL );
    ASSERT( p != NULL );
    ASSERT( world == NULL || sloc != NULL );

    m_lk_proof = p_lk;

    ProofLib::FormulaList miss_l, miss_r;

#if 0 // def DEBUG_PRINT_INFO
    if( world != NULL )
    {
      Print::PrintSequent( p->GetEndSequent(), std::cerr );
      std::cerr << std::endl;
      Print::PrintSequent( world, std::cerr );
      std::cerr << std::endl;
    }
#endif
    if( world != NULL &&
        !p->GetEndSequent()->IsSubSequentOf( world, &miss_l, &miss_r ) )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( sloc );
      e->Stream() << "End sequent";
//    Print::PrintSequent( p->GetEndSequent(), e->Stream() );
      e->Stream() << " of proof " << p->GetSymbol()->GetName()
                  << " is not a sub-sequent of the previous one through ";
//    Print::PrintSequent( world, e->Stream() );
      if( miss_l.GetChildCount() > 0 )
      {
        Print::PrintFormulaList( &miss_l, e->Stream() );
        e->Stream() << " on the left side";

        if( miss_r.GetChildCount() > 0 )
          e->Stream() << " and ";
      }
      if( miss_r.GetChildCount() > 0 )
      {
        Print::PrintFormulaList( &miss_r, e->Stream() );
        e->Stream() << " on the right side";
      }
      e->Stream() << ".";

      throw e;
    }

    ProofLib::Rule *rule = new ProofLib::Rule;

    p_lk->SetEndRule( rule );

    if( end_seq != NULL )
    {
#ifdef DEBUG_PRINT_INFO
      std::cerr << "not setting pseudo rule" << std::endl;
#endif
      rule->SetConclusion( end_seq->Clone() );
    } else
    {
#ifdef DEBUG_PRINT_INFO
      std::cerr << "setting pseudo rule" << std::endl;
#endif
      rule->SetConclusion( p->GetEndSequent()->Clone() );
     
      // this psuedo rule separates the world from the "real"
      // proof, it will contain later the first active fromula
      ProofLib::Rule *next_rule = new ProofLib::Rule;
      next_rule->SetConclusion( new ProofLib::Sequent( NULL, NULL ) );

      rule->AddRight( next_rule );

      rule = next_rule;
    }

    if( world == NULL )
      world = p->GetEndSequent()->Clone();

    if( p->GetRuleCount() == 0 ) // no rules here
    {
      ProofLib::Formula* f = FindAnAxiom( p->GetEndSequent(), world );

      if( f == NULL )
        return;

      AxiomToAtomic( rule, *p, f );

      return;
    }

    for( size_t i = 0; i < p->GetRuleCount(); i++ )
    {
      rule = ProcessRule( rule, p->GetRule(i), world );

      if( rule == NULL && i != p->GetRuleCount() - 1 )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( p->GetRule(i) );
        e->Stream() << "Additional rules after proof end step.";
        throw e;
      }
    }

    if( rule != NULL )
    {
      ProofLib::Formula* f = FindAnAxiom( world );

      if( f == NULL )
        return;

      AxiomToAtomic( rule, *p->GetRule(p->GetRuleCount()-1), f );
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessRule( ProofLib::Rule *rule,
                               const ProofRule *r,
                               ProofLib::Sequent *world )
  {
    ASSERT ( r != NULL );
    switch( r->GetType() )
    {
    case RESWORD_NOT:
      return ProcessNotRule( rule, r, world );

    case RESWORD_OR:
      if( r->IsLeftRule() )
        return ProcessBinaryAndOrImplRule( rule, r, world );
      else
        return ProcessUnaryAndOrRule( rule, r, world );

    case RESWORD_AND:
      if( r->IsLeftRule() )
        return ProcessUnaryAndOrRule( rule, r, world );
      else
        return ProcessBinaryAndOrImplRule( rule, r, world );

    case RESWORD_IMPL:
      if( r->IsLeftRule() )
        return ProcessBinaryAndOrImplRule( rule, r, world );
      else
        return ProcessImplRightRule( rule, r, world );

    case RESWORD_ALL:
    case RESWORD_EX:
      return ProcessQuantorRule( rule, r, world );

    case RESWORD_CONTINUED:
      ProcessContinued( rule, r, world );
      return NULL;

    case RESWORD_CUT:
      return ProcessCut( rule, r, world );

    case RESWORD_PARAMOD:
      return ProcessParamod( rule, r, world );

    case RESWORD_AXIOM:
      ProcessAxiom( rule, r, world );
      return NULL;

    case RESWORD_UNDEF:
      return ProcessUndef( rule, r, world );

    case RESWORD_AUTO:
      return ProcessAuto( rule, r, world );

    case RESWORD_ALL2:
    case RESWORD_EX2:
      return ProcessSecondOrderQuantorRule( rule, r, world );

    default:
#ifdef DEBUG
      std::cerr << "Got rule " << r->GetType() << std::endl;
#endif
      ASSERT( false );
    }

    return NULL;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessAxiom(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_AXIOM );

    ProofLib::FormulaList miss_l, miss_r;

    if( !r->GetSequent()->IsSubSequentOf( world, &miss_l, &miss_r ) )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException( r );
      e->Stream() << "Axiom sequent";
//    Print::PrintSequent( p->GetEndSequent(), e->Stream() );
//    e->Stream() << " of proof " << p->GetSymbol()->GetName();
      e->Stream() << " is not a sub-sequent of the previous one through ";
//    Print::PrintSequent( world, e->Stream() );
      if( miss_l.GetChildCount() > 0 )
      {
        Print::PrintFormulaList( &miss_l, e->Stream() );
        e->Stream() << " on the left side";

        if( miss_r.GetChildCount() > 0 )
          e->Stream() << " and ";
      }
      if( miss_r.GetChildCount() > 0 )
      {
        Print::PrintFormulaList( &miss_r, e->Stream() );
        e->Stream() << " on the right side";
      }
      e->Stream() << ".";

      throw e;
    }

    rule->AddRight( r->GetSequent()->Clone() );

    return NULL; // end of proof
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessNotRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_NOT );

    // a   ... aux side of the "goal" sequent
    // e   ... other side of the "goal" sequent
    // a_0 ... aux side of the last sequent
    // e_0 ... other side of the last sequent
    ProofLib::SElementList *a,*e;
    ProofLib::SElementList *a_0,*e_0;

    if( r->IsLeftRule() )
    {
      e = r->GetSequent()->GetLeft();
      a = r->GetSequent()->GetRight();
    }
    else
    {
      e = r->GetSequent()->GetRight();
      a = r->GetSequent()->GetLeft();
    }

    // the other side of the goal sequent has to be empty
    if( e != NULL && e->GetChildCount() > 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Left" : "Right" )
                   << " side is not empty.";
      throw ex;
    }

    // we have to have goals ....
    if( a == NULL || a->GetChildCount() == 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Right" : "Left" )
                   << " side is empty.";
      throw ex;
    }

    // for each goal formula
    for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      if( r->IsLeftRule() )
      {
        a_0 = world->GetLeft();
        e_0 = world->GetRight();
      }
      else
      {
        a_0 = world->GetRight();
        e_0 = world->GetLeft();
      }

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
                                ( a->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();

      // exists f already in the last sequent (on the correspondent side)?
      bool found = false;
      if( e_0 != NULL )
      {
        for( size_t j = 0; j < e_0->GetChildCount(); j++ )
        {
          ProofLib::SFormula *sf_1 = dynamic_cast<ProofLib::SFormula*>
                                      ( e_0->GetSElement(j) );
          ASSERT( sf_1 != NULL );

          if( f == sf_1->GetFormula() )
          {
            found = true;
            break;
          }
        }
      }
        
      if( found ) // f is already there where we want it, done.
        continue;

      // this will hold the parent formula
      ProofLib::Formula *parent = NULL;

      // these vars will be used to find deep nested formulas
      size_t level = 0;
      size_t level_hits = 0;

      do
      {
        // one level higher
        level++;
        // in case of odd levels the parent is on the "normal" side
        // and for even levels it is on the dual side
        // so we have to count hits for both sides, otherwise we will 
        // terminate to early.
        if( level % 2 != 0 ) level_hits = 0;

        // the side we search for the parent
        ProofLib::SElementList *p_l = ( level % 2 != 0 ? a_0 : e_0 );

        if( p_l == NULL )
          continue; // here we can not find anything

        for( size_t j = 0; j < p_l->GetChildCount(); j++ )
        {
          ProofLib::SFormula *sf_1 = dynamic_cast<ProofLib::SFormula*>
                                      ( p_l->GetSElement(j) );
          ASSERT( sf_1 != NULL );

          ProofLib::Formula *f_1 = sf_1->GetFormula();

          ProofLib::Formula *f_2 = f_1;

          for( size_t l = level; l > 0; l-- )
          {
            ProofLib::UnaryConFormula *uf = 
              dynamic_cast<ProofLib::UnaryConFormula*>( f_2 );

            if( uf == NULL || !uf->GetOperatorSymbol()->IsNegJunctor() )
            {
              f_2 = NULL;
              break;
            }

            f_2 = uf->GetFormula();
          }

          if( f_2 != NULL ) // a formula at this level found.
            level_hits++;

          if( f_2 == f ) // match !!
          {
            parent = f_1;
            break;
          }
        }
      // repeat do loop as long as there is no parent found and still
      // more deep nested formulas left.
      } while( parent == NULL && level_hits > 0 );

      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        throw ex;
      }

      // now we build the rule(s)
      while( parent != f )
      {
        ProofLib::UnaryConFormula *uf = 
          dynamic_cast<ProofLib::UnaryConFormula*>( parent );

        bool left = r->IsLeftRule();
  
        if( level % 2 == 0 )
          left = !left;
    
        level--;

        rule = BuildNotRule( rule, uf, left, world );

        parent = uf->GetFormula();
      }
    }

    return rule;
  }

  ProofLib::Rule* BuildNotRule( 
    ProofLib::Rule *rule,
    ProofLib::UnaryConFormula *uf,
    bool left,
    ProofLib::Sequent *world )
  {
    ASSERT( uf != NULL && uf->GetOperatorSymbol()->IsNegJunctor() );

    ProofLib::Formula *parent = uf->GetFormula();

    // new sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    if( left )
    {
      seq->AddRight( new ProofLib::SFormula( parent ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( parent ) );
      UpdateConculsionForUsedParent( rule, uf, true );
    }
    else
    {
      seq->AddLeft( new ProofLib::SFormula( parent ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( parent ) );
      UpdateConculsionForUsedParent( rule, uf, false );
    }

#ifdef DEBUG_PRINT_INFO
    std::cerr << "BY NOT: ";
    Print::PrintSequent( seq, std::cerr );
    std::cerr << std::endl;
#endif

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    SetRuleName( rule, ( left ?  RULE_NEGL : RULE_NEGR ) );

    rule->AddRight( next_rule );

    return next_rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  // helper type for ProcessUnaryAndOrRule and ProcessImplRightRule
  typedef struct
  {
    ProofLib::Formula* f;
    ProofLib::Formula* parent;
    std::queue<bool> left_right;
  } tree_info;

  ProofLib::Rule* ProcessUnaryAndOrRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    const ProofLib::OperatorSymbol *symbol = NULL;

    switch( r->GetType() )
    {
    case RESWORD_AND:
      ASSERT( r->IsLeftRule() );

      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_AND );
      break;
    case RESWORD_OR:
      ASSERT( !r->IsLeftRule() );

      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_OR );
      break;
    default:
      ASSERT( false );
    }

    // a   ... aux side of the "goal" sequent
    // e   ... other side of the "goal" sequent
    // a_0 ... aux side of the last sequent
    // e_0 ... other side of the last sequent
    ProofLib::SElementList *a,*e;
    ProofLib::SElementList *a_0,*e_0;

    if( r->IsLeftRule() )
    {
      a = r->GetSequent()->GetLeft();
      e = r->GetSequent()->GetRight();
    }
    else
    {
      a = r->GetSequent()->GetRight();
      e = r->GetSequent()->GetLeft();
    }

    // right side has to be empty
    if( e != NULL && e->GetChildCount() > 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Right" : "Left" )
                   << " side is not empty.";
      throw ex;
    }

    // a least one formula we need
    if( a == NULL || a->GetChildCount() == 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Left" : "Right" )
                   << " side is empty.";
      throw ex;
    }

    // for each formula
    for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      if( r->IsLeftRule() )
      {
        a_0 = world->GetLeft();
        e_0 = world->GetRight();
      }
      else
      {
        a_0 = world->GetRight();
        e_0 = world->GetLeft();
      }
  
      if( a_0 == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No rules on "
                     << ( r->IsLeftRule() ? "left" : "right" )
                     << " side available.";
  
        throw ex;
      }

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
                                ( a->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();

      std::queue<tree_info> todo;

      // build queue
      bool found = false;
      for( size_t j = 0; j < a_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 = dynamic_cast<ProofLib::SFormula*>
                                    ( a_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        if( f == sf_1->GetFormula() )
        {
          found = true;
          break;
        }

        tree_info t;
        t.f = sf_1->GetFormula();
        t.parent = t.f;
        todo.push( t );
      }
        
      // exists f already on the goal side?
      if( found )
        continue; // nothing to do in this case...

      std::queue<bool> left_right;

      ProofLib::Formula *parent = NULL;

      // build bintree to solve
      while( !todo.empty() )
      {
        ProofLib::Formula *f_1 = todo.front().f;
        ProofLib::Formula *p_0 = todo.front().parent;
        left_right = todo.front().left_right;
        todo.pop();
#if 0
        Print::PrintFormula( f, std::cerr );
        std::cerr << " -> ";
        Print::PrintFormula( f_1, std::cerr );
        std::cerr << std::endl;
        std::queue<bool> x( left_right );
      while( !x.empty() )
      {
        std::cerr << "Is " << ( x.front() ? "Left" : "Right" )
                  << std::endl;
        x.pop();
      }
        std::cerr << "-------" << std::endl;
#endif

        if( f_1 == f )
        {
          parent = p_0;
          break;
        }

        ProofLib::BinaryConFormula *bf = 
          dynamic_cast<ProofLib::BinaryConFormula*>( f_1 );

        if( bf == NULL || bf->GetOperatorSymbol() != symbol )
          continue;

        tree_info t_l;
        t_l.f = bf->GetLeft();
        t_l.parent = p_0;
        t_l.left_right = left_right;
        t_l.left_right.push( true );

        todo.push( t_l );

        tree_info t_r;
        t_r.f = bf->GetRight();
        t_r.parent = p_0;
        t_r.left_right = left_right;
        t_r.left_right.push( false );

        todo.push( t_r );
      }
      
      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        throw ex;
      }

      // generate rules
      while( !left_right.empty() )
      {
#if 0
        std::cerr << "Is " << ( left_right.front() ? "Left" : "Right" )
                  << std::endl;
#endif

        ProofLib::BinaryConFormula *bf = 
          dynamic_cast<ProofLib::BinaryConFormula*>( parent );

        rule = BuildUnaryAndOrRule(
          rule, bf, left_right.front(), world );

        parent = ( left_right.front() ? bf->GetLeft() : bf->GetRight() );

        left_right.pop();
      }
    }

    return rule;
  }

  ProofLib::Rule* BuildUnaryAndOrRule( 
    ProofLib::Rule *rule,
    ProofLib::BinaryConFormula *bf,
    bool left_formula,
    ProofLib::Sequent *world )
  {
    ASSERT( rule != NULL );
    ASSERT( bf != NULL );
#if 0
    std::cerr << "BuildUnaryAndOrRule from formula ";
    Print::PrintFormula( bf, std::cerr );
    std::cerr << " use the " << ( left_formula ? "left" : "right" )
              << " subformula." << std::endl;
#endif

    const ProofLib::OperatorSymbol *symbol = bf->GetOperatorSymbol();

    ASSERT( symbol->IsAndJunctor() || symbol->IsOrJunctor() );

    int rule_left;
    int rule_right;

    ProofLib::Formula *parent = left_formula ?  bf->GetLeft() : bf->GetRight();

    // new sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    if( symbol->IsAndJunctor() )
    {
      seq->AddLeft( new ProofLib::SFormula( parent ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( parent ) );
      UpdateConculsionForUsedParent( rule, bf, true );
      rule_left = RULE_ANDL1;
      rule_right = RULE_ANDL2;
    }
    else
    {
      seq->AddRight( new ProofLib::SFormula( parent ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( parent ) );
      UpdateConculsionForUsedParent( rule, bf, false );
      rule_left = RULE_ORR1;
      rule_right = RULE_ORR2;
    }

#ifdef DEBUG_PRINT_INFO
    Print::PrintSequent( seq, std::cerr );
    std::cerr << std::endl;
#endif

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    SetRuleName( rule, ( left_formula ?  rule_left : rule_right ) );

    rule->AddRight( next_rule );

    return next_rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessImplRightRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_IMPL && !r->IsLeftRule() );

    // a   ... aux side of the "goal" sequent
    // e   ... other side of the "goal" sequent
    // a_0 ... aux side of the last sequent
    // e_0 ... other side of the last sequent
    ProofLib::SElementList *a,*e;
    ProofLib::SElementList *a_0,*e_0;

    a = r->GetSequent()->GetLeft();
    e = r->GetSequent()->GetRight();

    // a least one formula we need
    if( ( e == NULL || e->GetChildCount() == 0 ) &&
        ( a == NULL || a->GetChildCount() == 0 ) )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << "Right and left side are empty.";
      throw ex;
    }

    // for each formula on the left side of the goal
    if( a != NULL ) for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      e_0 = world->GetLeft();
      a_0 = world->GetRight();
  
      if( a_0 == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        ex->Stream() << "No rules on the right side available.";
        throw ex;
      }

      ProofLib::SFormula *sf =
        dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();

      std::queue<tree_info> todo;

      // exists this formula on the left side already?
      bool found = false;
      if( e_0 != NULL ) for( size_t j = 0; j < e_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 =
          dynamic_cast<ProofLib::SFormula*>( e_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        if( f == sf_1->GetFormula() )
        {
          found = true;
          break;
        }
      }
        
      // exists f already on the goal side?
      if( found )
        continue; // nothing to do in this case...

      // build queue
      for( size_t j = 0; j < a_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 =
          dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        tree_info t;
        t.f = sf_1->GetFormula();
        t.parent = t.f;
        todo.push( t );
      }
        
      std::queue<bool> left_right;

      ProofLib::Formula *parent = NULL;

      // build bintree to solve
      while( !todo.empty() )
      {
        ProofLib::Formula *f_1 = todo.front().f;
        ProofLib::Formula *p_0 = todo.front().parent;
        left_right = todo.front().left_right;
        todo.pop();
#if 0
        Print::PrintFormula( f_1, std::cerr );
        std::cerr << std::endl;
        std::queue<bool> x( left_right );
      while( !x.empty() )
      {
        std::cerr << "Is " << ( x.front() ? "Left" : "Right" )
                  << std::endl;
        x.pop();
      }
        std::cerr << "-------" << std::endl;
#endif

        // "our" formula can only be on the left side
        if( f_1 == f && ( !left_right.empty() && left_right.back() ) )
        {
          parent = p_0;
          break;
        }

        // if the formula appears on the left side, it can not be used futher
        // (implication left is a binary rule ... )
        if( !left_right.empty() && left_right.back() )
          continue;

        ProofLib::BinaryConFormula *bf = 
          dynamic_cast<ProofLib::BinaryConFormula*>( f_1 );

        if( bf == NULL || !bf->GetOperatorSymbol()->IsImplJunctor() )
          continue;

        tree_info t_l;
        t_l.f = bf->GetLeft();
        t_l.parent = p_0;
        t_l.left_right = left_right;
        t_l.left_right.push( true );

        todo.push( t_l );

        tree_info t_r;
        t_r.f = bf->GetRight();
        t_r.parent = p_0;
        t_r.left_right = left_right;
        t_r.left_right.push( false );

        todo.push( t_r );
      }
      
      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        throw ex;
      }

      // generate rules
      rule = BuildImplRightRules( rule, parent, left_right, world );
    }

    // for each formula on the right side of the goal
    if( e != NULL ) for( size_t i = 0; i < e->GetChildCount(); i++ )
    {
      e_0 = world->GetLeft();
      a_0 = world->GetRight();
  
      if( a_0 == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        ex->Stream() << "No rules on the right side available.";
        throw ex;
      }

      ProofLib::SFormula *sf =
        dynamic_cast<ProofLib::SFormula*>( e->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();

      std::queue<tree_info> todo;

      bool found = false;
      // exists this formula on the right side already?
      // and build queue
      for( size_t j = 0; j < a_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 =
          dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        if( f == sf_1->GetFormula() )
        {
          found = true;
          break;
        }

        tree_info t;
        t.f = sf_1->GetFormula();
        t.parent = t.f;
        todo.push( t );
      }
        
      // exists f already on the goal side?
      if( found )
        continue; // nothing to do in this case...
        
      std::queue<bool> left_right;

      ProofLib::Formula *parent = NULL;

      // build bintree to solve
      while( !todo.empty() )
      {
        ProofLib::Formula *f_1 = todo.front().f;
        ProofLib::Formula *p_0 = todo.front().parent;
        left_right = todo.front().left_right;
        todo.pop();
#if 0
        Print::PrintFormula( f_1, std::cerr );
        std::cerr << std::endl;
        std::queue<bool> x( left_right );
      while( !x.empty() )
      {
        std::cerr << "Is " << ( x.front() ? "Left" : "Right" )
                  << std::endl;
        x.pop();
      }
        std::cerr << "-------" << std::endl;
#endif

        // "our" formula can only be on the right side
        if( f_1 == f && ( left_right.empty() || !left_right.back() ) )
        {
          parent = p_0;
          break;
        }

        // if the formula appears on the left side, it can not be used futher
        // (implication left is a binary rule ... )
        if( !left_right.empty() && left_right.back() )
          continue;

        ProofLib::BinaryConFormula *bf = 
          dynamic_cast<ProofLib::BinaryConFormula*>( f_1 );

        if( bf == NULL || !bf->GetOperatorSymbol()->IsImplJunctor() )
          continue;

        tree_info t_l;
        t_l.f = bf->GetLeft();
        t_l.parent = p_0;
        t_l.left_right = left_right;
        t_l.left_right.push( true );

        todo.push( t_l );

        tree_info t_r;
        t_r.f = bf->GetRight();
        t_r.parent = p_0;
        t_r.left_right = left_right;
        t_r.left_right.push( false );

        todo.push( t_r );
      }
      
      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        throw ex;
      }

      // generate rules
      rule = BuildImplRightRules( rule, parent, left_right, world );
    }

    return rule;
  }

  ProofLib::Rule* BuildImplRightRules(
    ProofLib::Rule *rule,
    ProofLib::Formula *parent,
    std::queue<bool> &left_right,
    ProofLib::Sequent *world )
  {
    while( !left_right.empty() )
    {
#if 0
      std::cerr << "Is " << ( left_right.front() ? "Left" : "Right" )
                << std::endl;
#endif

      ProofLib::BinaryConFormula *bf = 
        dynamic_cast<ProofLib::BinaryConFormula*>( parent );

      rule = BuildImplRightRule( rule, bf, world );

      parent = left_right.front() ?  bf->GetLeft() : bf->GetRight();
  
      left_right.pop();
    }

    return rule;
  }

  ProofLib::Rule* BuildImplRightRule(
    ProofLib::Rule *rule,
    ProofLib::BinaryConFormula *bf,
    ProofLib::Sequent *world )
  {
    ASSERT( rule != NULL );
    ASSERT( bf != NULL && bf->GetOperatorSymbol()->IsImplJunctor() );

    SetRuleName( rule, RULE_IMPLR );
    UpdateConculsionForUsedParent( rule, bf, false );

    // new sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    seq->AddLeft( new ProofLib::SFormula( bf->GetLeft() ) );
    seq->AddRight( new ProofLib::SFormula( bf->GetRight() ) );

    if( world != NULL )
    {
      world->AddLeft( new ProofLib::SFormula( bf->GetLeft() ) );
      world->AddRight( new ProofLib::SFormula( bf->GetRight() ) );
    }

#ifdef DEBUG_PRINT_INFO
    std::cerr << "BY IMPL RIGHT: ";
    Print::PrintSequent( seq, std::cerr );
    std::cerr << std::endl;
#endif

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    rule->AddRight( next_rule );

    return next_rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  typedef struct
  {
    ProofLib::Formula* f;
    ProofLib::Formula* parent;
    std::queue<ProofLib::Variable*> vars;
  } qt_list_info;

  typedef struct
  {
    ProofLib::Formula* f;
    ProofLib::Formula* parent;
    std::queue<ProofLib::SecondOrderVariable*> vars;
  } qt2_list_info;

  ProofLib::Rule* ProcessSecondOrderQuantorRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {

  	Exception *exception = NULL;
    const ProofLib::OperatorSymbol *symbol = NULL;
    bool vb1_applies = false;

    switch( r->GetType() )
    {
    case RESWORD_ALL2:
#ifdef DEBUG_PRINT_INFO
          std::cerr << "Processing ALL2" <<std::endl;
#endif
      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_ALL2 );
      vb1_applies = !r->IsLeftRule();
      break;
    case RESWORD_EX2:
#ifdef DEBUG_PRINT_INFO
          std::cerr << "Processing EX2" <<std::endl;
#endif
      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_EX2 );
      vb1_applies = r->IsLeftRule();
      break;
    default:
      ASSERT( false );
    }

    // a   ... aux side of the "goal" sequent
    // e   ... other side of the "goal" sequent
    // a_0 ... aux side of the last sequent
    // e_0 ... other side of the last sequent
    ProofLib::SElementList *a,*e;
    ProofLib::SElementList *a_0,*e_0;

    if( r->IsLeftRule() )
    {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "rule is left" <<std::endl;
#endif
      a = r->GetSequent()->GetLeft();
      e = r->GetSequent()->GetRight();
    }
    else
    {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "rule is right" <<std::endl;
#endif
      a = r->GetSequent()->GetRight();
      e = r->GetSequent()->GetLeft();
    }

    // unactive side has to be empty
    if( e != NULL && e->GetChildCount() > 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Right" : "Left" )
                   << " side is not empty.";
      throw ex;
    }

    // a least one formula we need
    if( a == NULL || a->GetChildCount() == 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Left" : "Right" )
                   << " side is empty.";
      throw ex;
    }
    
    // for each formula
    for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      if( r->IsLeftRule() )
      {
        a_0 = world->GetLeft();
        e_0 = world->GetRight();
      }
      else
      {
        a_0 = world->GetRight();
        e_0 = world->GetLeft();
      }
  
      if( a_0 == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No rules on "
                     << ( r->IsLeftRule() ? "left" : "right" )
                     << " side available.";
  
        throw ex;
      }

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
                                ( a->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();
#if 0 //def DEBUG_PRINT_INFO
      Print::PrintFormula( f, std::cerr );
      std::cerr << std::endl;
#endif

      std::queue<qt2_list_info> todo;

      // build queue
      bool found = false;
      for( size_t j = 0; j < a_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 = dynamic_cast<ProofLib::SFormula*>
                                    ( a_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        if( f == sf_1->GetFormula() )
        {
          found = true;
          break;
        }

        if( sf_1->GetFormula()->CastTo<ProofLib::SecondOrderQuantorFormula>() == NULL )
          continue;

        qt2_list_info t;
        t.f = sf_1->GetFormula();
        t.parent = t.f;
        todo.push( t );
      }

      // exists f already on the goal side?
      if( found )
        continue; // nothing to do in this case...

      ProofLib::Formula *parent = NULL;

      // the quantor variabels of the fromulas to be processed
      std::queue<ProofLib::SecondOrderVariable*> vars;
      // the lambda terms which will replace the quantor variables.
      std::queue<const ProofLib::LambdaTerm*> rep_terms;

      ProofLib::LambdaSubstitution *sigma = NULL;

      // build bintree to solve
      while( !todo.empty() )
      {
        // f_1 is the possible final parent and p_0 is the "real" parent, i.e.
        // a s-formula in the world by now, while f_1 != p_0 indicates that
        // f_1 is a real subformula of p_0
        ProofLib::Formula *f_1 = todo.front().f;
        ProofLib::Formula *p_0 = todo.front().parent;

        // the quantor variables x_1,...,x_n so that 
        //    p_0 == (Q x_1) ... (Q x_n) f_1
        // holds.
        vars = todo.front().vars;

        todo.pop();
#if 0
        std::cerr << "TRY: ";
        Print::PrintFormula( f_1, std::cerr );
        std::cerr << std::endl;
#endif
        if ( r->GetLambdaTerm() != NULL )
        {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "Rule has LambdaTerm: ";
          r->GetLambdaTerm()->ToString( std::cerr );
          std::cerr << std::endl;
#endif
          if ( !vars.empty() )
          {
#if 0
          	std::cerr << "vars.front() = ";
          	vars.front()->ToString(std::cerr);
          	std::cerr << std::endl;
#endif          	
            ProofLib::OperationLambdaSubstitution *mysigma = 
                new ProofLib::OperationLambdaSubstitution( m_lk_pdb,
                                                           vars.front(),
                                                           r->GetLambdaTerm() );

            ProofLib::Formula *f1_subst = ProofLib::DataNode::ApplyTo( f_1, *mysigma );
            if ( f1_subst == f )
            {
              sigma = mysigma;	
            } else
            {
              delete mysigma;
              continue;
            }
          }
        } else if ( !vb1_applies ) // weak rule
        {
          // try to find this lambda term automatically
#if 0
          std::cerr << "weak quant rule" << std::endl;
#endif
          ProofLib::TermList bound_vars;
          ProofLib::OperationGetBoundedVariables op_get_vars( bound_vars );
          a_0->Apply( op_get_vars );
          ProofLib::OperationGetLambdaSubstitutions *subs =
            new ProofLib::OperationGetLambdaSubstitutions( m_lk_pdb,
                                                           m_doc->GetNodeDatabase(),
                                                           m_doc->GetSymbolTable(),
                                                           bound_vars );
          if ( subs->CheckForSubstitutions( f_1, f ) )
          {
            sigma = subs;
          } else
          {
            delete subs;
          }
        } else
        {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "looking for substitution"<< std::endl;
#endif
          ProofLib::OperationGetSecondOrderVariableSubstitutions *mysigma = new ProofLib::OperationGetSecondOrderVariableSubstitutions( m_lk_pdb, m_doc->GetNodeDatabase(), m_doc->GetSymbolTable() );
          if ( !mysigma->CheckForSubstitutions( f_1, f ) )
          {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "couldn't find any"<< std::endl;
#endif
          	delete mysigma;
          }
          else
          {
#ifdef DEBUG_PRINT_INFO
          std::cerr << "found one"<< std::endl;
#endif
          	sigma = mysigma;
          }
        }

        if( sigma != NULL )
        {
          bool ok = true;

          // check if the unification base on the quantor vars
          if( vars.size() < sigma->GetCount() )
            ok = false;

          std::queue<ProofLib::SecondOrderVariable*> check = vars;

          // we know that the are so many quantifier var as unifications,
          // but are those unifications of the quantifier vars?
          size_t missed = 0;
          while( ok && !check.empty() )
          {
            bool found = false;

            for( size_t k = 0; k < sigma->GetCount(); k++ )
            {
              if( sigma->GetVar(k) == check.front() )
                found = true;
            }

            if( !found )
            {
               missed++;
            
              // for vb1 (eigenvariable) each variable has to be replace;
              // substitutions like x <- x are not premitted.
              if( vb1_applies )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
    
                ex->Stream() << "Second Order Quantor variable '";
                check.front()->ToString( ex->Stream() );
                ex->Stream() << "' has to be substituted by another variable "
                             << "as requiered by VB1 quantifier rule "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }
            }


            check.pop();
          }

          if( vars.size() != sigma->GetCount() + missed )
            ok = false;


          // check if (as far as requested) VB1 is satisfied.
          if( ok && vb1_applies )
          {
            // to check:
            //  1. the replacement R is a variable
            //  2. R does not occures in given sequent.
            //  3. R is not the replacement of another quantor var.

            for( size_t k = 0; k < sigma->GetCount(); k++ )
            {
              const ProofLib::LambdaTerm *t = sigma->GetSubst(k);

              //  1. the replacement R is a second order variable
              const ProofLib::AtomFormula *f = dynamic_cast<ProofLib::AtomFormula*>(t->getFormula());
              if( f == NULL || f->GetSecondOrderVariable() == NULL)
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
    
                ex->Stream() << "Replacement ";
                t->ToString( ex->Stream() );
                ex->Stream() << " is not a second order variable as needed by "
                             << "VB1 quantifier rule "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }

              //  2. R does not occures in given sequent.
	      bool found = false;
	      ProofLib::SElementList *list = world->GetLeft();
              for ( size_t i = 0; i < list->GetChildCount(); i++)
	      {
		ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>(list->GetSElement( i ));
		ASSERT( sf != NULL );
		ProofLib::AtomFormula *af = dynamic_cast<ProofLib::AtomFormula*>(sf->GetFormula());
		if ( af != NULL )
		{
                  ProofLib::SecondOrderVariable *v = af->GetSecondOrderVariable();
		  if ( v == f->GetSecondOrderVariable() )
		  {
		    found = true;
		  }
		}
	      }

	      list = world->GetRight();
	      for ( size_t i = 0; i < list->GetChildCount(); i++)
	      {
                ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>(list->GetSElement( i ));
		ASSERT( sf != NULL );
		ProofLib::AtomFormula *af = dynamic_cast<ProofLib::AtomFormula*>(sf->GetFormula());
		if ( af != NULL )
		{
                  ProofLib::SecondOrderVariable *v = af->GetSecondOrderVariable();
		  if ( v == f->GetSecondOrderVariable() )
		  {
		    found = true;
		  }
		}
	      }

              if( found )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );
    
                exception = ex;
    
                ex->Stream() << "Variable ";
                ProofLib::AtomFormula *f = dynamic_cast<ProofLib::AtomFormula*>( t->getFormula() );
                ASSERT( f != NULL );
                ProofLib::SecondOrderVariable *v = f->GetSecondOrderVariable();
                ASSERT( v != NULL );
                v->ToString( ex->Stream() );
                ex->Stream() << " violates VB1 quantifier rule by "
                             << "occuring in the conclusion "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }

              //  3. R is not the replacement of another quantor var.
              for( size_t l = k+1; l < sigma->GetCount(); l++ )
              {
                if( sigma->GetSubst(l) == t )
                {
                  ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        
                  if( exception != NULL )
                    ex->AppendTo( exception );
      
                  exception = ex;
      
                  ex->Stream() << "Variable ";
                  t->ToString( ex->Stream() );
                  ex->Stream() << " is the replacement of two different "
                               << "quantor variables "
                               << "when using ";
                  Print::PrintFormula( p_0, ex->Stream() );
                  ex->Stream() << " as parent.";
      
                  ok = false;
                  break;
                }
              }

              if( !ok ) break;
            }
          }

          // check if (as far as requested) VB2 is satisfied.
          if( ok && !vb1_applies )
          {
            // to check:
            //  1. any variable occuring in the replacement R does no
            //     occur bounded in A(x) (if (Q x) A(x) is the active quantor
            //     formula).

            ProofLib::Formula *f_v = p_0;
 
            for( size_t k = 0; k < sigma->GetCount(); k++ )
            {
              ProofLib::SecondOrderQuantorFormula *qf = 
                dynamic_cast<ProofLib::SecondOrderQuantorFormula*>( f_v );
      
              ASSERT( qf != NULL && qf->GetOperatorSymbol() == symbol );

              f_v = qf->GetFormula();

              ProofLib::TermList bound_vars;
              ProofLib::OperationGetBoundedVariables op( bound_vars );
              f_v->Apply( op );
              
              const ProofLib::LambdaTerm *t = sigma->GetSubst(k);
              ProofLib::Formula *f = t->getFormula();

              ProofLib::TermList t_vars;
              ProofLib::OperationGetVariables op2( t_vars );
              f->Apply( op2 );
              
              for( size_t l = 0; l < t_vars.GetChildCount(); l++ )
              {
                if( bound_vars.FindTerm( t_vars.GetTerm(l) ) >= 0 )
                {
                  ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        
                  if( exception != NULL )
                    ex->AppendTo( exception );
      
                  exception = ex;
      
                  ex->Stream() << "Variable ";
                  Print::PrintTerm( t_vars.GetTerm(l), ex->Stream() );
                  ex->Stream() << " in replacement term ";
                  t->ToString( ex->Stream() );
                  ex->Stream() << " violates VB2 quantor rule by "
                               << "occuring bounded in ";
                  Print::PrintFormula( f_v, ex->Stream() );
                  ex->Stream() << " when using ";
                  Print::PrintFormula( p_0, ex->Stream() );
                  ex->Stream() << " as parent.";
      
                  ok = false;
                  break;
                }
              }

              if( !ok ) break;
            }
          }
#if 0
          std::cerr << "building replacement queue" << std::endl;
#endif
          // build replacement queue
          while( !rep_terms.empty() ) rep_terms.pop();
          std::queue<ProofLib::SecondOrderVariable*> left_vars = vars;
    
          while( ok && !left_vars.empty() )
          {
            std::queue<ProofLib::SecondOrderVariable*> check = left_vars;
            check.pop();
    
            // assert, that the quantor variables are distinct
            while( ok && !check.empty() )
            {
              if( left_vars.front() == check.front() )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
      
                ex->Stream() << "Second Order Quantifier variable ";
                left_vars.front()->ToString( ex->Stream() );
                ex->Stream() << " occures in two quantifier "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }
    
              check.pop();
            }
    
            if( !ok )
              break;

            bool inserted = false;
            for( size_t l = 0; l < sigma->GetCount(); l++ )
            {
              if( left_vars.front() == sigma->GetVar(l) )
              {
                rep_terms.push( sigma->GetSubst(l) );
                inserted = true;
              }
            }

            // FIXME: create lambda term?
            //if( !inserted )
            //  rep_terms.push( left_vars.front() );
    
            left_vars.pop();
          }

          if( ok )
          {
#if 0
          	std::cerr << "found parent:";
          	Print::PrintFormula( p_0, std::cerr ) ;
          	std::cerr << std::endl;
#endif
            parent = p_0;
            break;
          }
        }

        ProofLib::SecondOrderQuantorFormula *qf = 
          dynamic_cast<ProofLib::SecondOrderQuantorFormula*>( f_1 );

        if( qf == NULL || qf->GetOperatorSymbol() != symbol )
          continue;

        qt2_list_info t;
        t.f = qf->GetFormula();
        t.parent = p_0;
        t.vars = vars;
        t.vars.push( qf->GetSecondOrderVariable() );

        todo.push( t );
      }

      if ( sigma != NULL )
      {
        delete sigma;
        sigma = NULL;
      }

      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        ex->AppendTo( exception );

        throw ex;
      }

      // delete unused exceptions
      if( exception != NULL )
      {
        delete exception;
        exception = NULL;
      }

      // generate rules
      while( !vars.empty() )
      {
        ASSERT( !rep_terms.empty() );

        ProofLib::SecondOrderQuantorFormula *qf = 
          dynamic_cast<ProofLib::SecondOrderQuantorFormula*>( parent );

        rule = BuildSecondOrderQuantorRule( rule, qf, r->IsLeftRule(), rep_terms.front(),
                                 world, &parent );

        vars.pop();
        rep_terms.pop();
      }
    }

    return rule;
  }

  ProofLib::Rule* ProcessQuantorRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    Exception *exception = NULL;
    const ProofLib::OperatorSymbol *symbol = NULL;
    bool vb1_applies = false;

    switch( r->GetType() )
    {
    case RESWORD_ALL:
      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_ALL );
      vb1_applies = !r->IsLeftRule();
      break;
    case RESWORD_EX:
      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_EX );
      vb1_applies = r->IsLeftRule();
      break;
    default:
      ASSERT( false );
    }

    // a   ... aux side of the "goal" sequent
    // e   ... other side of the "goal" sequent
    // a_0 ... aux side of the last sequent
    // e_0 ... other side of the last sequent
    ProofLib::SElementList *a,*e;
    ProofLib::SElementList *a_0,*e_0;

    if( r->IsLeftRule() )
    {
      a = r->GetSequent()->GetLeft();
      e = r->GetSequent()->GetRight();
    }
    else
    {
      a = r->GetSequent()->GetRight();
      e = r->GetSequent()->GetLeft();
    }

    // unactive side has to be empty
    if( e != NULL && e->GetChildCount() > 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Right" : "Left" )
                   << " side is not empty.";
      throw ex;
    }

    // a least one formula we need
    if( a == NULL || a->GetChildCount() == 0 )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << ( r->IsLeftRule() ? "Left" : "Right" )
                   << " side is empty.";
      throw ex;
    }

    // for each formula
    for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      if( r->IsLeftRule() )
      {
        a_0 = world->GetLeft();
        e_0 = world->GetRight();
      }
      else
      {
        a_0 = world->GetRight();
        e_0 = world->GetLeft();
      }
  
      if( a_0 == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No rules on "
                     << ( r->IsLeftRule() ? "left" : "right" )
                     << " side available.";
  
        throw ex;
      }

      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>
                                ( a->GetSElement(i) );
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();
#if 0 //def DEBUG_PRINT_INFO
      Print::PrintFormula( f, std::cerr );
      std::cerr << std::endl;
#endif

      std::queue<qt_list_info> todo;

      // build queue
      bool found = false;
      for( size_t j = 0; j < a_0->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sf_1 = dynamic_cast<ProofLib::SFormula*>
                                    ( a_0->GetSElement(j) );
        ASSERT( sf_1 != NULL );

        if( f == sf_1->GetFormula() )
        {
          found = true;
          break;
        }

        if( sf_1->GetFormula()->CastTo<ProofLib::QuantorFormula>() == NULL )
          continue;

        qt_list_info t;
        t.f = sf_1->GetFormula();
        t.parent = t.f;
        todo.push( t );
      }
        
      // exists f already on the goal side?
      if( found )
        continue; // nothing to do in this case...

      ProofLib::Formula *parent = NULL;

      // the quantor variabels of the fromulas to be processed
      std::queue<ProofLib::Variable*> vars;
      // the terms which will replace the quantor variables.
      std::queue<const ProofLib::Term*> rep_terms;


      ProofLib::OperationGetVariableSubstitutions sigma;
        
      // build bintree to solve
      while( !todo.empty() )
      {
        // f_1 is the possible final parent and p_0 is the "real" parent, i.e.
        // a s-formula in the world by now, while f_1 != p_0 indicates that
        // f_1 is a real subformula of p_0
        ProofLib::Formula *f_1 = todo.front().f;
        ProofLib::Formula *p_0 = todo.front().parent;

        // the quantor variables x_1,...,x_n so that 
        //    p_0 == (Q x_1) ... (Q x_n) f_1
        // holds.
        vars = todo.front().vars;

        todo.pop();
#if 0
        std::cerr << "TRY: ";
        Print::PrintFormula( f_1, std::cerr );
        std::cerr << std::endl;
#endif

        if( sigma.CheckForSubstitutions( f_1, f ) )
        {
          // sigma.RemoveInvariants();

          bool ok = true;

          // check if the unification base on the quantor vars
          if( vars.size() < sigma.GetCount() )
            ok = false;

          std::queue<ProofLib::Variable*> check = vars;

          // we know that the are so many quantifier var as unifications,
          // but are those unifications of the quantifier vars?
          size_t missed = 0;
          while( ok && !check.empty() )
          {
            bool found = false;

            for( size_t k = 0; k < sigma.GetCount(); k++ )
            {
              if( sigma.GetVar(k) == check.front() )
                found = true;
            }

            if( !found )
            {
               missed++;
            
              // for vb1 (eigenvariable) each variable has to be replace;
              // substitutions like x <- x are not premitted.
              if( vb1_applies )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
    
                ex->Stream() << "Quantor variable '";
                Print::PrintTerm( check.front(), ex->Stream() );
                ex->Stream() << "' has to be substituted by another variable "
                             << "as requiered by VB1 quantifier rule "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }
            }


            check.pop();
          }

          if( vars.size() != sigma.GetCount() + missed )
            ok = false;

    
          // check if (as far as requested) VB1 is satisfied.
          if( ok && vb1_applies )
          {
            // to check:
            //  1. the replacement R is a variable
            //  2. R does not occures in given sequent.
            //  3. R is not the replacement of another quantor var.

            for( size_t k = 0; k < sigma.GetCount(); k++ )
            {
              const ProofLib::Term *t = sigma.GetSubst(k);

              //  1. the replacement R is a variable
              if( t->CastTo<ProofLib::Variable>() == NULL )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
    
                ex->Stream() << "Replacement ";
                Print::PrintTerm( t, ex->Stream() );
                ex->Stream() << " is not a variable as needed by "
                             << "VB1 quantifier rule "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }

              //  2. R does not occures in given sequent.
              if( world->ContainsPointer( t ) )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
                if( exception != NULL )
                  ex->AppendTo( exception );
    
                exception = ex;
    
                ex->Stream() << "Variable ";
                Print::PrintTerm( t, ex->Stream() );
                ex->Stream() << " violates VB1 quantifier rule by "
                             << "occuring in the conclusion "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }

              //  3. R is not the replacement of another quantor var.
              for( size_t l = k+1; l < sigma.GetCount(); l++ )
              {
                if( sigma.GetSubst(l) == t )
                {
                  ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        
                  if( exception != NULL )
                    ex->AppendTo( exception );
      
                  exception = ex;
      
                  ex->Stream() << "Variable ";
                  Print::PrintTerm( t, ex->Stream() );
                  ex->Stream() << " is the replacement of two different "
                               << "quantor variables "
                               << "when using ";
                  Print::PrintFormula( p_0, ex->Stream() );
                  ex->Stream() << " as parent.";
      
                  ok = false;
                  break;
                }
              }

              if( !ok ) break;
            }
          }

          // check if (as far as requested) VB2 is satisfied.
          if( ok && !vb1_applies )
          {
            // to check:
            //  1. any variable occuring in the replacement R does no
            //     occur bounded in A(x) (if (Q x) A(x) is the active quantor
            //     formula).

            ProofLib::Formula *f_v = p_0;
 
            for( size_t k = 0; k < sigma.GetCount(); k++ )
            {
              ProofLib::QuantorFormula *qf = 
                dynamic_cast<ProofLib::QuantorFormula*>( f_v );
      
              ASSERT( qf != NULL && qf->GetOperatorSymbol() == symbol );

              f_v = qf->GetFormula();

              ProofLib::TermList bound_vars;
              ProofLib::OperationGetBoundedVariables op( bound_vars );
              f_v->Apply( op );
             // f_v->GetBoundedVariables( bound_vars );
              
              ProofLib::Term *t = (ProofLib::Term*) sigma.GetSubst(k);

              ProofLib::TermList t_vars;
              ProofLib::OperationGetVariables op2( t_vars );
              t->Apply( op2 );
              // t->GetVariables( t_vars );
              
              for( size_t l = 0; l < t_vars.GetChildCount(); l++ )
              {
                if( bound_vars.FindTerm( t_vars.GetTerm(l) ) >= 0 )
                {
                  ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        
                  if( exception != NULL )
                    ex->AppendTo( exception );
      
                  exception = ex;
      
                  ex->Stream() << "Variable ";
                  Print::PrintTerm( t_vars.GetTerm(l), ex->Stream() );
                  ex->Stream() << " in replacement term ";
                  Print::PrintTerm( t, ex->Stream() );
                  ex->Stream() << " violates VB2 quantor rule by "
                               << "occuring bounded in ";
                  Print::PrintFormula( f_v, ex->Stream() );
                  ex->Stream() << " when using ";
                  Print::PrintFormula( p_0, ex->Stream() );
                  ex->Stream() << " as parent.";
      
                  ok = false;
                  break;
                }
              }

              if( !ok ) break;
            }
          }

          // build replacement queue
          while( !rep_terms.empty() ) rep_terms.pop();
          std::queue<ProofLib::Variable*> left_vars = vars;
    
          while( ok && !left_vars.empty() )
          {
            std::queue<ProofLib::Variable*> check = left_vars;
            check.pop();
    
            // assert, that the quantor variables are distinct
            while( ok && !check.empty() )
            {
              if( left_vars.front() == check.front() )
              {
                ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

                if( exception != NULL )
                  ex->AppendTo( exception );

                exception = ex;
      
                ex->Stream() << "Quantifier variable ";
                Print::PrintTerm( left_vars.front(), ex->Stream() );
                ex->Stream() << " occures in two quantifier "
                             << "when using ";
                Print::PrintFormula( p_0, ex->Stream() );
                ex->Stream() << " as parent.";
    
                ok = false;
                break;
              }
    
              check.pop();
            }
    
            if( !ok )
              break;

            bool inserted = false;
            for( size_t l = 0; l < sigma.GetCount(); l++ )
            {
              if( left_vars.front() == sigma.GetVar(l) )
              {
                rep_terms.push( sigma.GetSubst(l) );
                inserted = true;
              }
            }

            if( !inserted )
              rep_terms.push( left_vars.front() );
    
            left_vars.pop();
          }

          if( ok )
          {
            parent = p_0;
            break;
          }
        }

        ProofLib::QuantorFormula *qf = 
          dynamic_cast<ProofLib::QuantorFormula*>( f_1 );

        if( qf == NULL || qf->GetOperatorSymbol() != symbol )
          continue;

        qt_list_info t;
        t.f = qf->GetFormula();
        t.parent = p_0;
        t.vars = vars;
        t.vars.push( qf->GetVariable() );

        todo.push( t );
      }
      
      // no luck: we did not find the parent.
      if( parent == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "No parent rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << " found.";

        ex->AppendTo( exception );

        throw ex;
      }

      // delete unused exceptions
      if( exception != NULL )
      {
        delete exception;
        exception = NULL;
      }

      // generate rules
      while( !vars.empty() )
      {
        ASSERT( !rep_terms.empty() );

        ProofLib::QuantorFormula *qf = 
          dynamic_cast<ProofLib::QuantorFormula*>( parent );

        rule = BuildQuantorRule( rule, qf, r->IsLeftRule(), rep_terms.front(),
                                 world, &parent );

#if 0
        ASSERT( qf != NULL && qf->GetOperatorSymbol() == symbol );

        // new sequent
        ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

        ProofLib::OperationVariableSubstitution
          tr( m_lk_pdb, vars.front(), (ProofLib::Term*) rep_terms.front() );

        parent = ProofLib::DataNode::ApplyTo( qf->GetFormula(), tr );

        if( r->IsLeftRule() )
        {
          seq->AddLeft( new ProofLib::SFormula( parent ) );
          world->AddLeft( new ProofLib::SFormula( parent ) );
          UpdateConculsionForUsedParent( rule, qf, true );
        }
        else
        {
          seq->AddRight( new ProofLib::SFormula( parent ) );
          world->AddRight( new ProofLib::SFormula( parent ) );
          UpdateConculsionForUsedParent( rule, qf, false );
        }

#ifdef DEBUG_PRINT_INFO
        Print::PrintTerm( vars.front(), std::cerr );
        std::cerr << " -> ";
        Print::PrintTerm( rep_terms.front(), std::cerr );
        std::cerr << ": ";
        Print::PrintSequent( seq, std::cerr );
        std::cerr << std::endl;
#endif

        // build rule
        ProofLib::Rule *next_rule = new ProofLib::Rule;

        next_rule->SetConclusion( seq );

        SetRuleName( rule, ( r->IsLeftRule() ?  rule_left : rule_right ) );

        rule->AddRight( next_rule );

        rule = next_rule;
#endif

        vars.pop();
        rep_terms.pop();
      }
    }

    return rule;
  }

  ProofLib::Rule* BuildQuantorRule(
    ProofLib::Rule *rule,
    ProofLib::QuantorFormula *qf,
    bool left,
    const ProofLib::Term *rep,
    ProofLib::Sequent *world,
    ProofLib::Formula **f )
  {
    ASSERT( rule != NULL );
    ASSERT( qf != NULL );
    ASSERT( rep != NULL );
    ASSERT( f != NULL );

    const ProofLib::OperatorSymbol *symbol = qf->GetOperatorSymbol();

    ASSERT( symbol->IsExistsQuantor() || symbol->IsAllQuantor() );

    int rule_left;
    int rule_right;

    if( symbol->IsAllQuantor() )
    {
      rule_left = RULE_FORALLL;
      rule_right = RULE_FORALLR;
    }
    else
    {
      rule_left = RULE_EXISTSL;
      rule_right = RULE_EXISTSR;
    }

    // new sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    ProofLib::OperationVariableSubstitution
      tr( m_lk_pdb, qf->GetVariable(), rep );

    *f = ProofLib::DataNode::ApplyTo( qf->GetFormula(), tr );

    if( left )
    {
      seq->AddLeft( new ProofLib::SFormula( *f ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( *f ) );
      UpdateConculsionForUsedParent( rule, qf, true );
    }
    else
    {
      seq->AddRight( new ProofLib::SFormula( *f ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( *f ) );
      UpdateConculsionForUsedParent( rule, qf, false );
    }

#ifdef DEBUG_PRINT_INFO
    Print::PrintFormula( qf->GetFormula(), std::cerr );
    std::cerr << " -> ";
    Print::PrintTerm( rep, std::cerr );
    std::cerr << ": ";
    Print::PrintSequent( seq, std::cerr );
    std::cerr << std::endl;
#endif

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    SetRuleName( rule, ( left ?  rule_left : rule_right ) );

    rule->AddRight( next_rule );

    return next_rule;
  }

  ProofLib::Rule* BuildSecondOrderQuantorRule(
    ProofLib::Rule *rule,
    ProofLib::SecondOrderQuantorFormula *qf,
    bool left,
    const ProofLib::LambdaTerm *rep,
    ProofLib::Sequent *world,
    ProofLib::Formula **f )
  {
    ASSERT( rule != NULL );
    ASSERT( qf != NULL );
    ASSERT( rep != NULL );
    ASSERT( f != NULL );
#if 0
    std::cerr << "Building second order quantor rule" << std::endl;
#endif
    const ProofLib::OperatorSymbol *symbol = qf->GetOperatorSymbol();

    ASSERT( symbol->IsSecondOrderExistsQuantor() || symbol->IsSecondOrderAllQuantor() );

    int rule_left;
    int rule_right;

    if( symbol->IsSecondOrderAllQuantor() )
    {
      rule_left = RULE_FORALLL2;
      rule_right = RULE_FORALLR2;
      
      if ( left )
      {
      	rule->SetLambdaTerm( rep );
      }
      
    }
    else
    {
      rule_left = RULE_EXISTSL2;
      rule_right = RULE_EXISTSR2;
      
      if ( !left )
      {
      	rule->SetLambdaTerm( rep );
      }
    }

    // new sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    ProofLib::OperationLambdaSubstitution
      tr( m_lk_pdb, qf->GetSecondOrderVariable(), rep );

    *f = ProofLib::DataNode::ApplyTo( qf->GetFormula(), tr );
#if 0
    std::cerr << "applied lambda substitution" << std::endl;
#endif
    if( left )
    {
      seq->AddLeft( new ProofLib::SFormula( *f ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( *f ) );
      UpdateConculsionForUsedParent( rule, qf, true );
    }
    else
    {
      seq->AddRight( new ProofLib::SFormula( *f ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( *f ) );
      UpdateConculsionForUsedParent( rule, qf, false );
    }

#ifdef DEBUG_PRINT_INFO
    Print::PrintFormula( qf->GetFormula(), std::cerr );
    std::cerr << " -> ";
    rep->ToString( std::cerr ) ;
    std::cerr << ": ";
    Print::PrintSequent( seq, std::cerr );
    std::cerr << std::endl;
#endif

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    SetRuleName( rule, ( left ?  rule_left : rule_right ) );

    rule->AddRight( next_rule );

    return next_rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessBinaryAndOrImplRule(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( (r->GetRefLeft() != NULL && r->GetRefRight() != NULL)
            || r->GetSequent() != NULL );

    const Proof* p_l = NULL;
    const Proof* p_r = NULL;

    ProofLib::Sequent *seq_l = NULL;
    ProofLib::Sequent *seq_r = NULL;

    if( r->GetRefLeft() != NULL )
    {
      p_l = GetProof( r->GetRefLeft() );

      if( p_l != NULL )
        seq_l = p_l->GetEndSequent();
      else if( r->GetRefLeft()->IsExplicitAxiom() )
        seq_l = r->GetRefLeft()->GetExplicitAxiom();
    }
    else if( r->GetByFormula() == NULL )
      seq_l = (ProofLib::Sequent *) r->GetSequent();

    if( r->GetRefRight() != NULL )
    {
      p_r = GetProof( r->GetRefRight() );

      if( p_r != NULL )
        seq_r = p_r->GetEndSequent();
      else if( r->GetRefRight()->IsExplicitAxiom() )
        seq_r = r->GetRefRight()->GetExplicitAxiom();
    }
    else
      seq_r = (ProofLib::Sequent *) r->GetSequent();

    const ProofLib::OperatorSymbol *symbol = NULL;
    ProofLib::SElementList *a = NULL;
    ProofLib::SElementList *b = NULL;
    ProofLib::SElementList *a_0 = NULL;
    ProofLib::SElementList *b_0 = NULL;
    ProofLib::SElementList *e_0 = NULL;

    switch( r->GetType() )
    {
    case RESWORD_AND:
      ASSERT( !r->IsLeftRule() );

      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_AND );

      if( seq_l != NULL ) a = seq_l->GetRight();
      if( seq_r != NULL ) b = seq_r->GetRight();
      a_0 = world->GetRight();
      b_0 = world->GetRight();
      e_0 = world->GetRight();
      break;
    case RESWORD_OR:
      ASSERT( r->IsLeftRule() );

      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_OR );

      if( seq_l != NULL ) a = seq_l->GetLeft();
      if( seq_r != NULL ) b = seq_r->GetLeft();
      a_0 = world->GetLeft();
      b_0 = world->GetLeft();
      e_0 = world->GetLeft();
      break;
    case RESWORD_IMPL:
      ASSERT( r->IsLeftRule() );

      symbol = GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_IMPL );

      if( seq_l != NULL ) a = seq_l->GetRight();
      if( seq_r != NULL ) b = seq_r->GetLeft();
      a_0 = world->GetRight();
      b_0 = world->GetLeft();
      e_0 = world->GetLeft();
      break;
    default:
      ASSERT( false );
    }

    // find the new formulas in the upper formulas
    ProofLib::Formula *f_l = FindFirstNewFormula( a, a_0 );
    ProofLib::Formula *f_r = FindFirstNewFormula( b, b_0 );

    // find the conjunctive formula in the lower formula
    ProofLib::BinaryConFormula *parent = NULL;

    if( e_0 ) for( size_t i_e_0 = 0; i_e_0 < e_0->GetChildCount(); i_e_0++ )
    {
      ProofLib::SFormula *sf_e_0 =
        dynamic_cast<ProofLib::SFormula*>( e_0->GetSElement(i_e_0) );

      ASSERT( sf_e_0 != NULL );

      ProofLib::BinaryConFormula *bf =
        sf_e_0->GetFormula()->CastTo<ProofLib::BinaryConFormula>();

#if 0
Print::PrintFormula( sf_e_0->GetFormula(), std::cerr );
std::cerr << std::endl;
if( f_l != NULL )
{
  Print::PrintFormula( f_l, std::cerr );
  std::cerr << std::endl;
}
if( f_r != NULL )
{
  Print::PrintFormula( f_r, std::cerr );
  std::cerr << std::endl;
}
  std::cerr << " ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
#endif
      if( bf == NULL ||
          bf->GetOperatorSymbol() != symbol ||
          ( f_l != NULL && bf->GetLeft() != f_l ) ||
          ( f_r != NULL && bf->GetRight() != f_r ) )
        continue;

      if( f_l == NULL ) f_l = bf->GetLeft();
      if( f_r == NULL ) f_r = bf->GetRight();

      parent = bf;
      break;
    }

    // no luck: we did not find the parent.
    if( parent == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      if( f_l == NULL && f_r == NULL )
      {
        ex->Stream() << "No " << symbol->GetName()
                     << " rule found in lower sequent.";
      }
      else if( f_l != NULL && f_r != NULL )
      {
        ex->Stream() << "No parent for ";
        Print::PrintFormula( f_l, ex->Stream() );
        ex->Stream() << " " << symbol->GetName() << " ";
        Print::PrintFormula( f_r, ex->Stream() );
        ex->Stream() << " found in lower sequent.";
      }
      else
      {
        ex->Stream() << "No " << symbol->GetName() << " rule with subformula ";
        if( f_l != NULL )
        {
          Print::PrintFormula( f_l, ex->Stream() );
          ex->Stream() << " on the left side";
        }
        else
        {
          Print::PrintFormula( f_r, ex->Stream() );
          ex->Stream() << " on the right side";
        }
        ex->Stream() << " found in lower sequent.";
      }

      throw ex;
    }

    ProofLib::Rule *next_rule_l = NULL;
    ProofLib::Rule *next_rule_r = NULL;
    ProofLib::Sequent *world_l = NULL;
    ProofLib::Sequent *world_r = NULL;

    if( r->GetSequent() != NULL && r->GetRefLeft() == NULL )
      world_l = world;
    if( r->GetSequent() != NULL && r->GetRefRight() == NULL )
      world_r = world;

    ASSERT( world_l != world_r || world_l == NULL );

    BuildBinaryAndOrImplRule( rule, parent, world,
                              &next_rule_l, &next_rule_r,
                              &world_l, &world_r );
#if 0
std::cerr << "WORLD LEFT: ";
Print::PrintSequent( world_l, std::cerr );
std::cerr << std::endl;
std::cerr << "WORLD RIGHT: ";
Print::PrintSequent( world_r, std::cerr );
std::cerr << std::endl;
#endif

    if( r->GetRefLeft() != NULL )
    {
      if( r->GetRefLeft()->IsExplicitAxiom() )
        InsertExplicitAxiom( r->GetRefLeft()->GetExplicitAxiom(),
                             next_rule_l, r, world_l );
      else
        ContinueProofWith( p_l, next_rule_l, r, world_l, r->GetRefLeft()->GetTautology() );

      delete world_l;
    }

    if( r->GetRefRight() != NULL )
    {
      if( r->GetRefRight()->IsExplicitAxiom() )
        InsertExplicitAxiom( r->GetRefRight()->GetExplicitAxiom(),
                             next_rule_r, r, world_r );
      else
        ContinueProofWith( p_r, next_rule_r, r, world_r, r->GetRefRight()->GetTautology() );
      delete world_r;
    }

    if( r->GetSequent() != NULL && r->GetRefLeft() == NULL )
      return next_rule_l;
    if( r->GetSequent() != NULL && r->GetRefRight() == NULL )
      return next_rule_r;

    return NULL;
  }

  ProofLib::Formula* FindFirstNewFormula( ProofLib::SElementList *a,
                                          ProofLib::SElementList *a_0 )
  {
    if( a == NULL )
      return NULL;

    if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
    {
      ProofLib::SFormula *sf_a =
        dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );

      ASSERT( sf_a != NULL );

      ProofLib::Formula *f_a = sf_a->GetFormula();

      bool found = false;
      if( a_0 != NULL )
        for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
      {
        ProofLib::SFormula *sf_a_0 =
          dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );

        ASSERT( sf_a_0 != NULL );

        if( f_a == sf_a_0->GetFormula() )
        {
          found = true;
          break;
        }
      }

      if( !found )
        return f_a;
    }

    return NULL;
  }

  void BuildBinaryAndOrImplRule(
    ProofLib::Rule *rule,
    ProofLib::BinaryConFormula *bf,
    const ProofLib::Sequent *world,
    ProofLib::Rule **next_rule_l,
    ProofLib::Rule **next_rule_r,
    ProofLib::Sequent **world_l,
    ProofLib::Sequent **world_r )
  {
    ASSERT( rule != NULL );
    ASSERT( bf != NULL );
    ASSERT( next_rule_l != NULL );
    ASSERT( next_rule_r != NULL );
    ASSERT( world_l == NULL || world != NULL );
    ASSERT( world_r == NULL || world != NULL );

    bool left_rule = false, left_add_left = false, right_add_left = false;
    int rule_name = 0;

    if( bf->GetOperatorSymbol()->IsAndJunctor() )
    {
      rule_name = RULE_ANDR;
      left_rule = false;
      left_add_left = false;
      right_add_left = false;
    }
    else if( bf->GetOperatorSymbol()->IsOrJunctor() )
    {
      rule_name = RULE_ORL;
      left_rule = true;
      left_add_left = true;
      right_add_left = true;
    }
    else if( bf->GetOperatorSymbol()->IsImplJunctor() )
    {
      rule_name = RULE_IMPLL;
      left_rule = true;
      left_add_left = false;
      right_add_left = true;
    }
    else
      ASSERT( false ); // can not handle whatever it is ...

    SetRuleName( rule, rule_name );
    UpdateConculsionForUsedParent( rule, bf, left_rule );

    ProofLib::Formula *f_l = bf->GetLeft();
    ProofLib::Formula *f_r = bf->GetRight();

    if( world_l != NULL && *world_l != world )
      *world_l = world->Clone();
    if( world_r != NULL && *world_r != world )
      *world_r = world->Clone();

    // left side
    ProofLib::Sequent *seq_l = new ProofLib::Sequent( NULL, NULL );
    
    if( left_add_left )
    {
      seq_l->AddLeft( new ProofLib::SFormula( f_l ) );
      if( world_l != NULL )
        (*world_l)->AddLeft( new ProofLib::SFormula( f_l ) );
    }
    else
    {
      seq_l->AddRight( new ProofLib::SFormula( f_l ) );
      if( world_l != NULL )
        (*world_l)->AddRight( new ProofLib::SFormula( f_l ) );
    }
    
    *next_rule_l = new ProofLib::Rule;
    (*next_rule_l)->SetConclusion( seq_l );

    rule->AddRight( *next_rule_l );

    // right side
    ProofLib::Sequent *seq_r = new ProofLib::Sequent( NULL, NULL );

    if( right_add_left )
    {
      seq_r->AddLeft( new ProofLib::SFormula( f_r ) );
      if( world_r != NULL )
        (*world_r)->AddLeft( new ProofLib::SFormula( f_r ) );
    }
    else
    {
      seq_r->AddRight( new ProofLib::SFormula( f_r ) );
      if( world_r != NULL )
        (*world_r)->AddRight( new ProofLib::SFormula( f_r ) );
    }
    
    *next_rule_r = new ProofLib::Rule;
    (*next_rule_r)->SetConclusion( seq_r );

    rule->AddRight( *next_rule_r );
  }

  //////////////////////////////////////////////////////////////////////////////

  typedef struct
  {
    ProofLib::Formula* parent;
    ProofLib::ParamodulationSubstitutions* sigma;
  } para_list_info;

  ProofLib::Rule* ProcessParamod(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
#if 0
        std::cerr << "================================================================================" << std::endl;
#endif
    ASSERT( r->GetType() == RESWORD_PARAMOD );
    ASSERT( r->GetRefLeft() != NULL || r->GetRefRight() != NULL ||
            r->GetByFormula() != NULL );
    ASSERT( (r->GetRefLeft() != NULL && r->GetRefRight() != NULL)
            || r->GetSequent() != NULL );

    const Proof* p_l = NULL;
    const Proof* p_r = NULL;

    const ProofLib::Sequent *u_seq_l = NULL;
    const ProofLib::Sequent *u_seq_r = NULL;

    if( r->GetRefLeft() != NULL )
    {
      if( r->GetRefLeft()->IsExplicitAxiom() )
        u_seq_l = r->GetRefLeft()->GetExplicitAxiom();
      else
        p_l = GetProof( r->GetRefLeft() );
    }
    else if( r->GetByFormula() == NULL )
      u_seq_l = r->GetSequent();

    if( r->GetRefRight() != NULL )
    {
      if( r->GetRefRight()->IsExplicitAxiom() )
        u_seq_r = r->GetRefRight()->GetExplicitAxiom();
      else
        p_r = GetProof( r->GetRefRight() );
    }
    else
      u_seq_r = r->GetSequent();

    if( p_r == NULL && u_seq_r == NULL )
    {
      ProofLib::GenericSLocException *ex =
        new ProofLib::GenericSLocException( r );

      ex->Stream() << "Auto propositional for right proof not allowed "
                   << "when using paramodulation.";

      throw ex;
    }

    ProofLib::Sequent *seq_r = ( p_r != NULL ? p_r->GetEndSequent() :
                                               u_seq_r->Clone() );
#if 0
        std::cerr << "seq_r := ";
        Print::PrintSequent( seq_r, std::cerr );
        std::cerr << std::endl;
#endif
    ProofLib::SElementList *a = seq_r->GetLeft();
    ProofLib::SElementList *a_0 = world->GetLeft();
    ProofLib::Formula *f = NULL;
    bool on_left_side = true;

    std::queue<para_list_info> possibilities;

    // search on both sides for formulas (f_a) and then check if there is
    // a formula (f_a_0) in the world that could be the parent of it.
    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
#endif

        // search in the world ...
        bool found = false;
        if( a_0 != NULL )
          for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
        {
          ProofLib::SFormula *sf_a_0 =
            dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );
    
          ASSERT( sf_a_0 != NULL );
    
          ProofLib::Formula *f_a_0 = sf_a_0->GetFormula();
#if 0
        std::cerr << "TRY: ";
        Print::PrintFormula( f_a_0, std::cerr );
        std::cerr << std::endl;
#endif

          // f_a alread is in the "world"
          if( f_a == f_a_0 )
          {
            found = true;
            break;
          }

          // Compare these formulas
          ProofLib::ParamodulationSubstitutions *sigma =
            new ProofLib::ParamodulationSubstitutions;

          if( !sigma->CheckForSubstitutions( f_a_0, f_a ) ||
              sigma->GetCount() == 0 )
          {
            //delete sigma;
            continue;
          }

#if 0
        Print::PrintTerm( sigma->GetTerm(0), std::cerr );
        std::cerr << " <--- ";
        Print::PrintTerm( sigma->GetSubst(0), std::cerr );
        std::cerr << std::endl;
#endif

          if( sigma->GetCount() > 1 )
            for( size_t i = 1; i < sigma->GetCount(); i++ )
          {
#if 0
        Print::PrintTerm( sigma->GetTerm(i), std::cerr );
        std::cerr << " <--- ";
        Print::PrintTerm( sigma->GetSubst(i), std::cerr );
        std::cerr << std::endl;
#endif
            if( sigma->GetTerm(0) != sigma->GetTerm(i) ||
                sigma->GetSubst(0) != sigma->GetSubst(i) )
            {
              //delete sigma;
              sigma = NULL;
              break;
            }
          }
#if 0
        std::cerr << "--------------------------------------------------------------------------------" << std::endl;
#endif

          if( sigma == NULL )
            continue;

          para_list_info info;
          info.parent = f_a_0;
          info.sigma = sigma;

          possibilities.push( info );
        }

        if( !found ) // we have something that is not in the world,
                     // so we assume it is the active formula
        {
          f = f_a;

          if( possibilities.empty() )
          {
            ProofLib::GenericSLocException *ex;

            if( r->GetRefRight() )
              ex = new ProofLib::GenericSLocException( r->GetRefRight() );
            else
              ex = new ProofLib::GenericSLocException( r );
      
            ex->Stream() << "Found no formula to match ";
            Print::PrintFormula( f, ex->Stream() );
            ex->Stream() << ".";
      
            throw ex;
          }

          break;
        }
      }

      a = seq_r->GetRight();
      a_0 = world->GetRight();
      if( f != NULL || !on_left_side )
        break; // finally done

      on_left_side = false;
    }

    if( f == NULL )
    {
      if( ( seq_r->GetRight() != NULL &&
            seq_r->GetRight()->GetChildCount() > 0 )
          ||
          ( seq_r->GetLeft() != NULL &&
            seq_r->GetLeft()->GetChildCount() > 0 ) )
        return rule;// we found at least something that is already in the world.

      ProofLib::GenericSLocException *ex =
        new ProofLib::GenericSLocException( r->GetRefRight() );

      ex->Stream() << "Need one formula to apply paramodulation.";

      throw ex;
    }

    ProofLib::Sequent *seq_l = NULL;
    ProofLib::SElementList *b = NULL;

    if( p_l != NULL )
    {
      seq_l = p_l->GetEndSequent();
      b = seq_l->GetRight();
    }
    else if( u_seq_l != NULL )
    {
      seq_l = u_seq_l->Clone();
      b = seq_l->GetRight();
    }
    else if( r->GetByFormula() != NULL )
    {
      seq_l = new ProofLib::Sequent( NULL, NULL );
      seq_l->AddRight( (ProofLib::Formula*) r->GetByFormula() );
      b = seq_l->GetRight();
    }

    ProofLib::Formula *parent = NULL;
    ProofLib::Formula *eq = NULL;
    ProofLib::ParamodulationSubstitutions *sigma = NULL;
    bool type_one = true;

    if( b == NULL )
    {
      parent = possibilities.front().parent;
      sigma = possibilities.front().sigma;
      possibilities.pop();
    }
    else while( !possibilities.empty() )
    {
      sigma = possibilities.front().sigma;

#if 0
    std::cerr << "--------------------" << std::endl;
    for( size_t i = 0; i < sigma->GetCount(); i++ )
    {
      if( i > 0 ) std::cerr << ", ";

      Print::PrintNodePointer( &sigma->GetNodePointer(i), std::cerr );
    }

    std::cerr << std::endl;
#endif

      bool found = false;
      for( size_t i_b = 0; i_b < b->GetChildCount(); i_b++ )
      {
        ProofLib::SFormula *sf_b =
          dynamic_cast<ProofLib::SFormula*>( b->GetSElement(i_b) );
  
        ASSERT( sf_b != NULL );
  
        ProofLib::AtomFormula *f_b =
          sf_b->GetFormula()->CastTo<ProofLib::AtomFormula>();

        if( f_b == NULL )
          continue;
  
        const PredicateSymbol *ps =
          dynamic_cast< const PredicateSymbol*>( f_b->GetSymbol() );

        ASSERT( ps != NULL );

        if( !ps->IsEqualPred() )
          continue;

#if 0
        Print::PrintTerm( f_b->GetTerm(0), std::cerr );
        std::cerr << std::endl;
        Print::PrintTerm( sigma->GetTerm(0), std::cerr );
        std::cerr << std::endl;
        Print::PrintTerm( f_b->GetTerm(1), std::cerr );
        std::cerr << std::endl;
        Print::PrintTerm( sigma->GetSubst(0), std::cerr );
        std::cerr << std::endl;
#endif
        if( f_b->GetTerm(0) == sigma->GetTerm(0) &&
            f_b->GetTerm(1) == sigma->GetSubst(0) )
        {
          type_one = true;
          found = true;
        }
        else if( f_b->GetTerm(1) == sigma->GetTerm(0) &&
                 f_b->GetTerm(0) == sigma->GetSubst(0) )
        {
          type_one = false;
          found = true;
        }

        if( found )
        {
#if 0
    std::cerr << "- FOUND -" << std::endl;
#endif
          eq = f_b;
          parent = possibilities.front().parent;
          break;
        }
      }

      possibilities.pop();

      if( !found )
        delete sigma;
      else
        break;
    }

#if 0
    for( size_t i = 0; i < sigma->GetCount(); i++ )
    {
      if( i > 0 ) std::cerr << ", ";

      Print::PrintNodePointer( &sigma->GetNodePointer(i), std::cerr );
    }

    std::cerr << std::endl;
#endif

    while( !possibilities.empty() )
    {
      ASSERT( possibilities.front().sigma != sigma );
      delete possibilities.front().sigma;
      possibilities.pop();
    }

    if( parent == NULL )
    {
      ProofLib::GenericSLocException *ex;

      if( r->GetRefRight() != NULL )
        ex = new ProofLib::GenericSLocException( r->GetRefRight() );
      else
        ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << "No equation found to create ";
      Print::PrintFormula( f, ex->Stream() );
      ex->Stream() << ".";

      throw ex;
    }

    int rule_name;

    if( on_left_side )
      if( type_one )
        rule_name = RULE_EQL2;
      else
        rule_name = RULE_EQL1;
    else
      if( type_one )
        rule_name = RULE_EQR2;
      else
        rule_name = RULE_EQR1;

    std::string param;

    for( size_t i = 0; i < sigma->GetCount(); i++ )
      param = PrintXMLNodePointer( sigma->GetNodePointer(i), param );

#if 0
std::cerr << param << std::endl;
Print::PrintFormula( parent, std::cerr );
std::cerr << std::endl;
#endif
    SetRuleName( rule, rule_name, param );
    UpdateConculsionForUsedParent( rule, parent, on_left_side );

#ifdef DEBUG_PRINT_INFO
    for( size_t i = 0; i < sigma->GetCount(); i++ )
    {
      if( i > 0 ) std::cerr << ", ";

      Print::PrintNodePointer( &sigma->GetNodePointer(i), std::cerr );
    }

    std::cerr << std::endl;
#endif

    // left side
    seq_l = new ProofLib::Sequent( NULL, NULL );

    // create equation
    if( eq == NULL )
    {
      ProofLib::TermList tlist;
  
      if( type_one )
      {
        tlist.AddRight( (ProofLib::Term*) sigma->GetTerm(0) );
        tlist.AddRight( (ProofLib::Term*) sigma->GetSubst(0) );
      }
      else
      {
        tlist.AddRight( (ProofLib::Term*) sigma->GetSubst(0) );
        tlist.AddRight( (ProofLib::Term*) sigma->GetTerm(0) );
      }
  
      const PredicateSymbol *ps =
        GetDoc()->GetSymbolTable()->GetEqualPredicateSymbol();
  
      eq = ProofLib::AtomFormula::
        New( GetDoc()->GetNodeDatabase(), ps, &tlist );
    }

    delete sigma;

    seq_l->AddRight( new ProofLib::SFormula( eq ) );

    //ASSERT( u_seq_l == NULL || u_seq_r == NULL );
    
    ProofLib::Sequent *world_l = NULL;
    ProofLib::Sequent *world_r = NULL;
    ProofLib::Rule *next_rule_l = NULL;
    ProofLib::Rule *next_rule_r = NULL;

    if( u_seq_l != NULL )
      world_l = world;
    else if( p_l != NULL )
      world_l = world->Clone();

    if( u_seq_r != NULL )
      world_r = world;
    else
      world_r = world->Clone();

    if( p_l == NULL && u_seq_l == NULL )
    {
      rule->AddRight( seq_l );
#if 0
Print::PrintSequent( seq_l, std::cerr );
std::cerr << std::endl;
#endif
    }
    else
    {
      world_l->AddRight( new ProofLib::SFormula( eq ) );

      next_rule_l = new ProofLib::Rule;
      next_rule_l->SetConclusion( seq_l );
  
      rule->AddRight( next_rule_l );
  
  #if 0 // def DEBUG_PRINT_INFO
      Print::PrintSequent( seq_l, std::cerr );
      std::cerr << std::endl;
      Print::PrintSequent( world_l, std::cerr );
      std::cerr << std::endl;
  #endif
      if( u_seq_l == NULL )
        ContinueProofWith( p_l, next_rule_l, r, world_l, r->GetRefLeft()->GetTautology() );
      else if( r->GetRefLeft() != NULL && r->GetRefLeft()->IsExplicitAxiom() )
        InsertExplicitAxiom( r->GetRefLeft()->GetExplicitAxiom(),
                             next_rule_l, r, world_l );
    }
  
    // right side
    seq_r = new ProofLib::Sequent( NULL, NULL );

    if( on_left_side )
    {
      seq_r->AddLeft( new ProofLib::SFormula( f ) );
      world_r->AddLeft( new ProofLib::SFormula( f ) );
    }
    else
    {
      seq_r->AddRight( new ProofLib::SFormula( f ) );
      world_r->AddRight( new ProofLib::SFormula( f ) );
    }
    
    next_rule_r = new ProofLib::Rule;
    next_rule_r->SetConclusion( seq_r );

    rule->AddRight( next_rule_r );

    if( u_seq_r == NULL )
      ContinueProofWith( p_r, next_rule_r, r, world_r, r->GetRefRight()->GetTautology() );
    else if( r->GetRefRight() != NULL && r->GetRefRight()->IsExplicitAxiom() )
      InsertExplicitAxiom( r->GetRefRight()->GetExplicitAxiom(),
                           next_rule_r, r, world_r );

    if( u_seq_l == NULL && p_l != NULL )
      delete world_l;
    if( u_seq_r == NULL )
      delete world_r;

    if( u_seq_l != NULL &&
        ( r->GetRefLeft() == NULL || r->GetRefLeft()->IsExplicitAxiom() ) )
      return next_rule_l;
    if( u_seq_r != NULL &&
        ( r->GetRefRight() == NULL || r->GetRefRight()->IsExplicitAxiom() ) )
      return next_rule_r;

    return NULL;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessCut(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_CUT );
    ASSERT( r->GetCutFormula() != NULL );
    ASSERT( (r->GetRefLeft() != NULL && r->GetRefRight() != NULL)
            || r->GetSequent() != NULL );

    ProofLib::Sequent *u_seq_l = NULL;
    ProofLib::Sequent *u_seq_r = NULL;

    if( r->GetRefLeft() == NULL )
      u_seq_l = (ProofLib::Sequent *) r->GetSequent();

    if( r->GetRefRight() == NULL )
      u_seq_r = (ProofLib::Sequent *) r->GetSequent();

    if( u_seq_l != NULL || u_seq_r != NULL )
    {
      ProofLib::SElementList *a = r->GetSequent()->GetLeft();
      bool on_left_side = true;
  
      // just to do both sides of the sequent ...
      while( true )
      {
        if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
        {
          ProofLib::SFormula *sf_a =
            dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
    
          ASSERT( sf_a != NULL );
    
          ProofLib::Formula *f_a = sf_a->GetFormula();
  
          // ignore cut-formula
          if( u_seq_l != NULL && !on_left_side && f_a == r->GetCutFormula() )
            continue;

          if( u_seq_r != NULL && on_left_side && f_a == r->GetCutFormula() )
            continue;

          ProofLib::SElementList *a_0 =
            ( on_left_side ? world->GetLeft() : world->GetRight() );
    
          bool found = false;
    
          if( a_0 != NULL )
            for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
          {
            ProofLib::SFormula *sf_a_0 =
              dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );
      
            ASSERT( sf_a_0 != NULL );
      
            if( f_a == sf_a_0->GetFormula() )
            {
              found = true;
              break;
            }
          }

          if( !found )
          {
            ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
            ex->Stream() << "The formula ";
            Print::PrintFormula( f_a, ex->Stream() );
            ex->Stream() << " on the "
                         << ( on_left_side ? "left" : "right" )
                         << " side of the sequent ";
            if( ( u_seq_l != NULL && !on_left_side ) ||
                ( u_seq_r != NULL && on_left_side ) )
              ex->Stream() << "is neither the cut-formula nor ";
            else
              ex->Stream() << "is not ";
            ex->Stream() << "in the world.";
            throw ex;
          }
        }

        a = r->GetSequent()->GetRight();
  
        if( !on_left_side )
          break; // finally done
  
        on_left_side = false;
      }
    }

    SetRuleName( rule, RULE_CUT );

    ProofLib::Sequent *world_l;
    ProofLib::Sequent *world_r;

    if( u_seq_l != NULL )
      world_l = world;
    else
      world_l = world->Clone();

    if( u_seq_r != NULL )
      world_r = world;
    else
      world_r = world->Clone();

    // left side
    ProofLib::Sequent *seq_l = new ProofLib::Sequent( NULL, NULL );

    seq_l->AddRight( (ProofLib::Formula*) r->GetCutFormula() );
    world_l->AddRight( (ProofLib::Formula*) r->GetCutFormula() );
    
    ProofLib::Rule *next_rule_l = new ProofLib::Rule;
    next_rule_l->SetConclusion( seq_l );

    rule->AddRight( next_rule_l );

    if( u_seq_l == NULL )
    {
      ContinueProofWith( r->GetRefLeft(), next_rule_l, r, world_l );

      delete world_l;
    }
  
    // right side
    ProofLib::Sequent *seq_r = new ProofLib::Sequent( NULL, NULL );

    seq_r->AddLeft( (ProofLib::Formula*) r->GetCutFormula() );
    world_r->AddLeft( (ProofLib::Formula*) r->GetCutFormula() );
    
    ProofLib::Rule *next_rule_r = new ProofLib::Rule;
    next_rule_r->SetConclusion( seq_r );

    rule->AddRight( next_rule_r );

    if( u_seq_r == NULL )
    {
      ContinueProofWith( r->GetRefRight(), next_rule_r, r, world_r );

      delete world_r;
    }

    if( u_seq_l != NULL )
      return next_rule_l;

    if( u_seq_r != NULL )
      return next_rule_r;

    return NULL;
  }

  //////////////////////////////////////////////////////////////////////////////

  ProofLib::Rule* ProcessUndef(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_UNDEF );

    if( r->GetDefIndent().empty() )
      return ProcessUndefPred(
        GetDoc()->GetSymbolTable()->GetEqualPredicateSymbol(),
        rule, r, world );

    ASSERT( !r->GetDefIndent().empty() );

    const Symbol *s =
      GetDoc()->GetSymbolTable()->GetSymbol( r->GetDefIndent() );

    if( s == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << "Undefined symbol " << r->GetDefIndent()
                   << " can not be undefined.";

      throw ex;
    }

    const PredicateSymbol *ps = dynamic_cast< const PredicateSymbol*>( s );

    if( ps != NULL )
      return ProcessUndefPred( ps, rule, r, world );

    const FunctionSymbol *fs = dynamic_cast< const FunctionSymbol*>( s );

    if( fs == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

      ex->Stream() << "Symbol " << r->GetDefIndent() << " has to be either "
                   << "function or predicate symbol.";

      throw ex;
    }

    if( fs->GetFunctionDefinition() != NULL || fs->IsRecursive() )
      return ProcessUndefFunc( fs, rule, r, world );
      
    if( fs->GetThroughFormula() != NULL )
      return ProcessUndefFuncByPred( fs, rule, r, world );

    ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );

    ex->Stream() << "Function symbol " << r->GetDefIndent()
                 << " is not defined through another function or predicate "
                 << "symbol";

    throw ex;
  }

  class pred_definition
  {
  public:
    pred_definition( ProofLib::OperationGetVariableSubstitutions *sigma,
                     ProofLib::Formula *f,
                     bool on_left_side )
      { this->sigma = sigma; this->f = f; this->on_left_side = on_left_side; }

    ProofLib::OperationGetVariableSubstitutions *sigma;
    ProofLib::Formula *f;
    bool on_left_side;
  };

  ProofLib::Rule* ProcessUndefPred(
    const PredicateSymbol *ps,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( ps != NULL );

    if( !ps->IsEqualPred() && ps->GetDefinedFormula() == NULL &&
        !ps->IsRecursive() )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "Predicate " << ps->GetName() << " is atomic and can "
                   << "not be undefined.";
      throw ex;
    }

    if( ps->IsRecursive() && r->GetRecursiveExpr() == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "Predicate " << ps->GetName() << " is recursive and "
                   << "can only be undefined using an recursive expresion.";
      throw ex;
    }

    // D(x) ist defined by an abritrary formula F(x) (x denotes the free vars
    // of F).
    // check the goal sequent if it constis only of formulas that are the
    // definition of 'ps' (i.e. the F(x) sigma, sigma is the unifcator for the
    // free variables), and put them togehter in a queue.

    std::queue< pred_definition > todo;

    ProofLib::SElementList *a = r->GetSequent()->GetLeft();
    bool on_left_side = true;

    ProofLib::Formula *F_0 = NULL;
    USTRING ps_name;

    unsigned int level = 0;

    if( ps->IsEqualPred() )
    {
      F_0 = GetDoc()->GetSymbolTable()->GetEqualityDefinitionOf(
        r->GetTypeSymbolLeft(), r->GetTypeSymbolRight() );
      ASSERT( F_0 != NULL );
      ps_name.FromUtf8( "equality" );
    }
    else if( ps->IsRecursive() )
    {
      ASSERT( r->GetRecursiveExpr()->GetVariableName() == NULL ); // const expr

      level = r->GetRecursiveExpr()->GetValueAtLevel( 0 );

      F_0 = ps->GetFormulaAtLevel(
        level,
        GetDoc()->GetNodeDatabase(),
        GetDoc()->GetSymbolTable(),
        *r );

      ps_name = ps->GetName();
    }
    else
    {
      F_0 = ps->GetDefinedFormula();
      ps_name = ps->GetName();
    }

    // just to do both sides ...
    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();

        ProofLib::OperationGetVariableSubstitutions *sigma =
          new ProofLib::OperationGetVariableSubstitutions( true );

        if( !sigma->CheckForSubstitutions( F_0, f_a ) )
        {
          delete sigma;

          ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      
          ex->Stream() << "Formula ";
          Print::PrintFormula( f_a, ex->Stream() );
          ex->Stream() << " is not the definition of " << ps_name << ".";
      
          throw ex;
        }

        todo.push( pred_definition( sigma, f_a, on_left_side ) );
      }

      a = r->GetSequent()->GetRight();

      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }

    if( todo.empty() )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "Empty sequent not allowed here.";
      throw ex;
    }

    // create all rules
    while( !todo.empty() )
    {
      ProofLib::OperationGetVariableSubstitutions *sigma = todo.front().sigma;
      ProofLib::Formula *f = todo.front().f;
      bool on_left_side = todo.front().on_left_side;

      // check that D(x) sigma is in the world
      ProofLib::Formula *D = NULL;

      if( ps->IsEqualPred() )
        D = GetDoc()->GetSymbolTable()->GetEqualityInstance(
          r->GetTypeSymbolLeft(), r->GetTypeSymbolRight(), *sigma,
          GetDoc()->GetNodeDatabase() );
      else
        D = ps->CreateDefineInstance( sigma, GetDoc()->GetNodeDatabase(),
                                      r->GetRecursiveExpr() );

      ProofLib::SElementList *a_0 =
        ( on_left_side ? world->GetLeft() : world->GetRight() );

      bool found_parent = false;

      if( a_0 != NULL )
        for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
      {
        ProofLib::SFormula *sf_a_0 =
          dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );
  
        ASSERT( sf_a_0 != NULL );
  
        if( sf_a_0->GetFormula() == D )
        {
          found_parent = true;
          break;
        }
      }

      if( !found_parent )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        ex->Stream() << "Can not find ";
        Print::PrintFormula( D, ex->Stream() );
        ex->Stream() << " on "
                     << ( on_left_side ? "left" : "right" )
                     << " side of the sequent to build rule for ";
        Print::PrintFormula( f, ex->Stream() );
        ex->Stream() << ".";
        throw ex;
      }

      std::ostringstream str;

      str << "();";

      for( size_t i = 0; i < sigma->GetCount(); i++ )
      {
        str << "[";

        if( ps->GetVariableMapping() == NULL )
        {
          str << "(" << i + 1 << ")";
        }
        else
        {
          bool found = false;
          for( size_t k = 0; k < ps->GetVariableMapping()->GetChildCount();
               k++ )
          {
            if( ps->GetVariableMapping()->GetTerm(k) == sigma->GetVar(i) )
            {
              str << "(" << k + 1 << ")";
              found = true;
              break;
            }
          }
          ASSERT( found );
        }

        for( size_t j = 0; j < sigma->GetSubstPos(i).GetCount(); j++ )
        {
          std::string s;
          PrintXMLNodePointer( sigma->GetSubstPos(i)[j], s );
          str << s; // PrintXMLNodePointer( sigma->GetSubstPos(i)[j], str );
        }

        str << "]";
      }

      str << ";()";
#if 0
std::cerr << str.str() << std::endl;
#endif
      std::string param( str.str() );

      if( Compiler::simulate_pred_undefs )
        rule = BuildPredUndefSimulation( rule, *r, D, f, on_left_side, world );
      else
        rule = BuildPredUndef( rule, *r, ps, level, D, f, param,
                               on_left_side, world );

      delete sigma;
      todo.pop();
    }

    return rule;
  }

  ProofLib::Rule* BuildPredUndef(
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    const ProofLib::Symbol *symbol,
    unsigned int level,
    ProofLib::Formula *D,
    ProofLib::Formula *f,
    const std::string &param,
    bool on_left_side,
    ProofLib::Sequent *world )
  {
    SetRuleName( rule, on_left_side ? RULE_UNDEFL : RULE_UNDEFR, param );

    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    if( on_left_side )
    {
      seq->AddLeft( f );
      if( world != NULL )
        world->AddLeft( f );
    }
    else
    {
      seq->AddRight( f );
      if( world != NULL )
        world->AddRight( f );
    }

    GetUsedDefinitionsListImpl()->AddSymbol( symbol, level );

    UpdateConculsionForUsedParent( rule, D, on_left_side );

    // build rule
    ProofLib::Rule *next_rule = new ProofLib::Rule;

    next_rule->SetConclusion( seq );

    rule->AddRight( next_rule );

    return next_rule;
  }

  ProofLib::Rule* BuildPredUndefSimulation(
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    ProofLib::Formula *D,
    ProofLib::Formula *f,
    bool on_left_side,
    ProofLib::Sequent *world )
  {
    // create ( D(x) <-> F(x) ) sigma (only the way needed) ...
    ProofLib::Formula* def = NULL;

    const ProofLib::OperatorSymbol *os =
      GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_IMPL );

    if( on_left_side )
      def = ProofLib::BinaryConFormula::
        New( GetDoc()->GetNodeDatabase(), os, D, f );
    else
      def = ProofLib::BinaryConFormula::
        New( GetDoc()->GetNodeDatabase(), os, f, D );

    // ... and add it to the end sequent (left side of it) of the root proof.
    AddToGlobalEndSequent( def, true );

    // create the cut
    SetRuleName( rule, RULE_CUT );

    // left side: build def :- f
    ProofLib::Sequent *seq_l = new ProofLib::Sequent( NULL, NULL );
//  ProofLib::Sequent *world_l = world->Clone();

    if( on_left_side )
    {
      seq_l->AddRight( new ProofLib::SFormula( f ) );
//    world_l->AddRight( new ProofLib::SFormula( f ) );
    }
    else
    {
      seq_l->AddLeft( new ProofLib::SFormula( f ) );
//    world_l->AddLeft( new ProofLib::SFormula( f ) );
    }
  
    seq_l->AddLeft( new ProofLib::SFormula( def ) ); // next active formula
//  world_l->AddLeft( new ProofLib::SFormula( def ) );
  
    ProofLib::Rule *impl_rule_l = new ProofLib::Rule;
    impl_rule_l->SetConclusion( seq_l );
    SetRuleName( impl_rule_l, RULE_IMPLL );

    rule->AddRight( impl_rule_l );

    // "apply" implication left, part D :- D
    seq_l = new ProofLib::Sequent( NULL, NULL );

    seq_l->AddLeft( new ProofLib::SFormula( D ) );
    seq_l->AddRight( new ProofLib::SFormula( D ) );

    impl_rule_l->AddRight( seq_l );

    // F :- F
    seq_l = new ProofLib::Sequent( NULL, NULL );

    seq_l->AddLeft( new ProofLib::SFormula( f ) );
//  world_l->AddLeft( new ProofLib::SFormula( f ) );
    seq_l->AddRight( new ProofLib::SFormula( f ) );
//  world_l->AddRight( new ProofLib::SFormula( f ) );

    ProofLib::Rule *ff_rule = new ProofLib::Rule;
    ff_rule->SetConclusion( seq_l );

    if( on_left_side )
      impl_rule_l->AddRight( ff_rule );
    else
      impl_rule_l->AddLeft( ff_rule );

    // make out F :- F a proof with atomic initial sequents
    AxiomToAtomic( ff_rule, sloc, f );
//  AxiomToAtomic( ff_rule, sloc, f, world_l );

//  delete world_l;

    // right (left) side
    ProofLib::Sequent *seq_r = new ProofLib::Sequent( NULL, NULL );

    if( on_left_side )
    {
      seq_r->AddLeft( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( f ) );
    }
    else
    {
      seq_r->AddRight( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( f ) );
    }

    ProofLib::Rule *next_rule = new ProofLib::Rule;
    next_rule->SetConclusion( seq_r );

    if( on_left_side )
      rule->AddRight( next_rule );
    else
      rule->AddLeft( next_rule );

    return next_rule;
  }

  ProofLib::Rule* ProcessUndefFunc(
    const FunctionSymbol *fs,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( fs != NULL );

    // d(x) ist defined by an abritrary term t(x) (x denotes the vars
    // of t).
    
    ProofLib::Term *T = NULL;

    unsigned int level = 0;

    if( fs->IsRecursive() )
    {
      if( r->GetRecursiveExpr() == NULL )
      {
        ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
        ex->Stream() << "In order to undefine a recursive function a recursive "
                     << "expression is required.";
        throw ex;
      }

      ASSERT( r->GetRecursiveExpr()->GetVariableName() == NULL ); // const expr

      level = r->GetRecursiveExpr()->GetValueAtLevel( 0 );

      T = fs->GetTermAtLevel(
        level,
        GetDoc()->GetNodeDatabase(),
        GetDoc()->GetSymbolTable(),
        *r );
    }
    else
      T = fs->GetFunctionDefinition();

#if 0
    std::cerr << "T: ";
    Print::PrintTerm( T, std::cerr );
    std::cerr << std::endl;
#endif

    // find for each s-formula in the goal sequent a parent in the world which
    // it could be unified using the definition.

    if( r->GetSequent()->GetLeft() == NULL &&
        r->GetSequent()->GetRight() == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "Empty sequent not allowed here.";
      throw ex;
    }

    ProofLib::SElementList *a = r->GetSequent()->GetLeft();
    bool on_left_side = true;

    // just to do both sides ...
    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();

        ProofLib::SElementList *a_0 =
          ( on_left_side ? world->GetLeft() : world->GetRight() );
  
        bool found_itself = false;
        bool found_parent = false;
        ProofLib::TermDiffernces *sigma = new ProofLib::TermDiffernces;
  
        if( a_0 != NULL )
          for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
        {
          ProofLib::SFormula *sf_a_0 =
            dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );
    
          ASSERT( sf_a_0 != NULL );
    
          ProofLib::Formula *f_a_0 = sf_a_0->GetFormula();

          if( f_a == f_a_0 )
          {
            found_itself = true;
            break;
          }

          sigma->Clear();

#if 0
        std::cerr << "-------" << std::endl;
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
        Print::PrintFormula( f_a_0, std::cerr );
        std::cerr << std::endl;
#endif

          if( sigma->BuildDifferences( f_a_0, f_a ) )
          {
#if 0
            std::cerr << "found " << sigma->GetCount() << " differences.";
            std::cerr << std::endl;
#endif
            for( size_t i_s = 0; i_s < sigma->GetCount(); i_s++ )
            {
              ProofLib::OperationGetVariableSubstitutions omega( true );

#if 0
              std::cerr << "to check T: ";
              Print::PrintTerm( T, std::cerr );
              std::cerr << std::endl;
              std::cerr << "to check found subst: ";
              Print::PrintTerm(
                (ProofLib::Term*) sigma->GetSubst(i_s), std::cerr );
              std::cerr << std::endl;
#endif
              if( !omega.CheckForSubstitutions(
                    T, (ProofLib::Term*) sigma->GetSubst(i_s) ) )
                continue;
              
#if 0
        Print::PrintVariableSubstitution( &omega, std::cerr );
        std::cerr << std::endl;
#endif
              ProofLib::Term *d =
                fs->CreateDefineInstance( &omega, GetDoc()->GetNodeDatabase(),
                                          r->GetRecursiveExpr() );
#if 0
        std::cerr << "d: ";
        Print::PrintTerm( d, std::cerr );
        std::cerr << std::endl;
#endif
#if 0
        Print::PrintTerm( sigma->GetTerm(i_s), std::cerr );
        std::cerr << std::endl;
#endif
              
              if( d != sigma->GetTerm(i_s) )
                continue;

#if 0
        for( size_t i = 0; i < sigma->GetOccurrences( i_s ).GetCount(); i++ )
        {
          Print::PrintNodePointer( &(sigma->GetOccurrences( i_s )[i]), std::cerr );
        }
        std::cerr << std::endl;
#endif
              ProofLib::OperationTermSubstitution
                tr( m_lk_pdb->GetNodeDatabase(),
                    d, (ProofLib::Term*) sigma->GetSubst(i_s),
                    &(sigma->GetOccurrences( i_s )) );

#if 0
  Print::PrintFormula( ProofLib::DataNode::ApplyTo( f_a_0, tr ), std::cerr );
  std::cerr << std::endl;
#endif
              if( ProofLib::DataNode::ApplyTo( f_a_0, tr ) != f_a )
                continue;
    
              if( Compiler::simulate_pred_undefs )
              {
                std::string param;
  
                for( size_t i = 0;
                     i < sigma->GetOccurrences(i_s).GetCount(); i++ )
                  param = PrintXMLNodePointer(
                            sigma->GetOccurrences(i_s)[i],
                            param );
  
                rule = BuildParamodRule(
                  rule, *r,
                  d, (ProofLib::Term*) sigma->GetSubst(i_s),
                  f_a_0, f_a,
                  param.c_str(),
                  on_left_side, world );
              }
              else
              {
                std::ostringstream str;
        
                for( size_t j = 0;
                     j < sigma->GetOccurrences( i_s ).GetCount(); j++ )
                {
                  std::string s;
                  PrintXMLNodePointer( sigma->GetOccurrences(i_s)[j], s );
                  str << s;
                }

                str << ";";
  
#ifdef DEF_RULE_PARAM_MORE_ACCURATE
                for( size_t j = 0;
                     j < sigma->GetOccurrences( i_s ).GetCount(); j++ )
                {
                  for( size_t i = 0; i < omega.GetCount(); i++ )
                  {
                    str << "[";
            
                    int pos =
                      fs->GetVariableMapping()->FindTerm( omega.GetVar(i) );
                    ASSERT( pos >= 0 );

                    std::string s;
                    PrintXMLNodePointer(
                      sigma->GetOccurrences(i_s)[j] + size_t(pos), s );
                    str << s;
            
                    for( size_t k = 0;
                         k < omega.GetSubstPos(i).GetCount(); k++ )
                    {
                      std::string s;
                      PrintXMLNodePointer(
                        sigma->GetOccurrences(i_s)[j] + omega.GetSubstPos(i)[k],
                        s );
                      str << s;
                    }
            
                    str << "]";
                  }
                }
#else
                for( size_t i = 0; i < omega.GetCount(); i++ )
                {
                  str << "[";
          
                  int pos =
                    fs->GetVariableMapping()->FindTerm( omega.GetVar(i) );
                  ASSERT( pos >= 0 );

                  str << "(" << pos + 1 << ")";
          
                  for( size_t k = 0;
                       k < omega.GetSubstPos(i).GetCount(); k++ )
                  {
                    std::string s;
                    PrintXMLNodePointer( omega.GetSubstPos(i)[k], s );
                    str << s;
                  }
          
                  str << "]";
                }
#endif

                str << ";()";
#if 0
std::cerr << str.str() << std::endl;
#endif
  
                rule = BuildFuncUndef( 
                  rule, *r,
                  fs, level,
                  d, (ProofLib::Term*) sigma->GetSubst(i_s),
                  f_a_0, f_a,
                  str.str().c_str(),
                  on_left_side, world );
              }

              found_parent = true;
              break; // we did it ;-)
            }

            if( found_parent )
              break;
          }
        }
  
        if( found_itself )
          continue; // ignore it and get the next one.

        if( !found_parent )
        {
          ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
          ex->Stream() << "Can not find ";
          ex->Stream() << "a parent on the "
                       << ( on_left_side ? "left" : "right" )
                       << " side of the sequent to build rule for ";
          Print::PrintFormula( f_a, ex->Stream() );
          ex->Stream() << " by undefinig " << fs->GetName();
          if( fs->IsRecursive() )
            ex->Stream() << " at level " << level;
          ex->Stream() << ".";
          throw ex;
        }

      }

      a = r->GetSequent()->GetRight();

      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }

    return rule;
  }

  ProofLib::Rule* BuildParamodRule(
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    ProofLib::Term *d,
    ProofLib::Term *s,
    ProofLib::Formula *f0,
    ProofLib::Formula *f,
    const char *param,
    bool on_left_side,
    ProofLib::Sequent *world )
  {
    // create rule
    if( on_left_side )
      SetRuleName( rule, RULE_EQL2, param );
    else
      SetRuleName( rule, RULE_EQR2, param );

    // parent in the conclusion
    UpdateConculsionForUsedParent( rule, f0, on_left_side );

    // create equation predicate
    ProofLib::TermList args;
    args.AddRight( d );
    args.AddRight( s );

    ProofLib::Formula *eq =
      ProofLib::AtomFormula::New( GetDoc()->GetNodeDatabase(),
        GetDoc()->GetSymbolTable()->GetEqualPredicateSymbol(),
        &args );
      

    // equality axiom for the definition
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );
    
    seq->AddRight( new ProofLib::SFormula( eq ) );

    rule->AddRight( seq );

    // continue proof
    seq = new ProofLib::Sequent( NULL, NULL );

    if( on_left_side )
    {
      seq->AddLeft( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( f ) );
    }
    else
    {
      seq->AddRight( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( f ) );
    }
    
    ProofLib::Rule *next_rule = new ProofLib::Rule;
    next_rule->SetConclusion( seq );

    rule->AddRight( next_rule );

    return next_rule;
  }

  ProofLib::Rule* BuildFuncUndef(
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    const ProofLib::Symbol *symbol,
    unsigned int level,
    ProofLib::Term *d,
    ProofLib::Term *s,
    ProofLib::Formula *f0,
    ProofLib::Formula *f,
    const char *param,
    bool on_left_side,
    ProofLib::Sequent *world )
  {
    // create rule
    if( on_left_side )
      SetRuleName( rule, RULE_UNDEFL, param );
    else
      SetRuleName( rule, RULE_UNDEFR, param );

    // parent in the conclusion
    UpdateConculsionForUsedParent( rule, f0, on_left_side );

    // build sequent
    ProofLib::Sequent *seq = new ProofLib::Sequent( NULL, NULL );

    if( on_left_side )
    {
      seq->AddLeft( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( f ) );
    }
    else
    {
      seq->AddRight( new ProofLib::SFormula( f ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( f ) );
    }
    
    GetUsedDefinitionsListImpl()->AddSymbol( symbol, level );

    ProofLib::Rule *next_rule = new ProofLib::Rule;
    next_rule->SetConclusion( seq );

    rule->AddRight( next_rule );

    return next_rule;
  }

  ProofLib::Rule* ProcessUndefFuncByPred(
    const FunctionSymbol *fs,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( fs != NULL );
    ASSERT( fs->GetThroughFormula() != NULL );
    ASSERT( fs->GetByFormula() != NULL );

    // d(x) ist defined through an atom predicate P(y) by a abritrary fromula
    // F(y) in the following way:
    //      P(d(x),y') = F(x,y')
    // x,y' symobilze the free variable of F, which all and only them occur in
    // P(d(x),y').
    
    // find for each s-formula in the goal sequent a parent in the world which
    // it could be unified using the definition.

    if( r->GetSequent()->GetLeft() == NULL &&
        r->GetSequent()->GetRight() == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "Empty sequent not allowed here.";
      throw ex;
    }

    ProofLib::SElementList *a = r->GetSequent()->GetLeft();
    bool on_left_side = true;

    // just to do both sides ...
    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();

        ProofLib::OperationGetVariableSubstitutions omega( true );

        // try to find a match
        if( !omega.CheckForSubstitutions( fs->GetByFormula(), f_a ) )
        {
          ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
          ex->Stream() << "The formula ";
          Print::PrintFormula( f_a, ex->Stream() );
          ex->Stream() << " on the "
                       << ( on_left_side ? "left" : "right" )
                       << " side of the sequent can not match the definition "
                       << "of " << fs->GetName() << ", which is ";
          Print::PrintFormula( fs->GetByFormula(), ex->Stream() );
          ex->Stream() << ".";
          throw ex;
        }
              
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
        Print::PrintVariableSubstitution( &omega, std::cerr );
        std::cerr << std::endl;
#endif

        // build the parent ...
        ProofLib::Formula *parent =
          fs->CreateThroughInstance( &omega, GetDoc()->GetNodeDatabase() );

#if 0
        Print::PrintFormula( parent, std::cerr );
        std::cerr << std::endl;
#endif

        // ... and search the world for it ...
        ProofLib::SElementList *a_0 =
          ( on_left_side ? world->GetLeft() : world->GetRight() );
  
        bool found_itself = false;
        bool found_parent = false;
  
        if( a_0 != NULL )
          for( size_t i_a_0 = 0; i_a_0 < a_0->GetChildCount(); i_a_0++ )
        {
          ProofLib::SFormula *sf_a_0 =
            dynamic_cast<ProofLib::SFormula*>( a_0->GetSElement(i_a_0) );
    
          ASSERT( sf_a_0 != NULL );
    
          ProofLib::Formula *f_a_0 = sf_a_0->GetFormula();

          if( f_a_0 == f_a )
          {
            found_itself = true;
            break;
          }

          if( f_a_0 == parent )
          {
            found_parent = true;
            break;
          }
        }

        if( found_itself )
          continue; // nothing to do!

        if( !found_parent )
        {
          ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
          ex->Stream() << "Can not find the parent ";
          Print::PrintFormula( parent, ex->Stream() );
          ex->Stream() << " for the formula ";
          Print::PrintFormula( f_a, ex->Stream() );
          ex->Stream() << " on the "
                       << ( on_left_side ? "left" : "right" )
                       << " side of the sequent in order to undefine "
                       << fs->GetName() << ".";
          throw ex;
        }

        // ... and if found so build the rules:
        ProofLib::TermList org_vars; 
        ProofLib::OperationGetFreeVariablesAndPositions sigma( org_vars );

        fs->GetThroughFormula()->Apply( sigma );

        std::ostringstream str;
  
        str << "();";
  
        for( size_t i = 0; i < omega.GetCount(); i++ )
        {
          str << "[";
  
          //str << "(" << i + 1 << ")";

#if 0
        Print::PrintFormula( fs->GetByFormula(), std::cerr );
        std::cerr << std::endl;
        Print::PrintTerm( omega.GetVar(i), std::cerr );
        std::cerr << std::endl;
        Print::PrintTerm( omega.GetSubst(i), std::cerr );
        std::cerr << std::endl;
#endif
          int pos = org_vars.FindTerm( omega.GetVar(i) ); 
          ASSERT( pos >= 0 );
          ASSERT( sigma.GetPositions( pos ).GetCount() == 1 );

          std::string s;
          PrintXMLNodePointer( sigma.GetPositions( pos )[0], s );
          str << s; // PrintXMLNodePointer( sigma.GetPositions( pos )[0], str );
  
          for( size_t j = 0; j < omega.GetSubstPos(i).GetCount(); j++ )
          {
            std::string s;
            PrintXMLNodePointer( omega.GetSubstPos(i)[j], s );
            str << s; // PrintXMLNodePointer( omega.GetSubstPos(i)[j], str );
          }
  
          str << "]";
        }

        str << ";";
        ProofLib::OperationFindFirstFuncSymbolOccurence op( fs );
        parent->Apply( op );
        std::string s;
        PrintXMLNodePointer( op.GetOccurence(), s );
        str << s;
#if 0
std::cerr << str.str() << std::endl;
#endif

        std::string param( str.str() );

        if( Compiler::simulate_pred_undefs )
          rule = BuildFuncByPredUndefSimulation(
            rule, *r, parent, f_a, on_left_side, world );
        else
          rule = BuildPredUndef( rule, *r, fs, 0, parent, f_a, param,
                                 on_left_side, world );
      }

      a = r->GetSequent()->GetRight();

      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }

    return rule;
  }

  ProofLib::Rule* BuildFuncByPredUndefSimulation(
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    ProofLib::Formula *P,
    ProofLib::Formula *F,
    bool on_left_side,
    ProofLib::Sequent *world )
  {
    // create P(d(x),y') = F(x,y') omega (only the way needed) ...
    ProofLib::Formula* def = NULL;

    const ProofLib::OperatorSymbol *os =
      GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_IMPL );

    if( on_left_side )
      def = ProofLib::BinaryConFormula::
        New( GetDoc()->GetNodeDatabase(), os, P, F );
    else
      def = ProofLib::BinaryConFormula::
        New( GetDoc()->GetNodeDatabase(), os, F, P );

    // ... and add it to the end sequent (left side of it) of the root
    // proof.
    AddToGlobalEndSequent( def, true );

    // create the cut
    SetRuleName( rule, RULE_CUT );

    // left side: build def :- F
    ProofLib::Sequent *seq_l = new ProofLib::Sequent( NULL, NULL );
//  ProofLib::Sequent *world_l = world->Clone();

    if( on_left_side )
    {
      seq_l->AddRight( new ProofLib::SFormula( F ) );
//    world_l->AddRight( new ProofLib::SFormula( F ) );
    }
    else
    {
      seq_l->AddLeft( new ProofLib::SFormula( F ) );
//    world_l->AddLeft( new ProofLib::SFormula( F ) );
    }
  
    seq_l->AddLeft( new ProofLib::SFormula( def ) ); // next active formula
//  world_l->AddLeft( new ProofLib::SFormula( def ) );

    ProofLib::Rule *impl_rule_l = new ProofLib::Rule;
    impl_rule_l->SetConclusion( seq_l );
    SetRuleName( impl_rule_l, RULE_IMPLL );

    rule->AddRight( impl_rule_l );

    // "apply" implication left, part P :- P
    seq_l = new ProofLib::Sequent( NULL, NULL );

    seq_l->AddLeft( new ProofLib::SFormula( P ) );
    seq_l->AddRight( new ProofLib::SFormula( P ) );

    impl_rule_l->AddRight( seq_l );

    // F :- F
    seq_l = new ProofLib::Sequent( NULL, NULL );
  
    seq_l->AddLeft( new ProofLib::SFormula( F ) );
//  world_l->AddLeft( new ProofLib::SFormula( F ) );
    seq_l->AddRight( new ProofLib::SFormula( F ) );
//  world_l->AddRight( new ProofLib::SFormula( F ) );
  
    ProofLib::Rule *ff_rule = new ProofLib::Rule;
    ff_rule->SetConclusion( seq_l );
  
    if( on_left_side )
      impl_rule_l->AddRight( ff_rule );
    else
      impl_rule_l->AddLeft( ff_rule );

    // make out F :- F a proof with atomic initial sequents
    AxiomToAtomic( ff_rule, sloc, F );
//  AxiomToAtomic( ff_rule, sloc, F, world_l );
  
//  delete world_l;
  
    // right (left) side
    ProofLib::Sequent *seq_r = new ProofLib::Sequent( NULL, NULL );
  
    if( on_left_side )
    {
      seq_r->AddLeft( new ProofLib::SFormula( F ) );
      if( world != NULL )
        world->AddLeft( new ProofLib::SFormula( F ) );
    }
    else
    {
      seq_r->AddRight( new ProofLib::SFormula( F ) );
      if( world != NULL )
        world->AddRight( new ProofLib::SFormula( F ) );
    }
  
    ProofLib::Rule *next_rule = new ProofLib::Rule;
    next_rule->SetConclusion( seq_r );
  
    if( on_left_side )
      rule->AddRight( next_rule );
    else
      rule->AddLeft( next_rule );
  
    return next_rule;
  }

  //////////////////////////////////////////////////////////////////////////////

  void ProcessContinued(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_CONTINUED );

    ContinueProofWith( r->GetRef(), rule, r, world );
  }

  const Proof* GetProof( const ProofRef *ref )
  {
    ASSERT( ref != NULL );

    return ref->GetProof( GetDoc()->GetSymbolTable(),
                          GetDoc()->GetProofTable(),
                          GetDoc()->GetNodeDatabase() );
#if 0
    if( ref->IsAutoProp() )
      return NULL;

    ASSERT( ref->GetName() != NULL );
    // check if proof exists and retrieve proof symbol table entry
    const ProofSymbol *ps = dynamic_cast<const ProofSymbol*>
      ( GetDoc()->GetSymbolTable()->GetSymbol( ref->GetName() ) );

    if( ps == NULL )
    {
      ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException(ref);
      e->Stream() << "No proof '" << ref->GetName() << "' found!";
      throw e;
    }

    return GetDoc()->GetProofTable()->GetProof( ps );
#endif
  }

  void InsertExplicitAxiom(
    const ProofLib::Sequent *ea,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
      ProofLib::FormulaList miss_l, miss_r;
  
      if( world != NULL &&
          !ea->IsSubSequentOf( world, &miss_l, &miss_r ) )
      {
        ProofLib::GenericSLocException *e =
          new ProofLib::GenericSLocException( r );
        e->Stream() << "Explicit Axiom ";
        Print::PrintSequent( ea, e->Stream() );
        e->Stream() << " is not a sub-sequent of the previous one through ";
        if( miss_l.GetChildCount() > 0 )
        {
          Print::PrintFormulaList( &miss_l, e->Stream() );
          e->Stream() << " on the left side";
  
          if( miss_r.GetChildCount() > 0 )
            e->Stream() << " and ";
        }
        if( miss_r.GetChildCount() > 0 )
        {
          Print::PrintFormulaList( &miss_r, e->Stream() );
          e->Stream() << " on the right side";
        }
        e->Stream() << ".";
  
        throw e;
      }

/*
      if( rule->GetParent() == NULL )
      {
        ASSERT( m_lk_proof->GetEndRule() == rule );
        m_lk_proof->SetEndRule( ea->Clone() );
      }
      else
      {
        ProofLib::Rule *rule_p = rule->GetParent()->CastTo<ProofLib::Rule>();
        ASSERT( rule_p != NULL );
  
        rule_p->ReplaceRuleItem( rule, ea->Clone() );
      }
*/
      rule->AddRight( ea->Clone() );
  }

  void ContinueProofWith(
    const ProofRef *ref,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    if( ref->IsExplicitAxiom() )
    {
      InsertExplicitAxiom( ref->GetExplicitAxiom(), rule, r, world );
      return;
    }

    const Proof *p = GetProof( ref );

    ContinueProofWith( p, rule, r, world, ref->GetTautology() );
  }

  void ContinueProofWith(
    const Proof *p,
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world,
    ProofLib::Sequent *taut )
  {
    if( p == NULL ) // auto prop
    {
#ifdef NEW_AUTO_PROP
      ASSERT( taut != NULL );
      if( !taut->IsSubSequentOf( world ) )
      {
          ProofLib::GenericSLocException *ex =
            new ProofLib::GenericSLocException( r );
    
          ex->Stream() << "Sequent ";
          Print::PrintSequent( taut, ex->Stream() );
          ex->Stream() << " is not a sub-sequent of the world "
	    << "(and can therefore not be automatically proved propositionally).";
    
          throw ex;
      }
      NewDeductCompletePropositional( taut, rule, r, world );
#else
      DeductCompletePropositional( rule, r, world );
#endif
      return;
    }

    // loop detection
    for( size_t i = 0; i < m_proof_stack.size(); i++ )
    {
      if( m_proof_stack[i] == p )
      {
        ProofLib::GenericSLocException *e = new ProofLib::GenericSLocException(r);
        e->Stream() << "Proof recursion by calling '"
                    << p->GetSymbol()->GetName()
                    << "' detected!";
        throw e;
      }
    }

    m_proof_stack.push_back( p );

    ProofLib::Proof *p_lk = NULL;

    const ProofInstance* pi = dynamic_cast<const ProofInstance*>(p);

    // check if proof name is already used, if yes: generate the next one!
    int num = 0;
    while( p_lk == NULL )
    {
      std::wostringstream str;

      if( num > 0 )
        str << "{";

      str << p->GetSymbol()->GetName();

      if( pi != NULL )
      {
        const ProofRecursionsExpander *recexp =
          pi->GetProofRecursionsExpander();

        if( recexp != NULL && recexp->GetLevel() > 0 )
          str << "[" << recexp->GetLevel() << "]";
      }

      const ProofParamArgMap* metaparas = pi->GetProofParamArgMap();

      if( metaparas->GetCount() > 0 )
      {
        bool only_terms = true;

        for( size_t i = 0; i < metaparas->GetCount(); i++ )
          if( dynamic_cast<ProofLib::Formula*>(metaparas->GetArg(i)) != NULL )
          {
            only_terms = false;
            break;
          }
        
        if( only_terms )
        {
          str << "(";
          for( size_t i = 0; i < metaparas->GetCount(); i++ )
          {
            if( i != 0 )
              str << ", ";

            ASSERT( dynamic_cast<ProofLib::Term*>
                      (metaparas->GetArg(i)) != NULL );

            HLK::Print::PrintTerm(
              dynamic_cast<ProofLib::Term*>(metaparas->GetArg(i)), str );
          }
          str << ")";
        }
      }

      if( num > 0 )
        str << "}_{" << num << "}";

#if 0
      const ProofSymbol *ps_1 = dynamic_cast<const ProofSymbol*>
        ( GetDoc()->GetSymbolTable()->GetSymbol( str.str().c_str() ) );

      if( ps_1 != NULL )
      {
        p_lk = m_lk_pdb->NewProof( ps_1 );

        if( p_lk != NULL )
          break;

        num++;
        continue;
      }

      try
      {
        GetDoc()->GetSymbolTable()->NewProof( str.str().c_str(), *r );
      }
      catch( SymbolExistsException* )
      {
        num++;
        continue;
      }

      ps_1 = dynamic_cast<const ProofSymbol*>
        ( GetDoc()->GetSymbolTable()->GetSymbol( str.str().c_str() ) );

      p_lk = m_lk_pdb->NewProof( ps_1 );

      num++;
#endif
      try
      {
        GetDoc()->GetSymbolTable()->NewProof( str.str().c_str(), *r );
      }
      catch( SymbolExistsException* )
      {
      }

      const ProofSymbol *ps_1 = dynamic_cast<const ProofSymbol*>
        ( GetDoc()->GetSymbolTable()->GetSymbol( str.str().c_str() ) );

      p_lk = m_lk_pdb->NewProof( ps_1 );

      num++;
    }
#if 0
    int num = 0;
    char *str = new char[strlen(p->GetSymbol()->GetName())+16];
    while( p_lk == NULL )
    {
      num++;
      sprintf( str, "%s_{%d}", p->GetSymbol()->GetName(), num );

      if( GetDoc()->GetSymbolTable()->GetSymbol( str ) != NULL )
        continue;

      GetDoc()->GetSymbolTable()->NewProof( str, *r );

      const ProofSymbol *ps_1 = dynamic_cast<const ProofSymbol*>
        ( GetDoc()->GetSymbolTable()->GetSymbol( str ) );

      p_lk = m_lk_pdb->NewProof( ps_1 );
    }
    delete [] str;
#endif

#if 0
    std::cerr << "Compile Subproof: " << p_lk->GetName() << std::endl;
#endif
  
    // create proof link
    ProofLib::ProofLink *p_link = p_lk->NewProofLink( m_lk_proof );

    //rule->AddRight( p_link );
    
    if( rule->GetParent() == NULL )
    {
#ifdef DEBUG_PRINT_INFO
      std::cerr << "setting end rule" << std::endl;
#endif
      ASSERT( m_lk_proof->GetEndRule() == rule );
      m_lk_proof->SetEndRule( p_link );
    }
    else
    {
#ifdef DEBUG_PRINT_INFO
      std::cerr << "replacing end rule" << std::endl;
#endif
      ProofLib::Rule *rule_p = rule->GetParent()->CastTo<ProofLib::Rule>();
      ASSERT( rule_p != NULL );

      rule_p->ReplaceRuleItem( rule, p_link );
    }

    // build next proof
    ProcessProof( p_lk, p, world, rule->GetConclusion(), r );

    delete rule;

    m_proof_stack.pop_back();
  }

  //////////////////////////////////////////////////////////////////////////////

  void AxiomToAtomic( 
    ProofLib::Rule *rule,
    const ProofLib::SourceLocation &sloc,
    ProofLib::Formula *f )
  {
    ASSERT( rule != NULL );
    ASSERT( f != NULL );

    ProofLib::AtomFormula *af = f->CastTo<ProofLib::AtomFormula>();

    if( af != NULL )
    {
      if( Compiler::simulate_pred_undefs )
        return;

      // check if af is a definition:
      const PredicateSymbol *ps =
        dynamic_cast< const PredicateSymbol*>( af->GetSymbol() );

      // 1. predicate symbol is a definition?
      if( ps->GetDefinedFormula() != NULL )
      {
        // ok, f = D( x_0, ..., x_n )

#if 0
Print::PrintFormula( f, std::cerr );
std::cerr << std::endl;
Print::PrintFormula( ps->GetDefinedFormula(), std::cerr );
std::cerr << std::endl;
#endif

        std::vector<ProofLib::NodePointers> positions;

        ProofLib::Formula* D = f;
        ProofLib::Formula* F =
          ps->CreateBodyInstance( D, GetDoc()->GetNodeDatabase(), positions );

        std::ostringstream str;

        str << "();";
 
        for( size_t i = 0; i < positions.size(); i++ )
        {
          str << "[";
          str << "(" << i + 1 << ")";

          for( size_t j = 0; j < positions[i].GetCount(); j++ )
          {
            std::string s;
            PrintXMLNodePointer( positions[i][j], s );
            str << s;
          }

          str << "]";
        }

        str << ";()";

        std::string param( str.str() );
#if 0
std::cerr << "==" << std::endl;
Print::PrintFormula( D, std::cerr );
std::cerr << std::endl;
Print::PrintFormula( F, std::cerr );
std::cerr << std::endl;
std::cerr << "==" << std::endl;
#endif

        if( Compiler::simulate_pred_undefs )
        {
          rule = BuildPredUndefSimulation( rule, sloc, D, F, true, NULL );
          rule = BuildPredUndefSimulation( rule, sloc, D, F, false, NULL );
        }
        else
        {
          rule = BuildPredUndef( rule, sloc, ps, 0, D, F, param, true, NULL );
          rule = BuildPredUndef( rule, sloc, ps, 0, D, F, param, false, NULL );
        }

        AxiomToAtomic( rule, sloc, F );
        return;
      }

      // 2. check for function symbols that are defined.
      ProofLib::OperationFindFirstDefinedFunction op_f;

      f->Apply( op_f );

      if( op_f.GetFunction() != NULL )
      {
        ProofLib::Function *fu = op_f.GetFunction();

        const FunctionSymbol *fs =
          dynamic_cast< const FunctionSymbol*>( fu->GetSymbol() );
        ASSERT( fs != NULL );

        if( fs->GetFunctionDefinition() != NULL )
        {
#if 0
          ASSERT( fs->GetVariableMapping() != NULL );
          ASSERT( fs->GetVariableMapping()->GetChildCount() ==
                  fu->GetChildCount() );

          ProofLib::OperationVariableSubstitution
            omega( GetDoc()->GetNodeDatabase() );

          for( size_t i = 0;
               i < fs->GetVariableMapping()->GetChildCount(); i++ )
          {
            omega.AddSubstitution(
             fs->GetVariableMapping()->GetTerm(i)->CastTo<ProofLib::Variable>(),
             fu->GetTerm(i) );
          }
#if 0
        Print::PrintTerm( fu, std::cerr );
        std::cerr << std::endl;
#endif
#if 0
          ProofLib::OperationGetVariableSubstitutions omega( true );

          if( !omega.CheckForSubstitutions( fs->GetFunctionDefinition(), fu ) )
          {
            ASSERT( false );
          }
              
#endif
#if 0
        Print::PrintVariableSubstitution( &omega, std::cerr );
        std::cerr << std::endl;
#endif

         ProofLib::Term *d =
           fs->CreateDefineInstance( &omega, GetDoc()->GetNodeDatabase() );
#endif
          std::vector<ProofLib::NodePointers> positions;

          ProofLib::Term *d =
            fs->CreateBodyInstance( fu, GetDoc()->GetNodeDatabase(),
                                    positions );
#if 0
        Print::PrintTerm( d, std::cerr );
        std::cerr << std::endl;
#endif
#if 0
        Print::PrintTerm( sigma->GetTerm(i_s), std::cerr );
        std::cerr << std::endl;
#endif
              
          ProofLib::OperationTermSubstitution
            tr( m_lk_pdb->GetNodeDatabase(), fu, d );

          ProofLib::Formula *f1 = ProofLib::DataNode::ApplyTo( f, tr );
    
#if 0
  Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl;
  Print::PrintFormula( f1, std::cerr );
  std::cerr << std::endl;
#endif

          if( Compiler::simulate_pred_undefs )
          {
            std::string param;
  
            for( size_t i = 0; i < tr.GetPositions().GetCount(); i++ )
              param = PrintXMLNodePointer( tr.GetPositions()[i], param );
  
            rule = BuildParamodRule(
              rule, sloc,
              d, fu,
              f, f1,
              param.c_str(),
              true, NULL );
            rule = BuildParamodRule(
              rule, sloc,
              d, fu,
              f, f1,
              param.c_str(),
              false, NULL );
          }
          else
          {
            std::ostringstream str;

#if 0
std::cerr << "tr.GetPositions().GetCount() = " << tr.GetPositions().GetCount();
std::cerr << std::endl;
#endif
            for( size_t j = 0; j < tr.GetPositions().GetCount(); j++ )
            {
              std::string s;
              PrintXMLNodePointer( tr.GetPositions()[j], s );
              str << s;
            }

            str << ";";
  
#ifdef DEF_RULE_PARAM_MORE_ACCURATE
            for( size_t j = 0; j < tr.GetPositions().GetCount(); j++ )
            {
              for( size_t i = 0; i < positions.size(); i++ )
              {
                str << "[";
        
                std::string s;
                PrintXMLNodePointer( tr.GetPositions()[j] + i, s );
                str << s;
        
                for( size_t k = 0; k < positions[i].GetCount(); k++ )
                {
                  std::string s;
                  PrintXMLNodePointer(
                    tr.GetPositions()[j] + positions[i][k], s );
                  str << s;
                }
        
                str << "]";
              }
            }
#else
            for( size_t i = 0; i < positions.size(); i++ )
            {
              str << "[";
      
              str << "(" << i + 1 << ")";
      
              for( size_t k = 0; k < positions[i].GetCount(); k++ )
              {
                std::string s;
                PrintXMLNodePointer( positions[i][k], s );
                str << s;
              }
      
              str << "]";
            }
#endif

            str << ";()";

            rule = BuildFuncUndef( 
              rule, sloc,
              fs, 0,
              d, fu,
              f, f1,
              str.str().c_str(),
              true, NULL );
            rule = BuildFuncUndef( 
              rule, sloc,
              fs, 0,
              d, fu,
              f, f1,
              str.str().c_str(),
              false, NULL );
          }

          AxiomToAtomic( rule, sloc, f1 );
          return;
        }
        else if( fs->GetThroughFormula() != NULL )
        {
          ProofLib::OperationGetVariableSubstitutions sigma;

          if( !sigma.CheckForSubstitutions( fs->GetThroughFormula(), f ) )
          {
            std::cerr << "Warning: Can not introduce definition for ";
            Print::PrintTerm( fu, std::cerr );
            std::cerr << ".";
            return;
          }

#if 0
  Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl;
  std::cerr << "sigma: ";
  Print::PrintVariableSubstitution( &sigma, std::cerr );
  std::cerr << std::endl;
#endif
          std::vector<ProofLib::NodePointers> positions;

          ProofLib::Formula* D = fs->CreateByInstance(
            &sigma,
            GetDoc()->GetNodeDatabase(),
            positions );

#if 0
  std::cerr << "through: ";
  Print::PrintFormula( fs->GetThroughFormula(), std::cerr );
  std::cerr << std::endl;
  std::cerr << "by: ";
  Print::PrintFormula( fs->GetByFormula(), std::cerr );
  std::cerr << std::endl;
  Print::PrintFormula( D, std::cerr );
  std::cerr << std::endl;
#endif
          if( Compiler::simulate_pred_undefs )
          {
            rule = BuildFuncByPredUndefSimulation(
              rule, sloc, f, D, true, NULL );
            rule = BuildFuncByPredUndefSimulation(
              rule, sloc, f, D, false, NULL );
          }
          else
          {
            std::ostringstream str;
      
            str << "();";
      
            for( size_t i = 0; i < sigma.GetCount(); i++ )
            {
              str << "[";
    
              std::string s;
              PrintXMLNodePointer( sigma.GetVarPos(i), s );
              str << s;
      
              for( size_t j = 0; j < positions[i].GetCount(); j++ )
              {
                std::string s;
                PrintXMLNodePointer( positions[i][j], s );
                str << s;
              }
      
              str << "]";
            }

            str << ";";
            ProofLib::OperationFindFirstFuncSymbolOccurence op( fs );
            fs->GetThroughFormula()->Apply( op );
            std::string s;
            PrintXMLNodePointer( op.GetOccurence(), s );
            str << s;

            rule = BuildPredUndef( rule, sloc, fs, 0, f, D, str.str(), true, NULL );
            rule = BuildPredUndef( rule, sloc, fs, 0, f, D, str.str(), false, NULL );
          }
  
          AxiomToAtomic( rule, sloc, D );
          return;
        }
        else if( fs->IsRecursive() )
        {

#if 0
std::cerr << "To expand: ";
HLK::Print::PrintTerm( fu, std::cerr );
std::cerr << std::endl;
#endif

          const ProofLib::RecursiveFunction *f_r =
            fu->CastTo<ProofLib::RecursiveFunction>();

          ASSERT( f_r != NULL );

          ASSERT( f_r->GetRecursiveExpr()->GetVariableName() == NULL );
            // const expr

          unsigned int level = f_r->GetRecursiveExpr()->GetValueAtLevel( 0 );

/*
int level = 0;
          ProofLib::Term *T = fs->GetTermAtLevel(
            level,
            GetDoc()->GetNodeDatabase(),
            GetDoc()->GetSymbolTable(),
            NULL );

              ProofLib::OperationGetVariableSubstitutions omega( true );

              if( !omega.CheckForSubstitutions(
                    T, (ProofLib::Term*) sigma->GetSubst(i_s) ) )
                ASSERT( false );
*/
              
          std::vector<ProofLib::NodePointers> positions;

          ProofLib::Term *d =
            fs->CreateBodyInstance( fu, GetDoc()->GetNodeDatabase(),
                                    positions, level,
                                    GetDoc()->GetSymbolTable(), &sloc );
              
          ProofLib::OperationTermSubstitution
            tr( m_lk_pdb->GetNodeDatabase(), fu, d );

          ProofLib::Formula *f1 = ProofLib::DataNode::ApplyTo( f, tr );
    
#if 0
  Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl;
  Print::PrintFormula( f1, std::cerr );
  std::cerr << std::endl;
#endif

          if( Compiler::simulate_pred_undefs )
          {
            std::string param;
  
            for( size_t i = 0; i < tr.GetPositions().GetCount(); i++ )
              param = PrintXMLNodePointer( tr.GetPositions()[i], param );
  
            rule = BuildParamodRule(
              rule, sloc,
              d, fu,
              f, f1,
              param.c_str(),
              true, NULL );
            rule = BuildParamodRule(
              rule, sloc,
              d, fu,
              f, f1,
              param.c_str(),
              false, NULL );
          }
          else
          {
            std::ostringstream str;

#if 0
std::cerr << "tr.GetPositions().GetCount() = " << tr.GetPositions().GetCount();
std::cerr << std::endl;
#endif
            for( size_t j = 0; j < tr.GetPositions().GetCount(); j++ )
            {
              std::string s;
              PrintXMLNodePointer( tr.GetPositions()[j], s );
              str << s;
            }

            str << ";";
  
#ifdef DEF_RULE_PARAM_MORE_ACCURATE
            for( size_t j = 0; j < tr.GetPositions().GetCount(); j++ )
            {
              for( size_t i = 0; i < positions.size(); i++ )
              {
                str << "[";
        
                std::string s;
                PrintXMLNodePointer( tr.GetPositions()[j] + i, s );
                str << s;
        
                for( size_t k = 0; k < positions[i].GetCount(); k++ )
                {
                  std::string s;
                  PrintXMLNodePointer(
                    tr.GetPositions()[j] + positions[i][k], s );
                  str << s;
                }
        
                str << "]";
              }
            }
#else
            for( size_t i = 0; i < positions.size(); i++ )
            {
              str << "[";
      
              str << "(" << i + 1 << ")";
      
              for( size_t k = 0; k < positions[i].GetCount(); k++ )
              {
                std::string s;
                PrintXMLNodePointer( positions[i][k], s );
                str << s;
              }
      
              str << "]";
            }
#endif

            str << ";()";

            rule = BuildFuncUndef( 
              rule, sloc,
              fs, level,
              d, fu,
              f, f1,
              str.str().c_str(),
              true, NULL );
            rule = BuildFuncUndef( 
              rule, sloc,
              fs, level,
              d, fu,
              f, f1,
              str.str().c_str(),
              false, NULL );
          }

          AxiomToAtomic( rule, sloc, f1 );
          return;
        }
        else
          ASSERT( false );
      }

      return; // we are done here!
    }

    ProofLib::UnaryConFormula *uf = f->CastTo<ProofLib::UnaryConFormula>();

    if( uf != NULL )
    {
      rule = BuildNotRule( rule, uf, true, NULL );
      rule = BuildNotRule( rule, uf, false, NULL );
      AxiomToAtomic( rule, sloc, uf->GetFormula() );
      return;
    }

    ProofLib::BinaryConFormula *bf = f->CastTo<ProofLib::BinaryConFormula>();

    if( bf != NULL )
    {
      if( bf->GetOperatorSymbol()->IsAndJunctor() ||
          bf->GetOperatorSymbol()->IsOrJunctor() )
      {
        ProofLib::Rule *rule_l = NULL;
        ProofLib::Rule *rule_r = NULL;

        BuildBinaryAndOrImplRule( rule, bf, NULL, &rule_l, &rule_r,
                                  NULL, NULL );

        rule_l = BuildUnaryAndOrRule( rule_l, bf, true, NULL );
        rule_r = BuildUnaryAndOrRule( rule_r, bf, false, NULL );

        AxiomToAtomic( rule_l, sloc, bf->GetLeft() );
        AxiomToAtomic( rule_r, sloc, bf->GetRight() );
        return;
      }
      else if( bf->GetOperatorSymbol()->IsImplJunctor() )
      {
        rule = BuildImplRightRule( rule, bf, NULL );

        ProofLib::Rule *rule_l = NULL;
        ProofLib::Rule *rule_r = NULL;

        BuildBinaryAndOrImplRule( rule, bf, NULL, &rule_l, &rule_r,
                                  NULL, NULL );

        AxiomToAtomic( rule_l, sloc, bf->GetLeft() );
        AxiomToAtomic( rule_r, sloc, bf->GetRight() );
        return;
      }
      else
        ASSERT( false ); // can not handle whatever it is ...
    }

    ProofLib::QuantorFormula *qf = f->CastTo<ProofLib::QuantorFormula>();

    if( qf != NULL )
    {
      const VariableSymbol* qvs =
        dynamic_cast<const VariableSymbol*>
          ( qf->GetVariable()->GetSymbol() );

      const VariableSymbol* vs = GetDoc()->GetSymbolTable()->NewVariable(
        NULL,
        qvs->GetType()->GetName(),
        true,
        sloc );

      bool left_first = !qf->GetOperatorSymbol()->IsAllQuantor();
      
      rule = BuildQuantorRule( rule, qf, left_first,
        ProofLib::Variable::New( GetDoc()->GetNodeDatabase(), vs ),
        NULL, &f );
      rule = BuildQuantorRule( rule, qf, !left_first,
        ProofLib::Variable::New( GetDoc()->GetNodeDatabase(), vs ),
        NULL, &f );
      
      AxiomToAtomic( rule, sloc, f );
      return;
    }

    ASSERT( false );
  }

  ProofLib::Rule* ProcessAuto(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ASSERT( r->GetType() == RESWORD_AUTO );

    if( r->GetSequent() == NULL )
    {
      ASSERT( r->GetTautology() != NULL );

      if( !r->GetTautology()->IsSubSequentOf( world ) )
      {
        ProofLib::GenericSLocException *ex =
          new ProofLib::GenericSLocException( r );
  
        ex->Stream() << "Sequent ";
        Print::PrintSequent( r->GetTautology(), ex->Stream() );
        ex->Stream() << " is not a sub-sequent of the world "
	  << "(and can therefore not be automatically proved propositionally).";
    
        throw ex;
      }
      NewDeductCompletePropositional( r->GetTautology(), rule, r, world );
      return NULL;
    }

    ProofLib::OperationFormulasPosAndPol op;
    world->Apply( op );

    const ProofLib::Sequent *seq = r->GetSequent();
    ProofLib::SElementList *a = seq->GetLeft();
    bool on_left_side = true;

#if 0
    ProofLib::OperationFormulasPosAndPol::FormulaPositions::const_iterator 
      positions = op.m_positions.begin();

    for( ; positions != op.m_positions.end(); positions++ )
    {
      Print::PrintFormula( positions->first, std::cerr );
      std::cerr << std::endl;

      for( size_t i = 0; i < positions->second.size(); i++ )
      {
         std::cerr << "  "
                   << (positions->second[i].second ? "left " : "right" )
                   << ": ";
         Print::PrintNodePointer( &(positions->second[i].first), std::cerr );
         std::cerr << std::endl;
      }
    }
#endif

    ProofLib::Sequent add_world( NULL, NULL );

    while( true )
    {
      if( a != NULL ) for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
#endif

        bool found = false;

        ProofLib::OperationFormulasPosAndPol::FormulaPositions::const_iterator 
          positions = op.m_positions.find( f_a );

        long best_match = -1;

        if( positions != op.m_positions.end() )
          for( size_t i = 0; i < positions->second.size(); i++ )
        {
          if( positions->second[i].left != on_left_side )
            continue;

          if( best_match < 0 )
            best_match = i;

          if( positions->second[i].np.GetLength() <
              positions->second[best_match].np.GetLength() )
            best_match = i;

          found = true;
        }

        const ProofLib::OperationFormulasPosAndPol::PositionAndPolarity
          &parent_position = positions->second[best_match];


        if( !found ) // we have something that is not in the world,
                     // so we assume it is the active formula
        {
          ProofLib::GenericSLocException *ex =
            new ProofLib::GenericSLocException( r );
    
          ex->Stream() << "Formula ";
          Print::PrintFormula( f_a, ex->Stream() );
          ex->Stream() << " is no subformula with this polarisation "
                       << "in the world.";
    
          throw ex;
        }

        ProofLib::NodePointer np( parent_position.np );

        std::stack<size_t> todo;

        while( np.GetDataNode( world )->CastTo<ProofLib::Formula>() != NULL )
          todo.push( np.Pop() );

        bool left = ( np[0] == 0 );

        while( !todo.empty() )
        {
          np.Push( todo.top() );
          todo.pop();

          if( todo.empty() )
            break; // done! :-)

          ProofLib::DataNode *node =  np.GetDataNode( world );

          ProofLib::UnaryConFormula *uf = NULL;
          ProofLib::BinaryConFormula *bf = NULL;

          if( (uf = node->CastTo<ProofLib::UnaryConFormula>()) != NULL )
          {
            ASSERT( uf->GetOperatorSymbol()->IsNegJunctor() );

            if( !add_world.ContainsFormula( uf->GetFormula(), !left ) )
              rule = BuildNotRule( rule, uf, left, &add_world );

            left = !left; 
          }
          else if( (bf = node->CastTo<ProofLib::BinaryConFormula>()) != NULL )
          {
            bool left_formula = ( todo.top() == 0 );

            ProofLib::Formula *c =
              left_formula ? bf->GetLeft() : bf->GetRight();

            if( bf->GetOperatorSymbol()->IsImplJunctor() && left_formula )
              left = !left; 

            if( !add_world.ContainsFormula( c, left ) )
            {
              if( bf->GetOperatorSymbol()->IsAndJunctor() ||
                  bf->GetOperatorSymbol()->IsOrJunctor() )
              {
                rule =
                  BuildUnaryAndOrRule( rule, bf, left_formula, &add_world );
              }
              else if( bf->GetOperatorSymbol()->IsImplJunctor() )
              {
                rule = BuildImplRightRule( rule, bf, &add_world );
              }
              else
                ASSERT( false ); // can not handle whatever it is ...
            }
          }
          else
            ASSERT(false);;
        }
      }

      a = seq->GetRight();
      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }
 
    if( add_world.GetLeft() != NULL )
    {
      if( world->GetLeft() == NULL )
        world->SetLeft( add_world.GetLeft()->Clone() );
      else
        world->GetLeft()->AddLeft( add_world.GetLeft()->Clone() );
    }

    if( add_world.GetRight() != NULL )
    {
      if( world->GetRight() == NULL )
        world->SetRight( add_world.GetRight()->Clone() );
      else
        world->GetRight()->AddLeft( add_world.GetRight()->Clone() );
    }

#if 0
    ProofLib::Rule *next_rule = new ProofLib::Rule;
    rule->AddRight( next_rule ); 
    next_rule->SetConclusion( world->Clone() );
#endif

    return rule;
  }

  typedef stlhash::hash_map<
      ProofLib::Formula*,
      std::pair<int,int> >
    FormulaCosts;
  typedef stlhash::hash_map<
      ProofLib::Formula*,
      std::pair<ProofLib::NodePointer,ProofLib::NodePointer> >
    FormulaWays;
  typedef stlhash::hash_map<
      ProofLib::Formula*,
      std::pair<bool,bool> >
    FormulaDone;
  FormulaCosts m_deduct_complete_costs;
  FormulaWays  m_deduct_complete_ways;
  FormulaDone  m_deduct_complete_begun;

  /// Calculates the cost it would take to get the formula f as s-formula on
  /// side "left" (or right if not "left").
  int CalcFormulaCosts(
    ProofLib::Formula *f,
    const ProofLib::OperationFormulasPosAndPol &op,
    bool left,
    ProofLib::Sequent *world )
  {
    // get formula occurrences
    ProofLib::OperationFormulasPosAndPol::FormulaPositions::const_iterator i_f;

    i_f = op.m_positions.find(f);

    ASSERT( i_f != op.m_positions.end() );

    const ProofLib::OperationFormulasPosAndPol::Occurrences &o = i_f->second;

    // check is the result is already in the "cache".
    if( m_deduct_complete_costs.find(f) != m_deduct_complete_costs.end() )
    {
      // left polirization already done? 
      if( left && m_deduct_complete_costs[f].first >= -1 )
        return m_deduct_complete_costs[f].first;
      // right polirization already done? 
      if( !left && m_deduct_complete_costs[f].second >= -1 )
        return m_deduct_complete_costs[f].second;

      // left polirization already begun? (loop dedection!)
      if( left && m_deduct_complete_begun[f].first )
        return 0;
      else
        m_deduct_complete_begun[f].first = left;
      // right polirization already begun? (loop dedection!)
      if( !left && m_deduct_complete_begun[f].second )
        return 0;
      else
        m_deduct_complete_begun[f].second = !left;
    }
    else
    {
      // both sides are not calculated yet.
      m_deduct_complete_costs[f].first = -2;
      m_deduct_complete_costs[f].second = -2;
      m_deduct_complete_begun[f].first = left;
      m_deduct_complete_begun[f].second = !left;
    }

      
#if 0
Print::PrintFormula( f, std::cerr );
std::cerr << std::endl;
std::cerr << ( left ? "left" : "right" ) << std::endl;
#endif
#if 0
std::cerr << "left " <<  ( m_deduct_complete_begun[f].first ? "begun" : "not begun" ) << std::endl;
std::cerr << "right " << ( m_deduct_complete_begun[f].second ? "begun" : "not begun" ) << std::endl;
#endif

    int min_costs = -1;
    const ProofLib::NodePointer *best_np = NULL;

    // check each occurence of f
    bool first = true;
    for( size_t i = 0; i < o.size(); i++ )
    {
      // has this occurrence the polirization we seek for?
      if( o[i].left != left )
        continue;
#if 0
      Print::PrintNodePointer( &o[i].np, std::cerr );
      std::cerr << ": " << o[i].np.GetLength();
      std::cerr << " - " << o[i].first_formula_offset << " = ";
      std::cerr << o[i].np.GetLength() - o[i].first_formula_offset << std::endl;
#endif

      // get costs of f ist self.
      int costs = o[i].np.GetLength() - o[i].first_formula_offset;

      ASSERT( costs >= 0 );

      // check for need of binary rules, since each of them need to satisfy
      // the other formular which is to be proven in a new branch.
      bool child_left = left;
      ProofLib::NodePointer np( o[i].np );
      while( np.GetLength() >= o[i].first_formula_offset )
      {
        // on which side is the very formula?
        bool left_formula = np.Pop() == 0;

        // check the super formula.
        ProofLib::UnaryConFormula *uf =
          np.GetDataNode(world)->CastTo<ProofLib::UnaryConFormula>();

        if( uf != NULL )
        {
          child_left = !child_left;
          continue;
        }

        // check the super formula.
        ProofLib::BinaryConFormula *bf =
          np.GetDataNode(world)->CastTo<ProofLib::BinaryConFormula>();

        if( bf == NULL )
          continue;

        if( bf->GetOperatorSymbol()->IsAndJunctor() && child_left )
          continue; // unary rule ...

        if( bf->GetOperatorSymbol()->IsOrJunctor() && !child_left )
          continue; // unary rule ...

        if( bf->GetOperatorSymbol()->IsImplJunctor() )
        {
          // is this the unary version of the implaction rule? i.e. impl:r?
          if( left_formula == child_left )
          {
            if( left_formula )
              child_left = !child_left;

            continue; // unary rule ...
          }

          if( left_formula )
            child_left = !child_left;
        }

        // ok, here we know. this needs a binary rule!

        // get the other formula
        ProofLib::Formula *other = left_formula ? bf->GetRight() :
                                                  bf->GetLeft();

        bool other_left = child_left;

        if( !left_formula && bf->GetOperatorSymbol()->IsImplJunctor() )
          other_left = !child_left;

#if 0
      std::cerr << "OTHER (other_left=";
      std::cerr << ( other_left ? "left" : "right" ) << "): ";
      Print::PrintFormula( other, std::cerr );
      std::cerr << std::endl;
#endif
#if 0
      std::cerr << "WORLD: ";
      Print::PrintSequent( world, std::cerr );
      std::cerr << std::endl;
#endif
        // we need now the other formula on the opposite polarization.
        int this_costs = CalcFormulaCosts( other, op, !other_left, world );

        if( this_costs < 0 )
        {
          costs = -1;
          break;
        }

        costs += this_costs;
      }

      if( costs >= 0 && ( first || costs < min_costs ) )
      {
        best_np = &o[i].np;
        min_costs = costs;
        first = false;
      }
    }

    if( left )
    {
      m_deduct_complete_costs[f].first = min_costs;
      if( best_np != NULL ) m_deduct_complete_ways[f].first = *best_np;
    }
    else
    {
      m_deduct_complete_costs[f].second = min_costs;
      if( best_np != NULL ) m_deduct_complete_ways[f].second = *best_np;
    }

    return min_costs;
  }


void NewDeductCompletePropositional( ProofLib::Sequent *taut, 
                                     ProofLib::Rule *rule,
				     const ProofRule *r,
				     ProofLib::Sequent *world,
				     ProofLib::Formula *aux_left = NULL,
				     ProofLib::Formula *aux_right = NULL)
  {

#ifdef DEBUG_PRINT_INFO
    std::cerr << "NewDeductPropositional with taut: " ;
    HLK::Print::PrintSequent( taut, std::cerr ) ;
    std::cerr << std::endl ;
#endif

    ProofLib::SElementList *left = taut->GetLeft();
    ProofLib::SElementList *right = taut->GetRight();

    // check whether we find the same formula on both sides
    if (left != NULL && right != NULL) for ( size_t i = 0; i < left->GetChildCount(); i++)
    {
      ProofLib::SFormula *sfl = dynamic_cast<ProofLib::SFormula*>(left->GetChild( i ));
      ASSERT( sfl != NULL );
      ProofLib::Formula *fl = sfl->GetFormula();
      
      for ( size_t j = 0; j < right->GetChildCount(); j++ )
      {
        ProofLib::SFormula *sfr = dynamic_cast<ProofLib::SFormula*>(right->GetChild( j ));
        ASSERT( sfr != NULL );
        ProofLib::Formula *fr = sfr->GetFormula();

	if ( fl == fr )
	{
#ifdef DEBUG_PRINT_INFO
	  std::cerr << "found on both sides: ";
	  Print::PrintFormula( fl, std::cerr );
	  std::cerr << std::endl;
#endif
          AxiomToAtomic( rule, *r, fl ) ;
	  ProofLib::AtomFormula *af = dynamic_cast<ProofLib::AtomFormula*>( fl );
	  if ( af != NULL )
	  {
            ProofLib::Sequent *conc = new ProofLib::Sequent( new ProofLib::SElementList,
		                                             new ProofLib::SElementList );
            conc->AddRight( af );
	    conc->AddLeft( af );
	    // may need to add auxiliary formula of rule to sequent
	    if ( aux_right != NULL && af != aux_right )
	      conc->AddRight( aux_right );
	    if ( aux_left != NULL && af != aux_left )
	      conc->AddLeft( aux_left );
	    rule->SetConclusion( conc );
	  }
	  return;
	}
      }
    }

    // need to remember auxiliary formulas of unary rules
    // with two auxiliary formulas (i.e. impl right)
    ProofLib::Formula *new_aux_left = NULL, *new_aux_right = NULL;

    // decompose formulas
    // right side of the sequent
    if (right != NULL) for (size_t i = 0; i < right->GetChildCount(); i++)
    {
      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>(right->GetChild( i ));
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();
      if ( dynamic_cast<ProofLib::AtomFormula*>(f) != NULL )
	continue;

      ProofLib::UnaryConFormula *uf = dynamic_cast<ProofLib::UnaryConFormula*>(f);

      if ( uf != NULL )
      {
        rule = BuildNotRule( rule, uf, false, NULL );

        ProofLib::Formula *newf = uf->GetFormula();
        ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	ProofLib::SElementList *newleft = new ProofLib::SElementList;
	newleft->AddLeft( left );
	ProofLib::SElementList *newright = new ProofLib::SElementList;
	newright->AddLeft( right );
	newright->RemoveAt( i );
	newleft->AddLeft( new ProofLib::SFormula( newf ) ) ;
	newtaut->SetLeft( newleft );
	newtaut->SetRight( newright );

	NewDeductCompletePropositional( newtaut, rule, r, world );

	return;
      }

      ProofLib::BinaryConFormula *bf = dynamic_cast<ProofLib::BinaryConFormula*>(f);
      if ( bf != NULL )
      {
	if ( bf->GetOperatorSymbol()->IsImplJunctor() )
	{
          rule = BuildImplRightRule( rule, bf, NULL );

          new_aux_right = bf->GetRight();
	  new_aux_left = bf->GetLeft();

          ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  ProofLib::SElementList *newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newright->RemoveAt( i );
	  newleft->AddLeft( new ProofLib::SFormula( new_aux_left ) ) ;
	  newright->AddLeft( new ProofLib::SFormula( new_aux_right ) ) ;
	  newtaut->SetLeft( newleft );
	  newtaut->SetRight( newright );

	  NewDeductCompletePropositional( newtaut, rule, r, world, new_aux_left, new_aux_right );

	  return;
	}

	if ( bf->GetOperatorSymbol()->IsOrJunctor() )
	{
          ProofLib::Formula *leftf = bf->GetLeft();
          ProofLib::Formula *rightf = bf->GetRight();

          ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  ProofLib::SElementList *newright = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newtaut->SetLeft( newleft ) ;
          newright->AddLeft( right );
          newright->RemoveAt( i );

          rule = BuildUnaryAndOrRule( rule, bf, false, NULL ) ;
          newright->AddLeft( new ProofLib::SFormula( rightf ) ) ;

 	  // check if we are already done
	  if ( left == NULL || !left->ContainsFormula( rightf ) )
	  {
	    rule = BuildUnaryAndOrRule( rule, bf, true, NULL ) ;
	    newright->AddLeft( new ProofLib::SFormula( leftf ) ) ;
	  }
          newtaut->SetRight( newright );

	  NewDeductCompletePropositional( newtaut, rule, r, world );
	  return;
	}

	if ( bf->GetOperatorSymbol()->IsAndJunctor() )
	{
	  ProofLib::Rule *next_rule_l = NULL;
	  ProofLib::Rule *next_rule_r = NULL;

          BuildBinaryAndOrImplRule( rule, bf, NULL,
	                            &next_rule_l,
				    &next_rule_r,
				    NULL, NULL );

          ProofLib::Formula *rightf = bf->GetRight();
          ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newtaut->SetLeft( newleft ) ;
	  ProofLib::SElementList *newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newright->RemoveAt( i );
	  newright->AddLeft( new ProofLib::SFormula( rightf ) ) ;
	  newtaut->SetRight( newright );

	  NewDeductCompletePropositional( newtaut, next_rule_r, r, world );

          ProofLib::Formula *leftf = bf->GetLeft();
	  newtaut = new ProofLib::Sequent(NULL, NULL);
	  newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newtaut->SetLeft( newleft ) ;
	  newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newright->RemoveAt( i );
	  newright->AddLeft( new ProofLib::SFormula( leftf ) ) ;
	  newtaut->SetRight( newright );

	  NewDeductCompletePropositional( newtaut, next_rule_l, r, world );

	  return;
	}
      }
    }

    // left side of the sequent
    if (left != NULL) for (size_t i = 0; i < left->GetChildCount(); i++)
    {
      ProofLib::SFormula *sf = dynamic_cast<ProofLib::SFormula*>(left->GetChild( i ));
      ASSERT( sf != NULL );
      ProofLib::Formula *f = sf->GetFormula();
      if ( dynamic_cast<ProofLib::AtomFormula*>(f) != NULL )
	continue;


      ProofLib::UnaryConFormula *uf = dynamic_cast<ProofLib::UnaryConFormula*>(f);

      if ( uf != NULL )
      {
        rule = BuildNotRule( rule, uf, true, NULL );

        ProofLib::Formula *newf = uf->GetFormula();
        ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	ProofLib::SElementList *newleft = new ProofLib::SElementList;
	newleft->AddLeft( left );
	ProofLib::SElementList *newright = new ProofLib::SElementList;
	newright->AddLeft( right );
	newleft->RemoveAt( i );
	newright->AddLeft( new ProofLib::SFormula( newf ) ) ;
	newtaut->SetLeft( newleft );
	newtaut->SetRight( newright );

	NewDeductCompletePropositional( newtaut, rule, r, world );
	return;
      }

      ProofLib::BinaryConFormula *bf = dynamic_cast<ProofLib::BinaryConFormula*>(f);
      if ( bf != NULL )
      {
	if ( bf->GetOperatorSymbol()->IsImplJunctor() )
	{
	  ProofLib::Rule *next_rule_l = NULL;
	  ProofLib::Rule *next_rule_r = NULL;

          BuildBinaryAndOrImplRule( rule, bf, NULL,
	                            &next_rule_l,
				    &next_rule_r,
				    NULL, NULL );

          ProofLib::Formula *leftf = bf->GetLeft();
	  ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newleft->RemoveAt( i );
	  newtaut->SetLeft( newleft );
	  ProofLib::SElementList *newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newright->AddLeft( new ProofLib::SFormula( leftf ) ) ;
          newtaut->SetRight( newright );

	  NewDeductCompletePropositional( newtaut, next_rule_l, r, world );

          ProofLib::Formula *rightf = bf->GetRight();
          newtaut = new ProofLib::Sequent(NULL, NULL);
	  newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newtaut->SetRight( newright ) ;
	  newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newleft->RemoveAt( i );
	  newleft->AddLeft( new ProofLib::SFormula( rightf ) ) ;
	  newtaut->SetLeft( newleft );

	  NewDeductCompletePropositional( newtaut, next_rule_r, r, world );

	  return;
	}

	if ( bf->GetOperatorSymbol()->IsOrJunctor() )
	{
	  ProofLib::Rule *next_rule_l = NULL;
	  ProofLib::Rule *next_rule_r = NULL;

          BuildBinaryAndOrImplRule( rule, bf, NULL,
	                            &next_rule_l,
				    &next_rule_r,
				    NULL, NULL );

          ProofLib::Formula *rightf = bf->GetRight();
          ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
	  ProofLib::SElementList* newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newtaut->SetRight( newright ) ;
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newleft->RemoveAt( i );
	  newleft->AddLeft( new ProofLib::SFormula( rightf ) ) ;
	  newtaut->SetLeft( newleft );

	  NewDeductCompletePropositional( newtaut, next_rule_r, r, world );

          ProofLib::Formula *leftf = bf->GetLeft();
	  newtaut = new ProofLib::Sequent(NULL, NULL);
	  newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newtaut->SetRight( newright );
	  newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newleft->RemoveAt( i );
	  newleft->AddLeft( new ProofLib::SFormula( leftf ) ) ;
	  newtaut->SetLeft( newleft );

	  NewDeductCompletePropositional( newtaut, next_rule_l, r, world );

	  return;
	}

	if ( bf->GetOperatorSymbol()->IsAndJunctor() )
	{
          ProofLib::Formula *leftf = bf->GetLeft();
          ProofLib::Formula *rightf = bf->GetRight();

          ProofLib::Sequent *newtaut = new ProofLib::Sequent(NULL, NULL);
          ProofLib::SElementList *newright = new ProofLib::SElementList;
	  newright->AddLeft( right );
	  newtaut->SetRight( newright ) ;
	  ProofLib::SElementList *newleft = new ProofLib::SElementList;
	  newleft->AddLeft( left );
	  newleft->RemoveAt( i );

          rule = BuildUnaryAndOrRule( rule, bf, true, NULL ) ;
	  newleft->AddLeft( new ProofLib::SFormula( leftf ) ) ;

 	  // check if we are already done
	  if ( right == NULL || !right->ContainsFormula( leftf ) )
	  {
            rule = BuildUnaryAndOrRule( rule, bf, false, NULL ) ;
 
	    newleft->AddLeft( new ProofLib::SFormula( rightf ) ) ;
	  }

	  newtaut->SetLeft( newleft );

	  NewDeductCompletePropositional( newtaut, rule, r, world );
	  return;
	}
      }
    }

#ifdef DEBUG_PRINT_INFO
    std::cerr << "did not find same formula on both sides" << std::endl;
#endif

    ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
    ex->Stream() << "There is no propositional proof for ";
    Print::PrintSequent( taut, ex->Stream() ) ;
    ex->Stream() << ".";
    throw ex;
  }

  void DeductCompletePropositional(
    ProofLib::Rule *rule,
    const ProofRule *r,
    ProofLib::Sequent *world )
  {
    ProofLib::OperationFormulasPosAndPol op;
    op.SetSearchBinaryRule( true );
    world->Apply( op );

    // clear costs cache.
    m_deduct_complete_costs.clear();
    m_deduct_complete_ways.clear();
    ProofLib::OperationFormulasPosAndPol::FormulaPositions::const_iterator i;

#if 0
std::cerr << " ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << std::endl;
std::cerr << " ++ WORLD: " << std::endl;
Print::PrintSequent( world, std::cerr );
std::cerr << std::endl;
std::cerr << std::endl;
#endif
    bool first = true;
    ProofLib::Formula *our_hero = NULL;
    size_t costs = 0;
    for( i = op.m_positions.begin(); i != op.m_positions.end(); i++ )
    {
      int c_l = CalcFormulaCosts( i->first, op, true, world );
      int c_r = CalcFormulaCosts( i->first, op, false, world );
#if 0
      Print::PrintFormula( i->first, std::cerr );
      std::cerr << ": " << c_l << " x " << c_r << std::endl;
#endif

      if( c_l == 0 && c_r == 0 )
      {
        AxiomToAtomic( rule, *r, i->first );
        return;
      }

      if( c_l >= 0 && c_r >= 0 && ( first || size_t(c_l+c_r) < costs ) )
      {
#if 0
    std::cerr << "NEW HERO (" << costs << " > " << c_l + c_r << "): ";
    Print::PrintFormula( i->first, std::cerr );
    std::cerr << std::endl;
#endif
        costs = c_l + c_r;
        our_hero = i->first;
        first = false;
      }
    }

    if( our_hero == NULL )
    {
      ProofLib::GenericSLocException *ex = new ProofLib::GenericSLocException( r );
      ex->Stream() << "There is no propositional proof for the world.";
      throw ex;
    }

#if 0
    std::cerr << "START: ";
    Print::PrintFormula( our_hero, std::cerr );
    std::cerr << std::endl;
#endif

    rule = BuildProofFor( rule, our_hero, r, true, world );
    rule = BuildProofFor( rule, our_hero, r, false, world );
    AxiomToAtomic( rule, *r, our_hero );
  }

  typedef stlhash::hash_set<ProofLib::Formula*> FormulaSet;

  ProofLib::Rule* BuildProofFor( 
    ProofLib::Rule *rule,
    ProofLib::Formula *f,
    const ProofRule *r,
    bool left,
    ProofLib::Sequent *world,
    FormulaSet visited = FormulaSet() )
  {
    ASSERT( m_deduct_complete_ways.find(f) != m_deduct_complete_ways.end() );

    if( visited.find(f) != visited.end() )
      return rule;

    ProofLib::NodePointer np = left ? m_deduct_complete_ways[f].first :
                                      m_deduct_complete_ways[f].second;
#if 0
    std::cerr << "BuildProofFor: ";
    Print::PrintFormula( f, std::cerr );
    std::cerr << ": " << ( left ? "left" : "right" ) << " pos: ";
    Print::PrintNodePointer( &np, std::cerr );
    std::cerr << std::endl;
#endif

    std::stack<size_t> todo;

#if 0
    std::cerr << "todo: ";
#endif
    while( !np.IsEmpty() )
    {
      ProofLib::DataNode *node = np.GetDataNode( world );

      if( node->CastTo<ProofLib::Formula>() == NULL )
        break;

      todo.push( np.Pop() );
#if 0
    std::cerr << todo.top() << " ";
#endif
    }
#if 0
    std::cerr << std::endl;
#endif

    ASSERT( !np.IsEmpty() );

    left = np[0] == 0;

    while( !todo.empty() )
    {
      np.Push( todo.top() );
      todo.pop();

      if( todo.empty() )
        break;

#if 0
    std::cerr << "np: ";
    Print::PrintNodePointer( &np, std::cerr );
    std::cerr << std::endl;
#endif

      ProofLib::DataNode *node =  np.GetDataNode( world );

      ProofLib::UnaryConFormula *uf = NULL;
      ProofLib::BinaryConFormula *bf = NULL;

      if( (uf = node->CastTo<ProofLib::UnaryConFormula>()) != NULL )
      {
        ASSERT( uf->GetOperatorSymbol()->IsNegJunctor() );

        rule = BuildNotRule( rule, uf, left, NULL );

        left = !left; 
      }
      else if( (bf = node->CastTo<ProofLib::BinaryConFormula>()) != NULL )
      {
        bool left_formula = ( todo.top() == 0 );

#if 0
        ProofLib::Formula *c =
          left_formula ? bf->GetLeft() : bf->GetRight();
#endif

        if( ( bf->GetOperatorSymbol()->IsAndJunctor() && left ) ||
            ( bf->GetOperatorSymbol()->IsOrJunctor() && !left ) )
        {
          rule = BuildUnaryAndOrRule( rule, bf, left_formula, NULL );
        }
        else if( bf->GetOperatorSymbol()->IsImplJunctor() && !left )
        {
          rule = BuildImplRightRule( rule, bf, NULL );

          if( left_formula ) left = !left;
        }
        else if( ( bf->GetOperatorSymbol()->IsAndJunctor() && !left ) ||
                 ( bf->GetOperatorSymbol()->IsImplJunctor() && left ) ||
                 ( bf->GetOperatorSymbol()->IsOrJunctor() && left ) )
        {
          ProofLib::Rule *next_rule_l = NULL;
          ProofLib::Rule *next_rule_r = NULL;

          BuildBinaryAndOrImplRule( rule, bf, NULL,
                                    &next_rule_l, &next_rule_r,
                                    NULL, NULL );

          // we have now build the binary rule, on one side the subformula 
          // f will be brought up by applying nessecary rules. On the other
          // branch we need to complete the proof for the second rule of
          // bf.
          ProofLib::Formula *other =
            !left_formula ? bf->GetLeft() : bf->GetRight();
          ProofLib::Rule *other_rule = NULL;
          bool other_left = !left;

          if( left_formula )
          {
            rule = next_rule_l;
            other_rule = next_rule_r;
            if( bf->GetOperatorSymbol()->IsImplJunctor() )
              left = !left;
          }
          else
          {
            rule = next_rule_r;
            other_rule = next_rule_l;
            if( bf->GetOperatorSymbol()->IsImplJunctor() )
              other_left = !other_left;
          }

          visited.insert( f );
          other_rule = BuildProofFor( other_rule, other, r,
                                      other_left, world, visited );

          AxiomToAtomic( other_rule, *r, other );
        }
        else
          ASSERT( false ); // can not handle whatever it is ...
      }
    }

    return rule;
  }

  static ProofLib::Formula* FindAnAxiom( 
    ProofLib::Sequent *seq,
    ProofLib::Sequent *world = NULL )
  {
    ASSERT( seq != NULL );

    ProofLib::SElementList *a = seq->GetLeft();
    ProofLib::SElementList *b_0 = ( world == NULL ? NULL : world->GetRight() );
    ProofLib::SElementList *b_1 = seq->GetRight();
    bool on_left_side = true;

    while( true )
    {
      if( a != NULL && ( b_0 != NULL || b_1 != NULL ) )
        for( size_t i_a = 0; i_a < a->GetChildCount(); i_a++ )
      {
        ProofLib::SFormula *sf_a =
          dynamic_cast<ProofLib::SFormula*>( a->GetSElement(i_a) );
  
        ASSERT( sf_a != NULL );
  
        ProofLib::Formula *f_a = sf_a->GetFormula();
#if 0
        Print::PrintFormula( f_a, std::cerr );
        std::cerr << std::endl;
#endif

        ProofLib::SElementList *b = b_0;

        while( true )
        {
          if( b != NULL )
            for( size_t i_b = 0; i_b < b->GetChildCount(); i_b++ )
          {
            ProofLib::SFormula *sf_b =
              dynamic_cast<ProofLib::SFormula*>( b->GetSElement(i_b) );
      
            ASSERT( sf_b != NULL );
      
            ProofLib::Formula *f_b = sf_b->GetFormula();
#if 0
          std::cerr << "TRY: ";
          Print::PrintFormula( f_b, std::cerr );
          std::cerr << std::endl;
#endif
  
            if( f_a == f_b )
              return f_a;
          }

          if( b == b_1 )
            break;

          b = b_1;
        }
      }

      a = seq->GetRight();
      b_0 = ( world == NULL ? NULL : world->GetLeft() );
      b_1 = seq->GetLeft();

      if( !on_left_side )
        break; // finally done

      on_left_side = false;
    }

    return NULL;
  }
};

bool Compiler::simulate_pred_undefs = false;

Compiler* Compiler::NewCompiler()
{
  return new CompilerImpl;
}

}
